@@ -3,9 +3,10 @@
     Part of SWI-Prolog
 
     Author:        Jan Wielemaker
-    E-mail:        J.Wielemaker@uva.nl
+    E-mail:        J.Wielemaker@ca.vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (C): 1985-2009, University of Amsterdam
+    Copyright (C): 1985-2011, University of Amsterdam
+			      VU University Amsterdam
 
     This program is free software; you can redistribute it and/or
     modify it under the terms of the GNU General Public License
@@ -17,9 +18,9 @@
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
 
-    You should have received a copy of the GNU Lesser General Public
+    You should have received a copy of the GNU General Public
     License along with this library; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 
     As a special exception, if you link this library with other files,
     compiled with a Free Software compiler, to produce an executable, this
@@ -38,7 +39,9 @@
 :- multifile
 	prolog:message/3,		% entire message
 	prolog:error_message/3,		% 1-st argument of error term
-	prolog:message_context/3.	% Context of error messages
+	prolog:message_context/3,	% Context of error messages
+	prolog:message_location/3,	% (File) location of error messages
+	prolog:message_line_element/2.	% Extend printing
 :- discontiguous
 	prolog_message/3.
 
@@ -56,6 +59,8 @@
 %	    * flush
 %	    Used only as last element of the list.   Simply flush the
 %	    output instead of producing a final newline.
+%	    * at_same_line
+%	    Start the messages at the same line (instead of using ~N)
 
 translate_message(Term) -->
 	translate_message2(Term), !.
@@ -72,7 +77,7 @@ translate_message2(error(resource_error(stack), Name)) -->
 translate_message2(error(resource_error(Missing), _)) -->
 	[ 'Not enough resources: ~w'-[Missing] ].
 translate_message2(error(ISO, SWI)) -->
-	swi_context(SWI),
+	swi_location(SWI),
 	term_message(ISO),
 	swi_extra(SWI).
 translate_message2('$aborted') -->
@@ -115,8 +120,8 @@ iso_message(domain_error(Domain, Actual)) -->
 	[ 'Domain error: `~w'' expected, found `~p'''-[Domain, Actual] ].
 iso_message(instantiation_error) -->
 	[ 'Arguments are not sufficiently instantiated' ].
-iso_message(representation_error(variable)) -->
-	[ 'Cannot represent: argument must be unbound' ].
+iso_message(uninstantiation_error(Var)) -->
+	[ 'Uninstantiated argument expected, found ~p'-[Var] ].
 iso_message(representation_error(What)) -->
 	[ 'Cannot represent due to `~w'''-[What] ].
 iso_message(permission_error(Action, built_in_procedure, Pred)) -->
@@ -134,13 +139,8 @@ iso_message(permission_error(Action, Type, Object)) -->
 iso_message(evaluation_error(Which)) -->
 	[ 'Arithmetic: evaluation error: `~p'''-[Which] ].
 iso_message(existence_error(procedure, Proc)) -->
-	[ 'Undefined procedure: ~p'-[Proc] ],
-	{ dwim_predicates(Proc, Dwims) },
-	(   {Dwims \== []}
-	->  [nl, '    However, there are definitions for:', nl],
-	    dwim_message(Dwims)
-	;   []
-	).
+	[ 'Undefined procedure: ~q'-[Proc] ],
+	undefined_proc_msg(Proc).
 iso_message(existence_error(Type, Object)) -->
 	[ '~w `~p'' does not exist'-[Type, Object] ].
 iso_message(busy(Type, Object)) -->
@@ -151,6 +151,31 @@ iso_message(syntax_error(Id)) -->
 iso_message(occurs_check(Var, In)) -->
 	[ 'Cannot unify ~p with ~p: would create an infinite tree'-[Var, In] ].
 
+undefined_proc_msg(_:(^)/2) --> !,
+	undefined_proc_msg((^)/2).
+undefined_proc_msg((^)/2) --> !,
+	[nl, '  ^/2 can only appear as the 2nd argument of setof/3 and bagof/3'].
+undefined_proc_msg((:-)/2) --> !,
+	[nl, '  Rules must be loaded from a file'],
+	faq('ToplevelMode').
+undefined_proc_msg((:-)/1) --> !,
+	[nl, '  Directives must be loaded from a file'],
+	faq('ToplevelMode').
+undefined_proc_msg((?-)/1) --> !,
+	[nl, '  ?- is the Prolog prompt'],
+	faq('ToplevelMode').
+undefined_proc_msg(Proc) -->
+	{ dwim_predicates(Proc, Dwims) },
+	(   {Dwims \== []}
+	->  [nl, '  However, there are definitions for:', nl],
+	    dwim_message(Dwims)
+	;   []
+	).
+
+faq(Page) -->
+	[nl, '  See FAQ at http://www.swi-prolog.org/FAQ/', Page, '.txt' ].
+
+
 syntax_error(end_of_clause) -->
 	[ 'Unexpected end of clause' ].
 syntax_error(end_of_clause_expected) -->
@@ -175,6 +200,10 @@ syntax_error(operator_expected) -->
 	[ 'Operator expected' ].
 syntax_error(operator_balance) -->
 	[ 'Unbalanced operator' ].
+syntax_error(quoted_punctuation) -->
+	[ 'Operand expected, unquoted comma or bar found' ].
+syntax_error(list_rest) -->
+	[ 'Unexpected comma or bar in rest of list' ].
 syntax_error(cannot_start_term) -->
 	[ 'Illegal start of term' ].
 syntax_error(punct(Punct, End)) -->
@@ -188,17 +217,20 @@ dwim_predicates(Name/_Arity, Dwims) :-
 	findall(Dwim, dwim_predicate(user:Name, Dwim), Dwims).
 
 dwim_message([]) --> [].
-dwim_message([user:Head|T]) --> !,
-	{functor(Head, Name, Arity)},
-	[ '        ~w/~d'-[Name, Arity], nl ],
+dwim_message([M:Head|T]) -->
+	{ hidden_module(M), !,
+	  functor(Head, Name, Arity)
+	},
+	[ '        ~q'-[Name/Arity], nl ],
 	dwim_message(T).
 dwim_message([Module:Head|T]) --> !,
-	{functor(Head, Name, Arity)},
-	[ '        ~w:~w/~d'-[Module, Name, Arity], nl],
+	{ functor(Head, Name, Arity)
+	},
+	[ '        ~q'-[Module:Name/Arity], nl],
 	dwim_message(T).
 dwim_message([Head|T]) -->
 	{functor(Head, Name, Arity)},
-	[ '        ~w/~d'-[Name, Arity], nl],
+	[ '        ~q'-[Name/Arity], nl],
 	dwim_message(T).
 
 
@@ -245,21 +277,26 @@ swi_message(format_argument_type(Fmt, Arg)) -->
 swi_message(format(Msg)) -->
 	[ 'Format error: ~w'-[Msg] ].
 
-swi_context(X) -->
+swi_location(X) -->
 	{ var(X)
 	}, !,
 	[].
-swi_context(context(Caller, _Msg)) -->
+swi_location(Context) -->
+	prolog:message_location(Context), !.
+swi_location(context(Caller, _Msg)) -->
 	{ ground(Caller)
 	}, !,
 	caller(Caller).
-swi_context(file(Path, Line, -1, _CharNo)) --> !,
+swi_location(file(Path, Line, -1, _CharNo)) --> !,
 	[ '~w:~d: '-[Path, Line] ].
-swi_context(file(Path, Line, LinePos, _CharNo)) -->
+swi_location(file(Path, Line, LinePos, _CharNo)) -->
 	[ '~w:~d:~d: '-[Path, Line, LinePos] ].
-swi_context(stream(Stream, Line, LinePos, _CharNo)) -->
-	[ 'Stream ~w:~d:~d '-[Stream, LinePos, Line] ].
-swi_context(_) -->
+swi_location(stream(Stream, Line, LinePos, CharNo)) -->
+	(   { stream_property(Stream, file_name(File)) }
+	->  swi_location(file(File, Line, LinePos, CharNo))
+        ;   [ 'Stream ~w:~d:~d '-[Stream, LinePos, Line] ]
+	).
+swi_location(_) -->
 	[].
 
 caller(system:'$record_clause'/3) --> !,
@@ -292,15 +329,33 @@ swi_extra(string(String, CharPos)) -->
 	},
 	[ nl, '~w'-[Before], nl, '** here **', nl, '~w'-[After] ].
 swi_extra(_) -->
+	[].
+
+thread_context -->
+	{ thread_self(Me), Me \== main }, !,
+	['[Thread ~w] '-[Me]].
+thread_context -->
 	[].
 
 		 /*******************************
 		 *	  NORMAL MESSAGES	*
 		 *******************************/
 
-prolog_message(initialization_exception(Goal, E)) -->
+prolog_message(initialization_error(_, E, File:Line)) --> !,
+	[ '~w:~d: '-[File, Line],
+	  'Initialization goal raised exception:', nl
+	],
+	translate_message(E).
+prolog_message(initialization_error(Goal, E, _)) -->
 	[ 'Initialization goal ~p raised exception:'-[Goal], nl ],
 	translate_message(E).
+prolog_message(initialization_failure(_Goal, File:Line)) --> !,
+	[ '~w:~d: '-[File, Line],
+	  'Initialization goal failed'-[]
+	].
+prolog_message(initialization_failure(Goal, _)) -->
+	[ 'Initialization goal failed: ~p'-[Goal]
+	].
 prolog_message(initialization_exception(E)) -->
 	[ 'Prolog initialisation failed:', nl ],
 	translate_message(E).
@@ -369,6 +424,8 @@ prolog_message(declare_module(Module, abolish(Predicates))) -->
 	[ 'Loading module ~w abolished: ~p'-[Module, Predicates] ].
 prolog_message(undefined_export(Module, PI)) -->
 	[ 'Exported procedure ~q:~q is not defined'-[Module, PI] ].
+prolog_message(no_exported_op(Module, Op)) -->
+	[ 'Operator ~q:~q is not exported (still defined)'-[Module, Op] ].
 prolog_message(discontiguous((-)/2)) -->
 	prolog_message(minus_in_identifier).
 prolog_message(discontiguous(Proc)) -->
@@ -377,16 +434,16 @@ prolog_message(load_file(start(Level, File))) -->
 	[ '~|~t~*+Loading '-[Level] ],
 	load_file(File),
 	[ ' ...' ].
-prolog_message(load_file(done(Level, File, Action, Module, Time, Heap))) -->
+prolog_message(load_file(done(Level, File, Action, Module, Time, Clauses))) -->
 	[ '~|~t~*+'-[Level] ],
 	load_file(File),
 	[ ' ~w'-[Action] ],
 	load_module(Module),
-	[ ' ~2f sec, ~D bytes'-[Time, Heap] ].
+	[ ' ~2f sec, ~D clauses'-[Time, Clauses] ].
 prolog_message(dwim_undefined(Goal, Alternatives)) -->
 	{ goal_to_predicate_indicator(Goal, Pred)
 	},
-	[ 'Undefined procedure: ~p'-[Pred], nl,
+	[ 'Undefined procedure: ~q'-[Pred], nl,
 	  '    However, there are definitions for:', nl
 	],
 	dwim_message(Alternatives).
@@ -403,7 +460,15 @@ prolog_message(minus_in_identifier) -->
 	].
 prolog_message(qlf(removed_after_error(File))) -->
 	[ 'Removed incomplete QLF file ~w'-[File] ].
-
+prolog_message(redefine_module(Module, OldFile, File)) -->
+	[ 'Module "~q" already loaded from ~w.'-[Module, OldFile], nl,
+	  'Wipe and reload from ~w? '-[File], flush
+	].
+prolog_message(redefine_module_reply) -->
+	[ 'Please answer y(es), n(o) or a(bort)' ].
+prolog_message(reloaded_in_module(Absolute, OldContext, LM)) -->
+	[ '~w was previously loaded in module ~w'-[Absolute, OldContext], nl,
+	  '\tnow it is reloaded into module ~w'-[LM] ].
 
 used_search([]) -->
 	[].
@@ -450,6 +515,7 @@ prolog_message(file_search(tried(Spec, Cond), Path)) -->
 		 *******************************/
 
 prolog_message(gc(start)) -->
+	thread_context,
 	[ 'GC: ', flush ].
 prolog_message(gc(done(G, T, Time, UG, UT, RG, RT))) -->
 	[ at_same_line,
@@ -457,6 +523,7 @@ prolog_message(gc(done(G, T, Time, UG, UT, RG, RT))) -->
 	  [G, T, Time, UG, UT, RG, RT]
 	].
 prolog_message(shift_stacks(start(_L,_G,_T))) -->
+	thread_context,
 	[ 'Stack-shift: ', flush ].
 prolog_message(shift_stacks(done(Time, L, G, T))) -->
 	{ LKB is L//1024,
@@ -468,9 +535,8 @@ prolog_message(shift_stacks(done(Time, L, G, T))) -->
 	  [LKB, GKB, TKB, Time]
 	].
 prolog_message(agc(start)) -->
-	{ thread_self(Me)
-	},
-	[ 'AGC: [~w] '-[Me], flush ].
+	thread_context,
+	[ 'AGC: ', flush ].
 prolog_message(agc(done(Collected, Remaining, Time))) -->
 	[ at_same_line,
 	  'reclaimed ~D atoms in ~2f sec. (remaining: ~D)'-
@@ -492,7 +558,8 @@ prolog_message(make(done(_Files))) -->
 prolog_message(make(library_index(Dir))) -->
 	[ 'Updating index for library ~w'-[Dir] ].
 prolog_message(autoload(Pred, File)) -->
-	[ 'Autoloading ~p from ~w'-[Pred, File] ].
+	thread_context,
+	[ 'autoloading ~p from ~w'-[Pred, File] ].
 prolog_message(autoload(read_index(Dir))) -->
 	[ 'Loading autoload index for ~w'-[Dir] ].
 
@@ -522,7 +589,7 @@ prolog_message(threads) -->
 prolog_message(threads) -->
 	[].
 prolog_message(copyright) -->
-	[ 'Copyright (c) 1990-2009 University of Amsterdam.', nl,
+	[ 'Copyright (c) 1990-2011 University of Amsterdam, VU Amsterdam', nl,
 	  'SWI-Prolog comes with ABSOLUTELY NO WARRANTY. This is free software,', nl,
 	  'and you are welcome to redistribute it under certain conditions.', nl,
 	  'Please visit http://www.swi-prolog.org for details.'
@@ -550,10 +617,10 @@ prolog_message(about) -->
 	prolog_message(copyright).
 prolog_message(halt) -->
 	[ 'halt' ].
-prolog_message(break(enter(Level))) -->
+prolog_message(break(begin, Level)) -->
 	[ 'Break level ~d'-[Level] ].
-prolog_message(break(exit(Level))) -->
-	[ nl, 'Exit break level ~d'-[Level] ].
+prolog_message(break(end, Level)) -->
+	[ 'Exit break level ~d'-[Level] ].
 prolog_message(var_query(_)) -->
 	[ '... 1,000,000 ............ 10,000,000 years later', nl, nl,
 	  '~t~8|>> 42 << (last release gives the question)'
@@ -565,25 +632,21 @@ prolog_message(query(QueryResult)) -->
 	query_result(QueryResult).
 
 query_result(no) -->		% failure
-	[ 'false.'-[] ],
+	[ ansi([bold,fg(red)], 'false.', []) ],
 	extra_line.
 query_result(yes([])) --> !,	% prompt_alternatives_on: groundness
-	[ 'true.'-[] ],
+	[ ansi(bold, 'true.', []) ],
 	extra_line.
 query_result(yes(Residuals)) -->
-	residuals(Residuals),
+	result([], Residuals),
 	extra_line.
 query_result(done) -->		% user typed <CR>
 	extra_line.
 query_result(yes(Bindings, Residuals)) -->
-	bindings(Bindings),
-	bind_res_sep(Bindings, Residuals),
-	residuals(Residuals),
+	result(Bindings, Residuals),
 	prompt(yes, Bindings, Residuals).
 query_result(more(Bindings, Residuals)) -->
-	bindings(Bindings),
-	bind_res_sep(Bindings, Residuals),
-	residuals(Residuals),
+	result(Bindings, Residuals),
 	prompt(more, Bindings, Residuals).
 query_result(help) -->
 	[ nl, 'Actions:'-[], nl, nl,
@@ -608,38 +671,69 @@ prompt(Answer, _, _) --> !,
 	prompt(Answer, non_empty).
 
 prompt(yes, empty) --> !,
-	[ 'true.'-[] ],
+	[ ansi(bold, 'true.', []) ],
 	extra_line.
 prompt(yes, _) --> !,
 	[ full_stop ],
 	extra_line.
 prompt(more, empty) --> !,
-	[ 'true '-[], flush ].
+	[ ansi(bold, 'true ', []), flush ].
 prompt(more, _) --> !,
 	[ ' '-[], flush ].
 
-bindings([]) -->
-	[].
-bindings([Name = Value|T]) -->
-	{ current_prolog_flag(toplevel_print_options, Options),
-	  (   T == []
-	  ->  Cont = ''
-	  ;   Cont = (,)
-	  )
+result(Bindings, Residuals) -->
+	{ current_prolog_flag(toplevel_print_options, Options0),
+	  Options = [partial(true)|Options0]
 	},
-	[ '~w = ~W~a'-[Name, Value, [partial(true)|Options], Cont] ],
-	(   {T == []}
-	->  []
-	;   [nl]
-	),
-	bindings(T).
+	bindings(Bindings, [priority(699)|Options]),
+	bind_res_sep(Bindings, Residuals),
+	residuals(Residuals, [priority(999)|Options]).
 
-residuals([]) --> !, [].
-residuals([G|Gs]) -->
+bindings([], _) -->
+	[].
+bindings([binding(Names,Skel,Subst)|T], Options) -->
+	{ '$last'(Names, Name) },
+	var_names(Names), value(Name, Skel, Subst, Options),
+	(   { T \== [] }
+	->  [ ','-[], nl ],
+	    bindings(T, Options)
+	;   []
+	).
+
+var_names([Name]) --> !,
+	[ '~w = '-[Name] ].
+var_names([Name1,Name2|T]) --> !,
+	[ '~w = ~w, '-[Name1, Name2] ],
+	var_names([Name2|T]).
+
+
+value(Name, Skel, Subst, Options) -->
+	(   { var(Skel), Subst = [Skel=S] }
+	->  { Skel = '$VAR'(Name) },
+	    [ '~W'-[S, Options] ]
+	;   [ '~W'-[Skel, Options] ],
+	    substitution(Subst, Options)
+	).
+
+substitution([], _) --> !.
+substitution([N=V|T], Options) -->
+	[ ', ', ansi(fg(green), '% where', []), nl,
+	  '    ~w = ~W'-[N,V,Options] ],
+	substitutions(T, Options).
+
+substitutions([], _) --> [].
+substitutions([N=V|T], Options) -->
+	[ ','-[], nl, '    ~w = ~W'-[N,V,Options] ],
+	substitutions(T, Options).
+
+
+residuals([], _) -->
+	[].
+residuals([G|Gs], Options) -->
 	(   { Gs \== [] }
-	->  [ '~q,'-[G], nl ],
-	    residuals(Gs)
-	;   [ '~q'-[G] ]
+	->  [ '~W,'-[G, Options], nl ],
+	    residuals(Gs, Options)
+	;   [ '~W'-[G, Options] ]
 	).
 
 bind_res_sep(_, []) --> !,
@@ -655,9 +749,9 @@ extra_line -->
 extra_line -->
 	[].
 
-prolog_message(if_tty(Text)) -->
+prolog_message(if_tty(Message)) -->
 	(   {current_prolog_flag(tty_control, true)}
-	->  [ at_same_line, '~w'-[Text] ]
+	->  [ at_same_line | Message ]
 	;   []
 	).
 prolog_message(halt(Reason)) -->
@@ -753,7 +847,7 @@ tracing_list([trace(Head, Ports)|T]) -->
 prolog_message(frame(Frame, backtrace, _PC)) --> !,
 	{ prolog_frame_attribute(Frame, level, Level)
 	},
-	[ '~t[~D] ~10|'-[Level] ],
+	[ ansi(bold, '~t[~D] ~10|', [Level]) ],
 	frame_context(Frame),
 	frame_goal(Frame).
 prolog_message(frame(Frame, choice, PC)) --> !,
@@ -815,16 +909,16 @@ frame_flags(Frame) -->
 	[ '~w~w '-[T, S] ].
 
 port(Port) -->
-	{ port_name(Port, Name)
+	{ port_name(Port, Colour, Name)
 	}, !,
-	[ '~w: '-[Name] ].
+	[ ansi([bold,fg(Colour)], '~w: ', [Name]) ].
 
-port_name(call,	     'Call').
-port_name(exit,	     'Exit').
-port_name(fail,	     'Fail').
-port_name(redo,	     'Redo').
-port_name(unify,     'Unify').
-port_name(exception, 'Exception').
+port_name(call,	     green,   'Call').
+port_name(exit,	     green,   'Exit').
+port_name(fail,	     red,     'Fail').
+port_name(redo,	     yellow,  'Redo').
+port_name(unify,     blue,    'Unify').
+port_name(exception, magenta, 'Exception').
 
 clean_goal(M:Goal, Goal) :-
 	hidden_module(M), !.
@@ -855,6 +949,7 @@ prolog_message(abnormal_thread_completion(Goal, fail)) -->
 prolog_message(threads_not_died(Count)) -->
 	[ '~D threads wouldn\'t die'-[Count] ].
 
+
 		 /*******************************
 		 *	PRINTING MESSAGES	*
 		 *******************************/
@@ -864,7 +959,7 @@ prolog_message(threads_not_died(Count)) -->
 :- dynamic
 	user:message_hook/3.
 
-%	print_message(+Kind, +Term)
+%%	print_message(+Kind, +Term)
 %
 %	Print an error message using a term as generated by the exception
 %	system.
@@ -881,7 +976,7 @@ print_message(Level, Term) :-
 	;   true
 	).
 
-%	print_system_message(+Term, +Level, +Lines)
+%%	print_system_message(+Term, +Level, +Lines)
 %
 %	Print the message if the user did not intecept the message.
 %	The first is used for errors and warnings that can be related
@@ -893,75 +988,116 @@ print_system_message(_, informational, _) :-
 	current_prolog_flag(verbose, silent), !.
 print_system_message(_, banner, _) :-
 	current_prolog_flag(verbose, silent), !.
+print_system_message(_, _, []) :- !.
 print_system_message(Term, Level, Lines) :-
 	flush_output(user_output),
 	source_location(File, Line),
 	Term \= error(syntax_error(_), _),
-	prefix(Level, Prefix, LinePrefix, PostFix, Wait, Stream), !,
-	format(Stream, Prefix, [File, Line]),
-	print_message_lines(Stream, LinePrefix, Lines),
-	format(Stream, PostFix, []),
+	prefix(Level, Prefix, LinePrefix, Wait, Stream), !,
+	insert_prefix(Lines, LinePrefix, PrefixLines),
+	'$append'([ begin(Level, Ctx),
+		    Prefix-[File,Line],
+		    nl
+		  | PrefixLines
+		  ],
+		  [ end(Ctx)
+		  ],
+		  AllLines),
+	print_message_lines(Stream, AllLines),
 	(   Wait > 0
 	->  sleep(Wait)
 	;   true
 	).
 print_system_message(_, Level, Lines) :-
-	flush_output(user_output),
 	prefix(Level, LinePrefix, Stream), !,
-	print_message_lines(Stream, LinePrefix, Lines).
+	insert_prefix(Lines, LinePrefix, PrefixLines),
+	'$append'([ begin(Level, Ctx)
+		  | PrefixLines
+		  ],
+		  [ end(Ctx)
+		  ],
+		  AllLines),
+	print_message_lines(Stream, AllLines).
+print_system_message(_, Level, _) :-
+	\+ prefix(Level, _, _),
+	throw(error(domain_error(message_kind, Level), _)).
 
-prefix(error,	      'ERROR: ~w:~d:~n',   '\t', '', 0.5, user_error).
-prefix(warning,	      'Warning: ~w:~d:~n', '\t', '', 0,   user_error).
+prefix(error,	      '~NERROR: ~w:~d:',   '~N\t', 0.1, user_error).
+prefix(warning,	      '~NWarning: ~w:~d:', '~N\t', 0,   user_error).
 
-prefix(help,	      '',          user_error).
-prefix(query,	      '',          user_error).
-prefix(debug,	      '',          user_output).
+prefix(help,	      '~N',	   user_error).
+prefix(query,	      '~N',        user_error).
+prefix(debug,	      '~N',        user_output).
 prefix(warning,	      Prefix,      user_error) :-
 	thread_self(Id),
 	(   Id == main
-	->  Prefix = 'Warning: '
-	;   atomic_list_concat(['Warning: [Thread ', Id, '] '], Prefix)
+	->  Prefix = '~NWarning: '
+	;   Prefix = '~NWarning: [Thread ~w] '-Id
 	).
-prefix(error,	      Prefix,   user_error) :-
+prefix(error,	      Prefix,      user_error) :-
 	thread_self(Id),
 	(   Id == main
-	->  Prefix = 'ERROR: '
-	;   atomic_list_concat(['ERROR: [Thread ', Id, '] '], Prefix)
+	->  Prefix = '~NERROR: '
+	;   Prefix = '~NERROR: [Thread ~w] '-Id
 	).
-prefix(banner,	      '',	   user_error).
-prefix(informational, '% ',        user_error).
+prefix(banner,	      '~N',	   user_error).
+prefix(informational, '~N% ',	   user_error).
+prefix(information,   '~N% ',	   user_error).
 
-%	print_message_lines(+Stream, +Prefix, +Lines)
+%%	print_message_lines(+Stream, +Prefix, +Lines)
 %
 %	Quintus compatibility predicate to print message lines using
 %	a prefix.
 
-print_message_lines(_, _, []) :- !.
-print_message_lines(S, P, [at_same_line|Lines]) :- !,
-	print_message_line(S, Lines, Rest),
-	print_message_lines(S, P, Rest).
-print_message_lines(S, P, Lines) :-
-	atom_concat('~N', P, Prefix),
-	format(S, Prefix, []),
-	print_message_line(S, Lines, Rest),
-	print_message_lines(S, P, Rest).
+print_message_lines(Stream, Prefix, Lines) :-
+	insert_prefix(Lines, Prefix, PrefixLines),
+	print_message_lines(Stream, PrefixLines).
 
-print_message_line(S, [flush], []) :- !,
-	flush_output(S).
-print_message_line(S, [], []) :- !,
-	nl(S).
-print_message_line(S, [nl|T], T) :- !,
-	nl(S).
-print_message_line(S, [full_stop|T], T) :- !,
+%%	insert_prefix(+Lines, +Prefix, -PrefixedLines)
+
+insert_prefix([at_same_line|Lines0], Prefix, Lines) :- !,
+	prefix_nl(Lines0, Prefix, Lines).
+insert_prefix(Lines0, Prefix, [prefix(Prefix)|Lines]) :-
+	prefix_nl(Lines0, Prefix, Lines).
+
+prefix_nl([], _, [nl]).
+prefix_nl([nl], _, [nl]) :- !.
+prefix_nl([flush], _, [flush]) :- !.
+prefix_nl([nl|T0], Prefix, [nl, prefix(Prefix)|T]) :- !,
+	prefix_nl(T0, Prefix, T).
+prefix_nl([H|T0], Prefix, [H|T]) :-
+	prefix_nl(T0, Prefix, T).
+
+%%	print_message_lines(+Stream, +Lines)
+
+print_message_lines(_, []) :- !.
+print_message_lines(S, [H|T]) :-
+	line_element(S, H),
+	print_message_lines(S, T).
+
+line_element(S, E) :-
+	prolog:message_line_element(S, E), !.
+line_element(S, full_stop) :- !,
 	'$put_token'(S, '.').		% insert space if needed.
-print_message_line(S, [Fmt-Args|T0], T) :- !,
-	format(S, Fmt, Args),
-	print_message_line(S, T0, T).
-print_message_line(S, [Fmt|T0], T) :-
-	format(S, Fmt, []),
-	print_message_line(S, T0, T).
+line_element(S, nl) :- !,
+	nl(S).
+line_element(S, prefix(Fmt-Args)) :- !,
+	format(S, Fmt, Args).
+line_element(S, prefix(Fmt)) :- !,
+	format(S, Fmt, []).
+line_element(S, flush) :- !,
+	flush_output(S).
+line_element(S, Fmt-Args) :- !,
+	format(S, Fmt, Args).
+line_element(S, ansi(_, Fmt, Args)) :- !,
+	format(S, Fmt, Args).
+line_element(_, begin(_Level, _Ctx)) :- !.
+line_element(_, end(_Ctx)) :- !.
+line_element(S, Fmt) :-
+	format(S, Fmt, []).
 
-%	message_to_string(+Term, -String)
+
+%%	message_to_string(+Term, -String)
 %
 %	Translate an error term into a string
 
@@ -977,10 +1113,13 @@ actions_to_format([Term, nl], Fmt, Args) :- !,
 actions_to_format([nl|T], Fmt, Args) :- !,
 	actions_to_format(T, Fmt0, Args),
 	atom_concat('~n', Fmt0, Fmt).
+actions_to_format([Skip|T], Fmt, Args) :-
+	action_skip(Skip), !,
+	actions_to_format(T, Fmt, Args).
 actions_to_format([Fmt0-Args0|Tail], Fmt, Args) :- !,
         actions_to_format(Tail, Fmt1, Args1),
         atom_concat(Fmt0, Fmt1, Fmt),
-        '$append'(Args0, Args1, Args).
+	append_args(Args0, Args1, Args).
 actions_to_format([Term|Tail], Fmt, Args) :-
 	atomic(Term), !,
         actions_to_format(Tail, Fmt1, Args),
@@ -988,7 +1127,19 @@ actions_to_format([Term|Tail], Fmt, Args) :-
 actions_to_format([Term|Tail], Fmt, Args) :-
         actions_to_format(Tail, Fmt1, Args1),
         atom_concat('~w', Fmt1, Fmt),
-        '$append'([Term], Args1, Args).
+	append_args([Term], Args1, Args).
+
+action_skip(at_same_line).
+action_skip(ansi(_Attrs, _Fmt, _Args)).
+action_skip(begin(_Level, _Ctx)).
+action_skip(end(_Ctx)).
+
+append_args(M:Args0, Args1, M:Args) :- !,
+	strip_module(Args1, _, A1),
+	'$append'(Args0, A1, Args).
+append_args(Args0, Args1, Args) :-
+	strip_module(Args1, _, A1),
+	'$append'(Args0, A1, Args).
 
 
 		 /*******************************
@@ -998,13 +1149,13 @@ actions_to_format([Term|Tail], Fmt, Args) :-
 :- dynamic
 	printed/2.
 
-%	print_once(Message, Level)
+%%	print_once(Message, Level)
 %
 %	True for messages that must be printed only once.
 
 print_once(compatibility(_), _).
 
-%	must_print(+Level, +Message)
+%%	must_print(+Level, +Message)
 %
 %	True if the message must be printed.
 
