@@ -3,9 +3,10 @@
     Part of SWI-Prolog
 
     Author:        Jan Wielemaker
-    E-mail:        J.Wielemaker@uva.nl
+    E-mail:        J.Wielemaker@cs.vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (C): 1985-2009, University of Amsterdam
+    Copyright (C): 1985-2011, University of Amsterdam
+			      VU University Amsterdam
 
     This program is free software; you can redistribute it and/or
     modify it under the terms of the GNU General Public License
@@ -17,9 +18,9 @@
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
 
-    You should have received a copy of the GNU Lesser General Public
+    You should have received a copy of the GNU General Public
     License along with this library; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 
     As a special exception, if you link this library with other files,
     compiled with a Free Software compiler, to produce an executable, this
@@ -40,10 +41,11 @@
 	    nospyall/0,
 	    debugging/0,
 	    rational/3,
-	    atomic_list_concat/2,
 	    atom_prefix/2,
 	    dwim_match/2,
+	    source_file_property/2,
 	    source_file/1,
+	    unload_file/1,
 	    prolog_load_context/2,
 	    stream_position_data/3,
 	    current_predicate/2,
@@ -53,11 +55,9 @@
 	    clause_property/2,
 	    recorda/2,
 	    recordz/2,
-	    recorded/2,
 	    current_module/1,
 	    module_property/2,
 	    module/1,
-	    statistics/0,
 	    shell/1,
 	    shell/0,
 	    on_signal/3,
@@ -66,67 +66,72 @@
 	    open_shared_object/3,
 	    format/1,
 	    garbage_collect/0,
-	    arithmetic_function/1,
+	    set_prolog_stack/2,
+	    prolog_stack_property/2,
 	    absolute_file_name/2,
 	    require/1,
 	    call_with_depth_limit/3,
-	    length/2,			% ?List, ?Length
 	    numbervars/3,		% +Term, +Start, -End
-	    nb_setval/2,		% +Var, +Value
-	    create_prolog_flag/3
+	    nb_setval/2			% +Var, +Value
 	  ]).
 
 		/********************************
 		*           DEBUGGER            *
 		*********************************/
 
-:- meta_predicate
-	'$map_bits'(2, +, +, -).
+%%	map_bits(:Pred, +Modify, +OldBits, -NewBits)
 
-'$map_bits'(_, [], Bits, Bits) :- !.
-'$map_bits'(Pred, [H|T], Old, New) :-
-	'$map_bits'(Pred, H, Old, New0),
-	'$map_bits'(Pred, T, New0, New).
-'$map_bits'(Pred, +Name, Old, New) :- !,	% set a bit
-	call(Pred, Name, Bits), !,
+:- meta_predicate
+	map_bits(2, +, +, -).
+
+map_bits(_, [], Bits, Bits) :- !.
+map_bits(Pred, [H|T], Old, New) :-
+	map_bits(Pred, H, Old, New0),
+	map_bits(Pred, T, New0, New).
+map_bits(Pred, +Name, Old, New) :- !,	% set a bit
+	bit(Pred, Name, Bits), !,
 	New is Old \/ Bits.
-'$map_bits'(Pred, -Name, Old, New) :- !, 	% clear a bit
-	call(Pred, Name, Bits), !,
+map_bits(Pred, -Name, Old, New) :- !,	% clear a bit
+	bit(Pred, Name, Bits), !,
 	New is Old /\ (\Bits).
-'$map_bits'(Pred, ?(Name), Old, Old) :-		% ask a bit
-	call(Pred, Name, Bits),
+map_bits(Pred, ?(Name), Old, Old) :-		% ask a bit
+	bit(Pred, Name, Bits),
 	Old /\ Bits > 0.
 
-'$port_bit'(      call, 2'000000001).
-'$port_bit'(      exit, 2'000000010).
-'$port_bit'(      fail, 2'000000100).
-'$port_bit'(      redo, 2'000001000).
-'$port_bit'(     unify, 2'000010000).
-'$port_bit'(     break, 2'000100000).
-'$port_bit'(  cut_call, 2'001000000).
-'$port_bit'(  cut_exit, 2'010000000).
-'$port_bit'( exception, 2'100000000).
-'$port_bit'(       cut, 2'011000000).
-'$port_bit'(       all, 2'000111111).
-'$port_bit'(      full, 2'000101111).
-'$port_bit'(      half, 2'000101101).	% '
+bit(Pred, Name, Bits) :-
+	call(Pred, Name, Bits), !.
+bit(_:Pred, Name, _) :-
+	throw(error(domain_error(Pred, Name), _)).
+
+port_name(      call, 2'000000001).
+port_name(      exit, 2'000000010).
+port_name(      fail, 2'000000100).
+port_name(      redo, 2'000001000).
+port_name(     unify, 2'000010000).
+port_name(     break, 2'000100000).
+port_name(  cut_call, 2'001000000).
+port_name(  cut_exit, 2'010000000).
+port_name( exception, 2'100000000).
+port_name(       cut, 2'011000000).
+port_name(       all, 2'000111111).
+port_name(      full, 2'000101111).
+port_name(      half, 2'000101101).	% '
 
 leash(Ports) :-
 	'$leash'(Old, Old),
-	'$map_bits'('$port_bit', Ports, Old, New),
+	map_bits(port_name, Ports, Old, New),
 	'$leash'(_, New).
 
 visible(Ports) :-
 	'$visible'(Old, Old),
-	'$map_bits'('$port_bit', Ports, Old, New),
+	map_bits(port_name, Ports, Old, New),
 	'$visible'(_, New).
 
-'$map_style_check'(atom,     	    2'0000001).
-'$map_style_check'(singleton, 	    2'0000010).
-'$map_style_check'(dollar,   	    2'0000100).
-'$map_style_check'((discontiguous),   2'0001000).
-'$map_style_check'(dynamic,	    2'0010000).
-'$map_style_check'(charset,	    2'0100000).
+style_name(atom,	    2'0000001).
+style_name(singleton,	    2'0000010).
+style_name(discontiguous,   2'0001000).
+style_name(dynamic,	    2'0010000).
+style_name(charset,	    2'0100000).
 
 style_check(+string) :- !,
 	set_prolog_flag(double_quotes, string).
@@ -136,7 +141,7 @@ style_check(?(string)) :- !,
 	current_prolog_flag(double_quotes, string).
 style_check(Spec) :-
 	'$style_check'(Old, Old),
-	'$map_bits'('$map_style_check', Spec, Old, New),
+	map_bits(style_name, Spec, Old, New),
 	'$style_check'(_, New).
 
 %	prolog:debug_control_hook(+Action)
@@ -359,15 +364,74 @@ atom_prefix(Atom, Prefix) :-
 %	bound to the canonical name for it. If File is bound it succeeds
 %	if the canonical name  as   defined  by  absolute_file_name/2 is
 %	known as a loaded filename.
+%
+%	Note that Time = 0.0 is used by  PlDoc and other code that needs
+%	to create a file record without being interested in the time.
 
 source_file(File) :-
 	(   ground(File)
-	->  ( 	'$time_source_file'(File, _, user)
+	->  (	'$time_source_file'(File, Time, user)
 	    ;	absolute_file_name(File, Abs),
-		'$time_source_file'(Abs, _, user)
+		'$time_source_file'(Abs, Time, user)
 	    ), !
-	;   '$time_source_file'(File, _, user)
+	;   '$time_source_file'(File, Time, user)
+	),
+	Time > 0.0.
+
+%%	source_file_property(?File, ?Property) is nondet.
+%
+%	True if Property is a property of the loaded source-file File.
+
+source_file_property(File, P) :-
+	nonvar(File), !,
+	canonical_source_file(File, Path),
+	property_source_file(P, Path).
+source_file_property(File, P) :-
+	property_source_file(P, File).
+
+property_source_file(modified(Time), File) :-
+	'$time_source_file'(File, Time, user).
+property_source_file(module(M), File) :-
+	(   nonvar(M)
+	->  '$current_module'(M, File)
+	;   nonvar(File)
+	->  '$current_module'(ML, File),
+	    (	atom(ML)
+	    ->	M = ML
+	    ;	'$member'(M, ML)
+	    )
+	;   '$current_module'(M, File)
 	).
+property_source_file(load_context(Module, Location), File) :-
+	'$time_source_file'(File, _, user),
+	clause(system:'$load_context_module'(File, Module), true, Ref),
+	(   clause_property(Ref, file(FromFile)),
+	    clause_property(Ref, line_count(FromLine))
+	->  Location = FromFile:FromLine
+	;   Location = user
+	).
+property_source_file(includes(File2, Stamp), File) :-
+	system:'$included'(File, File2, Stamp).
+property_source_file(derived_from(DerivedFrom, Stamp), File) :-
+	system:'$derived_source'(File, DerivedFrom, Stamp).
+
+
+%%	canonical_source_file(+Spec, -File) is semidet.
+%
+%	File is the canonical representation of the source-file Spec.
+
+canonical_source_file(Spec, File) :-
+	source_file(Spec), !,
+	File = Spec.
+canonical_source_file(Spec, File) :-
+	absolute_file_name(Spec,
+			       [ file_type(prolog),
+				 access(read),
+				 file_errors(fail)
+			       ],
+			       File),
+	source_file(File).
+
 
 %%	prolog_load_context(+Key, -Value)
 %
@@ -393,6 +457,25 @@ prolog_load_context(dialect, D) :-
 	current_prolog_flag(emulated_dialect, D).
 prolog_load_context(term_position, '$stream_position'(0,L,0,0,0)) :-
 	source_location(_, L).
+prolog_load_context(script, Bool) :-
+	(   '$toplevel':loaded_init_file(script, Path),
+	    source_location(Path, _)
+	->  Bool = true
+	;   Bool = false
+	).
+
+
+%%	unload_file(+File) is det.
+%
+%	Remove all traces of loading file.
+
+unload_file(File) :-
+	(   canonical_source_file(File, Path)
+	->  '$unload_file'(Path),
+	    retractall(system:'$load_context_module'(Path, _))
+	;   true
+	).
+
 
 		 /*******************************
 		 *	      STREAMS		*
@@ -475,6 +558,7 @@ current_predicate(Name, Module:Head) :-
 	'$defined_predicate'(DefModule:Head), !.
 current_predicate(Name, Module:Head) :-
 	current_prolog_flag(autoload, true),
+	\+ current_prolog_flag(Module:unknown, fail),
 	functor(Head, Name, Arity),
 	'$find_library'(Module, Name, Arity, _LoadModule, _Library), !.
 
@@ -487,13 +571,29 @@ generate_current_predicate(Name, Module, Head) :-
 '$defined_predicate'(Head) :-
 	'$get_predicate_attribute'(Head, defined, 1), !.
 
+%%	predicate_property(?Predicate, ?Property) is nondet.
+%
+%	True when Property is a property of Predicate.
+
 :- meta_predicate
 	predicate_property(:, ?).
 
 :- '$iso'(predicate_property/2).
 
-predicate_property(Pred, Property) :-
-	Property == undefined, !,
+predicate_property(Pred, Property) :-		% Mode ?,+
+	nonvar(Property), !,
+	property_predicate(Property, Pred).
+predicate_property(Pred, Property) :-		% Mode +,-
+	define_or_generate(Pred),
+	'$predicate_property'(Property, Pred).
+
+%%	property_predicate(+Property, ?Pred)
+%
+%	First handle the special  cases  that   are  not  about querying
+%	normally  defined  predicates:   =undefined=,    =visible=   and
+%	=autoload=, followed by the generic case.
+
+property_predicate(undefined, Pred) :- !,
 	Pred = Module:Head,
 	current_module(Module),
 	'$c_current_predicate'(_, Pred),
@@ -501,23 +601,50 @@ predicate_property(Pred, Property) :-
 	\+ current_predicate(_, Pred),
 	functor(Head, Name, Arity),
 	\+ system_undefined(Module:Name/Arity).
-predicate_property(Pred, Property) :-
-	Pred = M:_,
-	M == system, !,				% do not autoload into system
-	'$c_current_predicate'(_, Pred),
-	'$defined_predicate'(Pred),
+property_predicate(visible, Pred) :- !,
+	visible_predicate(Pred).
+property_predicate(autoload(File), _:Head) :- !,
+	current_prolog_flag(autoload, true),
+	(   callable(Head)
+	->  functor(Head, Name, Arity),
+	    (	'$find_library'(_, Name, Arity, _, File)
+	    ->	true
+	    )
+	;   '$find_library'(_, Name, Arity, _, File),
+	    functor(Head, Name, Arity)
+	).
+property_predicate(Property, Pred) :-
+	define_or_generate(Pred),
 	'$predicate_property'(Property, Pred).
-predicate_property(Pred, Property) :-
+
+%%	define_or_generate(+Head) is semidet.
+%%	define_or_generate(-Head) is nondet.
+%
+%	If the predicate is known, try to resolve it. Otherwise generate
+%	the known predicate, but do not try to (auto)load the predicate.
+
+define_or_generate(M:Head) :-
+	callable(Head),
+	'$get_predicate_attribute'(M:Head, defined, 1), !.
+define_or_generate(M:Head) :-
+	callable(Head),
+	nonvar(M), M \== system, !,
+	'$define_predicate'(M:Head).
+define_or_generate(Pred) :-
 	current_predicate(_, Pred),
-	'$define_predicate'(Pred),		% autoload if needed
-	'$predicate_property'(Property, Pred).
+	'$define_predicate'(Pred).
+
 
 '$predicate_property'(interpreted, Pred) :-
 	'$get_predicate_attribute'(Pred, foreign, 0).
+'$predicate_property'(visible, Pred) :-
+	'$get_predicate_attribute'(Pred, defined, 1).
 '$predicate_property'(built_in, Pred) :-
 	'$get_predicate_attribute'(Pred, system, 1).
 '$predicate_property'(exported, Pred) :-
 	'$get_predicate_attribute'(Pred, exported, 1).
+'$predicate_property'(public, Pred) :-
+	'$get_predicate_attribute'(Pred, public, 1).
 '$predicate_property'(foreign, Pred) :-
 	'$get_predicate_attribute'(Pred, foreign, 1).
 '$predicate_property'((dynamic), Pred) :-
@@ -532,8 +659,6 @@ predicate_property(Pred, Property) :-
 	'$get_predicate_attribute'(Pred, imported, Module).
 '$predicate_property'(transparent, Pred) :-
 	'$get_predicate_attribute'(Pred, transparent, 1).
-'$predicate_property'(indexed(Pattern), Pred) :-
-	'$get_predicate_attribute'(Pred, indexed, Pattern).
 '$predicate_property'(meta_predicate(Pattern), Pred) :-
 	'$get_predicate_attribute'(Pred, meta_predicate, Pattern).
 '$predicate_property'(file(File), Pred) :-
@@ -546,14 +671,13 @@ predicate_property(Pred, Property) :-
 	'$get_predicate_attribute'(Pred, hide_childs, 1).
 '$predicate_property'(spying, Pred) :-
 	'$get_predicate_attribute'(Pred, spy, 1).
-'$predicate_property'(hashed(N), Pred) :-
-	'$get_predicate_attribute'(Pred, hashed, N),
-	N > 0.
 '$predicate_property'(references(N), Pred) :-
 	'$get_predicate_attribute'(Pred, references, N),
 	N \== 0.			% show negative for debugging!
 '$predicate_property'(number_of_clauses(N), Pred) :-
 	'$get_predicate_attribute'(Pred, number_of_clauses, N).
+'$predicate_property'(indexed(Indices), Pred) :-
+	'$get_predicate_attribute'(Pred, indexed, Indices).
 '$predicate_property'(noprofile, Pred) :-
 	'$get_predicate_attribute'(Pred, noprofile, 1).
 '$predicate_property'(iso, Pred) :-
@@ -564,6 +688,38 @@ system_undefined(user:prolog_exception_hook/4).
 system_undefined(system:'$c_call_prolog'/0).
 system_undefined(system:window_title/2).
 
+%%	visible_predicate(:Head) is nondet.
+%
+%	True when Head can be called without raising an existence error.
+%	This implies it is defined,  can   be  inherited  from a default
+%	module or can be autoloaded.
+
+visible_predicate(Pred) :-
+	Pred = M:Head,
+	current_module(M),
+	(   callable(Head)
+	->  (   '$get_predicate_attribute'(Pred, defined, 1)
+	    ->	\+ hidden_system_predicate(Pred)
+	    ;	\+ current_prolog_flag(M:unknown, fail),
+		functor(Head, Name, Arity),
+		'$find_library'(M, Name, Arity, _LoadModule, _Library)
+	    )
+	;   (   default_module(M, DefM),
+	        '$c_current_predicate'(_, DefM:Head),
+		\+ '$get_predicate_attribute'(DefM:Head, imported, _),
+		\+ hidden_system_predicate(Pred)
+	    ;	'$in_library'(Name, Arity, _),
+		functor(Head, Name, Arity),
+		\+ '$get_predicate_attribute'(Pred, defined, 1)
+	    )
+	).
+
+hidden_system_predicate(_:Head) :-
+	functor(Head, Name, _),
+	sub_atom(Name, 0, _, _, $),
+	\+ current_prolog_flag(access_level, system).
+
+
 %%	clause_property(+ClauseRef, ?Property) is nondet.
 %
 %	Provide information on individual clauses.  Defined properties
@@ -591,14 +747,6 @@ clause_property(Clause, Property) :-
 	'$get_clause_attribute'(Clause, erased, true).
 
 
-recorda(Key, Value) :-
-	recorda(Key, Value, _).
-recordz(Key, Value) :-
-	recordz(Key, Value, _).
-recorded(Key, Value) :-
-	recorded(Key, Value, _).
-
-
 		 /*******************************
 		 *	       REQUIRE		*
 		 *******************************/
@@ -650,19 +798,35 @@ current_module(Module) :-
 %	    The module declaration is on line Count of File.
 %	    * exports(ListOfPredicateIndicators)
 %	    The module exports ListOfPredicateIndicators
+%	    * exported_operators(ListOfOp3)
+%	    The module exports the operators ListOfOp3.
 
 module_property(Module, Property) :-
 	nonvar(Module), nonvar(Property), !,
-	'$module_property'(Module, Property).
-module_property(Module, Property) :-
+	property_module(Property, Module).
+module_property(Module, Property) :-	% -, file(File)
 	nonvar(Property), Property = file(File), !,
-	'$current_module'(Module, File),
-	File \== [].
+	(   nonvar(File)
+	->  '$current_module'(Modules, File),
+	    (	atom(Modules)
+	    ->	Module = Modules
+	    ;	'$member'(Module, Modules)
+	    )
+	;   '$current_module'(Module, File),
+	    File \== []
+	).
 module_property(Module, Property) :-
 	current_module(Module),
+	property_module(Property, Module).
+
+property_module(exported_operators(List), Module) :- !,
+	'$exported_ops'(Module, List, []),
+	List \== [].
+property_module(Property, Module) :-
 	module_property(Property),
 	'$module_property'(Module, Property).
 
+module_property(class(_)).
 module_property(file(_)).
 module_property(line_count(_)).
 module_property(exports(_)).
@@ -679,94 +843,6 @@ module(Module) :-
 	'$module'(_, Module),
 	print_message(warning, no_current_module(Module)).
 
-		/********************************
-		*          STATISTICS           *
-		*********************************/
-
-statistics :-
-	statistics(trail, Trail),
-	statistics(trailused, TrailUsed),
-	statistics(local, Local),
-	statistics(localused, LocalUsed),
-	statistics(global, Global),
-	statistics(globalused, GlobalUsed),
-	statistics(cputime, Cputime),
-	statistics(inferences, Inferences),
-	statistics(heapused, Heapused),
-	statistics(atoms, Atoms),
-	statistics(functors, Functors),
-	statistics(predicates, Predicates),
-	statistics(modules, Modules),
-	statistics(codes, Codes),
-	statistics(locallimit, LocalLimit),
-	statistics(globallimit, GlobalLimit),
-	statistics(traillimit, TrailLimit),
-
-	format('~2f seconds cpu time for ~D inferences~n',
-				    [Cputime, Inferences]),
-	format('~D atoms, ~D functors, ~D predicates, ~D modules, ~D VM-codes~n~n',
-				    [Atoms, Functors, Predicates, Modules, Codes]),
-	format('                       Limit    Allocated       In use~n'),
-	(   statistics(heap, Heap),
-	    statistics(heaplimit, HeapLimit)
-	->  format('Heap         :~t~D~28| ~t~D~41| ~t~D~54| Bytes~n',
-		   [HeapLimit, Heap, Heapused])
-	;   format('Heap         :                  ~t~D~54| Bytes~n',
-		   [Heapused])
-	),
-	format('Local  stack :~t~D~28| ~t~D~41| ~t~D~54| Bytes~n',
-	       [LocalLimit, Local, LocalUsed]),
-	format('Global stack :~t~D~28| ~t~D~41| ~t~D~54| Bytes~n',
-	       [GlobalLimit, Global, GlobalUsed]),
-	format('Trail  stack :~t~D~28| ~t~D~41| ~t~D~54| Bytes~n~n',
-	       [TrailLimit, Trail, TrailUsed]),
-
-	gc_statistics,
-	agc_statistics,
-	shift_statistics,
-	thread_statistics.
-
-gc_statistics :-
-	statistics(collections, Collections),
-	Collections > 0, !,
-	statistics(collected, Collected),
-	statistics(gctime, GcTime),
-
-	format('~D garbage collections gained ~D bytes in ~2f seconds.~n',
-	       [Collections, Collected, GcTime]).
-gc_statistics.
-
-agc_statistics :-
-	catch(statistics(agc, Agc), _, fail),
-	Agc > 0, !,
-	statistics(agc_gained, Gained),
-	statistics(agc_time, Time),
-	format('~D atom garbage collections gained ~D atoms in ~2f seconds.~n',
-	       [Agc, Gained, Time]).
-agc_statistics.
-
-shift_statistics :-
-	statistics(local_shifts, LS),
-	statistics(global_shifts, GS),
-	statistics(trail_shifts, TS),
-	(   LS > 0
-	;   GS > 0
-	;   TS > 0
-	), !,
-	format('Stack shifts: ~D local, ~D global, ~D trail.~n',
-	       [LS, GS, TS]).
-shift_statistics.
-
-thread_statistics :-
-	current_prolog_flag(threads, true), !,
-	statistics(threads, Active),
-	statistics(threads_created, Created),
-	statistics(thread_cputime, CpuTime),
-	Finished is Created - Active,
-	format('~D threads, ~D finished threads used ~2f seconds.~n',
-	       [Active, Finished, CpuTime]).
-thread_statistics.
-
 
 		/********************************
 		*      SYSTEM INTERACTION       *
@@ -789,7 +865,10 @@ shell :-
 		 *******************************/
 
 :- meta_predicate
-	on_signal(+, :, :).
+	on_signal(+, :, :),
+	current_signal(?, ?, :).
+
+%%	on_signal(+Signal, -OldHandler, :NewHandler) is det.
 
 on_signal(Signal, Old, New) :-
 	atom(Signal), !,
@@ -800,14 +879,24 @@ on_signal(Signal, Old, New) :-
 on_signal(Signal, _Old, _New) :-
 	(   var(Signal)
 	->  Err = instantiation_error
-	;   Err = type_error(signal, Signal)
+	;   Err = type_error(signal_name, Signal)
 	),
 	throw(error(Err, context(on_signal/3, _))).
 
+%%	current_signal(?Name, ?SignalNumber, :Handler) is nondet.
+
 current_signal(Name, Id, Handler) :-
 	between(1, 32, Id),
 	'$on_signal'(Id, Name, Handler, Handler).
 
+:- multifile
+	prolog:called_by/2.
+
+prolog:called_by(on_signal(_,_,New), [New+1]) :-
+	(   new == throw
+	;   new == default
+	), !, fail.
+
 
 		 /*******************************
 		 *	      DLOPEN		*
@@ -866,67 +955,53 @@ absolute_file_name(Name, Abs) :-
 	atomic(Name), !,
 	'$absolute_file_name'(Name, Abs).
 absolute_file_name(Term, Abs) :-
-	'$chk_file'(Term, [''], [access(read)], File), !,
+	'$chk_file'(Term, [''], [access(read)], true, File), !,
 	'$absolute_file_name'(File, Abs).
 absolute_file_name(Term, Abs) :-
-	'$chk_file'(Term, [''], [], File), !,
+	'$chk_file'(Term, [''], [], true, File), !,
 	'$absolute_file_name'(File, Abs).
 
 
 		/********************************
-		*         MISCELLENEOUS         *
+		*	 MEMORY MANAGEMENT      *
 		*********************************/
 
-%	Invoke the garbage collector.  The argument is the debugging level
-%	to use during garbage collection.  This only works if the system
-%	is compiled with the -DODEBUG cpp flag.  Only to simplify maintenance.
+%%	garbage_collect is det.
+%
+%	Invoke the garbage collector.  The   argument  of the underlying
+%	'$garbage_collect'/1  is  the  debugging  level  to  use  during
+%	garbage collection. This only works if   the  system is compiled
+%	with the -DODEBUG cpp flag. Only to simplify maintenance.
 
 garbage_collect :-
 	'$garbage_collect'(0).
 
-%%	arithmetic_function(:Spec)
+%%	set_prolog_stack(+Name, +Option) is det.
 %
-%	Register a predicate as an arithmetic function.  Takes Name/Arity
-%	and a term as argument.
+%	Set a parameter for one of the Prolog stacks.
 
-:- meta_predicate
-	arithmetic_function(1).
+set_prolog_stack(Stack, Option) :-
+	Option =.. [Name,Value0],
+	Value is Value0,
+	'$set_prolog_stack'(Stack, Name, _Old, Value).
 
-arithmetic_function(Module:Term) :-
-	(   Term = Name/Arity
-	;   functor(Term, Name, Arity)
-	), !,
-	PredArity is Arity + 1,
-	functor(Head, Name, PredArity),
-	'$arithmetic_function'(Module:Head, 0).
-
-		 /*******************************
-		 *	 LIST MANIPULATION	*
-		 *******************************/
-
-%%	length(?List, ?N)
+%%	prolog_stack_property(?Stack, ?Property) is nondet.
 %
-%	Is true when N is the length of List.
+%	Examine stack properties.
 
-:- '$iso'((length/2)).
+prolog_stack_property(Stack, Property) :-
+	stack_property(P),
+	stack_name(Stack),
+	Property =.. [P,Value],
+	'$set_prolog_stack'(Stack, P, Value, Value).
 
-length(List, Length) :-
-	(   nonvar(Length)
-	->  '$length'(List, Length)
-	;   '$skip_list'(Length0, List, Tail),
-	    (	Tail == []
-	    ->	Length = Length0
-	    ;	var(Tail)
-	    ->  length3(Tail, Length, Length0)
-	    ;	throw(error(type_error(list,Tail),
-			    context(length/2, _)))
-	    )
-	).
+stack_name(local).
+stack_name(global).
+stack_name(trail).
 
-length3([], N, N).
-length3([_|List], N, N0) :-
-        succ(N0, N1),
-        length3(List, N, N1).
+stack_property(limit).
+stack_property(spare).
+stack_property(min_free).
 
 
 		 /*******************************
@@ -935,7 +1010,7 @@ length3([_|List], N, N0) :-
 
 :- '$iso'((numbervars/3)).
 
-%	numbervars(+Term, +StartIndex, -EndIndex)
+%%	numbervars(+Term, +StartIndex, -EndIndex) is det.
 %
 %	Number all unbound variables in Term   using  '$VAR'(N), where the
 %	first N is StartIndex and EndIndex is  unified to the index that
@@ -949,17 +1024,10 @@ numbervars(Term, From, To) :-
 		 *	       GVAR		*
 		 *******************************/
 
-%	nb_setval(+Name, +Value)
+%%	nb_setval(+Name, +Value) is det.
 %
 %	Bind the non-backtrackable variable Name with a copy of Value
 
 nb_setval(Name, Value) :-
 	duplicate_term(Value, Copy),
 	nb_linkval(Name, Copy).
-
-%%	create_prolog_flag(Name, Value, Options)
-%
-%	`Forward' compatibility
-
-create_prolog_flag(Name, Value, _Options) :-
-	set_prolog_flag(Name, Value).
