@@ -17,9 +17,9 @@
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
 
-    You should have received a copy of the GNU Lesser General Public
+    You should have received a copy of the GNU General Public
     License along with this library; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 
     As a special exception, if you link this library with other files,
     compiled with a Free Software compiler, to produce an executable, this
@@ -32,31 +32,54 @@
 :- module('$toplevel',
 	  [ '$initialise'/0,		% start Prolog (does not return)
 	    '$toplevel'/0,		% Prolog top-level (re-entrant)
-	    '$abort'/0,			% restart after an abort
-	    '$break'/0,			% live in a break
 	    '$compile'/0,		% `-c' toplevel
 	    '$welcome'/0,		% banner
-	    prolog/0, 			% user toplevel predicate
-	    '$set_prompt'/1,		% set the main prompt
+	    prolog/0,			% user toplevel predicate
+	    '$query_loop'/0,		% toplevel predicate
 	    (initialization)/1,		% initialization goal (directive)
 	    '$thread_init'/0,		% initialise thread
 	    (thread_initialization)/1	% thread initialization goal
 	    ]).
 
 
+		 /*******************************
+		 *	 FILE_SEARCH_PATH	*
+		 *******************************/
+
+:- multifile user:file_search_path/2.
+
+user:file_search_path(user_profile, '.').
+user:file_search_path(user_profile, app_preferences('.')).
+:- if(current_prolog_flag(windows, true)).
+user:file_search_path(app_preferences, PrologAppData) :-
+	current_prolog_flag(windows, true),
+	catch(win_folder(appdata, AppData), _, fail),
+	atom_concat(AppData, '/SWI-Prolog', PrologAppData),
+	(   exists_directory(PrologAppData)
+	->  true
+	;   catch(make_directory(PrologAppData), _, fail)
+	).
+:- endif.
+user:file_search_path(app_preferences, UserHome) :-
+	catch(expand_file_name(~, [UserHome]), _, fail).
+
+
 		/********************************
 		*         INITIALISATION        *
 		*********************************/
 
+%	note: loaded_init_file/2 is used by prolog_load_context/2 to
+%	confirm we are loading a script.
+
 :- dynamic
-	loaded_init_file/1.		% already loaded init files
+	loaded_init_file/2.		% already loaded init files
 
 '$welcome' :-
 	print_message(banner, welcome).
 
 '$load_init_file'(none) :- !.
 '$load_init_file'(Base) :-
-	loaded_init_file(Base), !.
+	loaded_init_file(Base, _), !.
 '$load_init_file'(InitFile) :-
 	is_absolute_file_name(InitFile), !,
 	ensure_loaded(user:InitFile).
@@ -65,44 +88,45 @@
 			   [ access(read),
 			     file_errors(fail)
 			   ], InitFile),
-	asserta(loaded_init_file(Base)),
+	asserta(loaded_init_file(Base, InitFile)),
 	ensure_loaded(user:InitFile).
 '$load_init_file'(_).
 
 '$load_system_init_file' :-
-	loaded_init_file(system), !.
+	loaded_init_file(system, _), !.
 '$load_system_init_file' :-
-	'$option'(system_init_file, Base, Base),
-	(   Base == none
-	->  asserta(loaded_init_file(system))
-	;   current_prolog_flag(home, Home),
-	    file_name_extension(Base, rc, Name),
-	    atomic_list_concat([Home, '/', Name], File),
-	    access_file(File, read),
-	    asserta(loaded_init_file(system)),
-	    load_files(user:File, [silent(true)]), !
-	).
+	'$option'(system_init_file, Base),
+	Base \== none,
+	current_prolog_flag(home, Home),
+	file_name_extension(Base, rc, Name),
+	atomic_list_concat([Home, '/', Name], File),
+	absolute_file_name(File, Path,
+			   [ file_type(prolog),
+			     access(read),
+			     file_errors(fail)
+			   ]),
+	asserta(loaded_init_file(system, Path)),
+	load_files(user:Path, [silent(true)]), !.
 '$load_system_init_file'.
 
 '$load_script_file' :-
-	loaded_init_file(script), !.
+	loaded_init_file(script, _), !.
 '$load_script_file' :-
-	'$option'(script_file, OsFile, OsFile),
-	OsFile \== '',
-	prolog_to_os_filename(File, OsFile),
-	(   exists_file(File)		% avoid expanding on extensions
-	->  asserta(loaded_init_file(script)),
-	    load_files(user:File, [expand(false)])
-	;   throw(error(existence_error(script_file, File), _))
-	).
-'$load_script_file'.
+	'$option'(script_file, OsFiles),
+	load_script_files(OsFiles).
 
-'$load_gnu_emacs_interface' :-
-	(   getenv('EMACS', t),
-	    current_prolog_flag(argv, Args),
-	    memberchk('+C', Args)
-	->  ensure_loaded(user:library(emacs_interface))
-	;   true
+load_script_files([]).
+load_script_files([OsFile|More]) :-
+	prolog_to_os_filename(File, OsFile),
+	(   absolute_file_name(File, Path,
+			       [ file_type(prolog),
+				 access(read),
+				 file_errors(fail)
+			       ])
+	->  asserta(loaded_init_file(script, Path)),
+	    load_files(user:Path, []),
+	    load_files(More)
+	;   throw(error(existence_error(script_file, File), _))
 	).
 
 
@@ -253,13 +277,15 @@ set_associated_file :-
 	current_prolog_flag(argv, Argv),
 	'$append'(Pre, [OsFile], Argv),
 	\+ memberchk(--, Pre),
-	\+ '$append'(_, ['-f'], Pre),	% Avoid loading twice
+	\+ ( '$append'(_, [LOpt], Pre),
+	     load_option(LOpt)		% Avoid loading twice
+	   ),
 	prolog_to_os_filename(File, OsFile),
 	file_name_extension(_, Ext, File),
 	access_file(File, read), !,
 	file_directory_name(File, Dir),
 	working_directory(_, Dir),
-	set_prolog_flag(associated_file, File),
+	create_prolog_flag(associated_file, File, []),
 	(   current_predicate(system:window_title/2)
 	->  atom_concat('SWI-Prolog -- ', File, Title),
 	    system:window_title(_, Title)
@@ -267,6 +293,10 @@ set_associated_file :-
 	).
 set_associated_file.
 
+load_option('-s').
+load_option('-l').
+load_option('-f').
+
 
 %%	start_pldoc
 %
@@ -316,7 +346,7 @@ hkey('HKEY_LOCAL_MACHINE/Software/SWI/Prolog').
 	->  true
 	;   Ext = Ext0
 	),
-	set_prolog_flag(associate, Ext).
+	create_prolog_flag(associate, Ext, []).
 '$set_prolog_file_extension'.
 
 
@@ -339,20 +369,20 @@ initialise_prolog :-
 	set_associated_file,
 	'$set_file_search_paths',
 	once(print_predicate(_, [print], PrintOptions)),
-	set_prolog_flag(toplevel_print_options, PrintOptions),
-	set_prolog_flag(prompt_alternatives_on, determinism),
-	set_prolog_flag(toplevel_extra_white_line, true),
+	create_prolog_flag(toplevel_print_options, PrintOptions, []),
+	create_prolog_flag(prompt_alternatives_on, determinism, []),
+	create_prolog_flag(toplevel_extra_white_line, true, []),
+	create_prolog_flag(toplevel_print_factorized, false, []),
 	'$set_debugger_print_options'(print),
 	'$run_initialization',
 	'$load_system_init_file',
-	'$load_gnu_emacs_interface',
-	'$option'(init_file, OsFile, OsFile),
+	'$option'(init_file, OsFile),
 	prolog_to_os_filename(File, OsFile),
 	'$load_init_file'(File),
 	start_pldoc,
 	'$load_script_file',
 	load_associated_file,
-	'$option'(goal, GoalAtom, GoalAtom),
+	'$option'(goal, GoalAtom),
 	term_to_atom(Goal, GoalAtom),
 	(   Goal == '$welcome'
 	->  flag('$banner_goal', TheGoal, TheGoal)
@@ -360,44 +390,40 @@ initialise_prolog :-
 	),
 	ignore(user:TheGoal).
 
-'$abort' :-
-	see(user),
-	tell(user),
-	flag('$break_level', _, 0),
-	flag('$compilation_level', _, 0),
-	'$calleventhook'(abort),
-	print_message(informational, '$aborted'),
-	'$toplevel'.
-
-'$break' :-
-	flag('$break_level', Old, Old+1),
-	flag('$break_level', New, New),
-	print_message(informational, break(enter(New))),
-	'$runtoplevel',
-	print_message(informational, break(exit(New))),
-	flag('$break_level', _, Old), !.
-
 :- '$hide'('$toplevel'/0).		% avoid in the GUI stacktrace
-:- '$hide'('$abort'/0).			% same after an abort
+
+%%	'$toplevel'/0
+%
+%	Called from PL_toplevel()
 
 '$toplevel' :-
 	'$runtoplevel',
 	print_message(informational, halt).
 
-%	Actually run the toplevel.  If there is a syntax error in the
-%	goal there is no reason to persue.  Something like that should
-%	happen to repetitive exceptions in the toplevel as well, but
-%	how do we distinguish between a stupid user and a program
-%	crashing in a loop?
+%%	'$runtoplevel'
+%
+%	Actually run the toplevel. If there  is   a  syntax error in the
+%	goal there is no reason to   persue.  Something like that should
+%	happen to repetitive exceptions in the toplevel as well, but how
+%	do we distinguish between a stupid   user and a program crashing
+%	in a loop?
+%
+%	@see prolog/0 is the default interactive toplevel
 
 '$runtoplevel' :-
-	'$option'(toplevel, TopLevelAtom, TopLevelAtom),
-	catch(term_to_atom(TopLevel, TopLevelAtom), E,
+	'$option'(toplevel, TopLevelAtom),
+	catch(term_to_atom(TopLevel0, TopLevelAtom), E,
 	      (print_message(error, E),
 	       halt(1))),
+	toplevel_goal(TopLevel0, TopLevel),
 	user:TopLevel.
 
-%	'$compile'
+toplevel_goal(prolog, '$query_loop') :- !.
+toplevel_goal(Goal, Goal).
+
+
+%%	'$compile'
+%
 %	Toplevel called when invoked with -c option.
 
 '$compile' :-
@@ -412,8 +438,20 @@ initialise_prolog :-
 		*********************************/
 
 prolog :-
-	flag('$tracing', _, off),
-	flag('$break_level', BreakLev, BreakLev),
+	break.
+
+%%	'$query_loop'
+%
+%	Run the normal Prolog query loop.  Note   that  the query is not
+%	protected by catch/3. Dealing with  unhandled exceptions is done
+%	by the C-function query_loop().  This   ensures  that  unhandled
+%	exceptions are really unhandled (in Prolog).
+
+'$query_loop' :-
+	(   current_prolog_flag(break_level, BreakLev)
+	->  true
+	;   BreakLev = -1
+	),
 	repeat,
 	    (   '$module'(TypeIn, TypeIn),
 		(   stream_property(user_input, tty(true))
@@ -439,11 +477,11 @@ read_query(Prompt, Goal, Bindings) :-
 	repeat,				% over syntax errors
 	prompt1(Prompt1),
 	catch('$raw_read'(user_input, Line), E,
-	      (print_message(error, E),
-	       (   E = error(syntax_error(_), _)
-	       ->  fail
-	       ;   throw(E)
-	       ))),
+	      (   E = error(syntax_error(_), _)
+	      ->  print_message(error, E),
+		  fail
+	      ;   throw(E)
+	      )),
 	(   current_predicate(_, user:rl_add_history(_))
 	->  format(atom(CompleteLine), '~W~W',
 		   [ Line, [partial(true)],
@@ -458,14 +496,10 @@ read_query(Prompt, Goal, Bindings) :-
 	      )), !,
 	'$save_history'(Line).
 read_query(Prompt, Goal, Bindings) :-
-	seeing(Old), see(user_input),
-	(   read_history(h, '!h',
-			 [trace, end_of_file],
-			 Prompt, Goal, Bindings)
-	->  see(Old)
-	;   see(Old),
-	    fail
-	).
+	read_history(h, '!h',
+		     [trace, end_of_file],
+		     Prompt, Goal, Bindings).
+
 
 remove_history_prompt('', '') :- !.
 remove_history_prompt(Prompt0, Prompt) :-
@@ -475,7 +509,7 @@ remove_history_prompt(Prompt0, Prompt) :-
 	atom_chars(Prompt, Chars).
 
 clean_history_prompt_chars([], []).
-clean_history_prompt_chars(['%', !|T], T) :- !.
+clean_history_prompt_chars(['~', !|T], T) :- !.
 clean_history_prompt_chars([H|T0], [H|T]) :-
 	clean_history_prompt_chars(T0, T).
 
@@ -488,8 +522,8 @@ set_default_history :-
 	(   (   current_prolog_flag(readline, true)
 	    ;	current_prolog_flag(emacs_inferior_process, true)
 	    )
-	->  set_prolog_flag(history, 0)
-	;   set_prolog_flag(history, 25)
+	->  create_prolog_flag(history, 0, [])
+	;   create_prolog_flag(history, 25, [])
 	).
 
 :- initialization set_default_history.
@@ -507,7 +541,8 @@ save_debug :-
 	),
 	current_prolog_flag(debug, Debugging),
 	set_prolog_flag(debug, false),
-	set_prolog_flag(query_debug_settings, debug(Debugging, Tracing)).
+	create_prolog_flag(query_debug_settings,
+			   debug(Debugging, Tracing), []).
 
 restore_debug :-
 	current_prolog_flag(query_debug_settings, debug(Debugging, Tracing)),
@@ -518,40 +553,30 @@ restore_debug :-
 	).
 
 :- initialization
-	set_prolog_flag(query_debug_settings, debug(false, false)).
+	create_prolog_flag(query_debug_settings, debug(false, false), []).
 
 
 		/********************************
 		*            PROMPTING		*
 		********************************/
 
-:- dynamic
-	'$prompt'/1.
-
-'$prompt'("%m%d%l%! ?- ").
-
-'$set_prompt'(P) :-
-	atom_codes(P, S),
-	retractall('$prompt'(_)),
-	assert('$prompt'(S)).
-
-
 '$system_prompt'(Module, BrekLev, Prompt) :-
-	'$prompt'(P0),
+	current_prolog_flag(toplevel_prompt, PAtom),
+	atom_codes(PAtom, P0),
 	(    Module \== user
-	->   '$substitute'("%m", [Module, ": "], P0, P1)
-	;    '$substitute'("%m", [], P0, P1)
+	->   '$substitute'("~m", [Module, ": "], P0, P1)
+	;    '$substitute'("~m", [], P0, P1)
 	),
-	(    BrekLev \== 0
-	->   '$substitute'("%l", ["[", BrekLev, "] "], P1, P2)
-	;    '$substitute'("%l", [], P1, P2)
+	(    BrekLev > 0
+	->   '$substitute'("~l", ["[", BrekLev, "] "], P1, P2)
+	;    '$substitute'("~l", [], P1, P2)
 	),
 	current_prolog_flag(query_debug_settings, debug(Debugging, Tracing)),
 	(    Tracing == true
-	->   '$substitute'("%d", ["[trace] "], P2, P3)
+	->   '$substitute'("~d", ["[trace] "], P2, P3)
 	;    Debugging == true
-	->   '$substitute'("%d", ["[debug] "], P2, P3)
-	;    '$substitute'("%d", [], P2, P3)
+	->   '$substitute'("~d", ["[debug] "], P2, P3)
+	;    '$substitute'("~d", [], P2, P3)
 	),
 	atom_chars(Prompt, P3).
 
@@ -588,12 +613,12 @@ subst_chars([H|T]) -->
 	print_message(query, query(eof)).
 '$execute'(Goal, Bindings) :-
 	'$module'(TypeIn, TypeIn),
+	'$dwim_correct_goal'(TypeIn:Goal, Bindings, Corrected), !,
 	setup_call_cleanup('$set_source_module'(M0, TypeIn),
-			   expand_goal(Goal, Expanded),
+			   expand_goal(Corrected, Expanded),
 			   '$set_source_module'(_, M0)),
-	'$dwim_correct_goal'(TypeIn:Expanded, Bindings, Corrected), !,
-	print_message(silent, toplevel_goal(Goal, Bindings)),
-	'$execute_goal2'(Corrected, Bindings).
+	print_message(silent, toplevel_goal(Expanded, Bindings)),
+	'$execute_goal2'(Expanded, Bindings).
 '$execute'(_, _) :-
 	notrace,
 	print_message(query, query(no)),
@@ -628,13 +653,17 @@ write_bindings(Bindings, Det) :-
 	copy_term(Bindings, Bindings1, Residuals0),
 	'$module'(TypeIn, TypeIn),
 	omit_qualifiers(Residuals0, TypeIn, Residuals),
-	bind_vars(Bindings1),
-	filter_bindings(Bindings1, Bindings2),
-	write_bindings2(Bindings2, Residuals, Det).
+	join_same_bindings(Bindings1, Bindings2),
+	factorize_bindings(Bindings2, Bindings3),
+	bind_vars(Bindings3, Bindings4),
+	filter_bindings(Bindings4, Bindings5),
+	write_bindings2(Bindings5, Residuals, Det).
 write_bindings(Bindings, Det) :-
-	bind_vars(Bindings),
-	filter_bindings(Bindings, Bindings1),
-	write_bindings2(Bindings1, [], Det).
+	join_same_bindings(Bindings, Bindings1),
+	factorize_bindings(Bindings1, Bindings2),
+	bind_vars(Bindings2, Bindings3),
+	filter_bindings(Bindings3, Bindings4),
+	write_bindings2(Bindings4, [], Det).
 
 write_bindings2([], Residuals, _) :-
 	current_prolog_flag(prompt_alternatives_on, groundness), !,
@@ -655,6 +684,26 @@ write_bindings2(Bindings, Residuals, _Det) :-
 	).
 
 
+%%	join_same_bindings(Bindings0, Bindings)
+%
+%	Join variables that are bound to the   same  value. Note that we
+%	return the _last_ value. This is   because the factorization may
+%	be different and ultimately the names will   be  printed as V1 =
+%	V2, ... VN = Value. Using the  last, Value has the factorization
+%	of VN.
+
+join_same_bindings([], []).
+join_same_bindings([Name=Val|T0], [[Name|Names]=V|T]) :-
+	take_same_bindings(Val, V, T0, Names, T1),
+	join_same_bindings(T1, T).
+
+take_same_bindings(Val, V, [Name=V1|T0], L, T) :-
+	V1 == Val, !,
+	L = [Name|Names],
+	take_same_bindings(V1, V, T0, Names, T).
+take_same_bindings(V, V, L, [], L).
+
+
 %%	omit_qualifiers(+QGoals, +TypeIn, -Goals) is det.
 %
 %	Omit unneeded module qualifiers  from   QGoals  relative  to the
@@ -697,37 +746,140 @@ omit_meta_qualifiers(when(Cond, QGoal), TypeIn, when(Cond, Goal)) :-
 omit_meta_qualifiers(G, _, G).
 
 
-%%	bind_vars(+Bindings)
+%%	bind_vars(+BindingsIn, -Bindings)
 %
 %	Bind variables to '$VAR'(Name), so they are printed by the names
 %	used in the query. Note that by   binding  in the reverse order,
 %	variables bound to one another come out in the natural order.
 
-bind_vars([]).
-bind_vars([Name=Var|T]) :-
-	bind_vars(T),
-	(   var(Var), \+ attvar(Var)
-	->  Var = '$VAR'(Name)
-	;   true
+bind_vars(Bindings0, Bindings) :-
+	bind_query_vars(Bindings0, Bindings, SNames),
+	bind_skel_vars(Bindings, Bindings, SNames, 1, _).
+
+bind_query_vars([], [], []).
+bind_query_vars([binding(Names,Var,[Var2=Cycle])|T0],
+		[binding(Names,Cycle,[])|T], [Name|SNames]) :-
+	Var == Var2, !,			% also implies var(Var)
+	'$last'(Names, Name),
+	Var = '$VAR'(Name),
+	bind_query_vars(T0, T, SNames).
+bind_query_vars([B|T0], [B|T], AllNames) :-
+	B = binding(Names,Var,Skel),
+	bind_query_vars(T0, T, SNames),
+	(   var(Var), \+ attvar(Var), Skel == []
+	->  AllNames = [Name|SNames],
+	    '$last'(Names, Name),
+	    Var = '$VAR'(Name)
+	;   AllNames = SNames
 	).
 
+
+
+bind_skel_vars([], _, _, N, N).
+bind_skel_vars([binding(_,_,Skel)|T], Bindings, SNames, N0, N) :-
+	bind_one_skel_vars(Skel, Bindings, SNames, N0, N1),
+	bind_skel_vars(T, Bindings, SNames, N1, N).
+
+%%	bind_one_skel_vars(+Subst, +Bindings, +VarName, +N0, -N)
+%
+%	Give names to the factorized variables that   do not have a name
+%	yet. This introduces names  _S<N>,   avoiding  duplicates.  If a
+%	factorized variable shares with another binding, use the name of
+%	that variable.
+%
+%	@tbd	Consider the call below. We could remove either of the
+%		A = x(1).  Which is best?
+%
+%		==
+%		?- A = x(1), B = a(A,A).
+%		A = x(1),
+%		B = a(A, A), % where
+%		    A = x(1).
+%		==
+
+bind_one_skel_vars([], _, _, N, N).
+bind_one_skel_vars([Var=Value|T], Bindings, Names, N0, N) :-
+	(   var(Var)
+	->  (	'$member'(binding(Names, VVal, []), Bindings),
+	        same_term(Value, VVal)
+	    ->	'$last'(Names, VName),
+		Var = '$VAR'(VName),
+		N2 = N0
+	    ;	between(N0, infinite, N1),
+	        atom_concat('_S', N1, Name),
+		\+ memberchk(Name, Names), !,
+		Var = '$VAR'(Name),
+		N2 is N1 + 1
+	    )
+	;   N2 = N0
+	),
+	bind_one_skel_vars(T, Bindings, Names, N2, N).
+
+
+%%	factorize_bindings(+Bindings0, -Factorized)
+%
+%	Factorize cycles and sharing in the bindings.
+
+factorize_bindings([], []).
+factorize_bindings([Name=Value|T0], [binding(Name, Skel, Subst)|T]) :-
+	'$factorize_term'(Value, Skel, Subst0),
+	(   current_prolog_flag(toplevel_print_factorized, true)
+	->  Subst = Subst0
+	;   only_cycles(Subst0, Subst)
+	),
+	factorize_bindings(T0, T).
+
+
+only_cycles([], []).
+only_cycles([B|T0], List) :-
+	(   B = (Var=Value),
+	    Var = Value,
+	    acyclic_term(Var)
+	->  only_cycles(T0, List)
+	;   List = [B|T],
+	    only_cycles(T0, T)
+	).
+
+
 %%	filter_bindings(+Bindings0, -Bindings)
 %
-%	Remove bindings that must not be printed.
+%	Remove bindings that must not be printed. There are two of them:
+%	Variables whose name start with '_'  and variables that are only
+%	bound to themselves (or, unbound).
 
 filter_bindings([], []).
-filter_bindings([H|T0], T) :-
-	hidden_binding(H), !,
-	filter_bindings(T0, T).
-filter_bindings([H|T0], [H|T]) :-
-	filter_bindings(T0, T).
+filter_bindings([H0|T0], T) :-
+	hide_vars(H0, H),
+	(   (   arg(1, H, [])
+	    ;   self_bounded(H)
+	    )
+	->  filter_bindings(T0, T)
+	;   T = [H|T1],
+	    filter_bindings(T0, T1)
+	).
 
-hidden_binding(Name = _) :-
-	sub_atom(Name, 0, _, _, '_'),
-	current_prolog_flag(toplevel_print_anon, false).
-hidden_binding(Name = Value) :-
+hide_vars(binding(Names0, Skel, Subst), binding(Names, Skel, Subst)) :-
+	hide_names(Names0, Skel, Subst, Names).
+
+hide_names([], _, _, []).
+hide_names([Name|T0], Skel, Subst, T) :-
+	(   sub_atom(Name, 0, _, _, '_'),
+	    current_prolog_flag(toplevel_print_anon, false)
+	->  true
+	;   Subst == [],
+	    Skel == '$VAR'(Name)
+	), !,
+	hide_names(T0, Skel, Subst, T).
+hide_names([Name|T0], Skel, Subst, [Name|T]) :-
+	hide_names(T0, Skel, Subst, T).
+
+self_bounded(binding([Name], Value, [])) :-
 	Value == '$VAR'(Name).
 
+%%	get_respons(-Action)
+%
+%	Read the continuation entered by the user.
+
 get_respons(Action) :-
 	repeat,
 	    flush_output(user_output),
@@ -744,19 +896,20 @@ answer_respons(Char, again) :-
 	print_message(help, query(help)).
 answer_respons(Char, redo) :-
 	memberchk(Char, ";nrNR \t"), !,
-	print_message(query, if_tty(';')).
+	print_message(query, if_tty([ansi(bold, ';', [])])).
 answer_respons(Char, redo) :-
 	memberchk(Char, "tT"), !,
 	trace,
-	print_message(query, if_tty('; [trace]')).
+	save_debug,
+	print_message(query, if_tty([ansi(bold, '; [trace]', [])])).
 answer_respons(Char, continue) :-
 	memberchk(Char, "ca\n\ryY."), !,
-	print_message(query, if_tty('.')).
+	print_message(query, if_tty([ansi(bold, '.', [])])).
 answer_respons(0'b, show_again) :- !,
 	break.
 answer_respons(Char, show_again) :-
 	print_predicate(Char, Pred, Options), !,
-	print_message(query, if_tty(Pred)),
+	print_message(query, if_tty(['~w'-[Pred]])),
 	set_prolog_flag(toplevel_print_options, Options).
 answer_respons(-1, show_again) :- !,
 	print_message(query, halt('EOF')),
@@ -764,11 +917,13 @@ answer_respons(-1, show_again) :- !,
 answer_respons(Char, again) :-
 	print_message(query, no_action(Char)).
 
-print_predicate(0'w, [write], [ quoted(true)
+print_predicate(0'w, [write], [ quoted(true),
+				spacing(next_argument)
 			      ]).
 print_predicate(0'p, [print], [ quoted(true),
 				portray(true),
-				max_depth(10)
+				max_depth(10),
+				spacing(next_argument)
 			      ]).
 
 
