@@ -3,9 +3,10 @@
     Part of SWI-Prolog
 
     Author:        Jan Wielemaker
-    E-mail:        J.Wielemaker@uva.nl
+    E-mail:        J.Wielemaker@vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (C): 1985-2009, University of Amsterdam
+    Copyright (C): 1985-2011, University of Amsterdam
+			      VU University Amsterdam
 
     This program is free software; you can redistribute it and/or
     modify it under the terms of the GNU General Public License
@@ -19,7 +20,7 @@
 
     You should have received a copy of the GNU Lesser General Public
     License along with this library; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 
     As a special exception, if you link this library with other files,
     compiled with a Free Software compiler, to produce an executable, this
@@ -60,26 +61,36 @@ and initialise this using make_<record>/2. In addition to providing good
 performance, this also provides type-checking and central declaration of
 defaults.
 
-==
-:- record atts(width:integer=100, shape:oneof([box,circle])=box).
+  ==
+  :- record atts(width:integer=100, shape:oneof([box,circle])=box).
 
-process(Data, Options) :-
-	make_atts(Options, Attributes),
-	action(Data, Attributes).
+  process(Data, Options) :-
+	  make_atts(Options, Attributes),
+	  action(Data, Attributes).
 
-action(Data, Attributes) :-
-	atts_shape(Attributes, Shape),
-	...
-==
+  action(Data, Attributes) :-
+	  atts_shape(Attributes, Shape),
+	  ...
+  ==
+
+Options typically have exactly one argument.   The  library does support
+options  with  0  or  more  than    one  arguments  with  the  following
+restrictions:
+
+  - The predicate option/3 and select_option/4, involving default are
+    meaningless. They perform an arg(1, Option, Default), causing
+    failure without arguments and filling only the first option-argument
+    otherwise.
+  - meta_options/3 can only qualify options with exactly one argument.
 
 @tbd	We should consider putting many options in an assoc or record
 	with appropriate preprocessing to achieve better performance.
-@tbd	We should provide some standard to to automatic type-checking
-	on option lists.
 @see	library(record)
+@see	Option processing capabilities may be declared using the
+	directive predicate_options/3.
 */
 
-%%	option(?Option, +OptionList, +Default)
+%%	option(?Option, +OptionList, +Default) is semidet.
 %
 %	Get  an  option  from  a  OptionList.  OptionList  can  use  the
 %	Name=Value as well as the Name(Value) convention.
@@ -87,18 +98,15 @@ action(Data, Attributes) :-
 %	@param Option	Term of the form Name(?Value).
 
 option(Opt, Options, Default) :-	% make option processing stead-fast
-	arg(1, Opt, OptVal),
-	ground(OptVal), !,
-	functor(Opt, OptName, 1),
-	functor(Gen, OptName, 1),
-	option(Gen, Options, Default),
-	Opt = Gen.
-option(Opt, Options, _) :-
-	get_option(Opt, Options), !.
-option(Opt, _, Default) :-
-	arg(1, Opt, Default).
+	functor(Opt, Name, Arity),
+	functor(GenOpt, Name, Arity),
+	(   get_option(GenOpt, Options)
+	->  Opt = GenOpt
+	;   arg(1, Opt, Default)
+	).
 
-%%	option(?Option, +OptionList)
+
+%%	option(?Option, +OptionList) is semidet.
 %
 %	Get  an  option  from  a  OptionList.  OptionList  can  use  the
 %	Name=Value as well as the Name(Value) convention. Fails silently
@@ -107,15 +115,10 @@ option(Opt, _, Default) :-
 %	@param Option	Term of the form Name(?Value).
 
 option(Opt, Options) :-			% make option processing stead-fast
-	arg(1, Opt, OptVal),
-	nonvar(OptVal), !,
-	functor(Opt, OptName, 1),
-	functor(Gen, OptName, 1),
-	option(Gen, Options),
-	Opt = Gen.
-option(Opt, Options) :-
-	get_option(Opt, Options), !.
-
+	functor(Opt, Name, Arity),
+	functor(GenOpt, Name, Arity),
+	get_option(GenOpt, Options), !,
+	Opt = GenOpt.
 
 get_option(Opt, Options) :-
 	memberchk(Opt, Options), !.
@@ -132,22 +135,17 @@ get_option(Opt, Options) :-
 %	options with RestOptions.
 
 select_option(Opt, Options0, Options) :-	% stead-fast
-	arg(1, Opt, OptVal),
-	nonvar(OptVal), !,
-	functor(Opt, OptName, 1),
-	functor(Gen, OptName, 1),
-	select_option(Gen, Options0, Options),
-	Opt = Gen.
-select_option(Opt, Options0, Options) :-
-	get_option(Opt, Options0, Options), !.
-
+	functor(Opt, Name, Arity),
+	functor(GenOpt, Name, Arity),
+	get_option(GenOpt, Options0, Options),
+	Opt = GenOpt.
 
 get_option(Opt, Options0, Options) :-
-	select(Opt, Options0, Options), !.
+	selectchk(Opt, Options0, Options), !.
 get_option(Opt, Options0, Options) :-
 	functor(Opt, OptName, 1),
 	arg(1, Opt, OptVal),
-	select(OptName=OptVal, Options0, Options), !.
+	selectchk(OptName=OptVal, Options0, Options).
 
 %%	select_option(?Option, +Options, -RestOptions, +Default) is det.
 %
@@ -155,10 +153,14 @@ get_option(Opt, Options0, Options) :-
 %	but if Option is not  in  Options,   its  value  is unified with
 %	Default and RestOptions with Options.
 
-select_option(Option, Options, RestOptions, _Default) :-
-	select_option(Option, Options, RestOptions), !.
-select_option(Option, Options, Options, Default) :-
-	arg(1, Option, Default).
+select_option(Option, Options, RestOptions, Default) :-
+	functor(Option, Name, Arity),
+	functor(GenOpt, Name, Arity),
+	(   get_option(GenOpt, Options, RestOptions)
+	->  Option = GenOpt
+	;   RestOptions = Options,
+	    arg(1, Option, Default)
+	).
 
 
 %%	merge_options(+New, +Old, -Merged) is det.
@@ -166,9 +168,15 @@ select_option(Option, Options, Options, Default) :-
 %	Merge two option lists. Merged is a sorted list of options using
 %	the canonical format Name(Value) holding   all  options from New
 %	and Old, after removing conflicting options from Old.
+%
+%	Multi-values options (e.g.,  proxy(Host,   Port))  are  allowed,
+%	where both option-name and arity  define   the  identity  of the
+%	option.
 
-merge_options([], Old, Merged) :- !, Merged = Old.
-merge_options(New, [], Merged) :- !, Merged = New.
+merge_options([], Old, Merged) :- !,
+	canonise_options(Old, Merged).
+merge_options(New, [], Merged) :- !,
+	canonise_options(New, Merged).
 merge_options(New, Old, Merged) :-
 	canonise_options(New, NCanonical),
 	canonise_options(Old, OCanonical),
@@ -179,28 +187,30 @@ merge_options(New, Old, Merged) :-
 ord_merge([], L, L) :- !.
 ord_merge(L, [], L) :- !.
 ord_merge([NO|TN], [OO|TO], Merged) :-
-	functor(NO, NName, 1),
-	functor(OO, OName, 1),
-	compare(Diff, NName, OName),
-	ord_merge(Diff, NO, NName, OO, OName, TN, TO, Merged).
+	sort_key(NO, NKey),
+	sort_key(OO, OKey),
+	compare(Diff, NKey, OKey),
+	ord_merge(Diff, NO, NKey, OO, OKey, TN, TO, Merged).
 
 ord_merge(=, NO, _, _, _, TN, TO, [NO|T]) :-
 	ord_merge(TN, TO, T).
-ord_merge(<, NO, _, OO, OName, TN, TO, [NO|T]) :-
+ord_merge(<, NO, _, OO, OKey, TN, TO, [NO|T]) :-
 	(   TN = [H|TN2]
-	->  functor(H, NName, 1),
-	    compare(Diff, NName, OName),
-	    ord_merge(Diff, H, NName, OO, OName, TN2, TO, T)
+	->  sort_key(H, NKey),
+	    compare(Diff, NKey, OKey),
+	    ord_merge(Diff, H, NKey, OO, OKey, TN2, TO, T)
 	;   T = [OO|TO]
 	).
-ord_merge(>, NO, NName, OO, _, TN, TO, [OO|T]) :-
+ord_merge(>, NO, NKey, OO, _, TN, TO, [OO|T]) :-
 	(   TO = [H|TO2]
-	->  functor(H, OName, 1),
-	    compare(Diff, NName, OName),
-	    ord_merge(Diff, NO, NName, H, OName, TN, TO2, T)
+	->  sort_key(H, OKey),
+	    compare(Diff, NKey, OKey),
+	    ord_merge(Diff, NO, NKey, H, OKey, TN, TO2, T)
 	;   T = [NO|TN]
 	).
 
+sort_key(Option, Name-Arity) :-
+	functor(Option, Name, Arity).
 
 %%	canonise_options(+OptionsIn, -OptionsOut) is det.
 %
@@ -231,9 +241,15 @@ canonise_options2([H|T0], [H|T]) :- !,
 %
 %	is_meta(callback).
 %	==
+%
+%	Meta-options must have exactly one  argument. This argument will
+%	be qualified.
+%
+%	@tbd	Should be integrated with declarations from
+%		predicate_options/3.
 
 :- meta_predicate
-	meta_options(:, :, -).
+	meta_options(1, :, -).
 
 meta_options(IsMeta, Context:Options0, Options) :-
 	meta_options(Options0, IsMeta, Context, Options).
