@@ -3,9 +3,10 @@
     Part of SWI-Prolog
 
     Author:        Jan Wielemaker
-    E-mail:        wielemak@science.uva.nl
+    E-mail:        J.Wielemaker@vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (C): 1985-2005, University of Amsterdam
+    Copyright (C): 1985-2011, University of Amsterdam
+			      VU University Amsterdam
 
     This program is free software; you can redistribute it and/or
     modify it under the terms of the GNU General Public License
@@ -19,7 +20,7 @@
 
     You should have received a copy of the GNU Lesser General Public
     License along with this library; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 
     As a special exception, if you link this library with other files,
     compiled with a Free Software compiler, to produce an executable, this
@@ -35,29 +36,46 @@
 	    predicate_name/2,		% +Head, -Name
 	    clause_name/2		% +ClauseRef, -Name
 	  ]).
-:- use_module(library(debug), [debug/3]).
 :- use_module(library(lists), [append/3]).
 :- use_module(library(occurs), [sub_term/2]).
 :- use_module(library(debug)).
 :- use_module(library(listing)).
+:- use_module(library(prolog_source)).
+
+:- public				% called from library(trace/clause)
+	unify_term/2,
+	make_varnames/5,
+	do_make_varnames/3.
+
+:- multifile
+	make_varnames_hook/5.
+
+/** <module> Get detailed source-information about a clause
 
-/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 This module started life as part of the   GUI tracer. As it is generally
 useful for debugging  purposes  it  has   moved  to  the  general Prolog
-library. Being only applicable to debugging   and  sitting very close to
-the kernel, do not rely too much that its functionality will be stable.
+library.
 
-The tracer library  tracer/clause.pl  adds   caching  and  dealing  with
+The tracer library library(trace/clause) adds   caching and dealing with
 dynamic predicates using listing to  XPCE   objects  to  this. Note that
-clause_info/4 as below can be pretty slow.
-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
+clause_info/4 as below can be slow.
+*/
 
 %%	clause_info(+ClauseRef, -File, -TermPos, -VarNames)
 %
-%	Fetches source information for the given clause.
+%	Fetches source information for the  given   clause.  File is the
+%	file from which the clause  was   loaded.  TermPos describes the
+%	source layout in a format   compatible  to the subterm_positions
+%	option of read_term/2.  VarNames provides access to the variable
+%	allocation in a stack-frame.  See make_varnames/5 for details.
 
 clause_info(ClauseRef, File, TermPos, NameOffset) :-
-	debug(clause_info, 'clause_info(~w)... ', [ClauseRef]),
+	(   debugging(clause_info)
+	->  clause_name(ClauseRef, Name),
+	    debug(clause_info, 'clause_info(~w) (~w)... ',
+		  [ClauseRef, Name])
+	;   true
+	),
 	clause_property(ClauseRef, file(File)),
 	'$clause'(Head, Body, ClauseRef, VarOffset),
 	(   Body == true
@@ -73,9 +91,9 @@ clause_info(ClauseRef, File, TermPos, NameOffset) :-
 	debug(clause_info, 'from ~w:~d ... ', [File, LineNo]),
 	read_term_at_line(File, LineNo, Module, Clause, TermPos0, VarNames),
 	debug(clause_info, 'read ...', []),
-	unify_clause(Clause, DecompiledClause, TermPos0, TermPos),
+	unify_clause(Clause, DecompiledClause, Module, TermPos0, TermPos),
 	debug(clause_info, 'unified ...', []),
-	make_varnames(Clause, VarOffset, VarNames, NameOffset),
+	make_varnames(Clause, DecompiledClause, VarOffset, VarNames, NameOffset),
 	debug(clause_info, 'got names~n', []), !.
 
 %%	unify_term(+T1, +T2)
@@ -102,6 +120,8 @@ unify_term(X, Y) :-
 	string(X),
 	is_list(Y),
 	string_to_list(X, Y), !.
+unify_term(_, Y) :-
+	Y == '...', !.				% elipses left by max_depth
 unify_term(_:X, Y) :-
 	unify_term(X, Y), !.
 unify_term(X, _:Y) :-
@@ -122,75 +142,43 @@ unify_args(I, Arity, T1, T2) :-
 	unify_args(A, Arity, T1, T2).
 
 
-%	Must be a user-programmable hook!
-
-alternate_syntax(prolog, _,    true,
-			       true).
-alternate_syntax(pce_class, M, pce_expansion:push_compile_operators(M),
-			       pce_expansion:pop_compile_operators) :-
-	current_prolog_flag(xpce, true).
-alternate_syntax(system, _,    style_check(+dollar),
-			       style_check(-dollar)).
-
-system_module(system) :- !.
-system_module(Module) :-
-	sub_atom(Module, 0, _, _, $), !.
+%%	read_term_at_line(+File, +Line, +Module,
+%%			  -Clause, -TermPos, -VarNames) is semidet.
+%
+%	Read a term from File at Line.
 
 read_term_at_line(File, Line, Module, Clause, TermPos, VarNames) :-
 	catch(open(File, read, In), _, fail),
-	call_cleanup(read(Line, In, Module, Clause, TermPos, VarNames),
-		     close(In)).
-
-read(Line, Handle, Module, Clause, TermPos, VarNames) :-
-	seek_to_line(Handle, Line),
-	read(Handle, Module, Clause, TermPos, VarNames).
-
-%%	read(+Stream, +Module, -Clause, -TermPos, -VarNames)
-%
-%	Read clause from Stream at current position with unknown syntax.
-%	It returns the term read  at  that   position,  as  well  as the
-%	subterm position info and variable names.
-%
-%	NOTE: Called directly from  library(trace/clause)   for  the GUI
-%	tracer.
-
-read(Handle, Module, Clause, TermPos, VarNames) :-
-	(   system_module(Module)
-	->  Syntax = system
-	;   true
-	),
-	stream_property(Handle, position(Here)),
-	alternate_syntax(Syntax, Module, Setup, Restore),
-	peek_char(Handle, X),
-	debug(clause_info, 'Using syntax ~w (c=~w)', [Syntax, X]),
-	Setup,
-	catch(read_term(Handle, Clause,
-			[ subterm_positions(TermPos),
-			  variable_names(VarNames),
-			  module(Module)
-			]),
-	      Error,
-	      true),
-	Restore,
-	(   var(Error)
-	->  !
-	;   set_stream_position(Handle, Here),
-	    fail
-	).
+	call_cleanup(
+	    read_source_term_at_location(
+		In, Clause,
+		[ line(Line),
+		  module(Module),
+		  subterm_positions(TermPos),
+		  variable_names(VarNames)
+		]),
+	    close(In)).
 
 
-%%	make_varnames(+ReadClause, +Offsets, +Names, -Term) is det.
+%%	make_varnames(+ReadClause, +DecompiledClause,
+%%		      +Offsets, +Names, -Term) is det.
 %
 %	Create a Term varnames(...) where each argument contains the name
 %	of the variable at that offset.  If the read Clause is a DCG rule,
 %	name the two last arguments <DCG_list> and <DCG_tail>
 %
+%	This    predicate    calles     the      multifile     predicate
+%	make_varnames_hook/5 with the same arguments   to allow for user
+%	extensions. Extending this predicate  is   needed  if a compiler
+%	adds additional arguments to the clause   head that must be made
+%	visible in the GUI tracer.
+%
 %	@param Offsets	List of Offset=Var
 %	@param Names	List of Name=Var
-%
-%	@bug Called directly from library(trace/clause) for the GUI tracer.
 
-make_varnames((Head --> _Body), Offsets, Names, Bindings) :- !,
+make_varnames(ReadClause, DecompiledClause, Offsets, Names, Term) :-
+	make_varnames_hook(ReadClause, DecompiledClause, Offsets, Names, Term), !.
+make_varnames((Head --> _Body), _, Offsets, Names, Bindings) :- !,
 	functor(Head, _, Arity),
 	In is Arity,
 	memberchk(In=IVar, Offsets),
@@ -198,8 +186,8 @@ make_varnames((Head --> _Body), Offsets, Names, Bindings) :- !,
 	Out is Arity + 1,
 	memberchk(Out=OVar, Offsets),
 	Names2 = ['<DCG_tail>'=OVar|Names1],
-	make_varnames(xx, Offsets, Names2, Bindings).
-make_varnames(_, Offsets, Names, Bindings) :-
+	make_varnames(xx, xx, Offsets, Names2, Bindings).
+make_varnames(_, _, Offsets, Names, Bindings) :-
 	length(Offsets, L),
 	functor(Bindings, varnames, L),
 	do_make_varnames(Offsets, Names, Bindings).
@@ -219,26 +207,35 @@ find_varname(Var, [Name = TheVar|_], Name) :-
 find_varname(Var, [_|T], Name) :-
 	find_varname(Var, T, Name).
 
-%%	unify_clause(+Read, +Decompiled, +ReadTermPos, -RecompiledTermPos).
+%%	unify_clause(+Read, +Decompiled, +Module, +ReadTermPos,
+%%		     -RecompiledTermPos).
 %
 %	What you read isn't always what goes into the database. The task
 %	of this predicate is to establish  the relation between the term
 %	read from the file and the result from decompiling the clause.
 %
-%	This really must be  more  flexible,   dealing  with  much  more
-%	complex source-translations, falling back to  a heuristic method
-%	locating as much as possible.
+%	This predicate calls the multifile predicate unify_clause_hook/5
+%	with the same arguments to support user extensions.
+%
+%	@tbd	This really must be  more   flexible,  dealing with much
+%		more complex source-translations,  falling   back  to  a
+%		heuristic method locating as much as possible.
 
-unify_clause(Read, Read, TermPos, TermPos) :- !.
+:- multifile
+	unify_clause_hook/5.
+
+unify_clause(Read, Read, _, TermPos, TermPos) :- !.
 					% XPCE send-methods
-unify_clause(:->(Head, Body), (PlHead :- PlBody), TermPos0, TermPos) :- !,
+unify_clause(Read, Decompiled, Module, TermPoso, TermPos) :-
+	unify_clause_hook(Read, Decompiled, Module, TermPoso, TermPos), !.
+unify_clause(:->(Head, Body), (PlHead :- PlBody), _, TermPos0, TermPos) :- !,
 	pce_method_clause(Head, Body, PlHead, PlBody, TermPos0, TermPos).
 					% XPCE get-methods
-unify_clause(:<-(Head, Body), (PlHead :- PlBody), TermPos0, TermPos) :- !,
+unify_clause(:<-(Head, Body), (PlHead :- PlBody), _, TermPos0, TermPos) :- !,
 	pce_method_clause(Head, Body, PlHead, PlBody, TermPos0, TermPos).
 					% Unit test clauses
 unify_clause((TH :- Body),
-	     (_:'unit body'(_, _) :- !, Body),
+	     (_:'unit body'(_, _) :- !, Body), _,
 	     TP0, TP) :-
 	(   TH = test(_,_)
 	;   TH = test(_)
@@ -247,17 +244,17 @@ unify_clause((TH :- Body),
 	TP  = term_position(F,T,FF,FT,[HP,term_position(0,0,0,0,[FF-FT,BP])]).
 					% module:head :- body
 unify_clause((Head :- Read),
-	     (Head :- _M:Compiled), TermPos0, TermPos) :-
-	unify_clause((Head :- Read), (Head :- Compiled), TermPos0, TermPos1),
+	     (Head :- _M:Compiled), Module, TermPos0, TermPos) :-
+	unify_clause((Head :- Read), (Head :- Compiled), Module, TermPos0, TermPos1),
 	TermPos1 = term_position(TA,TZ,FA,FZ,[PH,PB]),
 	TermPos  = term_position(TA,TZ,FA,FZ,
 				 [ PH,
 				   term_position(0,0,0,0,[0-0,PB])
 				 ]).
-unify_clause(Read, Compiled1, TermPos0, TermPos) :-
+unify_clause(Read, Compiled1, Module, TermPos0, TermPos) :-
 	Read = (_ --> List, _),
 	is_list(List),
-	catch(expand_term(Read, Compiled2), E, expand_failed(E, Read)),
+	ci_expand(Read, Compiled2, Module),
 	Compiled2 = (DH :- _),
 	functor(DH, _, Arity),
 	DArg is Arity - 1,
@@ -269,11 +266,11 @@ unify_clause(Read, Compiled1, TermPos0, TermPos) :-
 	TermPos1 = term_position(F,T,FF,FT,[ HP, BP ]),
 	match_module(Compiled2, Compiled1, TermPos1, TermPos).
 					% general term-expansion
-unify_clause(Read, Compiled1, TermPos0, TermPos) :-
-	catch(expand_term(Read, Compiled2), E, expand_failed(E, Read)),
+unify_clause(Read, Compiled1, Module, TermPos0, TermPos) :-
+	ci_expand(Read, Compiled2, Module),
 	match_module(Compiled2, Compiled1, TermPos0, TermPos).
 					% I don't know ...
-unify_clause(_, _, _, _) :-
+unify_clause(_, _, _, _, _) :-
 	debug(clause_info, 'Could not unify clause', []),
 	fail.
 
@@ -281,6 +278,13 @@ unify_clause_head(H1, H2) :-
 	strip_module(H1, _, H),
 	strip_module(H2, _, H).
 
+ci_expand(Read, Compiled, Module) :-
+	catch(setup_call_cleanup('$set_source_module'(Old, Module),
+				 expand_term(Read, Compiled),
+				 '$set_source_module'(_, Old)),
+	      E,
+	      expand_failed(E, Read)).
+
 match_module((H1 :- B1), (H2 :- B2), Pos0, Pos) :- !,
 	unify_clause_head(H1, H2),
 	unify_body(B1, B2, Pos0, Pos).
@@ -305,12 +309,26 @@ expand_failed(E, Read) :-
 %	Pos0 and Pos still include the term-position of the head.
 
 unify_body(B, B, Pos, Pos) :-
-	\+ sub_term(brace_term_position(_,_,_), Pos), !.
+	does_not_dcg_after_binding(B, Pos), !.
 unify_body(R, D,
 	   term_position(F,T,FF,FT,[HP,BP0]),
 	   term_position(F,T,FF,FT,[HP,BP])) :-
 	ubody(R, D, BP0, BP).
 
+%%	does_not_dcg_after_binding(+ReadBody, +ReadPos) is semidet.
+%
+%	True  if  ReadPos/ReadPos  does   not    contain   DCG   delayed
+%	unifications.
+%
+%	@tbd	We should pass that we are in a DCG; if we are not there
+%		is no reason for this test.
+
+does_not_dcg_after_binding(B, Pos) :-
+	acyclic_term(B),		% X = call(X)
+	\+ sub_term(brace_term_position(_,_,_), Pos),
+	\+ (sub_term((Cut,_=_), B), Cut == !), !.
+
+
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 Some remarks.
 
@@ -326,7 +344,10 @@ a --> { x, y, z }.
 %	@param TermPosRead	Sub-term positions of source
 
 ubody(B, B, P, P) :-
-	\+ sub_term(brace_term_position(_,_,_), P), !.
+	does_not_dcg_after_binding(B, P), !.
+ubody(X, call(X),			% X = call(X)
+      From-To,
+      term_position(From, To, From, To, [From-To])) :- !.
 ubody(B0, B,
       brace_term_position(F,T,A0),
       Pos) :-
@@ -391,6 +412,9 @@ conj((A,B), brace_term_position(_,T,PA), GL, TG, PL, TP) :-
 	conj(A, PA, GL, TGA, PL, TPA),
 	T1 is T - 1,
 	conj(B, T1-T, TGA, TG, TPA, TP).
+conj((!,(S=SR)), F-T, [!,S=SR|TG], TG, [F-T,F1-T1|TP], TP) :-
+	F1 is F+1,
+	T1 is T+1.
 conj(A, P, [A|TG], TG, [P|TP], TP).
 
 
@@ -509,7 +533,7 @@ pce_method_body2(A0, A, TermPos0, TermPos) :-
 pce_method_body2(A0, A, TermPos0, TermPos) :-
 	expand_goal(A0, A, TermPos0, TermPos).
 
-control_op((,)).
+control_op(',').
 control_op((;)).
 control_op((->)).
 control_op((*->)).
@@ -531,6 +555,8 @@ To keep track of the source-locations, we   have to redo the analysis of
 the clause as defined in init.pl
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
+expand_goal(G, call(G), P, term_position(0,0,0,0,[P])) :-
+        var(G), !.
 expand_goal(G, G, P, P) :-
         var(G), !.
 expand_goal(M0, M, P0, P) :-
@@ -578,15 +604,27 @@ goal_expansion(SendSuperN, send_class(R, _, Msg), P, P) :-
 	compound(SendSuperN),
 	SendSuperN =.. [send_super, R, Sel | Args],
 	Msg =.. [Sel|Args].
+goal_expansion(SendN, send(R, Msg), P, P) :-
+	compound(SendN),
+	SendN =.. [send, R, Sel | Args],
+	atom(Sel), Args \== [],
+	Msg =.. [Sel|Args].
 goal_expansion(GetSuperN, get_class(R, _, Msg, Answer), P, P) :-
 	compound(GetSuperN),
 	GetSuperN =.. [get_super, R, Sel | AllArgs],
 	append(Args, [Answer], AllArgs),
 	Msg =.. [Sel|Args].
+goal_expansion(GetN, get(R, Msg, Answer), P, P) :-
+	compound(GetN),
+	GetN =.. [get, R, Sel | AllArgs],
+	append(Args, [Answer], AllArgs),
+	atom(Sel), Args \== [],
+	Msg =.. [Sel|Args].
 goal_expansion(G0, G, P, P) :-
 	user:goal_expansion(G0, G),	% TBD: we need the module!
 	G0 \== G.			% \=@=?
 
+
 		 /*******************************
 		 *	  PRINTABLE NAMES	*
 		 *******************************/
@@ -607,6 +645,10 @@ thaffix(1, st) :- !.
 thaffix(2, nd) :- !.
 thaffix(_, th).
 
+%%	predicate_name(:Head, -PredName:string) is det.
+%
+%	Describe a predicate as [Module:]Name/Arity.
+
 predicate_name(Predicate, PName) :-
 	strip_module(Predicate, Module, Head),
 	(   user:prolog_predicate_name(Module:Head, PName)
@@ -630,24 +672,3 @@ clause_name(Ref, Name) :-
 	thaffix(N, Th),
 	format(string(Name), '~d-~w clause of ~w', [N, Th, PredName]).
 clause_name(_, '<meta-call>').
-
-
-		 /*******************************
-		 *        LOW-LEVEL STUFF	*
-		 *******************************/
-
-/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-These predicates communicate about lines.  We   should  consider using a
-line-cache for this for speed.
-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
-
-%%	seek_to_line(+Stream, +Line)
-%
-%	Seek to indicated line-number.
-
-seek_to_line(Fd, N) :-
-	N > 1, !,
-	skip(Fd, 10),
-	NN is N - 1,
-	seek_to_line(Fd, NN).
-seek_to_line(_, _).
