@@ -3,9 +3,10 @@
     Part of SWI-Prolog
 
     Author:        Jan Wielemaker
-    E-mail:        jan@swi.psy.uva.nl
+    E-mail:        J.Wielemaker@cs.vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (C): 1985-2002, University of Amsterdam
+    Copyright (C): 1985-2011, University of Amsterdam
+			      VU University Amsterdam
 
     This program is free software; you can redistribute it and/or
     modify it under the terms of the GNU General Public License
@@ -19,7 +20,7 @@
 
     You should have received a copy of the GNU Lesser General Public
     License along with this library; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 
     As a special exception, if you link this library with other files,
     compiled with a Free Software compiler, to produce an executable, this
@@ -31,7 +32,8 @@
 
 
 :- module(prolog_statistics,
-	  [ time/1,			% :Goal
+	  [ statistics/0,
+	    time/1,			% :Goal
 	    profile/1,			% :Goal
 	    profile/3,			% :Goal, +Style, +Top
 	    show_profile/2,		% +Style, +Top
@@ -44,35 +46,157 @@
 	profile(0),
 	profile(0, +, +).
 
-%%	time(:Goal)
-%
-%	Time the execution of Goal.  Possible choice-points of Goal are
-%	removed.
+/** <module> Get information about resource usage
 
-time(Goal0) :-
-	expand_goal(Goal0, Goal),
-	get_time(OldWall),
-	statistics(cputime, OldTime),
-	statistics(inferences, OldInferences),
-	(   catch(Goal, E, true)
-	->  Result = yes
-	;   Result = no
-	),
-	statistics(inferences, NewInferences),
-	statistics(cputime, NewTime),
-	get_time(NewWall),
+This library provides predicates to   obtain  information about resource
+usage by your program. The predicates of  this library are for human use
+at the toplevel: information is _printed_.   All predicates obtain their
+information using public low-level primitives.   These primitives can be
+use to obtain selective statistics during execution.
+*/
+
+%%	statistics is det.
+%
+%	Print information about  resource  usage   to  the  =user_error=
+%	stream.
+%
+%	@see	All statistics printed are obtained through statistics/2.
+
+statistics :-
+	statistics(user_error).
+
+statistics(Out) :-
+	statistics(trail, Trail),
+	statistics(trailused, TrailUsed),
+	statistics(local, Local),
+	statistics(localused, LocalUsed),
+	statistics(global, Global),
+	statistics(globalused, GlobalUsed),
+	statistics(process_cputime, Cputime),
+	statistics(inferences, Inferences),
+	statistics(atoms, Atoms),
+	statistics(functors, Functors),
+	statistics(predicates, Predicates),
+	statistics(modules, Modules),
+	statistics(codes, Codes),
+	statistics(locallimit, LocalLimit),
+	statistics(globallimit, GlobalLimit),
+	statistics(traillimit, TrailLimit),
+
+	format(Out, '~3f seconds cpu time for ~D inferences~n',
+				    [Cputime, Inferences]),
+	format(Out, '~D atoms, ~D functors, ~D predicates, ~D modules, ~D VM-codes~n~n',
+				    [Atoms, Functors, Predicates, Modules, Codes]),
+	format(Out, '                       Limit    Allocated       In use~n', []),
+	format(Out, 'Local  stack :~t~D~28| ~t~D~41| ~t~D~54| Bytes~n',
+	       [LocalLimit, Local, LocalUsed]),
+	format(Out, 'Global stack :~t~D~28| ~t~D~41| ~t~D~54| Bytes~n',
+	       [GlobalLimit, Global, GlobalUsed]),
+	format(Out, 'Trail  stack :~t~D~28| ~t~D~41| ~t~D~54| Bytes~n~n',
+	       [TrailLimit, Trail, TrailUsed]),
+
+	gc_statistics(Out),
+	agc_statistics(Out),
+	shift_statistics(Out),
+	thread_statistics(Out).
+
+gc_statistics(Out) :-
+	statistics(collections, Collections),
+	Collections > 0, !,
+	statistics(collected, Collected),
+	statistics(gctime, GcTime),
+
+	format(Out, '~D garbage collections gained ~D bytes in ~3f seconds.~n',
+	       [Collections, Collected, GcTime]).
+gc_statistics(_).
+
+agc_statistics(Out) :-
+	catch(statistics(agc, Agc), _, fail),
+	Agc > 0, !,
+	statistics(agc_gained, Gained),
+	statistics(agc_time, Time),
+	format(Out, '~D atom garbage collections gained ~D atoms in ~3f seconds.~n',
+	       [Agc, Gained, Time]).
+agc_statistics(_).
+
+shift_statistics(Out) :-
+	statistics(local_shifts, LS),
+	statistics(global_shifts, GS),
+	statistics(trail_shifts, TS),
+	(   LS > 0
+	;   GS > 0
+	;   TS > 0
+	), !,
+	statistics(shift_time, Time),
+	format(Out, 'Stack shifts: ~D local, ~D global, ~D trail in ~3f seconds.~n',
+	       [LS, GS, TS, Time]).
+shift_statistics(_).
+
+thread_statistics(Out) :-
+	current_prolog_flag(threads, true), !,
+	statistics(threads, Active),
+	statistics(threads_created, Created),
+	statistics(thread_cputime, CpuTime),
+	Finished is Created - Active,
+	format(Out, '~D threads, ~D finished threads used ~3f seconds.~n',
+	       [Active, Finished, CpuTime]).
+thread_statistics(_).
+
+
+%%	time(:Goal) is nondet.
+%
+%	Execute Goal, reporting statistics to the user. If Goal succeeds
+%	non-deterministically,  retrying  reports  the   statistics  for
+%	providing the next answer.
+%
+%	Statistics  are  retrieved  using   thread_statistics/3  on  the
+%	calling   thread.   Note   that   not    all   systems   support
+%	thread-specific CPU time. Notable, this is lacking on MacOS X.
+%
+%	@bug Inference statistics are often a few off.
+%	@see statistics/2 for obtaining statistics in your program and
+%	     understanding the reported values.
+
+time(Goal) :-
+	time_state(State0),
+	(   call_cleanup(catch(Goal, E, (report(State0,10), throw(E))),
+			 Det = true),
+	    time_true(State0),
+	    (	Det == true
+	    ->	!
+	    ;	true
+	    )
+	;   report(State0, 11),
+	    fail
+	).
+
+report(t(OldWall, OldTime, OldInferences), Sub) :-
+	time_state(t(NewWall, NewTime, NewInferences)),
 	UsedTime is NewTime - OldTime,
-	UsedInf  is NewInferences - OldInferences - 3,
+	UsedInf  is NewInferences - OldInferences - Sub,
 	Wall     is NewWall - OldWall,
 	(   UsedTime =:= 0
 	->  Lips = 'Infinite'
 	;   Lips is integer(UsedInf / UsedTime)
 	),
-	print_message(informational, time(UsedInf, UsedTime, Wall, Lips)),
-	(   nonvar(E)
-	->  throw(E)
-	;   Result == yes
-	).
+	print_message(information, time(UsedInf, UsedTime, Wall, Lips)).
+
+time_state(t(Wall, Time, Inferences)) :-
+	get_time(Wall),
+	statistics(cputime, Time),
+	statistics(inferences, Inferences).
+
+time_true(State0) :-
+	report(State0, 12).		% leave choice-point
+time_true(State) :-
+	get_time(Wall),
+	statistics(cputime, Time),
+	statistics(inferences, Inferences0),
+	plus(Inferences0, -3, Inferences),
+	nb_setarg(1, State, Wall),
+	nb_setarg(2, State, Time),
+	nb_setarg(3, State, Inferences),
+	fail.
 
 
 		 /*******************************
@@ -178,17 +302,28 @@ prof_statistics(nodes, Term, Ticks) :-
 	arg(4, Term, Ticks).
 
 
-%	prof_node(+KeyOn
-%		  Key-node(Pred,
-%		           TimeSelf, TimeSiblings,
-%		           Calls, Redo, Recursive,
-%		           Parents))
+%%	prof_node(+Field, -Pairs) is nondet.
+%
+%	Collect data for each of the interesting predicate.
+%
+%	@param Field specifies the field to use as key in each pair.
+%	@param Pair is a term of the following format:
+%
+%	  ==
+%	  KeyValue-node(Pred,
+%		        TimeSelf, TimeSiblings,
+%			Calls, Redo, Recursive,
+%		        Parents)
+%	  ==
 %
-%	Collect data for each of the interesting predicates.
 
 prof_node(KeyOn, Node) :-
-	style_check(+dollar),
-	call_cleanup(get_prof_node(KeyOn, Node), style_check(-dollar)).
+	setup_call_cleanup(
+	    ( current_prolog_flag(access_level, Old),
+	      set_prolog_flag(access_level, system)
+	    ),
+	    get_prof_node(KeyOn, Node),
+	    set_prolog_flag(access_level, Old)).
 
 get_prof_node(KeyOn, Key-Node) :-
 	Node = node(M:H,
@@ -271,7 +406,7 @@ predicate_name(H, Name) :-
 	prolog:message/3.
 
 prolog:message(time(UsedInf, UsedTime, Wall, Lips)) -->
-	[ '~D inferences, ~2f CPU in ~2f seconds (~w% CPU, ~w Lips)'-
+	[ '~D inferences, ~3f CPU in ~3f seconds (~w% CPU, ~w Lips)'-
 	  [UsedInf, UsedTime, Wall, Perc, Lips] ],
 	{   Wall > 0
 	->  Perc is round(100*UsedTime/Wall)
