@@ -19,7 +19,7 @@
 
     You should have received a copy of the GNU Lesser General Public
     License along with this library; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 
     As a special exception, if you link this library with other files,
     compiled with a Free Software compiler, to produce an executable, this
@@ -32,7 +32,7 @@
 :- module(url,
 	  [ parse_url/2,		% +URL, -Parts | -URL +Parts
 	    parse_url/3,		% +URL|URI, +BaseURL, -Parts
-	    				% -URL, +BaseURL, +Parts
+					% -URL, +BaseURL, +Parts
 	    is_absolute_url/1,		% +URL
 	    global_url/3,		% +Local, +Base, -Global
 	    http_location/2,		% ?Parts, ?Location
@@ -64,6 +64,8 @@ of the standard encountered in practical use.
 
 @author	Jan Wielemaker
 @author Lukas Faulstich
+@deprecated New code should use library(uri), provided by the =clib=
+	    package.
 */
 
 		 /*******************************
@@ -108,7 +110,7 @@ is_absolute_url(URL) :-
 is_absolute_url(URL) :-
 	sub_atom(URL, 0, _, _, 'ftp://'), !.
 is_absolute_url(URL) :-
-	sub_atom(URL, 0, _, _, 'file:'), !.
+	sub_atom(URL, 0, _, _, 'file://'), !.
 is_absolute_url(URL) :-
 	atom_codes(URL, Codes),
 	phrase(absolute_url, Codes, _), !.
@@ -174,7 +176,7 @@ curi(A) -->
 cpath(A) -->
 	(   { memberchk(path(Path), A) }
 	->  { atom_codes(Path, Codes) },
-	    www_encode(Codes, "/+:")
+	    www_encode(Codes, "/+:,")
 	;   ""
 	).
 
@@ -217,7 +219,7 @@ csearch(A)-->
 csearch([], _) -->
 	[].
 csearch([Parameter|Parameters], Sep) --> !,
-	Sep,
+	codes(Sep),
 	cparam(Parameter),
 	csearch(Parameters, "&").
 
@@ -235,6 +237,8 @@ cparam(NameValue) -->			% allow to feed Name(Value)
 cparam(Name)-->
 	cname(Name).
 
+codes([]) --> [].
+codes([H|T]) --> [H], codes(T).
 
 cname(Atom) -->
 	{ atom_codes(Atom, Codes) },
@@ -251,6 +255,8 @@ cvalue(Value) -->
 	},
 	www_encode(Codes, "").
 cvalue(Codes) -->
+	{ must_be(codes, Codes)
+	},
 	www_encode(Codes, "").
 
 
@@ -271,7 +277,6 @@ cfragment(_) -->
 		 *******************************/
 
 %%	parse_url(+URL, -Attributes) is det.
-%%	parse_url(+URL, +BaseURL, -Attributes) is det.
 %
 %	Construct or analyse a URL. URL is an   atom  holding a URL or a
 %	variable. Parts is a list of   components.  Each component is of
@@ -292,8 +297,8 @@ cfragment(_) -->
 %	    * port(Port)
 %	    Integer port-number to access on   the \arg{Host}. This only
 %	    appears if the port is  explicitly   specified  in  the URL.
-%	    Implicit default ports (e.g.  80   for  HTTP)  do \emph{not}
-%	    appear in the part-list.
+%	    Implicit default ports (e.g., 80 for   HTTP) do _not_ appear
+%	    in the part-list.
 %
 %	    * path(Path)
 %	    (File-) path addressed by the URL. This is supported for the
@@ -337,6 +342,11 @@ parse_url(URL, Attributes) :-
 	phrase(curl(Attributes), Codes), !,
 	atom_codes(URL, Codes).
 
+%%	parse_url(+URL, +BaseURL, -Attributes) is det.
+%
+%	Similar to parse_url/2 for relative URLs.  If URL is relative,
+%	it is resolved using the absolute URL BaseURL.
+
 parse_url(URL, BaseURL, Attributes) :-
 	nonvar(URL), !,
 	atom_codes(URL, Codes),
@@ -751,7 +761,7 @@ fragment_char(0']) --> "]", !.
 %
 %	Performs UTF-8 decoding of percent encoded strings.
 
-pchar(0' ) --> "+", !.			%' ?
+pchar(0'\s) --> "+", !.
 pchar(C) -->
 	[C],
 	{   unreserved(C)
@@ -832,10 +842,16 @@ encoding used with the HTTP GET.
 %%	www_form_encode(+Value, -XWWWFormEncoded) is det.
 %%	www_form_encode(-Value, +XWWWFormEncoded) is det.
 %
-%	En/Decode        between        native          value        and
-%	application/x-www-form-encoded. Maps space to   +,  keeps alnum,
-%	maps anything else to =|%XX|= and   newlines to =|%OD%OA|=. When
-%	decoding, newlines appear as a single newline (10) character.
+%	En/decode   to/from   application/x-www-form-encoded.   Encoding
+%	encodes all characters  except  RFC   3986  _unreserved_  (ASCII
+%	=alnum= (see code_type/2)), and  one   of  "-._~"  using percent
+%	encoding.  Newline  is  mapped  to  =|%OD%OA|=.  When  decoding,
+%	newlines appear as a single newline (10) character.
+%
+%	Note that a space  is  encoded   as  =|%20|=  instead  of =|+|=.
+%	Decoding decodes both to a space.
+%
+%	@deprecated Use uri_encoded/3 for new code.
 
 www_form_encode(Value, Encoded) :-
 	atomic(Value), !,
@@ -869,7 +885,7 @@ percent_encode(C, Extra) -->
 	[C].
 %percent_encode(0' , _) --> !, "+".	% Deprecated: use %20
 percent_encode(C, _) -->
-	{ C =< 128 }, !,
+	{ C =< 127 }, !,
 	percent_byte(C).
 percent_encode(C, _) -->		% Unicode characters
 	{ current_prolog_flag(url_encoding, utf8), !,
@@ -877,7 +893,11 @@ percent_encode(C, _) -->		% Unicode characters
 	},
 	percent_bytes(Bytes).
 percent_encode(C, _) -->
+	{ C =< 255 }, !,
 	percent_byte(C).
+percent_encode(_C, _) -->
+	{ representation_error(url_character)
+	}.
 
 percent_bytes([]) -->
 	"".
@@ -943,7 +963,7 @@ utf8_cont([]) -->
 %	@tbd	Having a global flag is highly inconvenient, but a
 %		work-around for old sites using ISO Latin 1 encoding.
 
-:- set_prolog_flag(url_encoding, utf8).
+:- create_prolog_flag(url_encoding, utf8, [type(atom)]).
 
 set_url_encoding(Old, New) :-
 	current_prolog_flag(url_encoding, Old),
@@ -1009,6 +1029,7 @@ parse_url_search(Codes, Fields) :-
 	is_list(Codes), !,
 	phrase(search(Fields), Codes).
 parse_url_search(Codes, Fields) :-
+	must_be(list, Fields),
 	phrase(csearch(Fields, ""), Codes).
 
 
