@@ -19,7 +19,7 @@
 
     You should have received a copy of the GNU Lesser General Public
     License along with this library; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 
     As a special exception, if you link this library with other files,
     compiled with a Free Software compiler, to produce an executable, this
@@ -38,6 +38,7 @@
 	  lookup_ht1/4,
 	  lookup_ht2/4,
 	  insert_ht/3,
+	  insert_ht1/4,
 	  insert_ht/4,
 	  delete_ht/3,
 	  delete_ht1/4,
@@ -48,16 +49,12 @@
 	]).
 
 :- use_module(pairlist).
-:- use_module(hprolog).
+:- use_module(library(dialect/hprolog)).
 :- use_module(library(lists)).
 
 :- multifile user:goal_expansion/2.
 :- dynamic user:goal_expansion/2.
 
-user:goal_expansion(term_hash(Term,Hash),hash_term(Term,Hash)).
-
-% term_hash(Term,Hash) :-
-% 	hash_term(Term,Hash).
 initial_capacity(89).
 
 new_ht(HT) :-
