@@ -1,4 +1,4 @@
-/*  $Id: chr_swi.pl,v 1.33 2007/03/08 08:51:08 toms Exp $
+/*  $Id$
 
     Part of CHR (Constraint Handling Rules)
 
@@ -19,7 +19,7 @@
 
     You should have received a copy of the GNU Lesser General Public
     License along with this library; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 
     As a special exception, if you link this library with other files,
     compiled with a Free Software compiler, to produce an executable, this
@@ -43,6 +43,7 @@
 	    op(1190, xfx, pragma),
 	    op( 500, yfx, #),
 	    op(1150, fx, chr_type),
+	    op(1150, fx, chr_declaration),
 	    op(1130, xfx, --->),
 	    op(1150, fx, (?)),
 	    chr_show_store/1,		% +Module
@@ -52,6 +53,8 @@
 	    chr_leash/1			% +Ports
 	  ]).
 
+:- expects_dialect(swi).
+
 :- set_prolog_flag(generate_debug_info, false).
 
 :- multifile user:file_search_path/2.
@@ -81,26 +84,26 @@ user:file_search_path(chr, library(chr)).
 %%	chr_flag/3,
 %%	chr_show_store/1
 %%	]).
-%% 
+%%
 %% :- op(1180, xfx, ==>),
-%% 	op(1180, xfx, <=>),
-%% 	op(1150, fx, constraints),
-%% 	op(1150, fx, handler),
-%% 	op(1150, fx, rules),
-%% 	op(1100, xfx, \),
-%% 	op(1200, xfx, @),
-%% 	op(1190, xfx, pragma),
-%% 	op( 500, yfx, #),
-%% 	op(1150, fx, chr_type),
-%% 	op(1130, xfx, --->),
-%% 	op(1150, fx, (?)).
-%% 
+%%	op(1180, xfx, <=>),
+%%	op(1150, fx, constraints),
+%%	op(1150, fx, handler),
+%%	op(1150, fx, rules),
+%%	op(1100, xfx, \),
+%%	op(1200, xfx, @),
+%%	op(1190, xfx, pragma),
+%%	op( 500, yfx, #),
+%%	op(1150, fx, chr_type),
+%%	op(1130, xfx, --->),
+%%	op(1150, fx, (?)).
+%%
 %% :- multifile user:file_search_path/2.
 %% :- dynamic   chr_translated_program/1.
-%% 
+%%
 %% user:file_search_path(chr, library(chr)).
-%% 
-%% 
+%%
+%%
 %% :- use_module('chr/chr_translate').
 %% :- use_module('chr/chr_runtime').
 %% :- use_module('chr/chr_hashtable_store').
@@ -114,7 +117,7 @@ user:file_search_path(chr, library(chr)).
 :- dynamic chr_pp/2.		% File, Term
 
 %	chr_expandable(+Term)
-%	
+%
 %	Succeeds if Term is a  rule  that   must  be  handled by the CHR
 %	compiler. Ideally CHR definitions should be between
 %
@@ -123,13 +126,14 @@ user:file_search_path(chr, library(chr)).
 %		:- end_constraints.
 %
 %	As they are not we have to   use  some heuristics. We assume any
-%	file is a CHR after we've seen :- constraints ... 
+%	file is a CHR after we've seen :- constraints ...
 
 chr_expandable((:- constraints _)).
 chr_expandable((constraints _)).
 chr_expandable((:- chr_constraint _)).
 chr_expandable((:- chr_type _)).
 chr_expandable((chr_type _)).
+chr_expandable((:- chr_declaration _)).
 chr_expandable(option(_, _)).
 chr_expandable((:- chr_option(_, _))).
 chr_expandable((handler _)).
@@ -140,7 +144,7 @@ chr_expandable((_ ==> _)).
 chr_expandable((_ pragma _)).
 
 %	chr_expand(+Term, -Expansion)
-%	
+%
 %	Extract CHR declarations and rules from the file and run the
 %	CHR compiler when reaching end-of-file.
 
@@ -153,24 +157,24 @@ extra_declarations([(:- use_module(chr(chr_runtime)))
 
 %% SICStus begin
 %% extra_declarations([(:-use_module(chr(chr_runtime)))
-%% 		     , (:- use_module(chr(hprolog),[term_variables/2,term_variables/3]))
-%% 		     , (:-use_module(chr(hpattvars)))
-%% 		     | Tail], Tail).		   
+%%		     , (:- use_module(chr(hprolog),[term_variables/2,term_variables/3]))
+%%		     , (:-use_module(chr(hpattvars)))
+%%		     | Tail], Tail).
 %% SICStus end
 
 chr_expand(Term, []) :-
 	chr_expandable(Term), !,
-	prolog_load_context(file,File),
+	prolog_load_context(source,File),
 	prolog_load_context(term_position,'$stream_position'(_, LineNumber, _, _, _)),
 	add_pragma_to_chr_rule(Term,line_number(LineNumber),NTerm),
 	assert(chr_term(File, LineNumber, NTerm)).
 chr_expand(Term, []) :-
 	Term = (:- chr_preprocessor Preprocessor), !,
-	prolog_load_context(file,File),
+	prolog_load_context(source,File),
 	assert(chr_pp(File, Preprocessor)).
 chr_expand(end_of_file, FinalProgram) :-
 	extra_declarations(FinalProgram,Program),
-	prolog_load_context(file,File),
+	prolog_load_context(source,File),
 	findall(T, retract(chr_term(File,_Line,T)), CHR0),
 	CHR0 \== [],
 	prolog_load_context(module, Module),
@@ -226,7 +230,7 @@ add_optimise_decl(CHR, CHR).
 
 
 %	call_chr_translate(+File, +In, -Out)
-%	
+%
 %	The entire chr_translate/2 translation may fail, in which case we'd
 %	better issue a warning  rather  than   simply  ignoring  the CHR
 %	declarations.
@@ -247,7 +251,7 @@ call_chr_preprocessor(Preprocessor,CHR,_NCHR) :-
 	( call(Preprocessor,CHR,CHR0) ->
 		nb_setval(chr_preprocessed_program,CHR0),
 		fail
- 	).
+	).
 call_chr_preprocessor(_,_,NCHR)	:-
 	nb_current(chr_preprocessed_program,NCHR), !,
 	nb_delete(chr_preprocessed_program).
@@ -275,9 +279,9 @@ user:message_hook(trace_mode(OnOff), _, _) :-
 	fail.				% backtrack to other handlers
 
 %	chr:debug_event(+State, +Event)
-%	
+%
 %	Hook into the CHR debugger.  At this moment we will discard CHR
-%	events if we are in a Prolog `skip' and we ignore the 
+%	events if we are in a Prolog `skip' and we ignore the
 
 chr:debug_event(_State, _Event) :-
 	tracing,			% are we tracing?
@@ -288,7 +292,7 @@ chr:debug_event(_State, _Event) :-
 	Level > Skip, !.
 
 %	chr:debug_interact(+Event, +Depth, -Command)
-%	
+%
 %	Hook into the CHR debugger to display Event and ask for the next
 %	command to execute. This  definition   causes  the normal Prolog
 %	debugger to be used for the standard ports.
@@ -315,10 +319,10 @@ prolog:message(chr(CHR)) -->
 	chr_message(CHR).
 
 		 /*******************************
-		 *	 TOPLEVEL PRINTING	*	
+		 *	 TOPLEVEL PRINTING	*
 		 *******************************/
 
-:- set_prolog_flag(chr_toplevel_show_store,true).
+:- create_prolog_flag(chr_toplevel_show_store, true, []).
 
 prolog:message(query(YesNo)) --> !,
 	['~@'-[chr:print_all_stores]],
@@ -342,89 +346,89 @@ print_all_stores :-
 		 *	   MUST BE LAST!	*
 		 *******************************/
 
-:- multifile user:term_expansion/2.
-:- dynamic   user:term_expansion/2.
+:- multifile system:term_expansion/2.
+:- dynamic   system:term_expansion/2.
 
-user:term_expansion(In, Out) :-
+system:term_expansion(In, Out) :-
 	chr_expand(In, Out).
 %% SWI end
 
 %% SICStus begin
-% 
+%
 % :- dynamic
-% 	current_toplevel_show_store/1,
-% 	current_generate_debug_info/1,
-% 	current_optimize/1.
-% 
+%	current_toplevel_show_store/1,
+%	current_generate_debug_info/1,
+%	current_optimize/1.
+%
 % current_toplevel_show_store(on).
-% 
+%
 % current_generate_debug_info(false).
-% 
+%
 % current_optimize(off).
-% 
+%
 % chr_current_prolog_flag(generate_debug_info, X) :-
-% 	chr_flag(generate_debug_info, X, X).
+%	chr_flag(generate_debug_info, X, X).
 % chr_current_prolog_flag(optimize, X) :-
-% 	chr_flag(optimize, X, X).
-% 
+%	chr_flag(optimize, X, X).
+%
 % chr_flag(Flag, Old, New) :-
-% 	Goal = chr_flag(Flag,Old,New),
-% 	g must_be(Flag, oneof([toplevel_show_store,generate_debug_info,optimize]), Goal, 1),
-% 	chr_flag(Flag, Old, New, Goal).
-% 
+%	Goal = chr_flag(Flag,Old,New),
+%	g must_be(Flag, oneof([toplevel_show_store,generate_debug_info,optimize]), Goal, 1),
+%	chr_flag(Flag, Old, New, Goal).
+%
 % chr_flag(toplevel_show_store, Old, New, Goal) :-
-% 	clause(current_toplevel_show_store(Old), true, Ref),
-% 	(   New==Old -> true
-% 	;   must_be(New, oneof([on,off]), Goal, 3),
-% 	    erase(Ref),
-% 	    assertz(current_toplevel_show_store(New))
-% 	).
+%	clause(current_toplevel_show_store(Old), true, Ref),
+%	(   New==Old -> true
+%	;   must_be(New, oneof([on,off]), Goal, 3),
+%	    erase(Ref),
+%	    assertz(current_toplevel_show_store(New))
+%	).
 % chr_flag(generate_debug_info, Old, New, Goal) :-
-% 	clause(current_generate_debug_info(Old), true, Ref),
-% 	(   New==Old -> true
-% 	;   must_be(New, oneof([false,true]), Goal, 3),
-% 	    erase(Ref),
-% 	    assertz(current_generate_debug_info(New))
-% 	).
+%	clause(current_generate_debug_info(Old), true, Ref),
+%	(   New==Old -> true
+%	;   must_be(New, oneof([false,true]), Goal, 3),
+%	    erase(Ref),
+%	    assertz(current_generate_debug_info(New))
+%	).
 % chr_flag(optimize, Old, New, Goal) :-
-% 	clause(current_optimize(Old), true, Ref),
-% 	(   New==Old -> true
-% 	;   must_be(New, oneof([full,off]), Goal, 3),
-% 	    erase(Ref),
-% 	    assertz(current_optimize(New))
-% 	).
-% 
-% 
+%	clause(current_optimize(Old), true, Ref),
+%	(   New==Old -> true
+%	;   must_be(New, oneof([full,off]), Goal, 3),
+%	    erase(Ref),
+%	    assertz(current_optimize(New))
+%	).
+%
+%
 % all_stores_goal(Goal, CVAs) :-
-% 	chr_flag(toplevel_show_store, on, on), !,
-% 	findall(C-CVAs, find_chr_constraint(C), Pairs),
-% 	andify(Pairs, Goal, CVAs).
+%	chr_flag(toplevel_show_store, on, on), !,
+%	findall(C-CVAs, find_chr_constraint(C), Pairs),
+%	andify(Pairs, Goal, CVAs).
 % all_stores_goal(true, _).
-% 
+%
 % andify([], true, _).
 % andify([X-Vs|L], Conj, Vs) :- andify(L, X, Conj, Vs).
-% 
+%
 % andify([], X, X, _).
 % andify([Y-Vs|L], X, (X,Conj), Vs) :- andify(L, Y, Conj, Vs).
-% 
+%
 % :- multifile user:term_expansion/6.
-% 
+%
 % user:term_expansion(In, _, Ids, Out, [], [chr|Ids]) :-
-% 	nonvar(In),
-% 	nonmember(chr, Ids),
-% 	chr_expand(In, Out), !.
-% 
+%	nonvar(In),
+%	nonmember(chr, Ids),
+%	chr_expand(In, Out), !.
+%
 %% SICStus end
 
 %%% for SSS %%%
 
-add_pragma_to_chr_rule(Name @ Rule, Pragma, Result) :- !,
+add_pragma_to_chr_rule((Name @ Rule), Pragma, Result) :- !,
 	add_pragma_to_chr_rule(Rule,Pragma,NRule),
 	Result = (Name @ NRule).
-add_pragma_to_chr_rule(Rule pragma Pragmas, Pragma, Result) :- !,
+add_pragma_to_chr_rule((Rule pragma Pragmas), Pragma, Result) :- !,
 	Result = (Rule pragma (Pragma,Pragmas)).
-add_pragma_to_chr_rule(Head ==> Body, Pragma, Result) :- !,
+add_pragma_to_chr_rule((Head ==> Body), Pragma, Result) :- !,
 	Result = (Head ==> Body pragma Pragma).
-add_pragma_to_chr_rule(Head <=> Body, Pragma, Result) :- !,
+add_pragma_to_chr_rule((Head <=> Body), Pragma, Result) :- !,
 	Result = (Head <=> Body pragma Pragma).
 add_pragma_to_chr_rule(Term,_,Term).
