@@ -1,11 +1,10 @@
-/*  $Id$
-
-    Part of SWI-Prolog
+/*  Part of SWI-Prolog
 
     Author:        Jan Wielemaker
-    E-mail:        jan@swi.psy.uva.nl
+    E-mail:        J.Wielemaker@cs.vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (C): 1985-2002, University of Amsterdam
+    Copyright (C): 2002-2010, University of Amsterdam
+			      Vu University Amsterdam
 
     This library is free software; you can redistribute it and/or
     modify it under the terms of the GNU Lesser General Public
@@ -19,7 +18,7 @@
 
     You should have received a copy of the GNU Lesser General Public
     License along with this library; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */
 
 #ifdef HAVE_CONFIG_H
@@ -40,10 +39,52 @@
 #endif
 #include <errno.h>
 
+#ifdef __WINDOWS__
+
+#if defined(__MINGW32__)
+#define WINVER 0x0501
+#endif
+
+#define MAKE_FUNCTORS 1
+#include <windows.h>
+#include "win_error.c"
+#define statstruct struct _stati64
+#define statfunc _wstati64
+/*#define HAVE_UTIME 1:	Broken ...*/
+#ifdef HAVE_UTIME
+#include <sys/utime.h>
+/* FIXME: under __MINGW32__ we could use `utimbuf' (no leading underscore. */
+#define utimestruct struct _utimbuf
+#define utimefunc _wutime
+#endif
+#define FCHAR wchar_t
+#define PL_get_file_name PL_get_file_nameW
+
+#if defined(__MINGW32__)
+#include <winbase.h>
+#endif
+
+#else /*__WINDOWS__*/
+
+#define statstruct struct stat
+#define statfunc stat
+#define utimestruct struct utimbuf
+#define utimefunc utime
+#define FCHAR char
+
+#endif /*__WINDOWS__*/
+
 static functor_t FUNCTOR_access1;
 static functor_t FUNCTOR_modified1;
 static functor_t FUNCTOR_changed1;
 static atom_t    ATOM_now;
+static atom_t    ATOM_hard;
+static atom_t    ATOM_symbolic;
+
+
+		 /*******************************
+		 *	       TIME		*
+		 *******************************/
 
 static int
 add_time_option(term_t list, functor_t f, time_t time)
@@ -53,9 +94,9 @@ add_time_option(term_t list, functor_t f, time_t time)
   while(PL_get_list(tail, head, tail))
   { if ( PL_unify_functor(head, f) )
     { term_t a = PL_new_term_ref();
-      
-      PL_get_arg(1, head, a);
-      return PL_unify_float(a, (double)time);
+
+      return (PL_get_arg(1, head, a) &&
+	      PL_unify_float(a, (double)time));
     }
   }
 
@@ -87,8 +128,8 @@ get_time_option(term_t list, functor_t f, time_t def, time_t *tme)
   { if ( PL_is_functor(head, f) )
     { term_t a = PL_new_term_ref();
       double f;
-      
-      PL_get_arg(1, head, a);
+
+      _PL_get_arg(1, head, a);
       if ( !PL_get_float(a, &f) )
       { atom_t now;
 
@@ -110,19 +151,19 @@ get_time_option(term_t list, functor_t f, time_t def, time_t *tme)
 
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 set_file_time(+Spec, -Old, +New)
-    Query/set file-times.  
+    Query/set file-times.
 
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
 static foreign_t
 pl_set_time_file(term_t spec, term_t old, term_t new)
-{ char *name;
-  struct stat sbuf;
+{ FCHAR *name;
+  statstruct sbuf;
 
   if ( !PL_get_file_name(spec, &name, 0) )
     return FALSE;
 
-  if ( stat(name, &sbuf) )
+  if ( statfunc(name, &sbuf) )
     return pl_error(NULL, 0, NULL, ERR_ERRNO, errno, "stat", "file", spec);
 
   add_time_option(old, FUNCTOR_access1,   sbuf.st_atime);
@@ -132,7 +173,7 @@ pl_set_time_file(term_t spec, term_t old, term_t new)
 
   if ( !PL_get_nil(new) )
 #ifdef HAVE_UTIME
-  { struct utimbuf tbuf;
+  { utimestruct tbuf;
 
     if ( !get_time_option(new, FUNCTOR_access1,
 			  sbuf.st_atime, &tbuf.actime) )
@@ -141,7 +182,7 @@ pl_set_time_file(term_t spec, term_t old, term_t new)
 			  sbuf.st_mtime, &tbuf.modtime) )
       return FALSE;
 
-    if ( utime(name, &tbuf) != 0 )
+    if ( utimefunc(name, &tbuf) != 0 )
       return pl_error(NULL, 0, NULL, ERR_ERRNO, errno, "set_time", "file", spec);
   }
 #else
@@ -152,12 +193,77 @@ pl_set_time_file(term_t spec, term_t old, term_t new)
 }
 
 
+		 /*******************************
+		 *	       LINK		*
+		 *******************************/
+
+static foreign_t
+pl_link_file(term_t from, term_t to, term_t how)
+{ FCHAR *fname, *tname;
+  atom_t hname;
+
+  if ( !PL_get_file_name(from, &fname, PL_FILE_OSPATH) ||
+       !PL_get_file_name(to,   &tname, PL_FILE_OSPATH) )
+    return FALSE;
+
+  if ( !PL_get_atom(how, &hname) )
+    return pl_error(NULL, 0, NULL, ERR_TYPE, how, "atom");
+
+#ifdef __WINDOWS__
+
+  if ( hname == ATOM_hard )
+  { if ( !CreateHardLinkW(tname, fname, NULL) )
+      return win_error("CreateHardLink");
+  } else if ( hname == ATOM_symbolic )
+  { static int (*symlink)(wchar_t *new, wchar_t *existing, DWORD flags);
+    static int fetched = FALSE;
+
+    if ( !fetched )
+    { HMODULE hmod = GetModuleHandle("kernel32.dll");
+      void *addr = GetProcAddress(hmod, "CreateSymbolicLink");
+
+      symlink = addr;
+      fetched = TRUE;
+    }
+
+    if ( !symlink )
+      return pl_error(NULL, 0, NULL, ERR_DOMAIN, how, "link_type");
+
+    if ( !(*symlink)(tname, fname, 0) )
+      return win_error("CreateSymbolicLink");
+  } else
+    return pl_error(NULL, 0, NULL, ERR_DOMAIN, how, "link_type");
+
+#else /*__WINDOWS__*/
+
+  if ( hname == ATOM_hard )
+  { if ( link(fname, tname) != 0 )
+      return pl_error(NULL, 0, NULL, ERR_ERRNO, errno, "link", "file", to);
+  } else if ( hname == ATOM_symbolic )
+  { if ( symlink(fname, tname) != 0 )
+      return pl_error(NULL, 0, NULL, ERR_ERRNO, errno, "link", "file", to);
+  } else
+    return pl_error(NULL, 0, NULL, ERR_DOMAIN, how, "link_type");
+
+#endif /*__WINDOWS__*/
+
+  return TRUE;
+}
+
+
 install_t
 install_files()
 { FUNCTOR_access1   = PL_new_functor(PL_new_atom("access"), 1);
   FUNCTOR_modified1 = PL_new_functor(PL_new_atom("modified"), 1);
   FUNCTOR_changed1  = PL_new_functor(PL_new_atom("changed"), 1);
   ATOM_now          = PL_new_atom("now");
+  ATOM_hard         = PL_new_atom("hard");
+  ATOM_symbolic     = PL_new_atom("symbolic");
+
+#ifdef __WINDOWS__
+  win_init_errors();
+#endif
 
   PL_register_foreign("set_time_file", 3, pl_set_time_file, 0);
+  PL_register_foreign("link_file",     3, pl_link_file,     0);
 }
