@@ -19,7 +19,7 @@
 
     You should have received a copy of the GNU Lesser General Public
     License along with this library; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 
     As a special exception, if you link this library with other files,
     compiled with a Free Software compiler, to produce an executable, this
@@ -48,7 +48,7 @@
 
 #include <io.h>
 #include <winsock2.h>
-typedef size_t socklen_t;
+#include <ws2tcpip.h>
 
 #else /*__WINDOWS__*/
 
@@ -108,17 +108,17 @@ typedef enum
 } nbio_request;
 
 					/* nbio_get_flags() mask */
-#define SOCK_INSTREAM	0x01
-#define SOCK_OUTSTREAM	0x02
-#define SOCK_BIND	0x04		/* What have we done? */
-#define SOCK_LISTEN	0x08
-#define SOCK_CONNECT	0x10
-#define SOCK_ACCEPT	0x20		/* Set on accepted sockets */
-#define SOCK_NONBLOCK	0x40		/* Set to non-blocking mode */
-#define SOCK_DISPATCH   0x80		/* do not dispatch events */
-#define SOCK_CLOSE_SEEN	0x100		/* FD_CLOSE seen */
-#define SOCK_EOF_SEEN   0x200		/* Seen end-of-file */
-#define SOCK_WAITING	0x400		/* using nbio_wait() */
+#define PLSOCK_INSTREAM	  0x001
+#define PLSOCK_OUTSTREAM  0x002
+#define PLSOCK_BIND	  0x004		/* What have we done? */
+#define PLSOCK_LISTEN	  0x008
+#define PLSOCK_CONNECT	  0x010
+#define PLSOCK_ACCEPT	  0x020		/* Set on accepted sockets */
+#define PLSOCK_NONBLOCK	  0x040		/* Set to non-blocking mode */
+#define PLSOCK_DISPATCH   0x080		/* do not dispatch events */
+#define PLSOCK_CLOSE_SEEN 0x100		/* FD_CLOSE seen */
+#define PLSOCK_EOF_SEEN   0x200		/* Seen end-of-file */
+#define PLSOCK_WAITING	  0x400		/* using nbio_wait() */
 
 
 typedef int	nbio_sock_t;		/* socket handle (not a file-descr) */
@@ -128,27 +128,33 @@ typedef struct _plsocket *plsocket_ptr;	/* wrapped socket */
 		 *	 BASIC FUNCTIONS	*
 		 *******************************/
 
-extern int	nbio_init(const char *module);
-extern int	nbio_cleanup(void);
-extern int	nbio_debug(int level);
+#ifdef HAVE_DECLSPEC
+#define NBIO_EXPORT(type)		__declspec(dllexport) type
+#else
+#define NBIO_EXPORT(type)	 	extern type
+#endif
 
-extern nbio_sock_t
+NBIO_EXPORT(int)	nbio_init(const char *module);
+NBIO_EXPORT(int)	nbio_cleanup(void);
+NBIO_EXPORT(int)	nbio_debug(int level);
+
+NBIO_EXPORT(nbio_sock_t)
 		nbio_socket(int domain, int type, int protocol);
-extern int	nbio_connect(nbio_sock_t socket,
+NBIO_EXPORT(int)	nbio_connect(nbio_sock_t socket,
 			     const struct sockaddr *serv_addr,
 			     size_t addrlen);
-extern int	nbio_bind(nbio_sock_t socket,
+NBIO_EXPORT(int)	nbio_bind(nbio_sock_t socket,
 			  struct sockaddr *my_addr,
 			  size_t addrlen);
-extern int	nbio_listen(nbio_sock_t socket, int backlog);
-extern nbio_sock_t
+NBIO_EXPORT(int)	nbio_listen(nbio_sock_t socket, int backlog);
+NBIO_EXPORT(nbio_sock_t)
 		nbio_accept(nbio_sock_t master,
 			    struct sockaddr *addr,
 			    socklen_t *addrlen);
 
 extern ssize_t	nbio_read(nbio_sock_t socket, char *buf, size_t bufSize);
 extern ssize_t 	nbio_write(nbio_sock_t socket, char *buf, size_t bufSize);
-extern int	nbio_closesocket(nbio_sock_t socket);
+NBIO_EXPORT(int) nbio_closesocket(nbio_sock_t socket);
 extern int 	nbio_close_input(nbio_sock_t socket);
 extern int 	nbio_close_output(nbio_sock_t socket);
 extern ssize_t	nbio_recvfrom(int socket, void *buf, size_t bufSize, int flags,
@@ -156,8 +162,8 @@ extern ssize_t	nbio_recvfrom(int socket, void *buf, size_t bufSize, int flags,
 extern ssize_t	nbio_sendto(nbio_sock_t socket, void *buf, size_t bufSize, int flags,
 			    const struct sockaddr *to, socklen_t tolen);
 
-extern int	nbio_wait(nbio_sock_t socket, nbio_request);
-extern SOCKET	nbio_fd(nbio_sock_t socket);
+NBIO_EXPORT(int)	nbio_wait(nbio_sock_t socket, nbio_request);
+NBIO_EXPORT(SOCKET)	nbio_fd(nbio_sock_t socket);
 extern int	nbio_select(int n,
 			    fd_set *readfds,
 			    fd_set *writefds,
@@ -170,7 +176,7 @@ extern int	nbio_get_ip(term_t ip4, struct in_addr *ip);
 extern int	nbio_error(int code, nbio_error_map map);
 extern const char*
 		nbio_last_error(nbio_sock_t socket);
-extern int	nbio_setopt(int socket, nbio_option opt, ...);
+NBIO_EXPORT(int)	nbio_setopt(int socket, nbio_option opt, ...);
 extern int	nbio_get_flags(int socket);
 
 					/* support tipc_setopt() */
