@@ -19,7 +19,7 @@
 
     You should have received a copy of the GNU Lesser General Public
     License along with this library; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */
 
 #ifdef HAVE_CONFIG_H
@@ -79,6 +79,26 @@ static functor_t FUNCTOR_eq2;		/* =/2 */
 #define DEBUG(g) (void)0
 #endif
 
+#ifdef __WINDOWS__
+#include <windows.h>
+#include <stdio.h>
+#include <fcntl.h>
+#include <io.h>
+
+#if !defined(__MINGW32__)
+typedef DWORD  pid_t;
+#endif
+
+typedef wchar_t echar;			/* environment character */
+
+#ifndef CREATE_BREAKAWAY_FROM_JOB
+#define CREATE_BREAKAWAY_FROM_JOB 0x1000000
+#endif
+
+#else
+typedef char echar;
+#endif
+
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 ISSUES:
 	- Deal with child errors (no cwd, cannot execute, etc.)
@@ -93,42 +113,51 @@ ISSUES:
 
 static int
 type_error(term_t actual, const char *expected)
-{ term_t ex = PL_new_term_ref();
+{ term_t ex;
 
-  PL_unify_term(ex, PL_FUNCTOR, FUNCTOR_error2,
-		      PL_FUNCTOR, FUNCTOR_type_error2,
-		        PL_CHARS, expected,
-		        PL_TERM, actual,
-		      PL_VARIABLE);
+  if ( (ex=PL_new_term_ref()) &&
+       PL_unify_term(ex,
+		     PL_FUNCTOR, FUNCTOR_error2,
+		       PL_FUNCTOR, FUNCTOR_type_error2,
+		         PL_CHARS, expected,
+		         PL_TERM, actual,
+		       PL_VARIABLE) )
+    return PL_raise_exception(ex);
 
-  return PL_raise_exception(ex);
+  return FALSE;
 }
 
 
 static int
 domain_error(term_t actual, const char *expected)
-{ term_t ex = PL_new_term_ref();
+{ term_t ex;
 
-  PL_unify_term(ex, PL_FUNCTOR, FUNCTOR_error2,
-		      PL_FUNCTOR, FUNCTOR_domain_error2,
-		        PL_CHARS, expected,
-		        PL_TERM, actual,
-		      PL_VARIABLE);
+  if ( (ex=PL_new_term_ref()) &&
+       PL_unify_term(ex,
+		     PL_FUNCTOR, FUNCTOR_error2,
+		       PL_FUNCTOR, FUNCTOR_domain_error2,
+		         PL_CHARS, expected,
+		         PL_TERM, actual,
+		       PL_VARIABLE) )
+    return PL_raise_exception(ex);
 
-  return PL_raise_exception(ex);
+  return FALSE;
 }
 
 
 static int
 resource_error(const char *resource)
-{ term_t ex = PL_new_term_ref();
+{ term_t ex;
 
-  PL_unify_term(ex, PL_FUNCTOR, FUNCTOR_error2,
-		      PL_FUNCTOR, FUNCTOR_resource_error1,
-		        PL_CHARS, resource,
-		      PL_VARIABLE);
+  if ( (ex=PL_new_term_ref()) &&
+       PL_unify_term(ex,
+		     PL_FUNCTOR, FUNCTOR_error2,
+		       PL_FUNCTOR, FUNCTOR_resource_error1,
+		         PL_CHARS, resource,
+		       PL_VARIABLE) )
+    return PL_raise_exception(ex);
 
-  return PL_raise_exception(ex);
+  return FALSE;
 }
 
 
@@ -136,17 +165,6 @@ resource_error(const char *resource)
 		 *	       ADMIN		*
 		 *******************************/
 
-#ifdef __WINDOWS__
-#include <windows.h>
-#include <stdio.h>
-#include <fcntl.h>
-#include <io.h>
-typedef DWORD  pid_t;
-typedef wchar_t echar;			/* environment character */
-#else
-typedef char echar;
-#endif
-
 typedef enum std_type
 { std_std,
   std_null,
@@ -249,7 +267,8 @@ static int
 get_echars_arg_ex(int i, term_t from, term_t arg, echar **sp, size_t *lenp)
 { const echar *s, *e;
 
-  PL_get_arg(i, from, arg);
+  if ( !PL_get_arg(i, from, arg) )
+    return FALSE;
 
 #ifdef __WINDOWS__
   if ( !PL_get_wchars(arg, lenp, sp,
@@ -345,7 +364,7 @@ get_stream(term_t t, p_options *info, p_stream *stream)
     }
   } else if ( PL_is_functor(t, FUNCTOR_pipe1) )
   { stream->term = PL_new_term_ref();
-    PL_get_arg(1, t, stream->term);
+    _PL_get_arg(1, t, stream->term);
     stream->type = std_pipe;
     info->pipes++;
     return TRUE;
@@ -368,7 +387,7 @@ parse_options(term_t options, p_options *info)
 
     if ( !PL_get_name_arity(head, &name, &arity) || arity != 1 )
       return type_error(head, "option");
-    PL_get_arg(1, head, arg);
+    _PL_get_arg(1, head, arg);
 
     if ( name == ATOM_stdin )
     { if ( !get_stream(arg, info, &info->streams[0]) )
@@ -431,13 +450,17 @@ get_exe(term_t exe, p_options *info)
   if ( !PL_get_chars(arg, &info->exe, CVT_ATOM|CVT_EXCEPTION|BUF_MALLOC|REP_FN) )
     return FALSE;
 
-  info->argv = PL_malloc((arity+2)*sizeof(char*));
+  if ( !(info->argv = PL_malloc((arity+2)*sizeof(char*))) )
+    return PL_resource_error("memory");
   memset(info->argv, 0, (arity+2)*sizeof(char*));
-  info->argv[0] = strdup(info->exe);
+  if ( !(info->argv[0] = PL_malloc(strlen(info->exe)+1)) )
+    return PL_resource_error("memory");
+  strcpy(info->argv[0], info->exe);
+
   { int i;
 
     for(i=1; i<=arity; i++)
-    { PL_get_arg(i, exe, arg);
+    { _PL_get_arg(i, exe, arg);
 
       if ( !PL_get_chars(arg, &info->argv[i],
 			 CVT_ATOMIC|CVT_EXCEPTION|BUF_MALLOC|REP_FN) )
@@ -641,57 +664,7 @@ win_init()
 }
 
 
-static atom_t
-WinError()
-{ int id = GetLastError();
-  char *msg;
-  static WORD lang;
-  static lang_initialised = 0;
-
-  if ( !lang_initialised )
-    lang = MAKELANGID(LANG_ENGLISH, SUBLANG_ENGLISH_UK);
-
-again:
-  if ( FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER|
-		     FORMAT_MESSAGE_IGNORE_INSERTS|
-		     FORMAT_MESSAGE_FROM_SYSTEM,
-		     NULL,			/* source */
-		     id,			/* identifier */
-		     lang,
-		     (LPTSTR) &msg,
-		     0,				/* size */
-		     NULL) )			/* arguments */
-  { atom_t a = PL_new_atom(msg);
-
-    LocalFree(msg);
-    lang_initialised = 1;
-
-    return a;
-  } else
-  { if ( lang_initialised == 0 )
-    { lang = MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT);
-      lang_initialised = 1;
-      goto again;
-    }
-
-    return PL_new_atom("Unknown Windows error");
-  }
-}
-
-
-static int
-win_error(const char *op)
-{ atom_t msg = WinError();
-  term_t ex = PL_new_term_ref();
-
-  PL_unify_term(ex, PL_FUNCTOR, FUNCTOR_error2,
-		      PL_FUNCTOR, FUNCTOR_system_error2,
-		        PL_CHARS, op,
-		        PL_ATOM, msg,
-		      PL_VARIABLE);
-
-  return PL_raise_exception(ex);
-}
+#include "win_error.c"
 
 
 typedef struct arg_string
@@ -1061,6 +1034,9 @@ do_create_process(p_options *info)
       break;
   }
 
+  if ( info->detached )
+    flags |= CREATE_BREAKAWAY_FROM_JOB;
+
   memset(&si, 0, sizeof(si));
   si.cb = sizeof(si);
   si.dwFlags = STARTF_USESTDHANDLES;
@@ -1230,10 +1206,10 @@ wait_for_pid(pid_t pid, term_t code, wait_options *opts)
     { term_t PID;
 
     error:
-      PID = PL_new_term_ref();
-      PL_put_integer(PID, pid);
-
-      return pl_error(NULL, 0, "waitpid", ERR_ERRNO, errno, "wait", "process", PID);
+      return ((PID = PL_new_term_ref()) &&
+	      PL_put_integer(PID, pid) &&
+	      pl_error(NULL, 0, "waitpid", ERR_ERRNO,
+		       errno, "wait", "process", PID));
     }
   }
 
@@ -1282,17 +1258,19 @@ wait_success(atom_t name, pid_t pid)
     { if ( WIFEXITED(status) && WEXITSTATUS(status) == 0 )
       { return TRUE;
       } else
-      { term_t code = PL_new_term_ref();
-	term_t ex = PL_new_term_ref();
+      { term_t code, ex;
 
-	unify_exit_status(code, status);
-	PL_unify_term(ex,
-		      PL_FUNCTOR, FUNCTOR_error2,
-		        PL_FUNCTOR, FUNCTOR_process_error2,
-		          PL_ATOM, name,
-		          PL_TERM, code,
-		        PL_VARIABLE);
-	return PL_raise_exception(ex);
+	if ( (code = PL_new_term_ref()) &&
+	     (ex = PL_new_term_ref()) &&
+	     unify_exit_status(code, status) &&
+	     PL_unify_term(ex,
+			   PL_FUNCTOR, FUNCTOR_error2,
+			     PL_FUNCTOR, FUNCTOR_process_error2,
+			       PL_ATOM, name,
+			       PL_TERM, code,
+			     PL_VARIABLE) )
+	  return PL_raise_exception(ex);
+	return FALSE;
       }
     }
 
@@ -1304,16 +1282,22 @@ wait_success(atom_t name, pid_t pid)
 }
 
 
+#ifndef HAVE_VFORK
+#define vfork fork
+#endif
+
 static int
 do_create_process(p_options *info)
 { int pid;
 
-  if ( !(pid=fork()) )			/* child */
+  if ( !(pid=vfork()) )			/* child */
   { int fd;
-    int rc;
 
     PL_cleanup_fork();
 
+    if ( info->detached )
+      setsid();
+
     if ( info->cwd )
     { if ( chdir(info->cwd) )
       { perror(info->cwd);
@@ -1362,21 +1346,18 @@ do_create_process(p_options *info)
     }
 
     if ( info->envp )
-      rc = execve(info->exe, info->argv, info->envp);
+      execve(info->exe, info->argv, info->envp);
     else
-      rc = execv(info->exe, info->argv);
+      execv(info->exe, info->argv);
 
-    if ( rc )
-    { perror(info->exe);
-      exit(1);
-    }
+    perror(info->exe);
+    exit(1);
+  } else if ( pid < 0 )			/* parent */
+  { term_t exe = PL_new_term_ref();
+    PL_put_atom_chars(exe, info->exe);
 
-    { term_t exe = PL_new_term_ref();
-      PL_put_atom_chars(exe, info->exe);
-
-      return pl_error(NULL, 0, "execv", ERR_ERRNO, errno, "exec", "process", exe);
-    }
-  } else				/* parent */
+    return pl_error(NULL, 0, "fork", ERR_ERRNO, errno, "fork", "process", exe);
+  } else
   { if ( info->pipes > 0 && info->pid == 0 )
     { IOSTREAM *s;
       process_context *pc = PL_malloc(sizeof(*pc));
@@ -1445,7 +1426,7 @@ Basic process creation interface takes
 	* List of arguments
 	* standard streams		% std, null, pipe(S)
 	* Working directory
-	* detached			% Unix
+	* detached
 	* window			% Windows
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
@@ -1504,7 +1485,7 @@ process_wait(term_t pid, term_t code, term_t options)
 
     if ( !PL_get_name_arity(head, &name, &arity) || arity != 1 )
       return type_error(head, "option");
-    PL_get_arg(1, head, arg);
+    _PL_get_arg(1, head, arg);
     if ( name == ATOM_timeout )
     { atom_t a;
 
