@@ -3,9 +3,9 @@
     Part of SWI-Prolog
 
     Author:        Jan Wielemaker
-    E-mail:        jan@swi.psy.uva.nl
+    E-mail:        J.Wielemaker@cs.vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (C): 1985-2002, University of Amsterdam
+    Copyright (C): 1985-2009, University of Amsterdam
 
     This library is free software; you can redistribute it and/or
     modify it under the terms of the GNU Lesser General Public
@@ -19,7 +19,7 @@
 
     You should have received a copy of the GNU Lesser General Public
     License along with this library; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */
 
 #ifdef HAVE_CONFIG_H
@@ -58,6 +58,17 @@ typedef struct
   term_t	algorithm_term;
 } optval;
 
+#define CONTEXT_MAGIC (~ 0x53484163L)
+
+struct context
+{ int		magic;
+  optval	opts;
+  union {
+    sha1_ctx	sha1;
+    sha2_ctx    sha2;
+  } context;
+};
+
 static int
 sha_options(term_t options, optval *result)
 { term_t opts = PL_copy_term_ref(options);
@@ -75,7 +86,7 @@ sha_options(term_t options, optval *result)
     if ( PL_get_name_arity(opt, &aname, &arity) && arity == 1 )
     { term_t a = PL_new_term_ref();
 
-      PL_get_arg(1, opt, a);
+      _PL_get_arg(1, opt, a);
 
       if ( aname == ATOM_algorithm )
       { atom_t a_algorithm;
@@ -141,6 +152,70 @@ pl_sha_hash(term_t from, term_t hash, term_t options)
 }
 
 
+static foreign_t
+pl_sha_new_ctx(term_t ctx, term_t options)
+{ struct context c;
+  optval *op = &(c.opts);
+
+  if ( !sha_options(options, op) )
+    return FALSE;
+
+  c.magic = CONTEXT_MAGIC;
+
+  if ( op->algorithm == ALGORITHM_SHA1 )
+  { sha1_begin(&(c.context.sha1));
+  } else
+  { sha2_begin((unsigned long) op->digest_size, &(c.context.sha2));
+  }
+
+  /* NB: the context size depends on the digest size */
+  /* (e. g., sha512_ctx is twice as long as sha256_ctx) */
+  /* so there're extra data.  It will do no harm, though. */
+  /* . */
+  return PL_unify_string_nchars(ctx, sizeof(c), (char*)&c);
+}
+
+
+static foreign_t
+pl_sha_hash_ctx(term_t old_ctx, term_t from, term_t new_ctx, term_t hash)
+{ char *data;
+  size_t datalen;
+  struct context *cp;
+  size_t clen;
+  unsigned char hval[SHA2_MAX_DIGEST_SIZE];
+
+  if ( !PL_get_nchars(from, &datalen, &data,
+		      CVT_ATOM|CVT_STRING|CVT_LIST|CVT_EXCEPTION) )
+    return FALSE;
+
+  if ( !PL_get_string_chars(old_ctx, (char **)&cp, &clen) )
+    return FALSE;
+
+  if ( clen != sizeof (*cp)
+       || cp->magic != CONTEXT_MAGIC ) {
+    return pl_error(NULL, 0, "Invalid OldContext passed",
+		    ERR_DOMAIN, old_ctx, "algorithm");
+  }
+
+  if ( cp->opts.algorithm == ALGORITHM_SHA1 )
+  { sha1_ctx *c1p = &(cp->context.sha1);
+    sha1_hash((unsigned char*)data, (unsigned long)datalen, c1p);
+    if ( !PL_unify_string_nchars(new_ctx, sizeof(*cp), (char*)cp) )
+      return FALSE;
+    sha1_end((unsigned char *)hval, c1p);
+  } else
+  { sha2_ctx *c1p = &(cp->context.sha2);
+    sha2_hash((unsigned char*)data, (unsigned long)datalen, c1p);
+    if ( !PL_unify_string_nchars(new_ctx, sizeof(*cp), (char*)cp) )
+      return FALSE;
+    sha2_end((unsigned char *)hval, c1p);
+  }
+
+  /* . */
+  return PL_unify_list_ncodes(hash, cp->opts.digest_size, (char*)hval);
+}
+
+
 static foreign_t
 pl_hmac_sha(term_t key, term_t data, term_t mac, term_t options)
 { char *sdata, *skey;
@@ -190,5 +265,7 @@ install_sha4pl()
   MKATOM(algorithm);
 
   PL_register_foreign("sha_hash", 3, pl_sha_hash, 0);
+  PL_register_foreign("sha_new_ctx", 2, pl_sha_new_ctx, 0);
+  PL_register_foreign("sha_hash_ctx", 4, pl_sha_hash_ctx, 0);
   PL_register_foreign("hmac_sha", 4, pl_hmac_sha, 0);
 }
