@@ -1,4 +1,4 @@
-/*  $Id: streampool.pl,v 1.5 2007/02/26 11:20:47 jan Exp $
+/*  $Id$
 
     Part of SWI-Prolog
 
@@ -19,7 +19,7 @@
 
     You should have received a copy of the GNU Lesser General Public
     License along with this library; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 
     As a special exception, if you link this library with other files,
     compiled with a Free Software compiler, to produce an executable, this
@@ -72,7 +72,7 @@ close_stream_pool :-
 	    fail
 	;   true
 	).
-	
+
 %	dispatch_stream_pool(+TimeOut)
 %
 %	Wait for input on one or more streams and handle that.  Wait for
@@ -103,7 +103,7 @@ action(Stream) :-
 	;   print_message(warning,
 			  goal_failed(Action, stream_pool))
 	).
-	
+
 %	stream_pool_main_loop
 %
 %	Keep handling input from the streams in the pool until they have
