@@ -3,9 +3,10 @@
     Part of SWI-Prolog
 
     Author:        Jan Wielemaker and Anjo Anjewierden
-    E-mail:        J.Wielemak@uva.nl
+    E-mail:        J.Wielemaker@cs.vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (C): 1985-2008, University of Amsterdam
+    Copyright (C): 1985-2010, University of Amsterdam
+			      VU University Amsterdam
 
     This program is free software; you can redistribute it and/or
     modify it under the terms of the GNU General Public License
@@ -19,7 +20,7 @@
 
     You should have received a copy of the GNU Lesser General Public
     License along with this library; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 
     As a special exception, if you link this library with other files,
     compiled with a Free Software compiler, to produce an executable, this
@@ -31,6 +32,7 @@
 
 :- module(html_write,
 	  [ reply_html_page/2,		% :Head, :Body
+	    reply_html_page/3,		% +Style, :Head, :Body
 
 					% Basic output routines
 	    page//1,			% :Content
@@ -56,24 +58,30 @@
 					% Emitting the HTML code
 	    print_html/1,		% +List
 	    print_html/2,		% +Stream, +List
-	    html_print_length/2		% +List, -Length
+	    html_print_length/2,	% +List, -Length
+
+					% Extension support
+	    (html_meta)/1,		% +Spec
+	    op(1150, fx, html_meta)
 	  ]).
 :- use_module(library(error)).
+:- use_module(library(apply)).
 :- use_module(library(lists)).
 :- use_module(library(option)).
 :- use_module(library(pairs)).
 :- use_module(library(sgml)).		% Quote output
-:- use_module(library(url)).
+:- use_module(library(uri)).
 :- use_module(library(quintus)).	% for meta_predicate/1
 :- set_prolog_flag(generate_debug_info, false).
 
 :- meta_predicate
+	reply_html_page(+, :, :),
 	reply_html_page(:, :),
 	html(:, -, +),
 	page(:, -, +),
 	page(:, :, -, +),
-	pagehead(:, -, +),
-	pagebody(:, -, +),
+	pagehead(+, :, -, +),
+	pagebody(+, :, -, +),
 	html_receive(+, 3, -, +),
 	html_post(+, :, -, +).
 
@@ -126,7 +134,7 @@ encoding.
 %		page//2.
 %
 %		* content_type(+ContentType)
-%		Set the =|Content-type|= for reply_html_page/2
+%		Set the =|Content-type|= for reply_html_page/3
 %
 %	Note  that  the  doctype  is  covered    by  two  prolog  flags:
 %	=html_doctype= for the html dialect  and =xhtml_doctype= for the
@@ -197,14 +205,13 @@ init_options :-
 	(   option_default(Name, Value),
 	    (	current_prolog_flag(Name, _)
 	    ->	true
-	    ;	set_prolog_flag(Name, Value)
+	    ;	create_prolog_flag(Name, Value, [])
 	    ),
 	    fail
 	;   true
 	).
 
-:- initialization
-	init_options.
+:- init_options.
 
 %%	xml_header(-Header)
 %
@@ -235,10 +242,13 @@ page(Content) -->
 	html(html(Content)).
 
 page(Head, Body) -->
+	page(default, Head, Body).
+
+page(Style, Head, Body) -->
 	doctype,
 	html_begin(html),
-	pagehead(Head),
-	pagebody(Body),
+	pagehead(Style, Head),
+	pagebody(Style, Body),
 	html_end(html).
 
 %%	doctype//
@@ -257,35 +267,45 @@ doctype -->
 	[].
 
 
-pagehead(Head) -->
+pagehead(_, Head) -->
 	{ functor(Head, head, _)
 	}, !,
 	html(Head).
-pagehead(Head) -->
+pagehead(Style, Head) -->
 	{ strip_module(Head, M, _),
-	  hook_module(M, head//1)
-	}, !,
-	M:head(Head).
-pagehead(Head) -->
+	  hook_module(M, HM, head//2)
+	},
+	HM:head(Style, Head), !.
+pagehead(_, Head) -->
+	{ strip_module(Head, M, _),
+	  hook_module(M, HM, head//1)
+	},
+	HM:head(Head), !.
+pagehead(_, Head) -->
 	html(head(Head)).
 
 
-pagebody(Body) -->
+pagebody(_, Body) -->
 	{ functor(Body, body, _)
 	}, !,
 	html(Body).
-pagebody(Body) -->
+pagebody(Style, Body) -->
 	{ strip_module(Body, M, _),
-	  hook_module(M, body//1)
-	}, !,
-	M:body(Body).
-pagebody(Body) -->
+	  hook_module(M, HM, body//2)
+	},
+	HM:body(Style, Body), !.
+pagebody(_, Body) -->
+	{ strip_module(Body, M, _),
+	  hook_module(M, HM, body//1)
+	},
+	HM:body(Body), !.
+pagebody(_, Body) -->
 	html(body(Body)).
 
 
-hook_module(M, PI) :-
+hook_module(M, M, PI) :-
 	current_predicate(M:PI), !.
-hook_module(user, PI) :-
+hook_module(_, user, PI) :-
 	current_predicate(user:PI).
 
 %%	html(+Content:dom)// is det
@@ -336,7 +356,9 @@ do_expand(script(Content), _) --> !,	% general CDATA declared content elements?
 	],
 	html_end(script).
 do_expand(&(Entity), _) --> !,
-	{ format(string(String), '&~w;', [Entity])
+	{   integer(Entity)
+	->  format(string(String), '&#~d;', [Entity])
+	;   format(string(String), '&~w;', [Entity])
 	},
 	[ String ].
 do_expand(Token, _) -->
@@ -560,6 +582,8 @@ name(Name) -->
 %	  Emit URL-encoded version of V.  See www_form_encode/2.
 %	  * An option list
 %	  Emit ?Name1=encode(Value1)&Name2=encode(Value2) ...
+%	  * A term Format-Arguments
+%	  Use format/3 and emit the result as quoted value.
 %
 %	The hook html_write:expand_attribute_value//1 can  be defined to
 %	provide additional `function like'   translations.  For example,
@@ -570,24 +594,36 @@ name(Name) -->
 :- multifile
 	expand_attribute_value//1.
 
-attribute_value(Var) -->
+attribute_value(List) -->
+	{ is_list(List) },
+	attribute_value_m(List).
+attribute_value(Value) -->
+	attribute_value_s(Value).
+
+% emit a single attribute value
+
+attribute_value_s(Var) -->
 	{ var(Var), !,
 	  instantiation_error(Var)
 	}.
-attribute_value(A+B) --> !,
+attribute_value_s(A+B) --> !,
 	attribute_value(A),
-	attribute_value(B).
-attribute_value([]) --> !.
-attribute_value(List) -->
-	{ is_list(List) }, !,
-	[ ? ],
-	search_parameters(List).
-attribute_value(encode(Value)) --> !,
-	{ www_form_encode(Value, Encoded) },
+	(   { is_list(B) }
+	->  (   { B == [] }
+	    ->	[]
+	    ;	[?], search_parameters(B)
+	    )
+	;   attribute_value(B)
+	).
+attribute_value_s(encode(Value)) --> !,
+	{ uri_encoded(query_value, Value, Encoded) },
 	[ Encoded ].
-attribute_value(Value) -->
+attribute_value_s(Value) -->
 	expand_attribute_value(Value), !.
-attribute_value(Value) -->
+attribute_value_s(Fmt-Args) --> !,
+	{ format(string(Value), Fmt, Args) },
+	html_quoted_attribute(Value).
+attribute_value_s(Value) -->
 	html_quoted_attribute(Value).
 
 search_parameters([H|T]) -->
@@ -614,6 +650,26 @@ search_parameter(Term) -->
 	{ domain_error(search_parameter, Term)
 	}.
 
+%%	attribute_value_m(+List)//
+%
+%	Used for multi-valued attributes, such as class-lists.  E.g.,
+%
+%	  ==
+%		body(class([c1, c2]), Body)
+%	  ==
+%
+%	  Emits =|<body class="c1 c2"> ...|=
+
+attribute_value_m([]) -->
+	[].
+attribute_value_m([H|T]) -->
+	attribute_value_s(H),
+	(   { T == [] }
+	->  []
+	;   [' '],
+	    attribute_value_m(T)
+	).
+
 
 		 /*******************************
 		 *	   QUOTING RULES	*
@@ -658,7 +714,7 @@ html_quoted_attribute(Text) -->
 %
 %	Reposition HTML to  the  receiving   Id.  The  http_post//2 call
 %	processes HTML using html//1. Embedded   \-commands are executed
-%	by mainman/1 from  print_html/1   or  html_print_length/2. These
+%	by mailman/1 from  print_html/1   or  html_print_length/2. These
 %	commands are called in the calling   context of the html_post//2
 %	call.
 %
@@ -860,14 +916,14 @@ post_close(_) -->
 %			close-tag or =empty= to indicate the element has
 %			no content model.
 %
-% 	@tbd	Complete table
+%	@tbd	Complete table
 
 :- multifile
 	layout/3.
 
 layout(table,	   2-1,	1-2).
 layout(blockquote, 2-1,	1-2).
-layout(pre, 	   2-1,	1-2).
+layout(pre,	   2-1,	1-2).
 layout(center,	   2-1,	1-2).
 layout(dl,	   2-1,	1-2).
 layout(ul,	   1-1,	1-1).
@@ -1018,18 +1074,114 @@ html_print_length([H|T], L0, L) :-
 
 
 %%	reply_html_page(:Head, :Body) is det.
+%%	reply_html_page(+Style, :Head, :Body) is det.
 %
 %	Provide the complete reply as required  by http_wrapper.pl for a
 %	page constructed from Head and   Body. The HTTP =|Content-type|=
 %	is provided by html_current_option/1.
 
 reply_html_page(Head, Body) :-
+	reply_html_page(default, Head, Body).
+reply_html_page(Style, Head, Body) :-
 	html_current_option(content_type(Type)),
-	phrase(page(Head, Body), HTML),
+	phrase(page(Style, Head, Body), HTML),
 	format('Content-type: ~w~n~n', [Type]),
 	print_html(HTML).
 
 
+		 /*******************************
+		 *     META-PREDICATE SUPPORT	*
+		 *******************************/
+
+%%	html_meta(+Heads) is det.
+%
+%	This directive can be used  to   declare  that an HTML rendering
+%	rule takes HTML content as  argument.   It  has  two effects. It
+%	emits  the  appropriate  meta_predicate/1    and  instructs  the
+%	built-in editor (PceEmacs) to provide   proper colouring for the
+%	arguments.  The  arguments  in  Head  are    the   same  as  for
+%	meta_predicate or can be constant =html=.  For example:
+%
+%	  ==
+%	  :- html_meta
+%		page(html,html,?,?).
+%	  ==
+
+html_meta(Spec) :-
+	throw(error(context_error(nodirective, html_meta(Spec)), _)).
+
+html_meta_decls(Var, _, _) :-
+	var(Var), !,
+	instantiation_error(Var).
+html_meta_decls((A,B), (MA,MB), [MH|T]) :- !,
+	html_meta_decl(A, MA, MH),
+	html_meta_decls(B, MB, T).
+html_meta_decls(A, MA, [MH]) :-
+	html_meta_decl(A, MA, MH).
+
+html_meta_decl(Head, MetaHead,
+	       html_write:html_meta_head(GenHead, Module, Head)) :-
+	functor(Head, Name, Arity),
+	functor(GenHead, Name, Arity),
+	prolog_load_context(module, Module),
+	Head =.. [Name|HArgs],
+	maplist(html_meta_decl, HArgs, MArgs),
+	MetaHead =.. [Name|MArgs].
+
+html_meta_decl(html, :) :- !.
+html_meta_decl(Meta, Meta).
+
+system:term_expansion((:- html_meta(Heads)),
+		      [ (:- meta_predicate(Meta))
+		      | MetaHeads
+		      ]) :-
+	html_meta_decls(Heads, Meta, MetaHeads).
+
+:- multifile
+	html_meta_head/3.
+
+html_meta_colours(Head, Goal, built_in-Colours) :-
+	Head =.. [_|MArgs],
+	Goal =.. [_|Args],
+	maplist(meta_colours, MArgs, Args, Colours).
+
+meta_colours(html, HTML, Colours) :- !,
+	html_colours(HTML, Colours).
+meta_colours(I, _, Colours) :-
+	integer(I), I>=0, !,
+	Colours = meta(I).
+meta_colours(_, _, classify).
+
+html_meta_called(Head, Goal, Called) :-
+	Head =.. [_|MArgs],
+	Goal =.. [_|Args],
+	meta_called(MArgs, Args, Called, []).
+
+meta_called([], [], Called, Called).
+meta_called([html|MT], [A|AT], Called, Tail) :- !,
+	phrase(called_by(A), Called, Tail1),
+	meta_called(MT, AT, Tail1, Tail).
+meta_called([0|MT], [A|AT], [A|CT0], CT) :- !,
+	meta_called(MT, AT, CT0, CT).
+meta_called([I|MT], [A|AT], [A+I|CT0], CT) :-
+	integer(I), I>0, !,
+	meta_called(MT, AT, CT0, CT).
+meta_called([_|MT], [_|AT], Called, Tail) :- !,
+	meta_called(MT, AT, Called, Tail).
+
+
+:- html_meta
+	html(html,?,?),
+	page(html,?,?),
+	page(html,html,?,?),
+	page(+,html,html,?,?),
+	pagehead(+,html,?,?),
+	pagebody(+,html,?,?),
+	reply_html_page(html,html),
+	reply_html_page(+,html,html),
+	html_post(+,html,?,?).
+
+
 		 /*******************************
 		 *	PCE EMACS SUPPORT	*
 		 *******************************/
@@ -1040,36 +1192,17 @@ reply_html_page(Head, Body) :-
 	emacs_prolog_colours:identify/2,
 	prolog:called_by/2.
 
-emacs_prolog_colours:goal_colours(html(HTML,_,_),
-				  built_in-[Colours, classify, classify]) :-
-	html_colours(HTML, Colours).
-emacs_prolog_colours:goal_colours(page(HTML,_,_),
-				  built_in-[Colours, classify, classify]) :-
-	html_colours(HTML, Colours).
-emacs_prolog_colours:goal_colours(page(Head, Body,_,_),
-				  built_in-[HC, BC, classify, classify]) :-
-	html_colours(Head, HC),
-	html_colours(Body, BC).
-emacs_prolog_colours:goal_colours(pagehead(HTML,_,_),
-				  built_in-[Colours, classify, classify]) :-
-	html_colours(HTML, Colours).
-emacs_prolog_colours:goal_colours(pagebody(HTML,_,_),
-				  built_in-[Colours, classify, classify]) :-
-	html_colours(HTML, Colours).
-emacs_prolog_colours:goal_colours(reply_html_page(Head, Body),
-				  built_in-[HC, BC]) :-
-	html_colours(Head, HC),
-	html_colours(Body, BC).
-emacs_prolog_colours:goal_colours(html_post(_Id, HTML, _, _),
-				  built_in-[classify, Colours]) :-
-	html_colours(HTML, Colours).
+emacs_prolog_colours:goal_colours(Goal, Colours) :-
+	html_meta_head(Goal, _Module, Head),
+	html_meta_colours(Head, Goal, Colours).
 
 
 					% TBD: Check with do_expand!
 html_colours(Var, classify) :-
 	var(Var), !.
-html_colours(\List, classify) :-
-	is_list(List), !.
+html_colours(\List, built_in-Colours) :-
+	is_list(List), !,
+	list_colours(List, Colours).
 html_colours(\_, built_in-[dcg]) :- !.
 html_colours(_:Term, built_in-[classify,Colours]) :- !,
 	html_colours(Term, Colours).
@@ -1187,20 +1320,9 @@ emacs_prolog_colours:identify(http_no_location_for_id(ID), Summary) :-
 %	or callable+N to indicate (DCG) arglist extension.
 
 
-prolog:called_by(html(HTML,_,_), Called) :-
-	phrase(called_by(HTML), Called).
-prolog:called_by(page(HTML,_,_), Called) :-
-	phrase(called_by(HTML), Called).
-prolog:called_by(page(Head,Body,_,_), Called) :-
-	phrase(called_by([Head,Body]), Called).
-prolog:called_by(pagehead(HTML,_,_), Called) :-
-	phrase(called_by(HTML), Called).
-prolog:called_by(pagebody(HTML,_,_), Called) :-
-	phrase(called_by(HTML), Called).
-prolog:called_by(html_post(_,HTML,_,_), Called) :-
-	phrase(called_by(HTML), Called).
-prolog:called_by(reply_html_page(Head,Body), Called) :-
-	phrase(called_by([Head,Body]), Called).
+prolog:called_by(Goal, Called) :-
+	html_meta_head(Goal, _Module, Head),
+	html_meta_called(Head, Goal, Called).
 
 called_by(Term) -->
 	called_by(Term, _).
@@ -1210,9 +1332,9 @@ called_by(Var, _) -->
 	[].
 called_by(\G, M) --> !,
 	(   { is_list(G) }
-	->  []
+	->  called_by(G, M)
 	;   {atom(M)}
-	->  [M:G+2]
+	->  [(M:G)+2]
 	;   [G+2]
 	).
 called_by([], _) --> !,
@@ -1233,6 +1355,14 @@ called_by(Term, M) -->
 called_by(_, _) -->
 	[].
 
+:- multifile
+	prolog:hook/1.
+
+prolog:hook(body(_,_,_)).
+prolog:hook(body(_,_,_,_)).
+prolog:hook(head(_,_,_)).
+prolog:hook(head(_,_,_,_)).
+
 
 		 /*******************************
 		 *	      MESSAGES		*
