@@ -17,7 +17,7 @@
 
     You should have received a copy of the GNU Lesser General Public
     License along with this library; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 
     As a special exception, if you link this library with other files,
     compiled with a Free Software compiler, to produce an executable, this
@@ -32,20 +32,34 @@
 	    http_handler/3,		% +Path, +Predicate, +Options
 	    http_delete_handler/1,	% +Path
 	    http_reply_file/3,		% +File, +Options, +Request
+	    http_redirect/3,		% +How, +Path, +Request
+	    http_404/2,			% +Options, +Request
 	    http_current_handler/2,	% ?Path, ?Pred
 	    http_current_handler/3,	% ?Path, ?Pred
-	    http_location_by_id/2	% +ID, -Location
+	    http_location_by_id/2,	% +ID, -Location
+	    http_link_to_id/3,		% +ID, +Parameters, -HREF
+	    http_reload_with_parameters/3, % +Request, +Parameters, -HREF
+	    http_safe_file/2		% +Spec, +Options
 	  ]).
 :- use_module(library(option)).
 :- use_module(library(lists)).
 :- use_module(library(time)).
 :- use_module(library(error)).
 :- use_module(library(settings)).
+:- use_module(library(uri)).
 :- use_module(library(http/mimetype)).
 :- use_module(library(http/http_path)).
 :- use_module(library(http/http_header)).
 :- use_module(library(http/thread_httpd)).
 
+:- predicate_options(http_404/2, 1, [index(any)]).
+:- predicate_options(http_reply_file/3, 2,
+		     [ cache(boolean),
+		       mime_type(any),
+		       pass_to(http_safe_file/2, 2)
+		     ]).
+:- predicate_options(http_safe_file/2, 2, [unsafe(boolean)]).
+
 /** <module> Dispatch requests in the HTTP server
 
 This module can be placed between   http_wrapper.pl  and the application
@@ -65,30 +79,17 @@ server(Port, Options) :-
 write_index(Request) :-
 	...
 ==
-
-@author	Jan Wielemaker
 */
 
 :- setting(http:time_limit, nonneg, 300,
 	   'Time limit handling a single query (0=infinite)').
 
-%%	http_handler(+Path, :Pred, +Options) is det.
-%
-%	Register Pred as  a  handler  for   HTTP  requests.  Path  is is
-%	specification as provided by  http_path.pl.   Pred  is either an
-%	atom or one of the following reserved terms:
-%
-%		* reply_file(+File, +FileOptions)
-%		Reply contents of File using the given options.
-%		FileOptions include:
-%
-%			* mime_type(+Type)
-%			Mime-type specified with File.
-%
-%			* cache(+Boolean)
-%			If =false=, do not pass and process
-%			last_modification time.
+%%	http_handler(+Path, :Closure, +Options) is det.
 %
+%	Register Closure as a handler for HTTP requests. Path is a
+%	specification as provided by http_path.pl.  If an HTTP
+%	request arrives at the server that matches Path, Closure
+%	is called with one extra argument: the parsed HTTP request.
 %	Options is a list containing the following options:
 %
 %		* authentication(+Type)
@@ -101,10 +102,34 @@ write_index(Request) :-
 %		Use =|Transfer-encoding: chunked|= if the client
 %		allows for it.
 %
+%		* content_type(+Term)
+%		Specifies the content-type of the reply.  This value is
+%		currently not used by this library.  It enhances the
+%		reflexive capabilities of this library through
+%		http_current_handler/3.
+%
 %		* id(+Term)
 %		Identifier of the handler.  The default identifier is
 %		the predicate name.  Used by http_location_by_id/2.
 %
+%		* hide_children(+Bool)
+%		If =true= on a prefix-handler (see prefix), possible
+%		children are masked.  This can be used to (temporary)
+%		overrule part of the tree.
+%
+%		* prefix
+%		Call Pred on any location that is a specialisation of
+%		Path.  If multiple handlers match, the one with the
+%		longest path is used.  Options defined with a prefix
+%		handler are the default options for paths that start
+%		with this prefix.  Note that the handler acts as a
+%		fallback handler for the tree below it:
+%
+%		  ==
+%		  :- http_handler(/, http_404([index('index.html')]),
+%				  [spawn(my_pool),prefix]).
+%		  ==
+%
 %		* priority(+Integer)
 %		If two handlers handle the same path, the one with the
 %		highest priority is used.  If equal, the last registered
@@ -112,11 +137,6 @@ write_index(Request) :-
 %		multifile predicates can change due to reloading files.
 %		The default priority is 0 (zero).
 %
-%		* prefix
-%		Call Pred on any location that is a specialisation of
-%		Path.  If multiple handlers match, the one with the
-%		longest path is used.
-%
 %		* spawn(+SpawnOptions)
 %		Run the handler in a seperate thread.  If SpawnOptions
 %		is an atom, it is interpreted as a thread pool name
@@ -134,6 +154,10 @@ write_index(Request) :-
 %	proper update through make/0 when the specification is modified.
 %	We do not expand when the  cross-referencer is running to ensure
 %	proper handling of the meta-call.
+%
+%	@error	existence_error(http_location, Location)
+%	@see    http_reply_file/3 and http_redirect/3 are generic
+%		handlers to serve files and achieve redirects.
 
 :- dynamic handler/4.			% Path, Action, IsPrefix, Options
 :- multifile handler/4.
@@ -151,23 +175,39 @@ http_handler(Path, Pred, Options) :-
 	assert(Clause).
 
 :- multifile
-	user:term_expansion/2.
+	system:term_expansion/2.
 
-user:term_expansion((:- http_handler(Path, Pred, Options)), Clause) :-
+system:term_expansion((:- http_handler(Path, Pred, Options)), Clause) :-
 	\+ current_prolog_flag(xref, true),
 	prolog_load_context(module, M),
 	compile_handler(Path, M:Pred, Options, Clause),
 	next_generation.
 
 
-%%	http_delete_handler(+Path) is det.
+%%	http_delete_handler(+Spec) is det.
 %
-%	Delete handler for Path. Typically, this should only be used for
-%	handlers that are registered dynamically.
+%	Delete handler for Spec. Typically, this should only be used for
+%	handlers that are registered dynamically. Spec is one of:
+%
+%	    * id(Id)
+%	    Delete a handler with the given id.  The default id is the
+%	    handler-predicate-name.
+%
+%	    * path(Path)
+%	    Delete handler that serves the given path.
 
-http_delete_handler(Path) :-
+http_delete_handler(id(Id)) :- !,
+	clause(handler(_Path, _:Pred, _, Options), true, Ref),
+	functor(Pred, _, DefID),
+	option(id(Id0), Options, DefID),
+	Id == Id0,
+	erase(Ref),
+	next_generation.
+http_delete_handler(path(Path)) :- !,
 	retractall(handler(Path, _Pred, _, _Options)),
 	next_generation.
+http_delete_handler(Path) :-
+	http_delete_handler(path(Path)).
 
 
 %%	next_generation is det.
@@ -381,6 +421,37 @@ location_by_id_raw(ID, Location, Priority) :-
 	;   ID = PN
 	), !.
 
+
+%%	http_link_to_id(+HandleID, +Parameters, -HREF)
+%
+%	HREF is a link on the local server to a handler with given ID,
+%	passing the given Parameters.
+
+http_link_to_id(HandleID, Parameters, HREF) :-
+	http_location_by_id(HandleID, Location),
+	uri_data(path, Components, Location),
+	uri_query_components(String, Parameters),
+	uri_data(search, Components, String),
+	uri_components(HREF, Components).
+
+%%	http_reload_with_parameters(+Request, +Parameters, -HREF) is det.
+%
+%	Create a request on the current handler with replaced search
+%	parameters.
+
+http_reload_with_parameters(Request, NewParams, HREF) :-
+	memberchk(path(Path), Request),
+	(   memberchk(search(Params), Request)
+	->  true
+	;   Params = []
+	),
+	merge_options(NewParams, Params, AllParams),
+	uri_query_components(Search, AllParams),
+	uri_data(path, Data, Path),
+	uri_data(search, Data, Search),
+	uri_components(HREF, Data).
+
+
 %	hook into html_write:attribute_value//1.
 
 :- multifile
@@ -399,7 +470,7 @@ html_write:expand_attribute_value(location_by_id(ID)) -->
 %	The  library  http_authenticate.pl  provides  an  implementation
 %	thereof.
 %
-%	@error	permission_error(http_location, access, Location)
+%	@error	permission_error(access, http_location, Location)
 
 :- multifile
 	http:authenticate/3.
@@ -410,7 +481,7 @@ authentication([authentication(Type)|Options], Request, Fields) :- !,
 	->  append(XFields, More, Fields),
 	    authentication(Options, Request, More)
 	;   memberchk(path(Path), Request),
-	    throw(error(permission_error(http_location, access, Path), _))
+	    throw(error(permission_error(access, http_location, Path), _))
 	).
 authentication([_|Options], Request, Fields) :-
 	authentication(Options, Request, Fields).
@@ -447,7 +518,8 @@ find_handler(Path, Action, Options) :-
 find_handler([node(prefix(Prefix), PAction, POptions, Children)|_],
 	     Path, Action, Options) :-
 	sub_atom(Path, 0, _, After, Prefix), !,
-	(   find_handler(Children, Path, Action, Options)
+	(   option(hide_children(false), POptions, false),
+	    find_handler(Children, Path, Action, Options)
 	->  true
 	;   Action = PAction,
 	    path_info(After, Path, POptions, Options)
@@ -556,7 +628,7 @@ extend(G0, Extra, G) :-
 %	@throws http_reply(file(MimeType, Path))
 
 http_reply_file(File, Options, Request) :-
-	check_file_safeness(File, Options),
+	http_safe_file(File, Options),
 	absolute_file_name(File, Path,
 			   [ access(read)
 			   ]),
@@ -581,7 +653,7 @@ http_reply_file(File, Options, Request) :-
 	),
 	throw(http_reply(Reply)).
 
-%%	check_file_safeness(+FileSpec, +Options) is det.
+%%	http_safe_file(+FileSpec, +Options) is det.
 %
 %	True if FileSpec is considered _safe_.  If   it  is  an atom, it
 %	cannot  be  absolute  and  cannot   have  references  to  parent
@@ -591,20 +663,20 @@ http_reply_file(File, Options, Request) :-
 %	@error instantiation_error
 %	@error permission_error(read, file, FileSpec)
 
-check_file_safeness(File, _) :-
+http_safe_file(File, _) :-
 	var(File), !,
 	instantiation_error(File).
-check_file_safeness(_, Options) :-
+http_safe_file(_, Options) :-
 	option(unsafe(true), Options, false), !.
-check_file_safeness(File, _) :-
-	check_file_safeness(File).
+http_safe_file(File, _) :-
+	http_safe_file(File).
 
-check_file_safeness(File) :-
+http_safe_file(File) :-
 	compound(File),
 	functor(File, _, 1), !,
 	arg(1, File, Name),
 	safe_name(Name, File).
-check_file_safeness(Name) :-
+http_safe_file(Name) :-
 	(   is_absolute_file_name(Name)
 	->  permission_error(read, file, Name)
 	;   true
@@ -613,7 +685,8 @@ check_file_safeness(Name) :-
 
 safe_name(Name, _) :-
 	must_be(atom, Name),
-	\+ unsafe_name(Name), !.
+	prolog_to_os_filename(FileName, Name),
+	\+ unsafe_name(FileName), !.
 safe_name(_, Spec) :-
 	permission_error(read, file, Spec).
 
@@ -623,6 +696,56 @@ unsafe_name(Name) :- sub_atom(Name, _, _, _, '/../').
 unsafe_name(Name) :- sub_atom(Name, _, _, 0, '/..').
 
 
+%%	http_redirect(+How, +To, +Request) is det.
+%
+%	Redirect to a new  location.  The   argument  order,  using  the
+%	Request as last argument, allows for  calling this directly from
+%	the handler declaration:
+%
+%	    ==
+%	    :- http_handler(root(.),
+%			    http_redirect(moved, myapp('index.html')),
+%			    []).
+%	    ==
+%
+%	@param How is one of =moved=, =moved_temporary= or =see_other=
+%	@param To is an atom, a aliased path as defined by
+%	http_absolute_location/3. or a term location_by_id(Id). If To is
+%	not absolute, it is resolved relative to the current location.
+
+http_redirect(How, To, Request) :-
+	(   To = location_by_id(Id)
+	->  http_location_by_id(Id, URL)
+	;   memberchk(path(Base), Request),
+	    http_absolute_location(To, URL, [relative_to(Base)])
+	),
+	must_be(oneof([moved, moved_temporary, see_other]), How),
+	Term =.. [How,URL],
+	throw(http_reply(Term)).
+
+
+%%	http_404(+Options, +Request) is det.
+%
+%	Reply using an "HTTP  404  not   found"  page.  This  handler is
+%	intended as fallback handler  for   _prefix_  handlers.  Options
+%	processed are:
+%
+%	    * index(Location)
+%	    If there is no path-info, redirect the request to
+%	    Location using http_redirect/3.
+%
+%	@error http_reply(not_found(Path))
+
+http_404(Options, Request) :-
+	option(index(Index), Options),
+	\+ ( memberchk(path_info(PathInfo), Request),
+	     PathInfo \== ''
+	   ), !,
+	http_redirect(moved, Index, Request).
+http_404(_Options, Request) :-
+	memberchk(path(Path), Request),
+	throw(http_reply(not_found(Path))).
+
 
 		 /*******************************
 		 *	  PATH COMPILATION	*
@@ -713,7 +836,8 @@ add_paths_tree([path(Path, Action, Options)|T], Tree0, Tree) :-
 
 plain_path(Path, Action, Options) :-
 	handler(Spec, Action, false, Options),
-	http_absolute_location(Spec, Path, []).
+	catch(http_absolute_location(Spec, Path, []), E,
+	      (print_message(error, E), fail)).
 
 
 %%	add_path_tree(+Path, +Action, +Options, +Tree0, -Tree) is det.
@@ -780,8 +904,9 @@ prolog:meta_goal(http_current_handler(_, G), [G+1]).
 
 prolog_edit:locate(Path, Spec, Location) :-
 	atom(Path),
+	sub_atom(Path, 0, _, _, /),
 	Pred = _M:_H,
-	http_current_handler(Path, Pred),
+	catch(http_current_handler(Path, Pred), _, fail),
 	closure_name_arity(Pred, 1, PI),
 	prolog_edit:locate(PI, Spec, Location).
 
