@@ -19,7 +19,7 @@
 
     You should have received a copy of the GNU General Public
     License along with this library; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 
     As a special exception, if you link this library with other files,
     compiled with a Free Software compiler, to produce an executable, this
@@ -38,6 +38,7 @@
 :- use_module(library(pldoc)).
 :- use_module(library(readutil)).
 :- use_module(library(error)).
+:- use_module(library(apply)).
 :- use_module(library(option)).
 :- use_module(library(lists)).
 :- use_module(library(debug)).
@@ -46,10 +47,12 @@
 :- use_module(pldoc(doc_modes)).
 :- use_module(pldoc(doc_html),		% we cannot import all as the
 	      [ doc_file_objects/5,	% \commands have the same name
+		unquote_filespec/2,
 		doc_tag_title/2,
 		existing_linked_file/2,
 		pred_anchor_name/3,
 		private/2,
+		(multifile)/2,
 		is_pi/1,
 		is_op_type/2
 	      ]).
@@ -75,6 +78,27 @@ the same paradigm. The module can
 @author Jan Wielemaker
 */
 
+:- predicate_options(doc_latex/3, 3,
+		     [ stand_alone(boolean),
+		       public_only(boolean),
+		       section_level(oneof([section,subsection,subsubsection])),
+		       summary(atom)
+		     ]).
+:- predicate_options(latex_for_file/3, 3,
+		     [ stand_alone(boolean),
+		       public_only(boolean),
+		       section_level(oneof([section,subsection,subsubsection]))
+		     ]).
+:- predicate_options(latex_for_predicates/3, 3,
+		     [				% no options
+		     ]).
+:- predicate_options(latex_for_wiki_file/3, 3,
+		     [ stand_alone(boolean),
+		       public_only(boolean),
+		       section_level(oneof([section,subsection,subsubsection]))
+		     ]).
+
+
 :- thread_local
 	options/1,
 	documented/1.
@@ -97,8 +121,9 @@ current_options([]).
 %
 %		* File
 %		If File is a prolog file (as defined by
-%		prolog_file_type/2), process using latex_for_file/3,
-%		otherwise process using latex_for_wiki_file/3.
+%		user:prolog_file_type/2), process using
+%		latex_for_file/3, otherwise process using
+%		latex_for_wiki_file/3.
 %
 %	Typically Spec is either a  list  of   filenames  or  a  list of
 %	predicate indicators.   Defined options are:
@@ -160,7 +185,7 @@ process_items(FileSpec, Mode0, Mode, Options) -->
 	    file_name_extension(_Base, Ext, File)
 	},
 	need_mode(body, Mode0, Mode),
-	(   { prolog_file_type(Ext, prolog) }
+	(   { user:prolog_file_type(Ext, prolog) }
 	->  latex_tokens_for_file(File, Options)
 	;   latex_tokens_for_wiki_file(File, Options)
 	).
@@ -335,12 +360,18 @@ latex(code(CodeList)) -->
 	{ is_list(CodeList), !,
 	  atomic_list_concat(CodeList, Atom)
 	},
-	[ verb(Atom) ].
+	(   {fragile}
+	->  latex(cmd(const(Atom)))
+	;   [ verb(Atom) ]
+	).
 latex(code(Code)) -->
 	{ identifier(Code) }, !,
 	latex(cmd(const(Code))).
 latex(code(Code)) -->
-	[ verb(Code) ].
+	(   {fragile}
+	->  latex(cmd(const(Code)))
+	;   [ verb(Code) ]
+	).
 latex(b(Code)) -->
 	latex(cmd(textbf(Code))).
 latex(i(Code)) -->
@@ -407,7 +438,7 @@ indent(prefixop) --> !,	       [ nl(1), indent(4) ].
 indent(postfixop) --> !,       [ nl(1), indent(4) ].
 indent(predicatesummary) --> !,[ nl(1) ].
 indent(oppredsummary) --> !,   [ nl(1) ].
-indent(hline) --> !, 	       [ nl(1) ].
+indent(hline) --> !,	       [ nl(1) ].
 indent(_) -->		       [].
 
 outdent(begin) --> !,		[ nl_exact(1) ].
@@ -429,7 +460,7 @@ outdent(prefixop) --> !,	[ nl(1) ].
 outdent(postfixop) --> !,	[ nl(1) ].
 outdent(predicatesummary) --> !,[ nl(1) ].
 outdent(oppredsummary) --> !,   [ nl(1) ].
-outdent(hline) --> !, 	        [ nl(1) ].
+outdent(hline) --> !,	        [ nl(1) ].
 outdent(_) -->			[].
 
 %%	latex_special(String, Rest)// is semidet.
@@ -503,6 +534,8 @@ latex_arg(H) -->
 	  urldef_name(Atom, Name)
 	}, !,
 	latex(cmd(Name)).
+latex_arg(no_escape(Text)) --> !,
+	[no_escape(Text)].
 latex_arg(H) -->
 	latex(H).
 
@@ -652,6 +685,10 @@ file(File, _Options) -->
 %	Called  from  name/arity  or   name//arity    patterns   in  the
 %	documentation.
 
+predref(Module:Name/Arity) --> !,
+	latex(cmd(qpredref(Module, Name, Arity))).
+predref(Module:Name//Arity) -->
+	latex(cmd(qdcgref(Module, Name, Arity))).
 predref(Name/Arity) -->
 	latex(cmd(predref(Name, Arity))).
 predref(Name//Arity) -->
@@ -728,9 +765,9 @@ param(param(Name,Descr)) -->
 
 file_header(File, Options) -->
 	{ memberchk(file(Title, Comment), Options), !,
-	  file_base_name(File, Base)
+	  file_synopsis(File, Synopsis)
 	},
-	file_title([Base, ' -- ', Title], File, Options),
+	file_title([Synopsis, ': ', Title], File, Options),
 	{ is_structured_comment(Comment, Prefixes),
 	  string_to_list(Comment, Codes),
 	  indented_lines(Codes, Prefixes, Lines),
@@ -741,15 +778,21 @@ file_header(File, Options) -->
 	latex(DOM),
 	latex(cmd(vspace('0.7cm'))).
 file_header(File, Options) -->
-	{ file_base_name(File, Base)
+	{ file_synopsis(File, Synopsis)
 	},
-	file_title([Base], File, Options).
+	file_title([Synopsis], File, Options).
 
 tags_to_front(DOM0, DOM) :-
 	append(Content, [\tags(Tags)], DOM0), !,
 	DOM = [\tags(Tags)|Content].
 tags_to_front(DOM, DOM).
 
+file_synopsis(File, Synopsis) :-
+	file_name_on_path(File, Term),
+	unquote_filespec(Term, Unquoted),
+	format(atom(Synopsis), '~w', [Unquoted]).
+
+
 %%	file_title(+Title:list, +File, +Options)// is det
 %
 %	Emit the file-header and manipulation buttons.
@@ -795,6 +838,8 @@ object(Obj, Pos, Comment, Mode0, Mode, Options) -->
 	  process_modes(Lines, Module, Pos, Modes, Args, Lines1),
 	  (   private(Obj, Options)
 	  ->  Class = privdef		% private definition
+	  ;   multifile(Obj, Options)
+	  ->  Class = multidef
 	  ;   Class = pubdef		% public definition
 	  ),
 	  (   Obj = Module:_
@@ -804,16 +849,20 @@ object(Obj, Pos, Comment, Mode0, Mode, Options) -->
 	  DOM = [\pred_dt(Modes, Class, POptions), dd(class=defbody, DOM1)],
 	  wiki_lines_to_dom(Lines1, Args, DOM0),
 	  strip_leading_par(DOM0, DOM1),
-	  assert(documented(Obj))
+	  assert_documented(Obj)
 	},
 	need_mode(description, Mode0, Mode),
 	latex(DOM).
-object([Obj|_Same], Pos, Comment, Mode0, Mode, Options) --> !,
-	object(Obj, Pos, Comment, Mode0, Mode, Options).
+object([Obj|Same], Pos, Comment, Mode0, Mode, Options) --> !,
+	object(Obj, Pos, Comment, Mode0, Mode, Options),
+	{ maplist(assert_documented, Same) }.
 object(Obj, _Pos, _Comment, Mode, Mode, _Options) -->
 	{ debug(pldoc, 'Skipped ~p', [Obj]) },
 	[].
 
+assert_documented(Obj) :-
+	assert(documented(Obj)).
+
 
 %%	need_mode(+Mode:atom, +Stack:list, -NewStack:list)// is det.
 %
@@ -919,6 +968,7 @@ pred_head(//(Head), Options) --> !,
 	latex(cmd(dcg(opt(Atts), Functor, Arity, \pred_args(Args, 1)))).
 pred_head(Head, _Options) -->			% Infix operators
 	{ Head =.. [Functor,Left,Right],
+	  Functor \== (:),
 	  is_op_type(Functor, infix), !
 	},
 	latex(cmd(infixop(Functor, \pred_arg(Left, 1), \pred_arg(Right, 2)))).
@@ -932,6 +982,14 @@ pred_head(Head, _Options) -->			% Postfix operators
 	  is_op_type(Functor, postfix), !
 	},
 	latex(cmd(postfixop(Functor, \pred_arg(Arg, 1)))).
+pred_head(M:Head, Options) --> !,		% Qualified predicates
+	{ pred_attributes(Options, Atts),
+	  Head =.. [Functor|Args],
+	  length(Args, Arity)
+	},
+	latex(cmd(qpredicate(opt(Atts),
+			     M,
+			     Functor, Arity, \pred_args(Args, 1)))).
 pred_head(Head, Options) -->			% Plain terms
 	{ pred_attributes(Options, Atts),
 	  Head =.. [Functor|Args],
@@ -953,6 +1011,8 @@ pred_att(Options, Det) :-
 	option(det(Det), Options).
 pred_att(Options, private) :-
 	option(class(privdef), Options).
+pred_att(Options, multifile) :-
+	option(class(multidef), Options).
 
 insert_comma([H1,H2|T0], [H1, ','|T]) :- !,
 	insert_comma([H2|T0], T).
@@ -1078,7 +1138,7 @@ latex_table(_Attrs, Content) -->
 	},
 %	latex(cmd(begin(table, opt(h)))),
 	latex(cmd(begin(quote))),
-	latex(cmd(begin(tabular, Format))),
+	latex(cmd(begin(tabular, no_escape(Format)))),
 	latex(cmd(hline)),
 	rows(Content),
 	latex(cmd(hline)),
@@ -1139,8 +1199,12 @@ latex_summary(_) :-
 pi_sort_key(M:PI, PI-(M:PI)) :- !.
 pi_sort_key(PI, PI-PI).
 
-object_name_arity(_:Name/Arity, Name, Arity).
+object_name_arity(_:Term, Name, Arity) :-
+	nonvar(Term), !,
+	object_name_arity(Term, Name, Arity).
 object_name_arity(Name/Arity, Name, Arity).
+object_name_arity(Name//Arity0, Name, Arity) :-
+	Arity is Arity0 + 2.
 
 summarylist(Objs, Options) -->
 	latex(cmd(begin(summarylist, ll))),
@@ -1224,6 +1288,8 @@ print_latex_token(latex(Code), Out) :- !,
 	write(Out, Code).
 print_latex_token(w(Word), Out) :- !,
 	print_latex(Out, Word).
+print_latex_token(no_escape(Text), Out) :- !,
+	write(Out, Text).
 print_latex_token(Rest, Out) :-
 	(   atomic(Rest)
 	->  print_latex(Out, Rest)
@@ -1287,10 +1353,12 @@ print_char('>', Out) :- !, write(Out, '$>$').
 print_char('{', Out) :- !, write(Out, '\\{').
 print_char('}', Out) :- !, write(Out, '\\}').
 print_char('$', Out) :- !, write(Out, '\\$').
+print_char('&', Out) :- !, write(Out, '\\&').
 print_char('#', Out) :- !, write(Out, '\\#').
 print_char('%', Out) :- !, write(Out, '\\%').
 print_char('\\',Out) :- !, write(Out, '\\bsl{}').
-print_char('^', Out) :- !, write(Out, '$\\wedge$').
+print_char('^', Out) :- !, write(Out, '\\Shat{}').
+print_char('|', Out) :- !, write(Out, '\\Sbar{}').
 print_char(C,   Out) :- put_char(Out, C).
 
 
