@@ -19,7 +19,7 @@
 
     You should have received a copy of the GNU General Public
     License along with this library; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 
     As a special exception, if you link this library with other files,
     compiled with a Free Software compiler, to produce an executable, this
@@ -41,6 +41,7 @@
 :- use_module(library(lists)).
 :- use_module(library(memfile)).
 :- use_module(library(operators)).
+:- use_module(library(error)).
 
 /** <module> Analyse PlDoc mode declarations
 
@@ -65,11 +66,6 @@ operator declarations in this module.
 		 *	       MODES		*
 		 *******************************/
 
-%	TBD: Extensive tests on style and completeness
-
-:- dynamic
-	mode/3.				% ?Mode, ?Module, ?Det
-
 %%	process_modes(+Lines:lines, +Module, +FilePos,
 %%		      -Modes:list, -Args:list(atom),
 %%		      -RestLines:lines) is det.
@@ -91,14 +87,22 @@ process_modes(Lines, Module, FilePos, ModeDecls, Vars, RestLines) :-
 %%	mode_lines(+Lines, -ModeText:codes, ?ModeTail:codes, -Lines) is det.
 %
 %	Extract the formal header. For %%   comments these are all lines
-%	starting with %%. For /** comments these  are all lines upto the
-%	first blank line.
+%	starting with %%. For /** comments,   first skip empty lines and
+%	then take all lines upto the   first  blank line. Skipping empty
+%	lines allows for comments using this style:
+%
+%	  ==
+%	  /**
+%	   * predicate(+arg1:type1, ?arg2:type2) is det
+%	   ...
+%	  ==
 
 mode_lines(Lines0, ModeText, ModeTail, Lines) :-
 	percent_mode_line(Lines0, ModeText, ModeTail0, Lines1), !,
 	percent_mode_lines(Lines1, ModeTail0, ModeTail, Lines).
 mode_lines(Lines0, ModeText, ModeTail, Lines) :-
-	non_empty_lines(Lines0, ModeText, ModeTail, Lines).
+	empty_lines(Lines0, Lines1),
+	non_empty_lines(Lines1, ModeText, ModeTail, Lines).
 
 percent_mode_line([1-[0'%|L]|Lines], ModeText, ModeTail, Lines) :-	%'
 	append(L, [10|ModeTail], ModeText).
@@ -108,6 +112,10 @@ percent_mode_lines(Lines0, ModeText, ModeTail, Lines) :-
 	percent_mode_lines(Lines1, ModeTail1, ModeTail, Lines).
 percent_mode_lines(Lines, Mode, Mode, Lines).
 
+empty_lines([_-[]|Lines0], Lines) :- !,
+	empty_lines(Lines0, Lines).
+empty_lines(Lines, Lines).
+
 non_empty_lines([], ModeTail, ModeTail, []).
 non_empty_lines([_-[]|Lines], ModeTail, ModeTail, Lines) :- !.
 non_empty_lines([_-L|Lines0], ModeText, ModeTail, Lines) :-
@@ -134,18 +142,29 @@ modes(Text, FilePos, Decls) :-
 	->  !
 	;   E = error(syntax_error(end_of_file), _)
 	->  fail
-	;   !, print_message(warning, E),
+	;   !, mode_syntax_error(E),
 	    Decls = []
 	).
 modes(Text, FilePos, Decls) :-
 	catch(read_mode_terms(Text, FilePos, ' . ', Decls), E, true),
 	(   var(E)
 	->  !
-	;   print_message(warning, E),
+	;   mode_syntax_error(E),
 	    fail
 	).
 modes(_, _, []).
 
+%%	mode_syntax_error(+ErrorTerm) is det.
+%
+%	Print syntax errors in  mode   declarations.  Currently, this is
+%	suppressed unless the flag =pldoc_errors= is specified.
+
+mode_syntax_error(E) :-
+	current_prolog_flag(pldoc_errors, true), !,
+	print_message(warning, E).
+mode_syntax_error(_).
+
+
 read_mode_terms(Text, File:Line, End, Terms) :-
 	new_memory_file(MemFile),
 	open_memory_file(MemFile, write, Out),
@@ -208,13 +227,9 @@ unprepare_module_operators :-
 %
 %	List is the list of operators exported from Module through its
 %	module header.
-%
-%	@tbd	Must be provided by a public interface from the core
 
 public_operators(Module, List) :-
-	Goal = Module:'$exported_op'(P,A,N),
-	predicate_property(Goal, _), !,
-	findall(op(P,A,N), Module:'$exported_op'(P,A,N), List).
+	module_property(Module, exported_operators(List)), !.
 public_operators(_, []).
 
 
@@ -238,7 +253,7 @@ extract_varnames(Name=_, [Name|VN], VN).
 %
 %	Assert modes into the database with the given position.
 %
-%	@param Modes	  List if mode-terms.  See process_modes/5.
+%	@param Modes	  List if mode-terms.  See process_modes/6.
 %	@param SourcePos  Term File:Line
 
 store_modes([], _).
@@ -248,7 +263,8 @@ store_modes([mode(Mode, _Bindings)|T], Pos) :-
 
 store_mode(Var, _) :-
 	var(Var), !,
-	throw(error(instantiation_error, context(_, 'PlDoc: Mode declaration expected'))).
+	throw(error(instantiation_error,
+		    context(_, 'PlDoc: Mode declaration expected'))).
 store_mode(Head0 is Det, Pos) :- !,
 	dcg_expand(Head0, Head),
 	compile_clause('$mode'(Head, Det), Pos).
@@ -256,6 +272,9 @@ store_mode(Head0, Pos) :-
 	dcg_expand(Head0, Head),
 	compile_clause('$mode'(Head, unknown), Pos).
 
+dcg_expand(M:Head0, M:Head) :-
+	atom(M), !,
+	dcg_expand(Head0, Head).
 dcg_expand(//(Head0), Head) :- !,
 	Head0 =.. [Name|List0],
 	maplist(remove_argname, List0, List1),
@@ -264,6 +283,12 @@ dcg_expand(//(Head0), Head) :- !,
 dcg_expand(Head0, Head) :-
 	remove_argnames(Head0, Head).
 
+remove_argnames(Var, _) :-
+	var(Var), !,
+	instantiation_error(Var).
+remove_argnames(M:Head0, M:Head) :- !,
+	must_be(atom, M),
+	remove_argnames(Head0, Head).
 remove_argnames(Head0, Head) :-
 	functor(Head0, Name, Arity),
 	functor(Head, Name, Arity),
@@ -408,12 +433,15 @@ mode_to_pi(Head is _Det, PI) :- !,
 mode_to_pi(Head, PI) :-
 	head_to_pi(Head, PI).
 
+head_to_pi(M:Head, M:PI) :-
+	atom(M), !,
+	head_to_pi(Head, PI).
 head_to_pi(//(Head), Name//Arity) :- !,
 	functor(Head, Name, Arity).
 head_to_pi(Head, Name/Arity) :-
 	functor(Head, Name, Arity).
 
-%%	compile_clause(+Term, +FilePos) is det.
+%%	compile_clause(:Term, +FilePos) is det.
 %
 %	Add a clause to the  compiled   program.  Unlike  assert/1, this
 %	associates the clause with the   given source-location, makes it
@@ -425,15 +453,17 @@ head_to_pi(Head, Name/Arity) :-
 %	@param FilePos	Term of the form File:Line, where File is a
 %			canonical filename.
 
-compile_clause(Term, FilePos) :-
+compile_clause(Term, File:Line) :-
 	'$set_source_module'(SM, SM),
-	clause_head(Term, Head),
+	strip_module(SM:Term, M, Plain),
+	clause_head(Plain, Head),
 	functor(Head, Name, Arity),
-	(   SM == user
-	->  multifile(SM:(Name/Arity))
-	;   discontiguous(SM:(Name/Arity))
+	multifile(M:(Name/Arity)),
+	(   M == SM
+	->  Clause = Term
+	;   Clause = M:Term
 	),
-	'$record_clause'(Term, FilePos, _Ref).
+	'$store_clause'('$source_location'(File, Line):Clause, File).
 
 clause_head((Head :- _Body), Head) :- !.
 clause_head(Head, Head).
