@@ -3,9 +3,10 @@
     Part of SWI-Prolog
 
     Author:        Jan Wielemaker
-    E-mail:        wielemak@science.uva.nl
+    E-mail:        J.Wielemaker@cs.vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (C): 2006, University of Amsterdam
+    Copyright (C): 2011, University of Amsterdam
+			 VU University Amsterdam
 
     This program is free software; you can redistribute it and/or
     modify it under the terms of the GNU General Public License
@@ -19,7 +20,7 @@
 
     You should have received a copy of the GNU General Public
     License along with this library; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 
     As a special exception, if you link this library with other files,
     compiled with a Free Software compiler, to produce an executable, this
@@ -63,6 +64,15 @@ well formatted HTML documents.
 @license GPL
 */
 
+:- predicate_options(doc_file_name/3, 3,
+		     [ format(oneof([html,tex]))
+		     ]).
+
+%%	prolog:predicate_summary(+PI, -Summary) is semidet.
+%
+%	Provide    predicate    summaries    to     the    XPCE    class
+%	=prolog_predicate=, used by the IDE tools.
+
 :- multifile
 	prolog:predicate_summary/2.	% ?PI, -Summary
 
@@ -137,7 +147,7 @@ is_structured_comment(Comment, Prefixes) :-
 	\+ blanks_to_nl(Comment),
 	\+ sub_string(Comment, 2, _, _, ' SWI '),	% HACK
 	\+ sub_string(Comment, 2, _, _, ' SICStus '),	% HACK
-	\+ sub_string(Comment, 2, _, _, ' Mats '), 	% HACK
+	\+ sub_string(Comment, 2, _, _, ' Mats '),	% HACK
 	Prefixes = ["%"].
 is_structured_comment(Comment, Prefixes) :-
 	sub_string(Comment, 0, _, _, '/**'), !,
@@ -165,10 +175,7 @@ blanks_to_nl(_).
 
 structured_comment(["%"]) -->
 	"%%", space,
-	\+ blanks_to_nl,
-	\+ contains(" SWI "),
-	\+ contains(" SICStus "),
-	\+ contains(" Mats ").
+	\+ separator_line.
 structured_comment(Prefixes) -->
 	"/**", space,
 	{ Prefixes = ["/**", " *"]
@@ -178,25 +185,28 @@ space -->
 	[H],
 	{ code_type(H, space) }.
 
-blanks_to_nl -->
-	blank_or_percents,
-	"\n".
+%%	separator_line// is semidet.
+%
+%	Matches a line like %% SWI or %%%%%%%%%%%%%%%%%%%%%%%%%, etc.
 
-blank_or_percents -->
-	space, !,
-	blank_or_percents.
-blank_or_percents -->
-	"%", !,
-	blank_or_percents.
-blank_or_percents -->
-	"".
+separator_line -->
+	string(S), "\n", !,
+	{   maplist(blank_or_percent, S)
+	;   contains(S, " SWI ")
+	;   contains(S, " SICStus ")
+	;   contains(S, " Mats ")
+	}.
 
-contains(String) -->
-	...,
-	String, !.
+string([]) --> [].
+string([H|T]) --> [H], string(T).
 
-... --> "".
-... --> [_], ... .
+blank_or_percent(0'%) :- !.
+blank_or_percent(C) :-
+	code_type(C, space).
+
+contains(Haystack, Needle) :-
+	append(_, Start, Haystack),
+	append(Needle, _, Start), !.
 
 
 %%	doc_file_name(+Source:atom, -Doc:atom, +Options:list) is det.
@@ -207,8 +217,8 @@ contains(String) -->
 %	@param Doc	the name of the file documenting Source.
 %	@param Options	Option list:
 %
-%			* format(-Format)
-%			Output format.  One of =html= or =latex=
+%			* format(+Format)
+%			Output format.  One of =html= or =tex=
 %
 %	@error	permission_error(overwrite, Source)
 
@@ -234,8 +244,8 @@ doc_file_name(Source, Doc, Options) :-
 %		* Name//Arity
 %		DCG rule indicator.  Same as Name/Arity+2
 %
-%		* module(Module)
-%		Comment appearing in Module.
+%		* module(ModuleTitle)
+%		Comment appearing in a module.
 %
 %	If Object is  unbound  and  multiple   objects  share  the  same
 %	description, Object is unified with a   list  of terms described
@@ -246,21 +256,19 @@ doc_file_name(Source, Doc, Options) :-
 
 doc_comment(Object, Pos, Summary, Comment) :-
 	var(Object), !,
-	current_module(M),
-	'$c_current_predicate'(_, M:'$pldoc'(_,_,_,_)),
+	locally_defined(M:'$pldoc'/4),
 	M:'$pldoc'(Obj, Pos, Summary, Comment),
 	qualify(M, Obj, Object0),
-	(   '$c_current_predicate'(_, M:'$pldoc_link'(_, _)),
+	(   locally_defined(M:'$pldoc_link'/2),
 	    findall(L, M:'$pldoc_link'(L, Obj), Ls), Ls \== []
 	->  maplist(qualify(M),	Ls, QLs),
 	    Object = [Object0|QLs]
 	;   Object = Object0
 	).
 doc_comment(M:Object, Pos, Summary, Comment) :- !,
-	current_module(M),
-	'$c_current_predicate'(_, M:'$pldoc'(_,_,_,_)),
+	locally_defined(M:'$pldoc'/4),
 	(   M:'$pldoc'(Object, Pos, Summary, Comment)
-	;   '$c_current_predicate'(_, M:'$pldoc_link'(_, _)),
+	;   locally_defined(M:'$pldoc_link'/2),
 	    M:'$pldoc_link'(Object, Obj2),
 	    M:'$pldoc'(Obj2, Pos, Summary, Comment)
 	).
@@ -269,6 +277,13 @@ doc_comment(Name/Arity, Pos, Summary, Comment) :-
 	doc_comment(M:Name/Arity, Pos, Summary, Comment).
 
 
+locally_defined(M:Name/Arity) :-
+	current_module(M),
+	current_predicate(M:Name/Arity),
+	functor(Head, Name, Arity),
+	\+ predicate_property(M:Head, imported_from(_)).
+
+
 qualify(M, H, H) :- system_module(M), !.
 qualify(M, H, H) :- sub_atom(M, 0, _, _, $), !.
 qualify(M, H, M:H).
@@ -304,6 +319,10 @@ prolog:predicate_summary(PI, Summary) :-
 %		Generated from Name(Arg, ...)
 %		* Name//Arity
 %		Generated from Name(Arg, ...)//
+%
+%	@param Comments is a list Pos-Comment returned by read_term/3
+%	@param TermPos is the start-location of the actual term
+%	@param File is the file that is being loaded.
 
 process_comments([], _, _).
 process_comments([Pos-Comment|T], TermPos, File) :-
@@ -320,11 +339,13 @@ process_comment(Pos, Comment, File) :-
 	process_structured_comment(FilePos, Comment, Prefixes).
 process_comment(_, _, _).
 
-%%	process_structured_comment(+FilePos, +Comment:string) is det.
+%%	process_structured_comment(+FilePos,
+%%				   +Comment:string,
+%%				   +Prefixed:list) is det.
 
 process_structured_comment(FilePos, Comment, _) :- % already processed
 	prolog_load_context(module, M),
-	'$c_current_predicate'(_, M:'$pldoc'(_,_,_,_)),
+	locally_defined(M:'$pldoc'/4),
 	catch(M:'$pldoc'(_, FilePos, _, Comment), _, fail), !.
 process_structured_comment(FilePos, Comment, Prefixes) :-
 	string_to_list(Comment, CommentCodes),
@@ -336,15 +357,41 @@ process_structured_comment(FilePos, Comment, Prefixes) :-
 	;   prolog_load_context(module, Module),
 	    process_modes(Lines, Module, FilePos, Modes, _, RestLines)
 	->  store_modes(Modes, FilePos),
-	    modes_to_predicate_indicators(Modes, [PI0|PIs]),
+	    modes_to_predicate_indicators(Modes, AllPIs),
+	    decl_module(AllPIs, M, [PI0|PIs]),
 	    summary_from_lines(RestLines, Codes),
 	    string_to_list(Summary, Codes),
-	    compile_clause('$pldoc'(PI0, FilePos, Summary, Comment), FilePos),
+	    compile_clause(M:'$pldoc'(PI0, FilePos, Summary, Comment),
+			   FilePos),
 	    forall(member(PI, PIs),
-		   compile_clause('$pldoc_link'(PI, PI0), FilePos))
+		   compile_clause(M:'$pldoc_link'(PI, PI0), FilePos))
 	), !.
-process_structured_comment(File:Line, Comment, _) :-
+process_structured_comment(Location, Comment, _) :-
 	print_message(warning,
-		      format('~w:~d: Failed to process comment:~n~s~n',
-			     [File, Line, Comment])),
+		      pldoc(invalid_comment(Location, Comment))),
 	fail.
+
+decl_module([], M, []) :-
+	(   var(M)
+	->  prolog_load_context(module, M)
+	;   true
+	).
+decl_module([H0|T0], M, [H|T]) :-
+	(   H0 = M1:H
+	->  M = M1
+	;   H = H0
+	),
+	decl_module(T0, M, T).
+
+
+		 /*******************************
+		 *	     MESSAGES		*
+		 *******************************/
+
+:- multifile
+	prolog:message//1.
+
+prolog:message(pldoc(invalid_comment(File:Line, Comment))) -->
+	[ '~w:~d: PlDoc: failed to process structured comment:~n~s~n'-
+		[File, Line, Comment]
+	].
