@@ -3,7 +3,7 @@
     Part of SWI-Prolog
 
     Author:        Jan Wielemaker
-    E-mail:        wielemak@science.uva.nl
+    E-mail:        jan@swi-prolog.org
     WWW:           http://www.swi-prolog.org
     Copyright (C): 2006-2008, University of Amsterdam
 
@@ -32,6 +32,9 @@ please visit http://www.swi-prolog.org/pldoc/package/plunit.html.
 @license	artistic
 */
 
+:- meta_predicate valid_options(+, 1).
+
+
 		 /*******************************
 		 *    CONDITIONAL COMPILATION	*
 		 *******************************/
@@ -82,7 +85,7 @@ sicstus :- catch(current_prolog_flag(system_type, _), _, fail).
 
 :- if(swi).
 throw_error(Error_term,Impldef) :-
-	throw(error(Error_term,Impldef)).
+	throw(error(Error_term,context(Impldef,_))).
 
 :- set_prolog_flag(generate_debug_info, false).
 :- use_module(library(option)).
@@ -92,7 +95,7 @@ current_test_flag(Name, Value) :-
 	current_prolog_flag(Name, Value).
 
 set_test_flag(Name, Value) :-
-	set_prolog_flag(Name, Value).
+	create_prolog_flag(Name, Value, []).
 :- endif.
 
 :- if(sicstus).
@@ -176,6 +179,11 @@ user:term_expansion((:- thread_local(PI)), (:- dynamic(PI))) :-
 %		If =true= and supported by the hosting Prolog, code
 %		is run in all supported unification mode and reported
 %		if the results are inconsistent.
+%
+%		* cleanup(+Bool)
+%		If =true= (default =false), cleanup report at the end
+%		of run_tests/1.  Used to improve cooperation with
+%		memory debuggers such as dmalloc.
 
 set_test_options(Options) :-
 	valid_options(Options, global_test_option),
@@ -189,6 +197,8 @@ global_test_option(silent(Bool)) :-
 	must_be(boolean, Bool).
 global_test_option(sto(Bool)) :-
 	must_be(boolean, Bool).
+global_test_option(cleanup(Bool)) :-
+	must_be(boolean, Bool).
 
 
 %%	loading_tests
@@ -232,15 +242,13 @@ begin_tests(Unit, Options) :-
 begin_tests(Unit, Name, File:Line, Options) :-
 	loading_tests, !,
 	'$set_source_module'(Context, Context),
-	Supers = [Context],
-	(   current_unit(Unit, Name, Supers, Options)
+	(   current_unit(Unit, Name, Context, Options)
 	->  true
 	;   retractall(current_unit(Unit, Name, _, _)),
-	    assert(current_unit(Unit, Name, Supers, Options)),
-	    set_import_modules(Name, Supers)
+	    assert(current_unit(Unit, Name, Context, Options))
 	),
 	'$set_source_module'(Old, Name),
-	'$declare_module'(Name, File, Line, false),
+	'$declare_module'(Name, test, Context, File, Line, false),
 	discontiguous(Name:'unit test'/4),
 	'$set_predicate_attribute'(Name:'unit test'/4, trace, 0),
 	discontiguous(Name:'unit body'/2),
@@ -249,11 +257,6 @@ begin_tests(Unit, Name, File:_Line, _Options) :-
 	'$set_source_module'(Old, Old),
 	asserta(loading_unit(Unit, Name, File, Old)).
 
-set_import_modules(Module, Imports) :-
-	findall(I, import_module(Module, I), IL),
-	forall(member(I, IL), delete_import_module(Module, I)),
-	forall(member(I, Imports), add_import_module(Module, I, end)).
-
 :- else.
 
 % we cannot use discontiguous as a goal in SICStus Prolog.
@@ -267,10 +270,10 @@ user:term_expansion((:- begin_tests(Set)),
 
 begin_tests(Unit, Name, File:_Line, Options) :-
 	loading_tests, !,
-	(   current_unit(Unit, Name, Supers, Options)
+	(   current_unit(Unit, Name, _, Options)
 	->  true
 	;   retractall(current_unit(Unit, Name, _, _)),
-	    assert(current_unit(Unit, Name, Supers, Options))
+	    assert(current_unit(Unit, Name, -, Options))
 	),
 	asserta(loading_unit(Unit, Name, File, -)).
 begin_tests(Unit, Name, File:_Line, _Options) :-
@@ -306,7 +309,9 @@ unit_module(Unit, Module) :-
 make_unit_module(Unit, Module) :-
 	unit_module(Unit, Module),
 	(   current_module(Module),
-	    \+ current_unit(_, Module, _, _)
+	    \+ current_unit(_, Module, _, _),
+	    predicate_property(Module:H, _P),
+	    \+ predicate_property(Module:H, imported_from(_M))
 	->  throw_error(permission_error(create, plunit, Unit),
 			'Existing module')
 	;  true
@@ -348,8 +353,9 @@ expand_test(Name, Options0, Body,
 	->  Options1 = Options0
 	;   Options1 = [Options0]
 	),
-	maplist(expand_option, Options1, Options),
-	valid_options(Options, test_option).
+	maplist(expand_option, Options1, Options2),
+	valid_options(Options2, test_option),
+	valid_test_mode(Options2, Options).
 
 expand_option(Var, _) :-
 	var(Var), !,
@@ -358,8 +364,27 @@ expand_option(A == B, true(A==B)) :- !.
 expand_option(A = B, true(A=B)) :- !.
 expand_option(A =@= B, true(A=@=B)) :- !.
 expand_option(A =:= B, true(A=:=B)) :- !.
+expand_option(error(X), throws(error(X, _))) :- !.
+expand_option(exception(X), throws(X)) :- !. % SICStus 4 compatibility
+expand_option(error(F,C), throws(error(F,C))) :- !. % SICStus 4 compatibility
+expand_option(true, true(true)) :- !.
 expand_option(O, O).
 
+valid_test_mode(Options0, Options) :-
+	include(test_mode, Options0, Tests),
+	(   Tests == []
+	->  Options = [true(true)|Options0]
+	;   Tests = [_]
+	->  Options = Options0
+	;   throw_error(plunit(incompatible_options, Tests), _)
+	).
+
+test_mode(true(_)).
+test_mode(all(_)).
+test_mode(set(_)).
+test_mode(fail).
+test_mode(throws(_)).
+
 
 %%	expand(+Term, -Clauses) is semidet.
 
@@ -367,6 +392,8 @@ expand(end_of_file, _) :-
 	loading_unit(Unit, _, _, _), !,
 	end_tests(Unit),		% warn?
 	fail.
+expand((:-end_tests(_)), _) :- !,
+	fail.
 expand(_Term, []) :-
 	\+ loading_tests.
 expand((test(Name) :- Body), Clauses) :- !,
@@ -380,10 +407,10 @@ expand(test(Name, _Options), _) :- !,
 
 :- if(swi).
 :- multifile
-	user:term_expansion/2.
+	system:term_expansion/2.
 :- endif.
 
-user:term_expansion(Term, Expanded) :-
+system:term_expansion(Term, Expanded) :-
 	(   loading_unit(_, _, File, _)
 	->  source_location(File, _),
 	    expand(Term, Expanded)
@@ -442,9 +469,7 @@ test_option(Option) :-
 	test_set_option(Option), !.
 test_option(true(_)).
 test_option(fail).
-test_option(true).
 test_option(throws(_)).
-test_option(error(_)).
 test_option(all(_)).
 test_option(set(_)).
 test_option(nondet).
@@ -476,6 +501,7 @@ test_set_option(sto(V)) :-
 :- thread_local
 	passed/5,			% Unit, Test, Line, Det, Time
 	failed/4,			% Unit, Test, Line, Reason
+	failed_assertion/7,		% Unit, Test, Line, ALoc, STO, Reason, Goal
 	blocked/4,			% Unit, Test, Line, Reason
 	sto/4,				% Unit, Test, Line, Results
 	fixme/5.			% Unit, Test, Line, Reason, Status
@@ -485,17 +511,35 @@ test_set_option(sto(V)) :-
 
 %%	run_tests is semidet.
 %%	run_tests(+TestSet) is semidet.
+%
+%	Run  tests  and  report  about    the   results.  The  predicate
+%	run_tests/0 runs all known  tests  that   are  not  blocked. The
+%	predicate run_tests/1 takes a  specification   of  tests to run.
+%	This  is  either  a  single   specification    or   a   list  of
+%	specifications. Each single specification is  either the name of
+%	a test-unit or a term <test-unit>:<test>, denoting a single test
+%	within a unit.
 
 run_tests :-
 	cleanup,
-	forall(current_test_set(Set),
-	       run_unit(Set)),
-	report.
+	setup_call_cleanup(
+	    setup_trap_assertions(Ref),
+	    forall(current_test_set(Set),
+		   run_unit(Set)),
+	    ( cleanup_trap_assertions(Ref),
+	      report,
+	      cleanup_after_test
+	    )).
 
 run_tests(Set) :-
 	cleanup,
-	run_unit(Set),
-	report.
+	setup_call_cleanup(
+	    setup_trap_assertions(Ref),
+	    run_unit(Set),
+	    ( cleanup_trap_assertions(Ref),
+	      report,
+	      cleanup_after_test
+	    )).
 
 run_unit([]) :- !.
 run_unit([H|T]) :- !,
@@ -542,11 +586,20 @@ cleanup :-
 	thread_self(Me),
 	retractall(passed(_, _, _, _, _)),
 	retractall(failed(_, _, _, _)),
+	retractall(failed_assertion(_, _, _, _, _, _, _)),
 	retractall(blocked(_, _, _, _)),
 	retractall(sto(_, _, _, _)),
 	retractall(fixme(_, _, _, _, _)),
 	retractall(running(_,_,_,_,Me)).
 
+cleanup_after_test :-
+	current_test_flag(test_options, Options),
+	option(cleanup(Cleanup), Options, false),
+	(   Cleanup == true
+	->  cleanup
+	;   true
+	).
+
 
 %%	run_tests_in_files(+Files:list)	is det.
 %
@@ -622,6 +675,59 @@ unification_capability(_) :-
 :- endif.
 :- endif.
 
+		 /*******************************
+		 *	ASSERTION HANDLING	*
+		 *******************************/
+
+:- if(swi).
+
+:- dynamic prolog:assertion_failed/2.
+
+setup_trap_assertions(Ref) :-
+	asserta((prolog:assertion_failed(Reason, Goal) :-
+			test_assertion_failed(Reason, Goal)),
+		Ref).
+
+cleanup_trap_assertions(Ref) :-
+	erase(Ref).
+
+test_assertion_failed(Reason, Goal) :-
+	thread_self(Me),
+	running(Unit, Test, Line, STO, Me),
+	(   catch(get_prolog_backtrace(10, Stack), _, fail),
+	    assertion_location(Stack, AssertLoc)
+	->  true
+	;   AssertLoc = unknown
+	),
+	current_test_flag(test_options, Options),
+	report_failed_assertion(Unit, Test, Line, AssertLoc,
+				STO, Reason, Goal, Options),
+	assert_cyclic(failed_assertion(Unit, Test, Line, AssertLoc,
+				       STO, Reason, Goal)).
+
+assertion_location(Stack, File:Line) :-
+	append(_, [AssertFrame,CallerFrame|_], Stack),
+	prolog_stack_frame_property(AssertFrame,
+				    predicate(prolog_debug:assertion/1)), !,
+	prolog_stack_frame_property(CallerFrame, location(File:Line)).
+
+report_failed_assertion(Unit, Test, Line, AssertLoc,
+			STO, Reason, Goal, _Options) :-
+	print_message(
+	    error,
+	    plunit(failed_assertion(Unit, Test, Line, AssertLoc,
+				    STO, Reason, Goal))).
+
+:- else.
+
+setup_trap_assertions(_).
+cleanup_trap_assertions(_).
+
+:- endif.
+
+
+
+
 
 		 /*******************************
 		 *	   RUNNING A TEST	*
@@ -643,7 +749,10 @@ run_test(Unit, Name, Line, Options, Body) :-
 run_test_once(Unit, Name, Line, Options, Body) :-
 	current_test_flag(test_options, GlobalOptions),
 	option(sto(false), GlobalOptions, false), !,
+	current_unification_capability(Type),
+	begin_test(Unit, Name, Line, Type),
 	run_test_6(Unit, Name, Line, Options, Body, Result),
+	end_test(Unit, Name, Line, Type),
 	report_result(Result, Options).
 run_test_once(Unit, Name, Line, Options, Body) :-
 	current_unit(Unit, _Module, _Supers, UnitOptions),
@@ -661,7 +770,9 @@ run_test_cap(Unit, Name, Line, Options, Body) :-
 	(   option(sto(Type), Options)
 	->  unification_capability(Type),
 	    set_unification_capability(Type),
+	    begin_test(Unit, Name, Line, Type),
 	    run_test_6(Unit, Name, Line, Options, Body, Result),
+	    end_test(Unit, Name, Line, Type),
 	    report_result(Result, Options)
 	;   findall(Key-(Type+Result),
 		    test_caps(Type, Unit, Name, Line, Options, Body, Result, Key),
@@ -760,8 +871,11 @@ run_test_6(Unit, Name, Line, Options, Body, Result) :-
 	    ->  (   var(E)
 		->  statistics(runtime, [T1,_]),
 		    Time is (T1 - T0)/1000.0,
-		    (   catch(Cmp, _, fail)			% tbd: error
-		    ->  Result = success(Unit, Name, Line, Det, Time)
+		    (   catch(Module:Cmp, E, true)
+		    ->  (   var(E)
+			->  Result = success(Unit, Name, Line, Det, Time)
+			;   Result = failure(Unit, Name, Line, cmp_error(Cmp, E))
+			)
 		    ;   Result = failure(Unit, Name, Line, wrong_answer(Cmp))
 		    ),
 		    cleanup(Module, Options)
@@ -774,11 +888,7 @@ run_test_6(Unit, Name, Line, Options, Body, Result) :-
 	;   Result = setup_failed(Unit, Name, Line)
 	).
 run_test_6(Unit, Name, Line, Options, Body, Result) :-
-	(   option(throws(Expect), Options)
-	->  true
-	;   option(error(ErrorExpect), Options)
-	->  Expect = error(ErrorExpect, _)
-	), !,
+	option(throws(Expect), Options), !,
 	unit_module(Unit, Module),
 	(   setup(Module, test(Unit,Name,Line), Options)
 	->  statistics(runtime, [T0,_]),
@@ -799,24 +909,6 @@ run_test_6(Unit, Name, Line, Options, Body, Result) :-
 	    )
 	;   Result = setup_failed(Unit, Name, Line)
 	).
-run_test_6(Unit, Name, Line, Options, Body, Result) :-
-	unit_module(Unit, Module),
-	(   setup(Module, test(Unit,Name,Line), Options)
-	->  statistics(runtime, [T0,_]),
-	    (   catch(call_det(Module:Body, Det), E, true)
-	    ->  (   var(E)
-		->  statistics(runtime, [T1,_]),
-		    Time is (T1 - T0)/1000.0,
-		    Result = success(Unit, Name, Line, Det, Time),
-		    cleanup(Module, Options)
-		;   Result = failure(Unit, Name, Line, E),
-		    cleanup(Module, Options)
-		)
-	    ;   Result = failure(Unit, Name, Line, failed),
-		cleanup(Module, Options)
-	    )
-	;   Result = setup_failed(Unit, Name, Line)
-	).
 
 
 %%	non_det_test(+Expected, +Unit, +Name, +Line, +Options, +Body, -Result)
@@ -907,10 +999,10 @@ call_det(Goal, true) :-
 %%	match_error(+Expected, +Received) is semidet.
 %
 %	True if the Received errors matches the expected error. Matching
-%	is based on subsumes_chk/2.
+%	is based on subsumes_term/2.
 
 match_error(Expect, Rec) :-
-	subsumes_chk(Expect, Rec).
+	subsumes_term(Expect, Rec).
 
 %%	setup(+Module, +Context, +Options) is semidet.
 %
@@ -1083,16 +1175,18 @@ running_tests(Running) :-
 report :-
 	number_of_clauses(passed/5, Passed),
 	number_of_clauses(failed/4, Failed),
+	number_of_clauses(failed_assertion/7, FailedAssertion),
 	number_of_clauses(blocked/4, Blocked),
 	number_of_clauses(sto/4, STO),
-	(   Passed+Failed+Blocked+STO =:= 0
+	(   Passed+Failed+FailedAssertion+Blocked+STO =:= 0
 	->  info(plunit(no_tests))
-	;   Failed+Blocked+STO =:= 0
+	;   Failed+FailedAssertion+Blocked+STO =:= 0
 	->  report_fixme,
 	    info(plunit(all_passed(Passed)))
 	;   report_blocked,
 	    report_fixme,
 	    report_failed,
+	    report_failed_assertions,
 	    report_sto
 	).
 
@@ -1124,6 +1218,14 @@ report_failed :-
 report_failed :-
 	info(plunit(failed(0))).
 
+report_failed_assertions :-
+	number_of_clauses(failed_assertion/7, N),
+	N > 0, !,
+	info(plunit(failed_assertions(N))),
+	fail.
+report_failed_assertions :-
+	info(plunit(failed_assertions(0))).
+
 report_sto :-
 	number_of_clauses(sto/4, N),
 	N > 0, !,
@@ -1246,6 +1348,9 @@ locationprefix(unit(Unit)) --> !,
 locationprefix(FileLine) -->
 	{ throw_error(type_error(locationprefix,FileLine), _) }.
 
+:- discontiguous
+	message//1.
+
 message(error(context_error(plunit_close(Name, -)), _)) -->
 	[ 'PL-Unit: cannot close unit ~w: no open unit'-[Name] ].
 message(error(context_error(plunit_close(Name, Start)), _)) -->
@@ -1253,6 +1358,9 @@ message(error(context_error(plunit_close(Name, Start)), _)) -->
 message(plunit(nondet(File, Line, Name))) -->
 	locationprefix(File:Line),
 	[ 'PL-Unit: Test ~w: Test succeeded with choicepoint'- [Name] ].
+message(error(plunit(incompatible_options, Tests), _)) -->
+	[ 'PL-Unit: incompatible test-options: ~p'-[Tests] ].
+
 					% Unit start/end
 :- if(swi).
 message(plunit(begin(Unit))) -->
@@ -1300,6 +1408,12 @@ message(plunit(failed(1))) --> !,
 	[ '1 test failed'-[] ].
 message(plunit(failed(N))) -->
 	[ '~D tests failed'-[N] ].
+message(plunit(failed_assertions(0))) --> !,
+	[].
+message(plunit(failed_assertions(1))) --> !,
+	[ '1 assertion failed'-[] ].
+message(plunit(failed_assertions(N))) -->
+	[ '~D assertions failed'-[N] ].
 message(plunit(sto(0))) --> !,
 	[].
 message(plunit(sto(N))) -->
@@ -1312,13 +1426,55 @@ message(plunit(fixme(Failed,Passed,0))) -->
 	[ 'FIXME: ~D failed; ~D passed'-[Failed, Passed] ].
 message(plunit(fixme(Failed,Passed,Nondet))) -->
 	{ TotalPassed is Passed+Nondet },
-	[ 'FIXME: ~D failed; ~D passed; (~D nondet)'-[Failed, TotalPassed, Nondet] ].
+	[ 'FIXME: ~D failed; ~D passed; (~D nondet)'-
+	  [Failed, TotalPassed, Nondet] ].
 message(plunit(failed(Unit, Name, Line, Failure))) -->
        { unit_file(Unit, File) },
        locationprefix(File:Line),
        test_name(Name),
        [': '-[] ],
        failure(Failure).
+:- if(swi).
+message(plunit(failed_assertion(Unit, Name, Line, AssertLoc,
+				_STO, Reason, Goal))) -->
+       { unit_file(Unit, File) },
+       locationprefix(File:Line),
+       test_name(Name),
+       [ ': assertion'-[] ],
+       assertion_location(AssertLoc, File),
+       assertion_reason(Reason), ['\n\t'],
+       assertion_goal(Unit, Goal).
+
+assertion_location(File:Line, File) -->
+	[ ' at line ~w'-[Line] ].
+assertion_location(File:Line, _) -->
+	[ ' at ~w:~w'-[File, Line] ].
+assertion_location(unknown, _) -->
+	[].
+
+assertion_reason(fail) --> !,
+	[ ' failed'-[] ].
+assertion_reason(Error) -->
+	{ message_to_string(Error, String) },
+	[ ' raised "~w"'-[String] ].
+
+assertion_goal(Unit, Goal) -->
+	{ unit_module(Unit, Module),
+	  unqualify(Goal, Module, Plain)
+	},
+	[ 'Assertion: ~p'-[Plain] ].
+
+unqualify(Var, _, Var) :-
+	var(Var), !.
+unqualify(M:Goal, Unit, Goal) :-
+	nonvar(M),
+	unit_module(Unit, M), !.
+unqualify(M:Goal, _, Goal) :-
+	callable(Goal),
+	predicate_property(M:Goal, imported_from(system)), !.
+unqualify(Goal, _, Goal).
+
+:- endif.
 					% Setup/condition errors
 message(plunit(error(Where, Context, Exception))) -->
 	locationprefix(Context),
@@ -1432,6 +1588,9 @@ failure(wrong_answer(Cmp)) -->
 	[ 'wrong answer (compared using ~w)'-[Op], nl ],
 	expected_got_ops_(Ex, A, OPS, Goals).
 :- if(swi).
+failure(cmp_error(_Cmp, Error)) -->
+	{ message_to_string(Error, Message) },
+	[ 'Comparison error: ~w'-[Message] ].
 failure(Error) -->
 	{ Error = error(_,_), !,
 	  message_to_string(Error, Message)
@@ -1489,7 +1648,7 @@ user:generate_message_hook(Message) -->
 	message(Message),
 	[nl].				% SICStus requires nl at the end
 
-%	user:message_hook(+Severity, +Message, +Lines) is semidet.
+%%	user:message_hook(+Severity, +Message, +Lines) is semidet.
 %
 %	Redefine printing some messages. It appears   SICStus has no way
 %	to get multiple messages at the same   line, so we roll our own.
