@@ -1,4 +1,4 @@
-/*  $Id: rdf_persistency.pl,v 1.26 2007/07/09 13:56:38 jan Exp $
+/*  $Id$
 
     Part of SWI-Prolog
 
@@ -19,7 +19,7 @@
 
     You should have received a copy of the GNU Lesser General Public
     License along with this library; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 
     As a special exception, if you link this library with other files,
     compiled with a Free Software compiler, to produce an executable, this
@@ -44,6 +44,9 @@
 :- use_module(library(debug)).
 :- use_module(library(error)).
 :- use_module(library(thread)).
+:- use_module(library(pairs)).
+:- use_module(library(time)).
+
 
 /** <module> RDF persistency plugin
 
@@ -73,28 +76,42 @@ move the .new to the plain snapshot name as a means of recovery.
 
 :- volatile
 	rdf_directory/1,
+	rdf_lock/2,
 	rdf_option/1,
 	source_journal_fd/2,
-	db_file_base/2.
+	db_file_base/2,
+	file_base_db/2.
 :- dynamic
 	rdf_directory/1,		% Absolute path
+	rdf_lock/2,			% Dir, Lock
 	rdf_option/1,			% Defined options
 	source_journal_fd/2,		% DB, JournalFD
-	db_file_base/2.			% DB, FileBase
+	db_file_base/2,			% DB, FileBase
+	file_base_db/2.			% FileBase, DB
+
+:- meta_predicate
+	no_agc(0).
+
+:- predicate_options(rdf_attach_db/2, 2,
+		     [ concurrency(positive_integer),
+		       max_open_journals(positive_integer),
+		       silent(oneof([true,false,brief])),
+		       log_nested_transactions(boolean)
+		     ]).
 
 %%	rdf_attach_db(+Directory, +Options)
-%	
+%
 %	Start persistent operations using Directory   as  place to store
 %	files.   There are several cases:
-%	
+%
 %		* Empty DB, existing directory
 %		Load the DB from the existing directory
-%		
+%
 %		* Full DB, empty directory
 %		Create snapshots for all sources in directory
-%		
+%
 %	Options:
-%	
+%
 %		* concurrency(+Jobs)
 %		Number of threads to use for loading the initial
 %		database.  If not provided it is the number of CPUs
@@ -103,11 +120,12 @@ move the .new to the plain snapshot name as a means of recovery.
 %		* max_open_journals(+Count)
 %		Maximum number of journals kept open.  If not provided,
 %		the default is 10.  See limit_fd_pool/0.
-%		
-%		* silent(+Boolean)
+%
+%		* silent(+BoolOrBrief)
 %		If =true= (default =false=), do not print informational
-%		messages.
-%		
+%		messages.  Finally, if =brief= it will show minimal
+%		feedback.
+%
 %		* log_nested_transactions(+Boolean)
 %		If =true=, nested _log_ transactions are added to the
 %		journal information.  By default (=false=), no log-term
@@ -128,7 +146,7 @@ rdf_attach_db(DirSpec, Options) :-
 	    assert(rdf_directory(Directory)),
 	    assert_options(Options),
 	    stop_monitor,		% make sure not to register load
-	    load_db,
+	    no_agc(load_db),
 	    at_halt(rdf_detach_db),
 	    start_monitor
 	).
@@ -149,28 +167,41 @@ assert_options([H|T]) :-
 	(   option_type(H, Check)
 	->  Check,
 	    assert(rdf_option(H))
-	;   domain_error(rdf_option, H)
+	;   true			% ignore options we do not understand
 	),
 	assert_options(T).
-	
+
 option_type(concurrency(X),		must_be(positive_integer, X)).
 option_type(max_open_journals(X),	must_be(positive_integer, X)).
-option_type(silent(X),			must_be(boolean, X)).
+option_type(silent(X),	       must_be(oneof([true,false,brief]), X)).
 option_type(log_nested_transactions(X),	must_be(boolean, X)).
 
 
+%%	no_agc(:Goal)
+%
+%	Run Goal with atom garbage collection   disabled. Loading an RDF
+%	database creates large amounts  of  atoms   we  *know*  are  not
+%	garbage.
+
+no_agc(Goal) :-
+	current_prolog_flag(agc_margin, Old),
+	set_prolog_flag(agc_margin, 0),
+	call_cleanup(Goal, set_prolog_flag(agc_margin, Old)).
+
+
 %%	rdf_detach_db is det.
 %
 %	Detach from the  current  database.   Succeeds  silently  if  no
 %	database is attached. Normally called at  the end of the program
-%	through at_halt/1. 
+%	through at_halt/1.
 
 rdf_detach_db :-
-	debug(halt, 'Detaching database', []),
+	debug(halt, 'Detaching RDF database', []),
 	stop_monitor,
 	close_journals,
 	(   retract(rdf_directory(Dir))
 	->  debug(halt, 'DB Directory: ~w', [Dir]),
+	    save_prefixes(Dir),
 	    retractall(rdf_option(_)),
 	    retractall(source_journal_fd(_,_)),
 	    retractall(db_file_base(_,_)),
@@ -180,7 +211,7 @@ rdf_detach_db :-
 
 
 %%	rdf_current_db(?Dir)
-%	
+%
 %	True if Dir is the current RDF persistent database.
 
 rdf_current_db(Directory) :-
@@ -189,15 +220,16 @@ rdf_current_db(Directory) :-
 
 
 %%	rdf_flush_journals(+Options)
-%	
+%
 %	Flush dirty journals.  Options:
-%	
+%
 %		* min_size(+KB)
 %		Only flush if journal is over KB in size.
-%		TBD: sensible default size
+%
+%	@tbd Provide a default for min_size?
 
 rdf_flush_journals(Options) :-
-	forall(rdf_source(DB),
+	forall(rdf_graph(DB),
 	       rdf_flush_journal(DB, Options)).
 
 rdf_flush_journal(DB, Options) :-
@@ -224,15 +256,37 @@ rdf_flush_journal(DB, Options) :-
 
 load_db :-
 	rdf_directory(Dir),
+	load_prefixes(Dir),
 	working_directory(Old, Dir),
+	get_time(Wall0),
+	statistics(cputime, T0),
 	call_cleanup(find_dbs(DBs), working_directory(_, Old)),
-	make_goals(DBs, Goals),
+	length(DBs, DBCount),
+	verbosity(DBCount, Silent),
+	make_goals(DBs, Silent, 1, DBCount, Goals),
 	concurrency(Jobs),
-	concurrent(Jobs, Goals, []).
+	concurrent(Jobs, Goals, []),
+	statistics(cputime, T1),
+	get_time(Wall1),
+	T is T1 - T0,
+	Wall is Wall1 - Wall0,
+	message_level(Silent, Level),
+	print_message(Level, rdf(restore(attached(DBCount, T/Wall)))).
+
+
+make_goals([], _, _, _, []).
+make_goals([DB|T0], Silent, I,  Total,
+	   [load_source(DB, Silent, I, Total)|T]) :-
+	I2 is I + 1,
+	make_goals(T0, Silent, I2, Total, T).
+
+verbosity(_DBCount, Silent) :-
+	rdf_option(silent(Silent)), !.
+verbosity(DBCount, Silent) :-
+	DBCount > 25, !,
+	Silent = brief.
+verbosity(_DBCount, false).
 
-make_goals([], []).
-make_goals([DB|T0], [load_source(DB)|T]) :-
-	make_goals(T0, T).
 
 %%	concurrency(-Jobs)
 %
@@ -255,64 +309,69 @@ concurrency(1).
 
 find_dbs(DBs) :-
 	expand_file_name(*, Files),
-	dbs(Files, DBs0),
-	sort(DBs0, DBs1),		% remove duplicates
-	maplist(key_by_size, DBs1, DBK),
-	keysort(DBK, DBK1),
-	unkey(DBK1, DBs).
-	
-dbs([], []).
-dbs([H0|T0], [H|T]) :-
-	file_name_extension(Base, Ext, H0),
-	db_extension(Ext), !,
-	rdf_db_to_file(H, Base),
-	dbs(T0, T).
-dbs([_|T0], T) :-
-	dbs(T0, T).
+	phrase(scan_db_files(Files), Scanned),
+	sort(Scanned, ByDB),
+	join_snapshot_and_journals(ByDB, BySize),
+	keysort(BySize, SortedBySize),
+	pairs_values(SortedBySize, DBs).
+
+
+%%	scan_db_files(+Files)// is det.
+%
+%	Produces a list of db(DB,  Size,   File)  for all recognised RDF
+%	database files.
+
+scan_db_files([]) -->
+	[].
+scan_db_files([File|T]) -->
+	{ file_name_extension(Base, Ext, File),
+	  db_extension(Ext), !,
+	  rdf_db_to_file(DB, Base),
+	  size_file(File, Size)
+	},
+	[ db(DB, Size, File) ],
+	scan_db_files(T).
+scan_db_files([_|T]) -->
+	scan_db_files(T).
+
 
 db_extension(trp).
 db_extension(jrn).
 
-key_by_size(DB, Size-DB) :-
-	rdf_db_to_file(DB, Base),
-	ext_size(Base, trp, S1),
-	ext_size(Base, jrn, S2),
-	Size is S1 + S2.
+join_snapshot_and_journals([], []).
+join_snapshot_and_journals([db(DB,S0,_)|T0], [S-DB|T]) :- !,
+	same_db(DB, T0, T1, S0, S),
+	join_snapshot_and_journals(T1, T).
 
-ext_size(Base, Ext, Size) :-
-	file_name_extension(Base, Ext, File),
-	(   exists_file(File)
-	->  size_file(File, Size)
-	;   Size = 0
-	).
-
-unkey([], []).
-unkey([_-H|T0], [H|T]) :-
-	unkey(T0, T).
+same_db(DB, [db(DB,S1,_)|T0], T, S0, S) :- !,
+	S2 is S0+S1,
+	same_db(DB, T0, T, S2, S).
+same_db(_, L, L, S, S).
 
 
-%%	load_source(+DB) is det.
-%	
+:- public load_source/4.		% called through make_goals/5
+
+%%	load_source(+DB, +Silent, +Nth, +Total) is det.
+%
 %	Load triples and reload  journal   from  the  indicated snapshot
 %	file.
+%
+%	@param Silent One of =false=, =true= or =brief=
 
-load_source(DB) :-
-	(   rdf_option(silent(true))
-	->  Level = silent
-	;   Level = informational
-	),
+load_source(DB, Silent, Nth, Total) :-
+	message_level(Silent, Level),
 	db_files(DB, SnapshotFile, JournalFile),
 	rdf_retractall(_,_,_,DB),
 	statistics(cputime, T0),
-	print_message(Level, rdf(restore(source(DB)))),
-	(   exists_db(SnapshotFile)
-	->  print_message(Level, rdf(restore(snapshot(SnapshotFile)))),
-	    db_file(SnapshotFile, AbsFile),
-	    rdf_load_db(AbsFile)
+	print_message(Level, rdf(restore(Silent, source(DB, Nth, Total)))),
+	db_file(SnapshotFile, AbsSnapShot),
+	(   exists_file(AbsSnapShot)
+	->  print_message(Level, rdf(restore(Silent, snapshot(SnapshotFile)))),
+	    rdf_load_db(AbsSnapShot)
 	;   true
 	),
 	(   exists_db(JournalFile)
-	->  print_message(Level, rdf(restore(journal(JournalFile)))),
+	->  print_message(Level, rdf(restore(Silent, journal(JournalFile)))),
 	    load_journal(JournalFile, DB)
 	;   true
 	),
@@ -322,13 +381,17 @@ load_source(DB) :-
 	->  true
 	;   Count = 0
 	),
-	print_message(Level, rdf(restore(done(DB, T, Count)))).
-	
-	
+	print_message(Level, rdf(restore(Silent,
+					 done(DB, T, Count, Nth, Total)))).
+
+message_level(true, silent) :- !.
+message_level(_, informational).
+
+
 		 /*******************************
 		 *	   LOAD JOURNAL		*
 		 *******************************/
-	
+
 %%	load_journal(+File:atom, +DB:atom) is det.
 %
 %	Process transactions from the RDF journal File, adding the given
@@ -340,7 +403,7 @@ load_journal(File, DB) :-
 			process_journal(T0, In, DB)
 		     ),
 		     close(In)).
-	
+
 process_journal(end_of_file, _, _) :- !.
 process_journal(Term, In, DB) :-
 	(   process_journal_term(Term, DB)
@@ -367,7 +430,7 @@ process_journal_term(start(_), _).	% journal open/close
 process_journal_term(end(_), _).
 process_journal_term(begin(_), _).	% logged transaction (compatibility)
 process_journal_term(end, _).
-process_journal_term(begin(_,_,_,_), _).% logged transaction (current)
+process_journal_term(begin(_,_,_,_), _). % logged transaction (current)
 process_journal_term(end(_,_,_), _).
 
 
@@ -421,7 +484,7 @@ stop_monitor :-
 %
 %	Handle an rdf_monitor/2 callback to  deal with persistency. Note
 %	that the monitor calls that come   from rdf_db.pl that deal with
-%	database changes are synchronised. They   do come from different
+%	database changes are serialized.  They   do  come from different
 %	threads though.
 
 monitor(Msg) :-
@@ -453,7 +516,7 @@ monitor(retract(S,P,O,DB)) :-
 	sync_journal(DB, Fd).
 monitor(update(S,P,O,DB:Line,Action)) :- !,
 	\+ blocked_db(DB, _),
-	(   Action = source(NewDB)
+	(   Action = graph(NewDB)
 	->  monitor(assert(S,P,O,NewDB)),
 	    monitor(retract(S,P,O,DB:Line))
 	;   journal_fd(DB, Fd),
@@ -462,7 +525,7 @@ monitor(update(S,P,O,DB:Line,Action)) :- !,
 	).
 monitor(update(S,P,O,DB,Action)) :-
 	\+ blocked_db(DB, _),
-	(   Action = source(NewDB)
+	(   Action = graph(NewDB)
 	->  monitor(assert(S,P,O,NewDB)),
 	    monitor(retract(S,P,O,DB))
 	;   journal_fd(DB, Fd),
@@ -511,7 +574,7 @@ monitor_transaction(log(_), end(N)) :-
 	check_nested(N),
 	retract(transaction_message(N, _, _)), !,
 	findall(DB:Id, retract(transaction_db(N, DB, Id)), DBs),
-	end_transactions(DBs, N).    
+	end_transactions(DBs, N).
 monitor_transaction(log(Msg, DB), begin(N)) :- !,
 	check_nested(N),
 	get_time(Time),
@@ -525,8 +588,9 @@ monitor_transaction(reset, begin(L)) :-
 	       monitor_transaction(unload(DB), begin(L))).
 monitor_transaction(reset, end(L)) :-
 	forall(blocked_db(DB, unload),
-	       monitor_transaction(unload(DB), end(L))).
-	       
+	       monitor_transaction(unload(DB), end(L))),
+	retractall(current_transaction_id(_,_)).
+
 
 %%	check_nested(+Level) is semidet.
 %
@@ -553,7 +617,8 @@ open_transaction(DB, Fd) :-
 	->  true
 	;   next_transaction_id(DB, Id),
 	    assert(transaction_db(N, DB, Id)),
-	    format(Fd, 'begin(~q, ~q, ~2f, ~q).~n', [Id, N, Time, Msg])
+	    RoundedTime is round(Time*100)/100,
+	    format(Fd, '~q.~n', [begin(Id, N, RoundedTime, Msg)])
 	).
 open_transaction(_,_).
 
@@ -575,30 +640,29 @@ next_transaction_id(DB, Id) :-
 	assert(current_transaction_id(DB, Id)).
 next_transaction_id(DB, Id) :-
 	db_files(DB, _, Journal),
-	catch(open_db(Journal, read, In, []),
-	      error(existence_error(_,_),_),
-	      fail), !,
-	call_cleanup(iterative_expand(In, Last), close(In)),
+	exists_file(Journal), !,
+	size_file(Journal, Size),
+	open_db(Journal, read, In, []),
+	call_cleanup(iterative_expand(In, Size, Last), close(In)),
 	Id is Last + 1,
 	assert(current_transaction_id(DB, Id)).
 next_transaction_id(DB, 1) :-
 	assert(current_transaction_id(DB, 1)).
 
-iterative_expand(In, Last) :-
-	between(10, 62, Step),		% 62 is big enough and safer than infinite
-	    Offset is -(1<<Step),
-	    catch(seek(In, Offset, eof, _), E, true),
-	    (	var(E)
-	    ->	skip(In, 10)		% records are line-based
-	    ;	seek(In, 0, bof, _), !,
-		WholeFile = true
-	    ),
-	    read(In, T0),
-	    last_transaction_id(T0, In, 0, Last),
-	    (	WholeFile == true	% we scanned the whole file
-	    ;   Last > 0		% we found a transaction
-	    ), !.
-iterative_expand(_, 0).
+iterative_expand(_, 0, 0) :- !.
+iterative_expand(In, Size, Last) :-	% Scan growing sections from the end
+	Max is floor(log(Size)/log(2)),
+	between(10, Max, Step),
+	Offset is -(1<<Step),
+	seek(In, Offset, eof, _),
+	skip(In, 10),			% records are line-based
+	read(In, T0),
+	last_transaction_id(T0, In, 0, Last),
+	Last > 0, !.
+iterative_expand(In, _, Last) :-	% Scan the whole file
+	seek(In, 0, bof, _),
+	read(In, T0),
+	last_transaction_id(T0, In, 0, Last).
 
 last_transaction_id(end_of_file, _, Last, Last) :- !.
 last_transaction_id(end(Id, _, _), In, _, Last) :-
@@ -614,7 +678,7 @@ last_transaction_id(_, In, Id, Last) :-
 %	End a transaction that affected the  given list of databases. We
 %	write the list of other affected databases as an argument to the
 %	end-term to facilitate fast finding of the related transactions.
-%	
+%
 %	In each database, the transaction is   ended with a term end(Id,
 %	Nesting, Others), where  Id  and   Nesting  are  the transaction
 %	identifier and nesting (see open_transaction/2)  and Others is a
@@ -672,7 +736,7 @@ sync_state([DB-MD5|TA], Pre) :-
 		 *******************************/
 
 %%	journal_fd(+DB, -Stream) is det.
-%	
+%
 %	Get an open stream to a journal. If the journal is not open, old
 %	journals are closed to satisfy   the =max_open_journals= option.
 %	Then the journal is opened in   =append= mode. Journal files are
@@ -721,10 +785,10 @@ close_oldest_journal :-
 	debug(rdf_persistency, 'Closing old journal for ~q', [DB]),
 	close_journal(DB).
 close_oldest_journal.
-		       
+
 
 %%	sync_journal(+DB, +Fd)
-%	
+%
 %	Sync journal represented by database and   stream.  If the DB is
 %	involved in a transaction there is   no point flushing until the
 %	end of the transaction.
@@ -735,7 +799,7 @@ sync_journal(_, Fd) :-
 	flush_output(Fd).
 
 %%	close_journal(+DB) is det.
-%	
+%
 %	Close the journal associated with DB if it is open.
 
 close_journal(DB) :-
@@ -751,7 +815,7 @@ close_journal_(DB) :-
 	).
 
 %	close_journals
-%	
+%
 %	Close all open journals.
 
 close_journals :-
@@ -760,7 +824,7 @@ close_journals :-
 		     print_message(error, E))).
 
 %%	create_db(+DB)
-%	
+%
 %	Create a saved version of DB in corresponding file, close and
 %	delete journals.
 
@@ -780,10 +844,15 @@ create_db(DB) :-
 
 
 %%	delete_db(+DB)
-%	
+%
 %	Remove snapshot and journal file for DB.
 
 delete_db(DB) :-
+	with_mutex(rdf_journal_file,
+		   delete_db_(DB)).
+
+delete_db_(DB) :-
+	close_journal_(DB),
 	db_abs_files(DB, Snapshot, Journal),
 	(   exists_file(Journal)
 	->  delete_file(Journal)
@@ -800,53 +869,61 @@ delete_db(DB) :-
 		 *******************************/
 
 %%	lock_db(+Dir)
-%	
-%	Lock the database  directory.  This  isn't   safe  as  the  file
-%	operations are not  atomic.  Needs   re-thinking,  but  with the
-%	normal server setting it should be ok.
+%
+%	Lock the database directory Dir.
 
 lock_db(Dir) :-
 	lockfile(Dir, File),
-	exists_file(File), !,
+	catch(open(File, update, Out, [lock(write), wait(false)]),
+	      error(permission_error(lock, _, _), _),
+	      locked_error(Dir)),
+	(   current_prolog_flag(pid, PID)
+	->  true
+	;   PID = 0			% TBD: Fix in Prolog
+	),
+	time_stamp(Now),
+	gethostname(Host),
+	format(Out, '/* RDF Database is in use */~n~n', []),
+	format(Out, '~q.~n', [ locked([ time(Now),
+					pid(PID),
+					host(Host)
+				      ])
+			     ]),
+	flush_output(Out),
+	set_end_of_stream(Out),
+	assert(rdf_lock(Dir, lock(Out, File))),
+	at_halt(unlock_db(Out, File)).
+
+locked_error(Dir) :-
+	lockfile(Dir, File),
 	(   catch(read_file_to_terms(File, Terms, []), _, fail),
 	    Terms = [locked(Args)]
 	->  Context = rdf_locked(Args)
 	;   Context = context(_, 'Database is in use')
 	),
 	throw(error(permission_error(lock, rdf_db, Dir), Context)).
-lock_db(Dir) :-
-	lockfile(Dir, File),
-	open(File, write, Out),
-	(   current_prolog_flag(pid, PID)
-	->  true
-	;   PID = 0			% TBD: Fix in Prolog
-	),
-	time_stamp(Now),
-	format(Out, '/* RDF Database is in use */~n~n', []),
-	format(Out, '~q.~n', [ locked([ time(Now),
-					pid(PID)
-				      ])
-			     ]),
-	close(Out),
-	at_halt(unlock_db(Dir)).
+
+%%	unlock_db(+Dir) is det.
+%%	unlock_db(+Stream, +File) is det.
 
 unlock_db(Dir) :-
-	lockfile(Dir, File),
-	(   exists_file(File)
-	->  delete_file(File)
-	;   true
-	).
+	retract(rdf_lock(Dir, lock(Out, File))), !,
+	unlock_db(Out, File).
+
+unlock_db(Out, File) :-
+	close(Out),
+	delete_file(File).
 
 		 /*******************************
 		 *	     FILENAMES		*
 		 *******************************/
 
 lockfile(Dir, LockFile) :-
-	concat_atom([Dir, /, lock], LockFile).
+	atomic_list_concat([Dir, /, lock], LockFile).
 
 db_file(Base, File) :-
 	rdf_directory(Dir),
-	concat_atom([Dir, /, Base], File).
+	atomic_list_concat([Dir, /, Base], File).
 
 open_db(Base, Mode, Stream, Options) :-
 	db_file(Base, File),
@@ -888,10 +965,10 @@ db_abs_files(DB, Snapshot, Journal) :-
 
 %%	rdf_journal_file(+DB, -File) is semidet.
 %%	rdf_journal_file(-DB, -File) is nondet.
-%	
+%
 %	True if File is the absolute  file   name  of  an existing named
-%	graph DB. 
-%	
+%	graph DB.
+%
 %	@tbd	Avoid using private rdf_db:rdf_graphs_/1.
 
 rdf_journal_file(DB, Journal) :-
@@ -906,22 +983,31 @@ rdf_journal_file(DB, Journal) :-
 
 %%	rdf_db_to_file(+DB, -File) is det.
 %%	rdf_db_to_file(-DB, +File) is det.
-%	
+%
 %	Translate between database encoding (often an   file or URL) and
 %	the name we store in the  directory.   We  keep  a cache for two
 %	reasons. Speed, but much more important   is that the mapping of
 %	raw --> encoded provided by  www_form_encode/2 is not guaranteed
 %	to be unique by the W3C standards.
+%
+%	@tbd	We keep two predicates for exploiting Prolog indexing.
+%		Once multi-argument indexed is hashed we should clean
+%		this up.
 
 rdf_db_to_file(DB, File) :-
+	nonvar(File),
+	file_base_db(File, DB), !.
+rdf_db_to_file(DB, File) :-
+	nonvar(DB),
 	db_file_base(DB, File), !.
 rdf_db_to_file(DB, File) :-
 	url_to_filename(DB, File),
-	assert(db_file_base(DB, File)).
+	assert(db_file_base(DB, File)),
+	assert(file_base_db(File, DB)).
 
 %%	url_to_filename(+URL, -FileName) is det.
 %%	url_to_filename(-URL, +FileName) is det.
-%	
+%
 %	Turn  a  valid  URL  into  a  filename.  Earlier  versions  used
 %	www_form_encode/2, but this can produce  characters that are not
 %	valid  in  filenames.  We  will  use    the   same  encoding  as
@@ -955,7 +1041,7 @@ url_encode(Enc) -->
 	},
 	url_encode(T).
 url_encode([]) -->
-	[], !.
+	eos, !.
 url_encode([0'%,D1,D2|T]) -->
 	[C],
 	{ Dv1 is (C>>4 /\ 0xf),
@@ -965,13 +1051,63 @@ url_encode([0'%,D1,D2|T]) -->
 	},
 	url_encode(T).
 
+eos([], []).
+
 alphanum(C) -->
 	[C],
 	{ C < 128,			% US-ASCII
 	  code_type(C, alnum)
 	}.
 
-no_enc_extra(0'_) --> "_".		%'
+no_enc_extra(0'_) --> "_".
+
+
+		 /*******************************
+		 *	      PREFIXES		*
+		 *******************************/
+
+save_prefixes(Dir) :-
+	atomic_list_concat([Dir, /, 'prefixes.db'], PrefixFile),
+	setup_call_cleanup(open(PrefixFile, write, Out, [encoding(utf8)]),
+			   write_prefixes(Out),
+			   close(Out)).
+
+write_prefixes(Out) :-
+	format(Out, '% Snapshot of defined RDF prefixes~n~n', []),
+	forall(rdf_current_ns(Alias, URI),
+	       format(Out, 'prefix(~q, ~q).~n', [Alias, URI])).
+
+%%	load_prefixes(+RDFDBDir) is det.
+%
+%	If the file RDFDBDir/prefixes.db exists,  load the prefixes. The
+%	prefixes are registered using rdf_register_ns/3. Possible errors
+%	because the prefix  definitions  have   changed  are  printed as
+%	warnings, retaining the  old  definition.   Note  that  changing
+%	prefixes generally requires reloading all RDF from the source.
+
+load_prefixes(Dir) :-
+	atomic_list_concat([Dir, /, 'prefixes.db'], PrefixFile),
+	(   exists_file(PrefixFile)
+	->  setup_call_cleanup(open(PrefixFile, read, In, [encoding(utf8)]),
+			       read_prefixes(In),
+			       close(In))
+	;   true
+	).
+
+read_prefixes(Stream) :-
+	read_term(Stream, T0, []),
+	read_prefixes(T0, Stream).
+
+read_prefixes(end_of_file, _) :- !.
+read_prefixes(prefix(Alias, URI), Stream) :- !,
+	must_be(atom, Alias),
+	must_be(atom, URI),
+	catch(rdf_register_ns(Alias, URI, []), E,
+	      print_message(warning, E)),
+	read_term(Stream, T, []),
+	read_prefixes(T, Stream).
+read_prefixes(Term, _) :-
+	domain_error(prefix_term, Term).
 
 
 		 /*******************************
@@ -979,7 +1115,7 @@ no_enc_extra(0'_) --> "_".		%'
 		 *******************************/
 
 %%	mkdir(+Directory)
-%	
+%
 %	Create a directory if it does not already exist.
 
 mkdir(Directory) :-
@@ -988,7 +1124,7 @@ mkdir(Directory) :-
 	make_directory(Directory).
 
 %%	time_stamp(-Integer)
-%	
+%
 %	Return time-stamp rounded to integer.
 
 time_stamp(Int) :-
@@ -1004,23 +1140,46 @@ time_stamp(Int) :-
 	prolog:message/3,
 	prolog:message_context/3.
 
-prolog:message(rdf(restore(source(DB)))) -->
+prolog:message(rdf(Term)) -->
+	message(Term).
+
+message(restore(attached(Graphs, Time/Wall))) -->
+	{ catch(Percent is round(100*Time/Wall), _, Percent = 0) },
+	[ 'Attached ~D graphs in ~2f seconds (~d% CPU = ~2f sec.)'-
+	  [Graphs, Wall, Percent, Time] ].
+message(restore(true, Action)) --> !,
+	silent_message(Action).
+message(restore(brief, Action)) --> !,
+	brief_message(Action).
+message(restore(_, source(DB, Nth, Total))) -->
 	{ file_base_name(DB, Base) },
-	[ 'Restoring ~w ... '-[Base], flush ].
-prolog:message(rdf(restore(snapshot(_)))) -->
+	[ 'Restoring ~w ... (~D of ~D graphs) '-[Base, Nth, Total], flush ].
+message(restore(_, snapshot(_))) -->
 	[ at_same_line, '(snapshot) '-[], flush ].
-prolog:message(rdf(restore(journal(_)))) -->
+message(restore(_, journal(_))) -->
 	[ at_same_line, '(journal) '-[], flush ].
-prolog:message(rdf(restore(done(_, Time, Count)))) -->
+message(restore(_, done(_, Time, Count, _Nth, _Total))) -->
 	[ at_same_line, '~D triples in ~2f sec.'-[Count, Time] ].
-prolog:message(rdf(update_failed(S,P,O,Action))) -->
+message(update_failed(S,P,O,Action)) -->
 	[ 'Failed to update <~p ~p ~p> with ~p'-[S,P,O,Action] ].
 
+silent_message(_Action) --> [].
+
+brief_message(source(DB, Nth, Total)) -->
+	{ file_base_name(DB, Base) },
+	[ at_same_line,
+	  '\r~w~`.t ~D of ~D graphs~72|'-[Base, Nth, Total],
+	  flush
+	].
+brief_message(snapshot(_File)) --> [].
+brief_message(journal(_File))  --> [].
+brief_message(done(_DB, _Time, _Count, _Nth, _Total)) --> [].
+
 prolog:message_context(rdf_locked(Args)) -->
 	{ memberchk(time(Time), Args),
 	  memberchk(pid(Pid), Args),
 	  format_time(string(S), '%+', Time)
 	},
 	[ nl,
-	  'locked at ~s by process id ~w'-[S,Pid] 
+	  'locked at ~s by process id ~w'-[S,Pid]
 	].
