@@ -3,9 +3,10 @@
     Part of SWI-Prolog
 
     Author:        Jan Wielemaker
-    E-mail:        wielemak@science.uva.nl
+    E-mail:        J.Wielemaker@cs.vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (C): 1985-2005, University of Amsterdam
+    Copyright (C): 1985-2011, University of Amsterdam
+			      VU University Amsterdam
 
     This program is free software; you can redistribute it and/or
     modify it under the terms of the GNU General Public License
@@ -19,7 +20,7 @@
 
     You should have received a copy of the GNU Lesser General Public
     License along with this library; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 
     As a special exception, if you link this library with other files,
     compiled with a Free Software compiler, to produce an executable, this
@@ -58,11 +59,67 @@
 	    xml_quote_attribute/2,	% +In, -Quoted
 	    xml_quote_cdata/2,		% +In, -Quoted
 	    xml_name/1,			% +In
+	    xml_name/2,			% +In, +Encoding
+	    iri_xml_namespace/2,	% +IRI, -Namespace
+	    iri_xml_namespace/3,	% +IRI, -Namespace, -LocalName
 	    xml_is_dom/1		% +Term
 	  ]).
 :- use_module(library(lists)).
 :- use_module(library(option)).
 
+:- meta_predicate
+	load_structure(+, -, :).
+
+:- predicate_options(load_structure/3, 3,
+		     [ charpos(integer),
+		       defaults(boolean),
+		       dialect(oneof([sgml,xml,xmlns])),
+		       doctype(atom),
+		       dtd(any),
+		       encoding(encoding),
+		       entity(atom,atom),
+		       file(atom),
+		       line(integer),
+		       number(oneof([token,integer])),
+		       qualify_attributes(boolean),
+		       shorttag(boolean),
+		       space(oneof([sgml,preserve,defailt,remove])),
+		       xmlns(atom),
+		       xmlns(atom,atom),
+		       pass_to(sgml_parse/2, 2)
+		     ]).
+:- predicate_options(load_dtd/3, 3,
+		     [ dialect(oneof([sgml,xml,xmlns])),
+		       pass_to(open/4, 4)
+		     ]).
+:- predicate_options(sgml_parse/2, 2,
+		     [ call(oneof([begin,end,cdata,pi,decl,error,xmlns,urlns]),
+			    callable),
+		       content_length(integer),
+		       document(-any),
+		       max_errors(integer),
+		       parse(oneof([file,element,content,declaration,input])),
+		       source(any),
+		       syntax_errors(oneof([quiet,print,style])),
+		       xml_no_ns(oneof([error,quiet]))
+		     ]).
+:- predicate_options(new_sgml_parser/2, 2,
+		     [ dtd(any)
+		     ]).
+
+
+/** <module> SGML, XML and HTML parser
+
+This library allows you to parse SGML, XML   and HTML data into a Prolog
+data  structure.  The  high-level  interface  defined  in  library(sgml)
+provides  access  at  the  file-level,  while  the  low-level  interface
+(load_structure/3) defined in  the  foreign   module  works  with Prolog
+streams. Please examine the source as a  starting point for dealing with
+data from other sources  than  files,   such  as  SWI-Prolog  resources,
+network-sockets, character strings, etc. The   first example below loads
+an HTML file.
+*/
+
 :- multifile user:file_search_path/2.
 :- dynamic   user:file_search_path/2.
 
@@ -73,10 +130,7 @@ sgml_register_catalog_file(File, Location) :-
 	prolog_to_os_filename(File, OsFile),
 	'_sgml_register_catalog_file'(OsFile, Location).
 
-load_foreign :-
-	current_predicate(_, _:sgml_parse(_,_)), !.
-load_foreign :-
-	load_foreign_library(foreign(sgml2pl)).
+:- use_foreign_library(foreign(sgml2pl)).
 
 register_catalog(Base) :-
 	absolute_file_name(dtd(Base),
@@ -87,12 +141,8 @@ register_catalog(Base) :-
 			       SocFile),
 	sgml_register_catalog_file(SocFile, end).
 
-init :-
-	load_foreign,
-	ignore(register_catalog('HTML4')).
-
 :- initialization
-	init.
+	ignore(register_catalog('HTML4')).
 
 
 		 /*******************************
@@ -120,6 +170,16 @@ diagnosed to mess with the entity resolution by Fabien Todescato.
 
 dtd_alias(html, 'HTML4').
 
+%%	dtd(+Type, -DTD) is det.
+%
+%	DTD is a DTD object created from  the file dtd(Type). Loaded DTD
+%	objects are cached. Note that  DTD   objects  may  not be shared
+%	between threads. Therefore, dtd/2  maintains   the  pool  of DTD
+%	objects  using  a  thread_local  predicate.    DTD  objects  are
+%	destroyed if a thread terminates.
+%
+%	@error existence_error(source_sink, dtd(Type))
+
 dtd(Type, DTD) :-
 	current_dtd(Type, DTD), !.
 dtd(Type, DTD) :-
@@ -138,23 +198,28 @@ dtd(Type, DTD) :-
 
 %%	load_dtd(+DTD, +DtdFile, +Options)
 %
-%	Load file into a DTD.  Defined options are:
+%	Load DtdFile into a DTD.  Defined options are:
 %
 %		* dialect(+Dialect)
 %		Dialect to use (xml, xmlns, sgml)
 %
 %		* encoding(+Encoding)
 %		Encoding of DTD file
+%
+%	@param	DTD is a fresh DTD object, normally created using
+%		new_dtd/1.
 
 load_dtd(DTD, DtdFile) :-
 	load_dtd(DTD, DtdFile, []).
 load_dtd(DTD, DtdFile, Options) :-
 	split_dtd_options(Options, DTDOptions, FileOptions),
-	open_dtd(DTD, DTDOptions, DtdOut),
-	open(DtdFile, read, DtdIn, FileOptions),
-	copy_stream_data(DtdIn, DtdOut),
-	close(DtdIn),
-	close(DtdOut).
+	setup_call_cleanup(
+	    open_dtd(DTD, DTDOptions, DtdOut),
+	    setup_call_cleanup(
+		open(DtdFile, read, DtdIn, FileOptions),
+		copy_stream_data(DtdIn, DtdOut),
+		close(DtdIn)),
+	    close(DtdOut)).
 
 split_dtd_options([], [], []).
 split_dtd_options([H|T], [H|TD], S) :-
@@ -261,7 +326,19 @@ set_parser_options(Parser, Options, RestOptions) :-
 set_parser_options(_, Options, Options).
 
 
-load_structure(stream(In), Term, Options) :- !,
+%%	load_structure(+Source, -ListOfContent, :Options) is det.
+%
+%	Parse   Source   and   return   the   resulting   structure   in
+%	ListOfContent.  Source  is  either   a    term   of  the  format
+%	stream(StreamHandle) or a  file-name.  Options   is  a  list  of
+%	options controlling the conversion process.
+%
+%	A proper XML document contains only   a  single toplevel element
+%	whose name matches the document type.   Nevertheless,  a list is
+%	returned for consistency with  the   representation  of  element
+%	content.
+
+load_structure(stream(In), Term, M:Options) :- !,
 	(   select_option(offset(Offset), Options, Options1)
 	->  seek(In, Offset, bof, _)
 	;   Options1 = Options
@@ -274,8 +351,8 @@ load_structure(stream(In), Term, Options) :- !,
 	new_sgml_parser(Parser,
 			[ dtd(DTD)
 			]),
-	def_entities(Options2, DTD, Options3),
-	call_cleanup(parse(Parser, Options3, TermRead, In),
+	def_entities(Options2, Parser, Options3),
+	call_cleanup(parse(Parser, M:Options3, TermRead, In),
 		     free_sgml_parser(Parser)),
 	(   ExplicitDTD == true
 	->  (   DTD = dtd(_, DocType),
@@ -289,31 +366,46 @@ load_structure(stream(In), Term, Options) :- !,
 load_structure(Stream, Term, Options) :-
 	is_stream(Stream), !,
 	load_structure(stream(Stream), Term, Options).
-load_structure(File, Term, Options) :-
-	open(File, read, In, [type(binary)]),
-	load_structure(stream(In), Term, [file(File)|Options]),
-	close(In).
+load_structure(File, Term, M:Options) :-
+	setup_call_cleanup(
+	    open(File, read, In, [type(binary)]),
+	    load_structure(stream(In), Term, M:[file(File)|Options]),
+	    close(In)).
 
-parse(Parser, Options, Document, In) :-
+parse(Parser, M:Options, Document, In) :-
 	set_parser_options(Parser, Options, Options1),
+	parser_meta_options(Options1, M, Options2),
 	sgml_parse(Parser,
 		   [ document(Document),
 		     source(In)
-		   | Options1
+		   | Options2
 		   ]).
 
-def_entities([], _, []).
-def_entities([entity(Name, Value)|T], DTD, Opts) :- !,
-	def_entity(DTD, Name, Value),
-	def_entities(T, DTD, Opts).
-def_entities([H|T0], DTD, [H|T]) :-
-	def_entities(T0, DTD, T).
+parser_meta_options([], _, []).
+parser_meta_options([call(When, Closure)|T0], M, [call(When, M:Closure)|T]) :- !,
+	parser_meta_options(T0, M, T).
+parser_meta_options([H|T0], M, [H|T]) :-
+	parser_meta_options(T0, M, T).
 
-def_entity(DTD, Name, Value) :-
-	open_dtd(DTD, [], Stream),
+
+def_entities([], _, []).
+def_entities([H|T], Parser, Opts) :-
+	def_entity(H, Parser), !,
+	def_entities(T, Parser, Opts).
+def_entities([H|T0], Parser, [H|T]) :-
+	def_entities(T0, Parser, T).
+
+def_entity(entity(Name, Value), Parser) :-
+	get_sgml_parser(Parser, dtd(DTD)),
 	xml_quote_attribute(Value, QValue),
-	format(Stream, '<!ENTITY ~w "~w">~n', [Name, QValue]),
-	close(Stream).
+	setup_call_cleanup(open_dtd(DTD, [], Stream),
+			   format(Stream, '<!ENTITY ~w "~w">~n',
+				  [Name, QValue]),
+			   close(Stream)).
+def_entity(xmlns(URI), Parser) :-
+	set_sgml_parser(Parser, xmlns(URI)).
+def_entity(xmlns(NS, URI), Parser) :-
+	set_sgml_parser(Parser, xmlns(NS, URI)).
 
 
 		 /*******************************
@@ -339,8 +431,8 @@ load_html_file(File, Term) :-
 		 *	      ENCODING		*
 		 *******************************/
 
-%	xml_quote_attribute(+In, -Quoted)
-%	xml_quote_cdata(+In, -Quoted)
+%%	xml_quote_attribute(+In, -Quoted) is det.
+%%	xml_quote_cdata(+In, -Quoted) is det.
 %
 %	Backward  compatibility  for  versions  that  allow  to  specify
 %	encoding. All characters that cannot fit the encoding are mapped
@@ -353,6 +445,10 @@ xml_quote_attribute(In, Quoted) :-
 xml_quote_cdata(In, Quoted) :-
 	xml_quote_cdata(In, Quoted, ascii).
 
+%%	xml_name(+Atom) is semidet.
+%
+%	True if Atom is a valid XML name.
+
 xml_name(In) :-
 	xml_name(In, ascii).
 
@@ -361,7 +457,7 @@ xml_name(In) :-
 		 *	   TYPE CHECKING	*
 		 *******************************/
 
-%	xml_is_dome(@Term)
+%%	xml_is_dom(@Term) is semidet.
 %
 %	True  if  term  statisfies   the    structure   as  returned  by
 %	load_structure/3 and friends.
@@ -419,9 +515,26 @@ prolog:message(sgml(Parser, File, Line, Message)) -->
 	prolog:called_by/2.
 
 prolog:called_by(sgml_parse(_, Options), Called) :-
-	is_list(Options),
-	findall(G+3,
-		(   member(call(_, G), Options),
-		    callable(G)
-		),
-		Called).
+	findall(Meta, meta_call_term(_, Meta, Options), Called).
+
+meta_call_term(T, G+N, Options) :-
+	T = call(Event, G),
+	pmember(T, Options),
+	call_params(Event, Term),
+	functor(Term, _, N).
+
+pmember(X, List) :-			% member for partial lists
+	nonvar(List),
+	List = [H|T],
+	(   X = H
+	;   pmember(X, T)
+	).
+
+call_params(begin, begin(tag,attributes,parser)).
+call_params(end,   end(tag,parser)).
+call_params(cdata, cdata(cdata,parser)).
+call_params(pi,	   pi(cdata,parser)).
+call_params(decl,  decl(cdata,parser)).
+call_params(error, error(severity,message,parser)).
+call_params(xmlns, xmlns(namespace,url,parser)).
+call_params(urlns, urlns(url,url,parser)).
