@@ -19,7 +19,7 @@
 
     You should have received a copy of the GNU Lesser General Public
     License along with this library; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */
 
 
@@ -62,6 +62,7 @@ static functor_t FUNCTOR_public_key5;
 static functor_t FUNCTOR_private_key8;
 static functor_t FUNCTOR_key1;
 static functor_t FUNCTOR_hash1;
+static functor_t FUNCTOR_next_update1;
 static functor_t FUNCTOR_signature1;
 static functor_t FUNCTOR_equals2;
 static functor_t FUNCTOR_crl1;
@@ -454,10 +455,10 @@ unify_asn1_time(term_t term, ASN1_TIME *time)
   time_tm.tm_yday = 0;
   time_tm.tm_isdst = 0;  /* No DST adjustment requested, though mktime might do it anyway */
   result = mktime(&time_tm);
+  /* mktime assumes that the time_tm contains information for localtime. Convert back to UTC: */
   if ((time_t)-1 != result)
-  { if (time_tm.tm_isdst != 0)
-        result -= 3600; /* if mktime has adjusted the time for DST, adjust it back 1 hour*/
-     result += lSecondsFromUTC; /* Add in the UTC offset, or should we return UTC time? */
+  { result += lSecondsFromUTC; /* Add in the UTC offset of the original value */
+    result -= timezone; /* Adjust for localtime */
   } else
   { ssl_deb(2, "mktime() failed");
     return FALSE;
@@ -543,6 +544,8 @@ unify_crl(term_t term, X509_CRL* crl)
   term_t issuer = PL_new_term_ref();
   term_t revocations = PL_new_term_ref();
   term_t list = PL_copy_term_ref(revocations);
+  term_t next_update = PL_new_term_ref();
+
   int result = 1;
   long n;
   unsigned char* p;
@@ -556,23 +559,26 @@ unify_crl(term_t term, X509_CRL* crl)
   i2a_ASN1_INTEGER(mem, crl->signature);
   if (!(unify_name(issuer, X509_CRL_get_issuer(crl)) &&
         unify_hash(hash, crl->sig_alg->algorithm, i2d_X509_CRL_INFO_wrapper, crl->crl) &&
+        unify_asn1_time(next_update, X509_CRL_get_nextUpdate(crl)) &&
         PL_unify_term(term,
-                      PL_LIST, 4,
+                      PL_LIST, 5,
                       PL_FUNCTOR, FUNCTOR_issuername1,
                       PL_TERM, issuer,
                       PL_FUNCTOR, FUNCTOR_signature1,
                       PL_NCHARS, (size_t)crl->signature->length, crl->signature->data,
                       PL_FUNCTOR, FUNCTOR_hash1,
                       PL_TERM, hash,
+                      PL_FUNCTOR, FUNCTOR_next_update1,
+                      PL_TERM, next_update,
                       PL_FUNCTOR, FUNCTOR_revocations1,
                       PL_TERM, revocations)))
   {
      return FALSE;
   }
-
   for (i = 0; i < sk_X509_REVOKED_num(info->revoked); i++)
   {
      X509_REVOKED* revoked = sk_X509_REVOKED_value(info->revoked, i);
+     (void)BIO_reset(mem);
      i2a_ASN1_INTEGER(mem, revoked->serialNumber);
      result &= (((n = BIO_get_mem_data(mem, &p)) > 0) &&
                 PL_unify_list(list, item, list) &&
@@ -1130,20 +1136,40 @@ pl_ssl_close(PL_SSL_INSTANCE *instance)
 static int
 pl_ssl_control(PL_SSL_INSTANCE *instance, int action, void *data)
 { switch(action)
-  { case SIO_GETFILENO:
-    { if (instance->sread != NULL)
-    {  SOCKET fd = Sfileno(instance->sread);
-       SOCKET *fdp = data;
-       *fdp = fd;
-       return 0;
-    } else if (instance->swrite != NULL)
-    {  SOCKET fd = Sfileno(instance->swrite);
-       SOCKET *fdp = data;
-       *fdp = fd;
-       return 0;
-    }
-    }
-    return -1;
+  {
+#ifdef __WINDOWS__
+    case SIO_GETFILENO:
+      return -1;
+    case SIO_GETWINSOCK:
+      { if (instance->sread != NULL)
+        { (*instance->sread->functions->control)(instance->sread->handle,
+                                                 SIO_GETWINSOCK,
+                                                 data);
+          return 0;
+        } else if (instance->swrite != NULL)
+        { (*instance->swrite->functions->control)(instance->swrite->handle,
+                                                  SIO_GETWINSOCK,
+                                                  data);
+          return 0;
+        }
+      }
+      return -1;
+#else
+    case SIO_GETFILENO:
+      { if (instance->sread != NULL)
+        {  SOCKET fd = Sfileno(instance->sread);
+           SOCKET *fdp = data;
+           *fdp = fd;
+           return 0;
+        } else if (instance->swrite != NULL)
+        {  SOCKET fd = Sfileno(instance->swrite);
+           SOCKET *fdp = data;
+           *fdp = fd;
+           return 0;
+        }
+      }
+      return -1;
+#endif
     case SIO_SETENCODING:
     case SIO_FLUSHOUTPUT:
       return 0;
@@ -1200,19 +1226,19 @@ pl_ssl_negotiate(term_t config, term_t org_in, term_t org_out, term_t in, term_t
   if ( !get_conf(config, &conf) )
     return FALSE;
   if ( !PL_get_stream_handle(org_in, &sorg_in) )
-     return FALSE;
+    return FALSE;
   if ( !PL_get_stream_handle(org_out, &sorg_out) )
-     return FALSE;
+    return FALSE;
   if ( !(instance = ssl_ssl_bio(conf, sorg_in, sorg_out)) )
-  {  PL_release_stream(sorg_in);
-     PL_release_stream(sorg_out);
-     return FALSE;			/* TBD: error */
+  { PL_release_stream(sorg_in);
+    PL_release_stream(sorg_out);
+    return ssl_error("ssl_verify");
   }
 
   if ( !(i=Snew(instance, SIO_INPUT|SIO_RECORDPOS|SIO_FBUF, &ssl_funcs)) )
-  {  PL_release_stream(sorg_in);
-     PL_release_stream(sorg_out);
-    return FALSE;
+  { PL_release_stream(sorg_in);
+    PL_release_stream(sorg_out);
+    return PL_resource_error("memory");
   }
   instance->close_needed++;
   if ( !PL_unify_stream(in, i) )
@@ -1224,8 +1250,8 @@ pl_ssl_negotiate(term_t config, term_t org_in, term_t org_out, term_t in, term_t
   Sset_filter(sorg_in, i);
   PL_release_stream(sorg_in);
   if ( !(o=Snew(instance, SIO_OUTPUT|SIO_RECORDPOS|SIO_FBUF, &ssl_funcs)) )
-  {  PL_release_stream(sorg_out);
-    return FALSE;
+  { PL_release_stream(sorg_out);
+    return PL_resource_error("memory");
   }
   instance->close_needed++;
   if ( !PL_unify_stream(out, o) )
@@ -1433,6 +1459,7 @@ install_ssl4pl()
   FUNCTOR_public_key5     = PL_new_functor(PL_new_atom("public_key"), 5);
   FUNCTOR_private_key8    = PL_new_functor(PL_new_atom("private_key"), 8);
   FUNCTOR_hash1	          = PL_new_functor(PL_new_atom("hash"), 1);
+  FUNCTOR_next_update1    = PL_new_functor(PL_new_atom("next_update"), 1);
   FUNCTOR_signature1      = PL_new_functor(PL_new_atom("signature"), 1);
   FUNCTOR_equals2         = PL_new_functor(PL_new_atom("="), 2);
   FUNCTOR_crl1            = PL_new_functor(PL_new_atom("crl"), 1);
