@@ -1,4 +1,4 @@
-/*  $Id: ssl_test.pl,v 1.4 2006/06/19 17:28:33 jan Exp $
+/*  $Id$
 
     Part of SWI-Prolog
 
@@ -19,7 +19,7 @@
 
     You should have received a copy of the GNU Lesser General Public
     License along with this library; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 
     As a special exception, if you link this library with other files,
     compiled with a Free Software compiler, to produce an executable, this
@@ -29,14 +29,18 @@
     the GNU General Public License.
 */
 
+:- asserta(user:file_search_path(library, '.')).
+:- asserta(user:file_search_path(library, '../clib')).
 :- asserta(user:file_search_path(foreign, '.')).
+:- asserta(user:file_search_path(foreign, '../clib')).
 
-:- use_module(ssl).
+:- use_module(library(ssl)).
 :- use_module(library(debug)).
 :- use_module(library(readutil)).
 
 %:- debug(connection).
-:- debug(_).
+%:- debug(data).
+%:- debug(_).
 
 :- dynamic
 	option/1,			% Options to test
@@ -78,7 +82,7 @@ make_server(SSL) :-
 		   cacert_file('etc/demoCA/cacert.pem'),
 		   certificate_file('etc/server/server-cert.pem'),
 		   key_file('etc/server/server-key.pem'),
-               	   cert_verify_hook(get_cert_verify),
+		   cert_verify_hook(get_cert_verify),
 %		   password('apenoot1'),
 		   pem_password_hook(get_server_pwd)
 		 ]).
@@ -104,7 +108,9 @@ copy_client(In, Out) :-
 	read_line_to_codes(In, Line),
 	(   Line == end_of_file
 	->  true
-	;   debug(data, 'Got ~s~n', [Line]),
+	;   debug(data, 'SERVER: Got ~s~n', [Line]),
+	    sleep(1.5),
+	    debug(data, 'SERVER: writing ~s~n', [Line]),
 	    format(Out, '~s~n', [Line]),
 	    flush_output(Out),
 	    (	Line = "bye"
@@ -143,32 +149,41 @@ client :-
 
 client_loop(SSL) :-
 	ssl_open(SSL, In, Out),
+        set_stream(In, timeout(1)),
 	Message = 'Hello world',
 	write_server(Message, In, Out),
-	write_server(Message, In, Out),
 	(   option(timeout(T))
 	->  Wait is T*2,
 	    sleep(Wait)
 	;   true
 	),
-	numlist(1, 10000, L),
-	term_to_atom(L, Long),
-	write_server(Long, In, Out),
 	write_server(bye, In, Out),
 	close(In),
 	close(Out).
 
 write_server(Message, In, Out) :-
+	debug(data, 'CLIENT: writing: ~q~n', [Message]),
 	write(Out, Message), nl(Out),
-        flush_output(Out),
+	flush_output(Out),
+	sleep(0.1),
+	catch(read_from_server(In, Message),
+	      E,
+	      debug(data, 'CLIENT: exception: ~q~n', [E])),
+	(   var(E)
+	->  true
+	;   read_from_server(In, Message)
+        ).
+
+read_from_server(In, Message) :-
+	debug(data, 'CLIENT: attempting to read reply from stream~n', []),
 	read_line_to_codes(In, Line),
 	(   Line == end_of_file
 	->  true
 	;   atom_codes(Reply, Line),
-	    debug(data, 'Got ~q~n', [Reply]),
+	    debug(data, 'CLIENT: Got ~q~n', [Reply]),
 	    (	Reply == Message
 	    ->	true
-	    ;	format(user_error, 'ERROR: Sent ~q, Got ~q~n',
+	    ;	format(user_error, 'CLIENT: ERROR: Sent ~q, Got ~q~n',
 		       [Message, Reply])
 	    )
 	).
