@@ -19,7 +19,7 @@
 
     You should have received a copy of the GNU Lesser General Public
     License along with this library; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */
 
 #include <string.h>
@@ -918,6 +918,12 @@ ssl_lib_init(void)
 /*
  * Read function.  To allow for setting a timeout on the SSL stream, we
  * use the timeout of this stream if we do not have a timeout ourselves.
+ *
+ * Note that if the underlying stream received a timeout, we lift this
+ * error to the ssl stream and clear the error on the underlying stream.
+ * This way, the normal timeout-reset in pl-stream.c correctly resets
+ * a possible timeout.  See also test_ssl.c.  Patch and analysis by
+ * Keri Harris.
  */
 
 int bio_read(BIO* bio, char* buf, int len)
@@ -938,10 +944,7 @@ int bio_read(BIO* bio, char* buf, int len)
 
    if ( ssl_stream && (stream->flags & SIO_TIMEOUT) )
    { ssl_stream->flags |= (SIO_FERR|SIO_TIMEOUT);
-#ifdef ETIMEDOUT
-     errno = ETIMEDOUT;
-#endif
-     rc = -1;
+     Sclearerr(stream);
    }
 
    return rc;
@@ -975,12 +978,28 @@ int bio_gets(BIO* bio, char* buf, int len)
 
 int bio_write(BIO* bio, const char* buf, int len)
 {
-   IOSTREAM* stream;
+   IOSTREAM* stream = BIO_get_ex_data(bio, 0);
+   IOSTREAM* ssl_stream = stream->upstream;
    int r;
-   stream  = BIO_get_ex_data(bio, 0);
-   r = (int)Sfwrite(buf, sizeof(char), len, stream);
-   /* OpenSSL expects there to be no buffering when it writes. Flush here */
-   Sflush(stream);
+
+   if ( ssl_stream &&
+	stream->timeout < 0 &&
+	ssl_stream->timeout > 0 )
+   { stream->timeout = ssl_stream->timeout;
+     r = (int)Sfwrite(buf, sizeof(char), len, stream);
+     /* OpenSSL expects there to be no buffering when it writes. Flush here */
+     Sflush(stream);
+     stream->timeout = -1;
+   } else
+   { r = (int)Sfwrite(buf, sizeof(char), len, stream);
+     Sflush(stream);
+   }
+
+   if ( ssl_stream && (stream->flags & SIO_TIMEOUT) )
+   { ssl_stream->flags |= (SIO_FERR|SIO_TIMEOUT);
+     Sclearerr(stream);
+   }
+
    return r;
 }
 
@@ -1144,7 +1163,7 @@ ssl_ssl_bio(PL_SSL *config, IOSTREAM* sread, IOSTREAM* swrite)
     return NULL;
 }
 
-int
+ssize_t
 ssl_read(PL_SSL_INSTANCE *instance, char *buf, int size)
 /*
  * Perform read on SSL session
@@ -1172,7 +1191,7 @@ ssl_read(PL_SSL_INSTANCE *instance, char *buf, int size)
     } while (1);
 }
 
-int
+ssize_t
 ssl_write(PL_SSL_INSTANCE *instance, const char *buf, int size)
 /*
  * Perform write on SSL session
