@@ -19,7 +19,7 @@
 
     You should have received a copy of the GNU Lesser General Public
     License along with this library; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */
 
 #ifdef HAVE_CONFIG_H
@@ -47,7 +47,7 @@ static char *
 winerror(int id)
 { char *msg;
   static WORD lang;
-  static lang_initialised = 0;
+  static int lang_initialised = 0;
 
   if ( !lang_initialised )
     lang = MAKELANGID(LANG_ENGLISH, SUBLANG_ENGLISH_UK);
