@@ -1,4 +1,4 @@
-/*  $Id: table.c,v 1.19 2007/01/23 19:22:16 jan Exp $
+/*  $Id$
 
     Part of SWI-Prolog
 
@@ -19,7 +19,7 @@
 
     You should have received a copy of the GNU Lesser General Public
     License along with this library; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */
 
 #ifdef HAVE_CONFIG_H
@@ -47,6 +47,8 @@
 #include <sys/stat.h>
 #include <fcntl.h>
 #include <unistd.h>
+#endif
+#ifdef HAVE_CTYPE_H
 #include <ctype.h>
 #endif
 #ifdef HAVE_ALLOCA_H
@@ -105,7 +107,7 @@ open_table(+File, +Fields, +Options, -Handle)
 				32 (space), implying all blank
 				space.
 
- 		sorted(Field)
+		sorted(Field)
 				Defines the file is sorted in the
 				specified field name.  In this
 				case lookup will exploit binary
@@ -178,7 +180,7 @@ init_constants()
   ATOM_exact			= PL_new_atom("exact");
   ATOM_field			= PL_new_atom("field");
   ATOM_field_count		= PL_new_atom("field_count");
-  ATOM_field_separator  	= PL_new_atom("field_separator");
+  ATOM_field_separator		= PL_new_atom("field_separator");
   ATOM_file			= PL_new_atom("file");
   ATOM_float			= PL_new_atom("float");
   ATOM_functor			= PL_new_atom("functor");
@@ -187,8 +189,8 @@ init_constants()
   ATOM_key_field		= PL_new_atom("key_field");
   ATOM_map_space_to_underscore	= PL_new_atom("map_space_to_underscore");
   ATOM_prefix			= PL_new_atom("prefix");
-  ATOM_record	        	= PL_new_atom("record");
-  ATOM_record_separator 	= PL_new_atom("record_separator");
+  ATOM_record			= PL_new_atom("record");
+  ATOM_record_separator		= PL_new_atom("record_separator");
   ATOM_size			= PL_new_atom("size");
   ATOM_skip			= PL_new_atom("skip");
   ATOM_sorted			= PL_new_atom("sorted");
@@ -209,43 +211,51 @@ init_constants()
 
 static int
 domain_error(term_t actual, const char *expected)
-{ term_t ex = PL_new_term_ref();
+{ term_t ex;
 
-  PL_unify_term(ex, PL_FUNCTOR_CHARS, "error", 2,
-		      PL_FUNCTOR_CHARS, "domain_error", 2,
-		        PL_CHARS, expected,
-		        PL_TERM, actual,
-		      PL_VARIABLE);
+  if ( (ex=PL_new_term_ref()) &&
+       PL_unify_term(ex,
+		     PL_FUNCTOR_CHARS, "error", 2,
+		       PL_FUNCTOR_CHARS, "domain_error", 2,
+		         PL_CHARS, expected,
+		         PL_TERM, actual,
+		       PL_VARIABLE) )
+    return PL_raise_exception(ex);
 
-  return PL_raise_exception(ex);
+  return FALSE;
 }
 
 
 static int
 type_error(term_t actual, const char *expected)
-{ term_t ex = PL_new_term_ref();
+{ term_t ex;
 
-  PL_unify_term(ex, PL_FUNCTOR_CHARS, "error", 2,
-		      PL_FUNCTOR_CHARS, "type_error", 2,
-		        PL_CHARS, expected,
-		        PL_TERM, actual,
-		      PL_VARIABLE);
+  if ( (ex=PL_new_term_ref()) &&
+       PL_unify_term(ex,
+		     PL_FUNCTOR_CHARS, "error", 2,
+		       PL_FUNCTOR_CHARS, "type_error", 2,
+		         PL_CHARS, expected,
+		         PL_TERM, actual,
+		       PL_VARIABLE) )
+    return PL_raise_exception(ex);
 
-  return PL_raise_exception(ex);
+  return FALSE;
 }
 
 
 static int
 existence_error(term_t actual, const char *expected)
-{ term_t ex = PL_new_term_ref();
+{ term_t ex;
 
-  PL_unify_term(ex, PL_FUNCTOR_CHARS, "error", 2,
-		      PL_FUNCTOR_CHARS, "existence_error", 2,
-		        PL_CHARS, expected,
-		        PL_TERM, actual,
-		      PL_VARIABLE);
-
-  return PL_raise_exception(ex);
+  if ( (ex=PL_new_term_ref()) &&
+       PL_unify_term(ex,
+		     PL_FUNCTOR_CHARS, "error", 2,
+		       PL_FUNCTOR_CHARS, "existence_error", 2,
+		         PL_CHARS, expected,
+		         PL_TERM, actual,
+		       PL_VARIABLE) )
+    return PL_raise_exception(ex);
+  return FALSE;
 }
 
 
@@ -272,7 +282,7 @@ get_size_ex(term_t t, size_t *v)
   if ( PL_get_int64(t, &i) )
   { if ( i < 0 )
       return domain_error(t, "nonneg");
-      
+
     *v = (size_t)i;			/* TBD: Check on 32-bit systems */
     return TRUE;
   }
@@ -288,7 +298,7 @@ get_offset_ex(term_t t, table_offset_t *v)
   if ( PL_get_int64(t, &i) )
   { if ( i < 0 )
       return domain_error(t, "nonneg");
-      
+
     *v = (table_offset_t)i;		/* TBD: Check on 32-bit systems */
     return TRUE;
   }
@@ -311,7 +321,7 @@ get_table_ex(term_t handle, Table *table)
 
     return existence_error(handle, "table");
   }
-    
+
   return type_error(handle, "table");
 }
 
@@ -377,7 +387,7 @@ default_escape_table(Table t)
   { PL_warning("Not enough memory");
     return;
   }
-  
+
   for(i=0; i<256; i++)
     t->escape_table[i] = i;
 
@@ -405,7 +415,7 @@ get_char(term_t t, int *chr)
 
 static foreign_t
 pl_new_table(term_t file, term_t columns, term_t options, term_t handle)
-{ Table table = malloc(sizeof(struct tabletag));	/* table to create */
+{ Table table = malloc(sizeof(*table));	/* table to create */
   field fields[MAXFIELDS];		/* scratch field structures */
   int   nfields=0;			/* # collected fields */
   int   defarg=1;			/* default argument */
@@ -413,13 +423,17 @@ pl_new_table(term_t file, term_t columns, term_t options, term_t handle)
   term_t head = PL_new_term_ref();	/* scratch list-head */
   term_t arg  = PL_new_term_ref();	/* scratch argument */
 
+  memset(table, 0, sizeof(*table));
+
   table->record_functor = 0;		/* not filled */
   table->keyfield = -1;
   table->escape = -1;
   table->escape_table = NULL;
 
   if ( !PL_get_atom(file, &table->file) )
+  { free(table);
     return error(ERR_INSTANTIATION, "open_table/4", 1, file);
+  }
 
   while(PL_get_list(tail, head, tail))
   { int arity;
@@ -430,7 +444,9 @@ pl_new_table(term_t file, term_t columns, term_t options, term_t handle)
 	 !PL_get_arg(1, head, arg) ||
 	 !PL_get_atom(arg, &typename) ||
 	 !get_type(typename, &fields[nfields].type) )
+    { free(table);
       return error(ERR_INSTANTIATION, "new_table/4", 2, columns);
+    }
 
     fields[nfields].index = nfields;	/* index number (0..) */
     fields[nfields].width = 0;		/* variable-width field */
@@ -444,14 +460,14 @@ pl_new_table(term_t file, term_t columns, term_t options, term_t handle)
       term_t head2 = PL_new_term_ref();
       term_t arg2  = PL_new_term_ref();
 
-      PL_get_arg(2, head, tail2);
+      _PL_get_arg(2, head, tail2);
       while(PL_get_list(tail2, head2, tail2))
       { atom_t a;
 	int optarity;
 
 	if ( PL_get_name_arity(head2, &a, &optarity) )
 	{ if ( optarity == 1 )
-	  { PL_get_arg(1, head2, arg2);
+	  { _PL_get_arg(1, head2, arg2);
 
 	    if ( !get_field_flag(a, arg2, &fields[nfields]) )
 	      goto colerr;
@@ -461,12 +477,15 @@ pl_new_table(term_t file, term_t columns, term_t options, term_t handle)
 	  }
 	} else
 	{ colerr:
+	  free(table);
 	  return error(ERR_INSTANTIATION, "new_table/4", 2, columns);
 	}
       }
       if ( !PL_get_nil(tail2) )
+      { free(table);
 	return error(ERR_INSTANTIATION, "new_table/4", 2, columns);
-      
+      }
+
       if ( fields[nfields].flags & FIELD_SORTED )
 	table->keyfield = nfields;
 
@@ -479,7 +498,9 @@ pl_new_table(term_t file, term_t columns, term_t options, term_t handle)
     nfields++;
   }
   if ( !PL_get_nil(tail) )
+  { free(table);
     return error(ERR_INSTANTIATION, "new_table/4", 2, columns);
+  }
 
   table->record_sep = '\n';
   table->field_sep  = ' ';
@@ -495,13 +516,13 @@ pl_new_table(term_t file, term_t columns, term_t options, term_t handle)
     { term_t head2 = PL_new_term_ref();
       term_t tail2 = PL_new_term_ref();
 
-      PL_get_arg(1, head, arg);
+      _PL_get_arg(1, head, arg);
 
       if ( !PL_get_integer(arg, &table->escape) )
 	goto err3;
 
       default_escape_table(table);
-      PL_get_arg(2, head, tail2);
+      _PL_get_arg(2, head, tail2);
       while(PL_get_list(tail2, head2, tail2))
       { atom_t name;
 	int arity;
@@ -525,7 +546,7 @@ pl_new_table(term_t file, term_t columns, term_t options, term_t handle)
 
     if ( arity != 1 )
       goto err3;
-    PL_get_arg(1, head, arg);
+    _PL_get_arg(1, head, arg);
     if ( name == ATOM_record_separator )
     { if ( !PL_get_integer(arg, &table->record_sep) )
 	goto err3;
@@ -540,13 +561,14 @@ pl_new_table(term_t file, term_t columns, term_t options, term_t handle)
   }
   if ( !PL_get_nil(tail) )
   { err3:
+    free(table);
     return error(ERR_INSTANTIATION, "new_table/4", 3, options);
   }
 
   table->nfields = nfields;
   table->fields = malloc(sizeof(struct fieldtag)*nfields);
   memcpy(&table->fields[0], fields, sizeof(struct fieldtag)*nfields);
-  
+
   if ( !table->record_functor )
   { int	maxarg=0;
     int i;
@@ -575,7 +597,7 @@ pl_new_table(term_t file, term_t columns, term_t options, term_t handle)
 
 static int
 open_table(Table table)
-{ if ( !table->buffer )
+{ if ( !table->opened )
   {
 #ifdef __WINDOWS__
     BY_HANDLE_FILE_INFORMATION info;
@@ -592,28 +614,31 @@ open_table(Table table)
 
     if ( !GetFileInformationByHandle(table->hfile, &info) )
       goto errio;
-    
+
     table->size = info.nFileSizeLow;
 
-    table->hmap = CreateFileMapping(table->hfile,
-				    NULL,
-				    PAGE_READONLY,
-				    0L,
-				    (DWORD)table->size, /* Truncated? */
-				    NULL);
-    if ( !table->hmap )
-      goto errio;
+    if ( table->size > 0 )
+    { table->hmap = CreateFileMapping(table->hfile,
+				      NULL,
+				      PAGE_READONLY,
+				      0L,
+				      (DWORD)table->size, /* Truncated? */
+				      NULL);
+      if ( !table->hmap )
+	goto errio;
+
+      table->buffer = MapViewOfFile(table->hmap,
+				    FILE_MAP_READ,
+				    0L, 0L, /* offset */
+				    0L);	/* size (0=all) */
+
+      if ( !table->buffer )
+	goto errio;
+    }
 
-    table->buffer = MapViewOfFile(table->hmap,
-				  FILE_MAP_READ,
-				  0L, 0L, /* offset */
-				  0L);	/* size (0=all) */
-				  
-    if ( !table->buffer )
-      goto errio;
-    
     table->window      = table->buffer;
     table->window_size = table->size;
+    table->opened      = TRUE;
 
     return TRUE;
 
@@ -647,16 +672,18 @@ open_table(Table table)
       goto errio;
 
     table->size = buf.st_size;
-    
-    if ( (table->buffer = mmap(NULL, table->size,
-			       PROT_READ, MAP_SHARED|MAP_NORESERVE,
-			       table->fd, 0)) == (char *) -1 )
-      goto errio;
+    if ( table->size > 0 )
+    { if ( (table->buffer = mmap(NULL, table->size,
+				 PROT_READ, MAP_SHARED|MAP_NORESERVE,
+				 table->fd, 0)) == (char *) -1 )
+	goto errio;
+    }
     close(table->fd);
     table->fd = -1;
 
     table->window      = table->buffer;
     table->window_size = table->size;
+    table->opened      = TRUE;
 
     return TRUE;
 
@@ -712,7 +739,7 @@ pl_get_table_attribute(term_t handle, term_t name, term_t value)
   if ( !get_table_ex(handle, &table) )
     return FALSE;
 
-  
+
   if ( PL_get_name_arity(name, &n, &arity) )
   { if ( n == ATOM_file && arity == 0 )
       return PL_unify_atom(value, table->file);
@@ -720,8 +747,8 @@ pl_get_table_attribute(term_t handle, term_t name, term_t value)
     if ( n == ATOM_field && arity == 1 )
     { term_t a = PL_new_term_ref();
       int i;
-      
-      PL_get_arg(1, name, a);
+
+      _PL_get_arg(1, name, a);
       if ( PL_get_integer(a, &i) )
       { if ( i >= 1 && i <= table->nfields )
 	  return unify_field_info(value, &table->fields[i-1]);
@@ -794,53 +821,63 @@ unify_field_info(term_t t, Field field)	/* name(Type, Flags) */
   }
 
   if ( field->flags & FIELD_UNIQUE )
-  { PL_unify_list(tail, head, tail);
-    PL_unify_atom(head, ATOM_unique);
+  { if ( !PL_unify_list(tail, head, tail) ||
+	 !PL_unify_atom(head, ATOM_unique) )
+      return FALSE;
     options++;
   }
   if ( field->flags & FIELD_DOWNCASE )
-  { PL_unify_list(tail, head, tail);
-    PL_unify_atom(head, ATOM_downcase);
+  { if ( !PL_unify_list(tail, head, tail) ||
+	 !PL_unify_atom(head, ATOM_downcase) )
+      return FALSE;
     options++;
   }
   if ( field->flags & FIELD_ALLOWBADNUM )
-  { PL_unify_list(tail, head, tail);
-    PL_unify_atom(head, ATOM_syntax);
+  { if ( !PL_unify_list(tail, head, tail) ||
+	 !PL_unify_atom(head, ATOM_syntax) )
+      return FALSE;
     options++;
   }
   if ( field->flags & FIELD_MAPSPACETOUNDERSCORE )
-  { PL_unify_list(tail, head, tail);
-    PL_unify_atom(head, ATOM_map_space_to_underscore);
+  { if ( !PL_unify_list(tail, head, tail) ||
+	 !PL_unify_atom(head, ATOM_map_space_to_underscore) )
+      return FALSE;
     options++;
   }
   if ( field->flags & FIELD_SORTED )
-  { PL_unify_list(tail, head, tail);
+  { if ( !PL_unify_list(tail, head, tail) )
+      return FALSE;
     if ( field->ord )
-      PL_unify_term(head, PL_FUNCTOR, PL_new_functor(ATOM_sorted, 1),
-			    PL_ATOM, field->ord->name);
-    else
-      PL_unify_atom(head, ATOM_sorted);
+    { if ( !PL_unify_term(head, PL_FUNCTOR, PL_new_functor(ATOM_sorted, 1),
+			  PL_ATOM, field->ord->name) )
+	return FALSE;
+    } else
+    { if ( !PL_unify_atom(head, ATOM_sorted) )
+	return FALSE;
+    }
     options++;
   }
   if ( field->width > 0 )
-  { PL_unify_term(head, PL_FUNCTOR, PL_new_functor(ATOM_width, 1),
-			    PL_INT, field->width);
+  { if ( !PL_unify_term(head, PL_FUNCTOR, PL_new_functor(ATOM_width, 1),
+			PL_INT, field->width) )
+      return FALSE;
     options++;
   }
   if ( field->arg > 0 )
-  { PL_unify_term(head, PL_FUNCTOR, PL_new_functor(ATOM_arg, 1),
-			    PL_INT, field->arg);
+  { if ( !PL_unify_term(head, PL_FUNCTOR, PL_new_functor(ATOM_arg, 1),
+			PL_INT, field->arg) )
+      return FALSE;
     options++;
   }
 
   if ( options )
-  { PL_unify_nil(tail);
-    return PL_unify_term(t, PL_FUNCTOR, PL_new_functor(field->name, 2),
-			 	PL_ATOM, type,
-			 	PL_TERM, flags);
+  { return (PL_unify_nil(tail) &&
+	    PL_unify_term(t, PL_FUNCTOR, PL_new_functor(field->name, 2),
+			  PL_ATOM, type,
+			  PL_TERM, flags));
   } else
   { return PL_unify_term(t, PL_FUNCTOR, PL_new_functor(field->name, 1),
-			  	PL_ATOM, type);
+				PL_ATOM, type);
 
   }
 }
@@ -873,8 +910,6 @@ pl_close_table(term_t handle)
       CloseHandle(table->hmap);
     if ( table->hfile )
       CloseHandle(table->hfile);
-    table->size   = -1;
-    table->buffer = NULL;
     table->hfile  = NULL;
     table->hmap   = NULL;
 #endif
@@ -883,11 +918,13 @@ pl_close_table(term_t handle)
       munmap(table->buffer, table->size);
     if ( table->fd >= 0 )
       close(table->fd);
-    table->size   = -1;
-    table->buffer = NULL;
     table->fd     = -1;
 #endif
+
+    table->size   = -1;
+    table->buffer = NULL;
     table->window = NULL;
+    table->opened = FALSE;
   }
 
   PL_succeed;
@@ -903,11 +940,12 @@ pl_free_table(term_t handle)
 
   if ( !get_table_ex(handle, &table) )
     return FALSE;
-  
+
   table->magic = 0;			/* so it won't be recognised */
   if ( table->escape_table )
     free(table->escape_table);
 
+  free(table->fields);
   free(table);
 
   PL_succeed;
@@ -973,7 +1011,7 @@ find_start_of_record(Table t, table_offset_t start)
       s--;
 
     return s-t->window;
-  }    
+  }
 }
 
 
@@ -981,7 +1019,7 @@ static table_offset_t
 previous_record(Table t, table_offset_t start)
 { char *s;
   int er = t->record_sep;
-    
+
   if ( start < 0 || start > (table_offset_t)t->window_size )
     return -1;
 
@@ -1025,7 +1063,7 @@ start_of_record(+Table, +From, +To, -StartOfRecord)
 
 static foreign_t
 pl_start_of_record(term_t handle,		/* table */
-		   term_t from, term_t to, 	/* range  */
+		   term_t from, term_t to,	/* range  */
 		   term_t recstart,		/* return */
 		   control_t control)		/* backtracking control */
 { Table table;
@@ -1043,7 +1081,7 @@ pl_start_of_record(term_t handle,		/* table */
     case PL_REDO:
       f = PL_foreign_context(control);
       break;
-    case PL_CUTTED:
+    case PL_PRUNED:
     default:
       PL_succeed;
   }
@@ -1064,12 +1102,12 @@ pl_start_of_record(term_t handle,		/* table */
 					/* find start of search */
   if ( f <= 0 )
     start = table->window;
-  else 
+  else
     start = table->window + f;
-    
+
   if ( start > end )
     PL_fail;
-  
+
   er = table->record_sep;
 
   if ( start == table->window || start[-1] == er )
@@ -1085,10 +1123,10 @@ found:
     start++;
 
   n = start-table->window;
-  
+
   if ( PL_unify_integer(recstart, n) )
     PL_retry(n+1);
-  
+
   PL_fail;
 }
 
@@ -1196,7 +1234,7 @@ tab_memcpy(Table table, int flags, char *to, const char *from, size_t len)
   { strncpy(to, from, len);
     to[len] = EOS;
   }
-	
+
   if ( flags & FIELD_MAPSPACETOUNDERSCORE )
   { char *q;
 
@@ -1248,7 +1286,7 @@ unify_field_text(Table t, int flags, int type,
 #ifndef HAVE_ALLOCA
   if ( buf != tmp )
     free(tmp);
-#endif  
+#endif
   return rval;
 }
 
@@ -1343,7 +1381,7 @@ read_record(Table t, table_offset_t start, table_offset_t *end, term_t record)
 
   if ( !PL_unify_functor(record, t->record_functor) )
     return FALSE;
-  
+
   f = t->fields;
   for(n=1; n<=t->nfields; n++, f++)
   { if ( f->arg > 0 )
@@ -1448,11 +1486,12 @@ pl_read_fields(term_t handle, term_t from, term_t to, term_t fields)
 
     if ( !PL_get_name_arity(head, &a, &arity) || arity != 1 )
       return error(ERR_INSTANTIATION, "read_fields/4", 4, fields);
-    
+
     for(i=0; i<table->nfields; i++)
     { if ( table->fields[i].name == a )
       { argv[i] = PL_new_term_ref();
-	PL_get_arg(1, head, argv[i]);
+	if ( !PL_get_arg(1, head, argv[i]) )
+	  return FALSE;
 	goto cont;
       }
     }
@@ -1495,6 +1534,7 @@ pl_read_fields(term_t handle, term_t from, term_t to, term_t fields)
 #define MR_KEY_ONLY	0x02		/* only compare the key (=sorted) */
 					/* field */
 
+#define MATCH_ERROR    -3		/* Prolog (resource) error */
 #define MATCH_NORECORD -2		/* bad record */
 #define MATCH_LT       -1		/* < */
 #define MATCH_EQ	0		/* == */
@@ -1536,13 +1576,22 @@ match_field(Table t, Field f, QueryField q, table_offset_t start, table_offset_t
       if ( q->flags & QUERY_READ )
       { switch(f->type)
 	{ case FIELD_ATOM:
-	    PL_unify_atom_chars(q->value.term, tmp);
+	    if ( !PL_unify_atom_chars(q->value.term, tmp) )
+	    { DEALLOC();
+	      return MATCH_ERROR;
+	    }
 	    break;
 	  case FIELD_STRING:
-	    PL_unify_string_chars(q->value.term, tmp);
+	    if ( !PL_unify_string_chars(q->value.term, tmp) )
+	    { DEALLOC();
+	      return MATCH_ERROR;
+	    }
 	    break;
 	  case FIELD_CODELIST:
-	    PL_unify_list_chars(q->value.term, tmp);
+	    if ( !PL_unify_list_chars(q->value.term, tmp) )
+	    { DEALLOC();
+	      return MATCH_ERROR;
+	    }
 	    break;
 	}
 
@@ -1636,7 +1685,8 @@ match_field(Table t, Field f, QueryField q, table_offset_t start, table_offset_t
       }
 
       if ( q->flags & QUERY_READ )
-      { PL_unify_integer(q->value.term, l);
+      { if ( !PL_unify_integer(q->value.term, l) )
+	  return MATCH_ERROR;
 	return MATCH_EQ;
       }
 
@@ -1659,7 +1709,8 @@ match_field(Table t, Field f, QueryField q, table_offset_t start, table_offset_t
       }
 
       if ( q->flags & QUERY_READ )
-      { PL_unify_float(q->value.term, x);
+      { if ( !PL_unify_float(q->value.term, x) )
+	  return MATCH_ERROR;
 	return MATCH_EQ;
       }
 
@@ -1688,9 +1739,10 @@ match_record(Query q, table_offset_t start, table_offset_t *end, int flags)
       continue;
 
     match = match_field(t, f, qf, start, &start, (flags&MR_BIND));
-    
+
     switch(match)
     { case MATCH_NORECORD:
+      case MATCH_ERROR:
 	rval = match;
         goto out;
       case MATCH_EQ:
@@ -1726,15 +1778,22 @@ execute_binary_search(Query q)
   { table_offset_t next;
 
     switch( match_record(q, here, &next, MR_KEY_ONLY) )
-    { case MATCH_NORECORD:
+    { case MATCH_ERROR:
+	return -1;
+      case MATCH_NORECORD:
       { if ( here >= (table_offset_t)t->window_size )
 	  return FALSE;
 	here = next;
 	continue;
       }
       case MATCH_LT:
-      { high = here;
-	here = find_start_of_record(t, (low+high)/2);
+      { table_offset_t rc;
+	table_offset_t mid = (low+high)/2;
+
+	high = here;
+	while((rc=find_start_of_record(t, mid) == here) && mid > low)
+	  mid--;
+	here = rc;
 	DEBUG(Sdprintf("<, %d %d %d\n", low, here, high));
 	goto next;
       }
@@ -1751,13 +1810,17 @@ execute_binary_search(Query q)
 	  return here;
 	} else
 	{ table_offset_t first = here, prev = here;
-	
+
 	  while(prev > 0)		/* find the first */
 	  { prev = previous_record(t, prev);
-	    if ( match_record(q, prev, &next, MR_KEY_ONLY) == MATCH_EQ )
-	    { first = prev;
-	    } else
-	      break;
+	    switch( match_record(q, prev, &next, MR_KEY_ONLY) )
+	    { case MATCH_EQ:
+		first = prev;
+	        continue;
+	      case MATCH_ERROR:
+		return -1;
+	    }
+	    break;
 	  }
 	  return first;
 	}
@@ -1772,6 +1835,8 @@ execute_binary_search(Query q)
       { switch( match_record(q, here, &next, MR_KEY_ONLY) )
 	{ case MATCH_EQ:
 	    return here;
+	  case MATCH_ERROR:
+	    return -1;
 	}
 	here = next;
       }
@@ -1801,7 +1866,7 @@ field structures for a redo in pl_in_table().   No  checking needs to be
 done as the initial call has already done that.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
-static void
+static int
 rebind_query_vars(Query q, term_t from)
 { if ( q->nvars > 0 )
   { term_t tail = PL_copy_term_ref(from);
@@ -1810,26 +1875,30 @@ rebind_query_vars(Query q, term_t from)
     int varsleft = q->nvars;
 
     while(PL_get_list(tail, head, tail))
-    { PL_get_arg(1, head, arg);
+    { if ( !PL_get_arg(1, head, arg) )
+	return FALSE;
 
       if ( PL_is_variable(arg) )
       { atom_t name;
 	int arity;
 	int i;
-	
-	PL_get_name_arity(head, &name, &arity);
+
+	if ( !PL_get_name_arity(head, &name, &arity) )
+	  return FALSE;
 	for(i=0; i<q->table->nfields; i++)
 	{ if ( q->table->fields[i].name == name )
 	  { QueryField qf = &q->field[i];
 	    qf->value.term = PL_copy_term_ref(arg);
 	    if ( --varsleft == 0 )
-	      return;
+	      return TRUE;
 	    break;
 	  }
 	}
       }
     }
   }
+
+  return TRUE;
 }
 
 
@@ -1855,7 +1924,7 @@ make_query(Table t, term_t from)
 
     if ( !PL_get_name_arity(head, &name, &arity) || arity < 1 || arity > 2 )
       goto err2;
-    PL_get_arg(1, head, arg);
+    _PL_get_arg(1, head, arg);
 
     for(i=0; i<t->nfields; i++)
     { if ( t->fields[i].name == name )
@@ -1905,7 +1974,7 @@ make_query(Table t, term_t from)
 	  atom_t tab;
 	  int a;
 
-	  PL_get_arg(2, head, arg);
+	  _PL_get_arg(2, head, arg);
 	  if ( !PL_get_name_arity(arg, &opt, &a) || a > 1 )
 	    goto err2;
 	  if ( opt == ATOM_prefix )
@@ -1960,7 +2029,7 @@ unique_match(Query q)
   { if ( !(qf->flags & QUERY_DONTCARE) && f->flags & FIELD_UNIQUE )
       return TRUE;
   }
-  
+
   return FALSE;
 }
 
@@ -1976,7 +2045,7 @@ pl_in_table(term_t handle, term_t spec, term_t record, control_t control)
   switch(PL_foreign_control(control))
   { case PL_FIRST_CALL:
     { Table t;
-      
+
       if ( !get_table_ex(handle, &t) )
 	return FALSE;
       if ( !open_table(t) )
@@ -1999,7 +2068,7 @@ pl_in_table(term_t handle, term_t spec, term_t record, control_t control)
       q = PL_foreign_context_address(control);
       rebind_query_vars(q, spec);
       break;
-    case PL_CUTTED:
+    case PL_PRUNED:
     default:
       q = PL_foreign_context_address(control);
       if ( q )
@@ -2014,8 +2083,11 @@ pl_in_table(term_t handle, term_t spec, term_t record, control_t control)
     DEBUG(Sdprintf("Binary search, match at offset=%ld\n", q->offset));
 
     if ( q->technique & TECH_UNIQUE )
-    { if ( match_record(q, q->offset, &next, MR_BIND) == MATCH_EQ )
-      { PL_unify_integer(record, q->offset);
+    { int rc;
+
+      if ( (rc=match_record(q, q->offset, &next, MR_BIND)) == MATCH_EQ )
+      { if ( !PL_unify_integer(record, q->offset) )
+	  return FALSE;
 	if ( unique_match(q) )
 	{ free_query(q);
 	  return TRUE;
@@ -2023,16 +2095,24 @@ pl_in_table(term_t handle, term_t spec, term_t record, control_t control)
 	q->offset = next;
 
 	PL_retry_address(q);
+      } else if ( rc == MATCH_ERROR )
+      { free_query(q);
+	return FALSE;
       }
     } else
     { DEBUG(Sdprintf("Non-unique match\n"));
 
       do
-      { DEBUG(Sdprintf("Trying offset %ld\n", q->offset));
+      { int rc;
 
-	if ( match_record(q, q->offset, &next, 0) == MATCH_EQ )
-	{ match_record(q, q->offset, &next, MR_BIND);
-	  PL_unify_integer(record, q->offset);
+	DEBUG(Sdprintf("Trying offset %ld\n", q->offset));
+
+	if ( (rc=match_record(q, q->offset, &next, 0)) == MATCH_EQ )
+	{ if ( (match_record(q, q->offset, &next, MR_BIND) == MATCH_ERROR) ||
+	       !PL_unify_integer(record, q->offset) )
+	  { free_query(q);
+	    return FALSE;
+	  }
 	  if ( unique_match(q) )
 	  { free_query(q);
 	    return TRUE;
@@ -2040,6 +2120,9 @@ pl_in_table(term_t handle, term_t spec, term_t record, control_t control)
 	  q->offset = next;
 
 	  PL_retry_address(q);
+	} else if ( rc == MATCH_ERROR )
+	{ free_query(q);
+	  return FALSE;
 	}
 
 	q->offset = next;
@@ -2048,7 +2131,7 @@ pl_in_table(term_t handle, term_t spec, term_t record, control_t control)
 
     free_query(q);
     return FALSE;
-  } 
+  }
 
 
   while(q->offset < (table_offset_t)q->table->window_size)
@@ -2056,7 +2139,8 @@ pl_in_table(term_t handle, term_t spec, term_t record, control_t control)
 
     if ( match_record(q, q->offset, &next, 0) == MATCH_EQ )
     { match_record(q, q->offset, &next, MR_BIND);
-      PL_unify_integer(record, q->offset);
+      if ( !PL_unify_integer(record, q->offset) )
+	return FALSE;
       if ( unique_match(q) )
       { free_query(q);
 	return TRUE;
