@@ -19,7 +19,7 @@
 
     You should have received a copy of the GNU Lesser General Public
     License along with this library; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */
 
 #define O_DEBUG 1
@@ -40,6 +40,7 @@
 #include <sys/types.h>
 #include <assert.h>
 #include <string.h>
+#include <stdio.h>
 
 #ifdef HAVE_LINUX_TIPC_H
 #include <linux/tipc.h>
@@ -47,6 +48,8 @@
 #error "Cannot find <tipc.h>"
 #endif
 
+static int tipc_version = 2;
+
 static atom_t ATOM_scope;
 static atom_t ATOM_no_scope;
 static atom_t ATOM_node;
@@ -78,9 +81,11 @@ static atom_t ATOM_conn_timeout;	/* "conn_timeout" */
 
 static functor_t FUNCTOR_tipc_socket1;	/* $tipc_socket(Id) */
 static functor_t FUNCTOR_port_id;
-static functor_t FUNCTOR_name;
-static functor_t FUNCTOR_name_seq;
-static functor_t FUNCTOR_mcast;
+static functor_t FUNCTOR_name3;
+static functor_t FUNCTOR_name_seq3;
+static functor_t FUNCTOR_mcast3;
+
+#define V1_TIPC_SUB_SERVICE 0x02
 
 
 		 /*******************************
@@ -95,7 +100,7 @@ tipc_get_socket(term_t Socket, int *id)
   if ( PL_is_functor(Socket, FUNCTOR_tipc_socket1) )
   { term_t a = PL_new_term_ref();
 
-    PL_get_arg(1, Socket, a);
+    _PL_get_arg(1, Socket, a);
     if ( PL_get_integer(a, id) )
       return TRUE;
   }
@@ -125,6 +130,24 @@ tipc_unify_socket(term_t Socket, int id)
 #define tcp_get_socket(t, p) tipc_get_socket(t, p)
 #include "sockcommon.c"
 
+int
+get_uint(term_t term, unsigned *value)
+{ int64_t v0;
+
+  if ( !PL_get_int64(term, &v0))
+    return FALSE;
+
+#if 0
+	/* tipc users are somewhat cavalier about using -1 in place of 4294967295U */
+  if(v0 < -1 || v0 > UINT_MAX)
+	return FALSE;
+#endif
+
+  *value = (unsigned) v0 & 0xffffffff;
+
+  return TRUE;
+}
+
 static int
 nbio_get_tipc(term_t tipc, struct sockaddr_tipc *sockaddr)
 { term_t a = PL_new_term_ref();
@@ -134,14 +157,14 @@ nbio_get_tipc(term_t tipc, struct sockaddr_tipc *sockaddr)
   {
   if ( PL_is_functor(tipc, FUNCTOR_port_id) )
   {
-    int ref, node;
+    unsigned ref, node;
 
-    PL_get_arg(1, tipc, a);
-    if ( !PL_get_integer(a, &ref) )
+    _PL_get_arg(1, tipc, a);
+    if ( !get_uint(a, &ref) )
       break;
 
-    PL_get_arg(2, tipc, a);
-    if ( !PL_get_integer(a, &node) )
+    _PL_get_arg(2, tipc, a);
+    if ( !get_uint(a, &node) )
       break;
 
     sockaddr->addrtype     = TIPC_ADDR_ID;
@@ -151,20 +174,20 @@ nbio_get_tipc(term_t tipc, struct sockaddr_tipc *sockaddr)
     return TRUE;
   }
 
-  if ( PL_is_functor(tipc, FUNCTOR_name) )
+  if ( PL_is_functor(tipc, FUNCTOR_name3) )
   {
-    int arg1, arg2, arg3;
+    unsigned arg1, arg2, arg3;
 
-    PL_get_arg(1, tipc, a);
-    if ( !PL_get_integer(a, &arg1) )
+    _PL_get_arg(1, tipc, a);
+    if ( !get_uint(a, &arg1) )
       break;
 
-    PL_get_arg(2, tipc, a);
-    if ( !PL_get_integer(a, &arg2) )
+    _PL_get_arg(2, tipc, a);
+    if ( !get_uint(a, &arg2) )
       break;
 
-    PL_get_arg(3, tipc, a);
-    if ( !PL_get_integer(a, &arg3) )
+    _PL_get_arg(3, tipc, a);
+    if ( !get_uint(a, &arg3) )
       break;
 
     sockaddr->addrtype                = TIPC_ADDR_NAME;
@@ -176,21 +199,21 @@ nbio_get_tipc(term_t tipc, struct sockaddr_tipc *sockaddr)
 
   }
 
-  if ( PL_is_functor(tipc, FUNCTOR_name_seq) ||
-       PL_is_functor(tipc, FUNCTOR_mcast))
+  if ( PL_is_functor(tipc, FUNCTOR_name_seq3) ||
+       PL_is_functor(tipc, FUNCTOR_mcast3))
   {
-    int arg1, arg2, arg3;
+    unsigned arg1, arg2, arg3;
 
-    PL_get_arg(1, tipc, a);
-    if ( !PL_get_integer(a, &arg1) )
+    _PL_get_arg(1, tipc, a);
+    if ( !get_uint(a, &arg1) )
       break;
 
-    PL_get_arg(2, tipc, a);
-    if ( !PL_get_integer(a, &arg2) )
+    _PL_get_arg(2, tipc, a);
+    if ( !get_uint(a, &arg2) )
       break;
 
-    PL_get_arg(3, tipc, a);
-    if ( !PL_get_integer(a, &arg3) )
+    _PL_get_arg(3, tipc, a);
+    if ( !get_uint(a, &arg3) )
       break;
 
     sockaddr->addrtype           = TIPC_ADDR_NAMESEQ;
@@ -361,7 +384,7 @@ pl_tipc_setopt(term_t Socket, term_t opt)
       } else /*if ( arity == 1 )*/
       { term_t a = PL_new_term_ref();
 
-	PL_get_arg(1, opt, a);
+	_PL_get_arg(1, opt, a);
 	if ( !PL_get_bool(a, &enable) )
 	  return pl_error(NULL, 0, NULL, ERR_DOMAIN, a, "boolean");
       }
@@ -405,7 +428,7 @@ unify_tipc_address(term_t t, struct sockaddr_tipc *addr)
 
 
 static foreign_t
-pl_tipc_get_name(term_t Socket, term_t t)
+pl_tipc_basic_get_name(term_t Socket, term_t t, int peer)
 { struct sockaddr_tipc addr;
   int socket;
   SOCKET fd;
@@ -420,14 +443,26 @@ pl_tipc_get_name(term_t Socket, term_t t)
 
   fd = nbio_fd(socket);
 
-  if ( getsockname(fd, (struct sockaddr *) &addr, &alen) )
+  if ( (peer) ? getpeername(fd, (struct sockaddr *) &addr, &alen)
+			  : getsockname(fd, (struct sockaddr *) &addr, &alen) )
     return nbio_error(errno, TCP_ERRNO);
   else
     return unify_tipc_address(t, &addr);
 }
 
+static foreign_t
+pl_tipc_get_name(term_t Socket, term_t t)
+{
+	return pl_tipc_basic_get_name(Socket, t, 0);
+}
 
-#define TIPC_MAXDATA 4096
+static foreign_t
+pl_tipc_get_peer_name(term_t Socket, term_t t)
+{
+	return pl_tipc_basic_get_name(Socket, t, 1);
+}
+
+#define TIPC_MAXDATA TIPC_MAX_USER_MSG_SIZE
 
 static foreign_t
 pl_tipc_receive(term_t Socket, term_t Data, term_t From, term_t options)
@@ -457,7 +492,7 @@ pl_tipc_receive(term_t Socket, term_t Data, term_t From, term_t options)
 	if ( name == ATOM_as && arity == 1)
 	{ atom_t a;
 
-          PL_get_arg(1, head, arg);
+	  _PL_get_arg(1, head, arg);
 
 	  if ( !PL_get_atom(arg, &a) )
 	    return pl_error(NULL, 0, NULL, ERR_TYPE, head, "atom");
@@ -663,7 +698,7 @@ pl_tipc_subscribe(term_t Socket, term_t Address,
 { struct sockaddr_tipc sockaddr;
   struct tipc_subscr subscr;
   int socket;
-  int time, filt;
+  unsigned time, filt;
   char *handle;
   size_t handle_len;
   SOCKET fd;
@@ -678,21 +713,36 @@ pl_tipc_subscribe(term_t Socket, term_t Address,
   if(sockaddr.addrtype != TIPC_ADDR_NAMESEQ)
     return pl_error(NULL, 0, NULL, ERR_DOMAIN, Address, "name_seq/3");
 
-  if( !PL_get_integer(timeout, &time))
+  if( !get_uint(timeout, &time))
     return pl_error(NULL, 0, NULL, ERR_DOMAIN, timeout, "integer");
 
-  if( !PL_get_integer(filter, &filt))
+  if( !get_uint(filter, &filt))
     return pl_error(NULL, 0, NULL, ERR_DOMAIN, filter, "integer");
 
   if ( !PL_get_nchars(usr_handle, &handle_len, &handle, CVT_ALL|CVT_EXCEPTION) )
     return FALSE;
 
-  memcpy(&subscr.seq, &sockaddr.addr.nameseq, sizeof(subscr.seq));
-  subscr.timeout = (unsigned) time;
-  subscr.filter = (unsigned) filt;
-  memcpy(&subscr.usr_handle, handle,
-	 (handle_len < sizeof(subscr.usr_handle)) ? handle_len
-	 					  : sizeof(subscr.usr_handle));
+  if(tipc_version > 1)
+  { struct tipc_name_seq *p = &subscr.seq,
+                         *p1 = &sockaddr.addr.nameseq;
+
+    p->type = htonl(p1->type);
+    p->lower = htonl(p1->lower);
+    p->upper = htonl(p1->upper);
+
+    subscr.timeout = htonl(time);
+    subscr.filter = htonl((filt == V1_TIPC_SUB_SERVICE) 
+                          ? TIPC_SUB_SERVICE 
+                          : filt);
+  } else {
+    memcpy(&subscr.seq, &sockaddr.addr.nameseq, sizeof(subscr.seq));
+    subscr.timeout = time;
+    subscr.filter = filt;
+  }
+  memcpy(&subscr.usr_handle, handle, 
+         (handle_len < sizeof(subscr.usr_handle)) 
+          ? handle_len
+          : sizeof(subscr.usr_handle));
 
   fd = nbio_fd(socket);
 
@@ -702,10 +752,119 @@ pl_tipc_subscribe(term_t Socket, term_t Address,
     return TRUE;
 }
 
+static foreign_t
+pl_tipc_receive_subscr_event(term_t Socket, term_t Data)
+{ struct sockaddr_tipc sockaddr;
+#ifdef __WINDOWS__
+  int alen = sizeof(sockaddr);
+#else
+  socklen_t alen = sizeof(sockaddr);
+#endif
+  int socket;
+  int flags = 0;
+  union {
+     char asCodes[sizeof(struct tipc_event)];
+     struct tipc_event asEvent;
+  } buf;
+
+  ssize_t n;
+  struct tipc_event *event = &buf.asEvent;
+
+  if ( !tipc_get_socket(Socket, &socket))
+    return FALSE;
+
+  if ( (n=nbio_recvfrom(socket, buf.asCodes, sizeof(buf.asCodes), flags,
+			(struct sockaddr*)&sockaddr, &alen)) == -1 )
+    return nbio_error(errno, TCP_ERRNO);
+
+  if(n != sizeof(*event))
+     return FALSE;
+
+  if(tipc_version > 1)
+  { struct tipc_name_seq *p = &event->s.seq;
+
+    event->event = ntohl(event->event);
+    event->found_lower = ntohl(event->found_lower);
+    event->found_upper = ntohl(event->found_upper);
+
+    event->port.ref = ntohl(event->port.ref);
+    event->port.node = ntohl(event->port.node);
+
+    p->type = ntohl(p->type);
+    p->lower = ntohl(p->lower);
+    p->upper = ntohl(p->upper);
+
+    event->s.timeout = ntohl(event->s.timeout);
+    event->s.filter = ntohl(event->s.filter);
+
+    if(event->s.filter == TIPC_SUB_SERVICE) 
+        event->s.filter = V1_TIPC_SUB_SERVICE;
+  }
+
+  switch(event->event) 
+  {
+      case TIPC_PUBLISHED:
+      case TIPC_WITHDRAWN:
+        { term_t Found = PL_new_term_ref(),
+                 Port_id = PL_new_term_ref(),
+                 Subscr = PL_new_term_ref();
+          const char *event_chars = (event->event == TIPC_PUBLISHED)
+                                    ? "published"
+                                    : "withdrawn";
+
+          if(!PL_unify_term(Subscr, PL_FUNCTOR_CHARS, "name_seq", 3,
+			   IntArg(event->s.seq.type),
+			   IntArg(event->s.seq.lower),
+			   IntArg(event->s.seq.upper)))
+             return FALSE;
+
+          if(!PL_unify_term(Found, PL_FUNCTOR_CHARS, "name_seq", 3,
+			   IntArg(event->s.seq.type),
+			   IntArg(event->found_lower),
+			   IntArg(event->found_upper)))
+             return FALSE;
+
+          if(!PL_unify_term(Port_id, PL_FUNCTOR_CHARS, "port_id", 2,
+			   IntArg(event->port.ref),
+			   IntArg(event->port.node)))
+             return FALSE;
+
+          if(!PL_unify_term(Data, PL_FUNCTOR_CHARS, "tipc_event", 4,
+               AtomArg(event_chars),
+			   PL_TERM, Subscr,
+			   PL_TERM, Found,
+			   PL_TERM, Port_id))
+             return FALSE;
+          
+          return TRUE;
+        }
+
+      case TIPC_SUBSCR_TIMEOUT:
+        {
+          return PL_unify_term(Data, PL_FUNCTOR_CHARS, "subscr_timeout", 0);
+        }
+      default:
+          return FALSE;
+   };
+  return FALSE;
+}
+
 
 install_t
 install_tipc()
-{ nbio_init("tipc");
+{ FILE *fp = fopen("/sys/module/tipc/version", "r");
+
+  if(fp)
+    { char buf[32];
+      size_t n = fread(buf, sizeof(char), sizeof(buf), fp);
+
+      if(n > 0)
+         tipc_version = buf[0] - '0';
+
+      fclose(fp);
+    } 
+
+  nbio_init("tipc");
 
   ATOM_scope	       = PL_new_atom("scope");
   ATOM_no_scope	       = PL_new_atom("no_scope");
@@ -738,9 +897,9 @@ install_tipc()
 
   FUNCTOR_tipc_socket1 = PL_new_functor(PL_new_atom("$tipc_socket"), 1);
   FUNCTOR_port_id      = PL_new_functor(PL_new_atom("port_id"), 2);
-  FUNCTOR_name	       = PL_new_functor(PL_new_atom("name"), 3);
-  FUNCTOR_name_seq     = PL_new_functor(PL_new_atom("name_seq"), 3);
-  FUNCTOR_mcast	       = PL_new_functor(PL_new_atom("mcast"), 3);
+  FUNCTOR_name3	       = PL_new_functor(PL_new_atom("name"), 3);
+  FUNCTOR_name_seq3    = PL_new_functor(PL_new_atom("name_seq"), 3);
+  FUNCTOR_mcast3       = PL_new_functor(PL_new_atom("mcast"), 3);
 
   PL_register_foreign("tipc_socket",          2, tipc_socket,         0);
   PL_register_foreign("tipc_close_socket",    1, tipc_close_socket,   0);
@@ -751,7 +910,9 @@ install_tipc()
   PL_register_foreign("tipc_accept",          3, pl_tipc_accept,      0);
   PL_register_foreign("tipc_connect",         2, pl_tipc_connect,     0);
   PL_register_foreign("tipc_get_name",        2, pl_tipc_get_name,    0);
+  PL_register_foreign("tipc_get_peer_name",   2, pl_tipc_get_peer_name, 0);
   PL_register_foreign("tipc_receive",	      4, pl_tipc_receive,     0);
-  PL_register_foreign("tipc_send",	      4, pl_tipc_send,	      0);
+  PL_register_foreign("tipc_send",	          4, pl_tipc_send,	      0);
   PL_register_foreign("tipc_subscribe",	      5, pl_tipc_subscribe,   0);
+  PL_register_foreign("tipc_receive_subscr_event", 2, pl_tipc_receive_subscr_event,     0);
 }
