@@ -3,9 +3,9 @@
     Part of XPCE --- The SWI-Prolog GUI toolkit
 
     Author:        Jan Wielemaker and Anjo Anjewierden
-    E-mail:        jan@swi.psy.uva.nl
+    E-mail:        J.Wielemaker@cs.vu.nl
     WWW:           http://www.swi.psy.uva.nl/projects/xpce/
-    Copyright (C): 1985-2002, University of Amsterdam
+    Copyright (C): 1985-2010, University of Amsterdam, VU University Amsterdam
 
     This program is free software; you can redistribute it and/or
     modify it under the terms of the GNU General Public License
@@ -19,7 +19,7 @@
 
     You should have received a copy of the GNU Lesser General Public
     License along with this library; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 
     As a special exception, if you link this library with other files,
     compiled with a Free Software compiler, to produce an executable, this
@@ -53,10 +53,11 @@
 	  justify_paragraph        = key('\\eQ'),
 	  bookmark_line            = key('\\e@'),
 	  execute_extended_command = key('\\ex'),
-	  sticky_window		   = key('\\es'),
 	  write_region		   = key('\\e\\C-w'),
 	  compile		   = key('\\C-x\\C-m'),
 
+	  prefix		   = key('\\eg'),
+
 					% FILE menu
 	  show_buffer_menu	   = key('\\C-x\\C-b') + button(file),
 	  switch_to_buffer	   = key('\\C-xb') +
@@ -75,7 +76,7 @@
 	  ispell		   = button(file),
 	  shell			   = button(file),
 	  (mode)		   = key('\\em') +
-	  			     button(file, @emacs_mode?modes),
+				     button(file, @emacs_mode?modes),
 	  properties		   = button(file),
 	  -			   = button(file),
 	  quit			   = key('\\C-x\\C-c') + button(file),
@@ -87,8 +88,11 @@
 	  cut			   = button(edit),
 	  paste			   = button(edit),
 	  -			   = button(edit),
-	  find			   = button(edit),
+	  goto_line		   = button(edit) +
+				     key('\\eg\\eg') + key('\\egg'),
+	  find			   = key(key_top_4) + button(edit),
 	  replace		   = button(edit),
+	  preferences		   = button(edit),
 
 					% BROWSER menu
 	  split_window		   = key('\\C-x2') + button(browse),
@@ -98,6 +102,7 @@
 	  show_bookmarks	   = button(browse),
 	  -			   = button(browse),
 	  grep			   = button(browse),
+	  git_grep		   = button(browse),
 
 					% COMPILE menu
 	  compile		   = button(compile),
@@ -107,7 +112,9 @@
 	  help_on_mode		   = button(help),
 	  customise		   = button(help),
 	  -			   = button(help),
-	  show_key_bindings	   = key('\\C-hb') + button(help),
+	  show_key_bindings	   = key(key_top_1)
+					+ key('\\C-hb')
+					+ button(help),
 	  describe_key		   = key('\\C-hk') + button(help),
 	  describe_function	   = key('\\C-hf') + button(help),
 	  -			   = button(help),
@@ -117,20 +124,29 @@
 	[
 	]).
 
-class_variable(grep_command,	string, 'grep -n %s /dev/null',
+class_variable(grep_command, string,
+	       [ 'X'('grep -n %s /dev/null'),
+		 windows('grep -n %s NUL')
+	       ],
 	       "Command of M-x grep").
-class_variable(shell_command,	chain*,
+class_variable(git_grep_command, string,
+	       'git --no-pager grep -n %s',
+	       "Command of M-x git-grep").
+class_variable(shell_command, chain*,
 	       when(@pce?window_system == windows,
 		    @nil,
 		    chain('/bin/sh', '-c')),
 	       "Command for running grep, make, etc.").
 class_variable(auto_colourise_size_limit, int, 50000,
 	       "Auto-colourise if buffer is smaller then this").
-class_variable(print_command,	string,
+class_variable(print_command, string,
 	       [ 'X'('lpr %s'),
 		 windows('NOTEPAD.EXE /P %s')
 	       ],
 	       "Command to print a file").
+class_variable(tab_width, int,
+	       8,
+	       "Distance between tab-stops").
 
 variable(print_command, name, both,
 	 "Command to print a file").
@@ -163,7 +179,7 @@ auto_colourise_buffer(M) :->
 	    get(TB, size, Size),
 	    get(M, auto_colourise_size_limit, Limit),
 	    Size < Limit,
-	    get(M, focus_function, @nil) 	% don't colour in special modes
+	    get(M, focus_function, @nil)	% don't colour in special modes
 	->  (	send(M, colourise_buffer)
 	    ->	send(TB, coloured_generation, Generation)
 	    ;	true
@@ -221,7 +237,11 @@ remove_condition(From, To, Style,
 quit(M) :->
 	"Destroy the editor"::
 	ignore(send(M?text_buffer, save_if_modified)),
-	send(M?frame, destroy).
+	send(M, destroy).
+
+preferences(_M) :->
+	"Edit (xpce) preferences"::
+	prolog_edit_preferences(xpce).
 
 
 		 /*******************************
@@ -408,8 +428,8 @@ save_some_buffers(_M, Arg:[int]) :->
 
 find_file(_M, File:file) :->
 	"Find existing file or create new one"::
-	new(B, emacs_buffer(File)),
-	send(B, open).
+	new(Buffer, emacs_buffer(File)),
+	send(Buffer, open, tab).
 
 new(M, File:save_file) :->
 	"Create a new file"::
@@ -444,7 +464,7 @@ show_buffer_menu(_M) :->
 
 switch_to_buffer(_, Buffer:emacs_buffer) :->
 	"Switch this window to named buffer"::
-	send(Buffer, open).
+	send(Buffer, open, tab).
 %	send(M, text_buffer, Buffer).		% Always in same window
 
 
@@ -483,6 +503,13 @@ newline_mode(M, Mode:mode={posix,dos}) :->
 	get(TB, file, File),
 	send(File, newline_mode, Mode).
 
+tab_width(M, Width:width='2..') :->
+	"Set distance between tab-stops"::
+	get(M, text_buffer, TB),
+	get(M, editor, E),
+	send(TB, tab_width, Width),
+	send(E, tab_distance, Width).
+
 
 		 /*******************************
 		 *	       PRINT		*
@@ -583,7 +610,7 @@ replace_end(M) :->
 '_query_replace_regex'(M, Id:event_id) :->
 	"Focus function for query_replace_regex"::
 	(   Id == 0'y				% replace-and-continue
- 	->  replace_match(M),
+	->  replace_match(M),
 	    ignore(replace_find_and_mark(M))
 	;   Id == 0'.				% replace and done
 	->  replace_match(M),
@@ -684,7 +711,7 @@ split_command_string(String, List) :-
 	     message(Ch, append, ?(@arg1, register_value, @arg2, 0, name))),
 	chain_list(Ch, List).
 
-compile(M, Command:shell_command=string, Label:[name], Pool:[name]) :->
+compile(M, Command:shell_command=string, Label:[name]) :->
 	"Run Unix (compilation) process in buffer"::
 	send(M, save_some_buffers),
 	send(M, has_processes),
@@ -715,15 +742,13 @@ compile(M, Command:shell_command=string, Label:[name], Pool:[name]) :->
 	    Process =.. [process|Args]
 	),
 	new(P, Process),
-	default(Pool, compile, ThePool),
-	send(B, pool, ThePool),
 	send(P, directory, Dir),
 	send(B, clear),
 	send(B, format, 'cd %s\n%s\n', Dir?path, Command),
 	send(B, directory, Dir),
 	send(B, process, P),
 	send(B, start_process),
-	send(B, open),
+	send(B, open, tab),
 	send(B?editors, for_some,
 	     and(message(@arg1, report, status, 'Running ...'),
 		 message(@arg1, scroll_to, 0),
@@ -731,12 +756,19 @@ compile(M, Command:shell_command=string, Label:[name], Pool:[name]) :->
 
 
 grep(M, GrepArgs:grep_arguments=string) :->
-	"Run Unix grep in compilation buffer"::
+	"Run the OS grep command in compilation buffer"::
 	get(M, grep_command, GrepCommad),
 	send(M, compile,
 	     string(GrepCommad, GrepArgs),
-	     string('grep %s', GrepArgs),
-	     grep).
+	     string('grep %s', GrepArgs)).
+
+
+git_grep(M, GrepArgs:grep_arguments=string) :->
+	"Run the git grep in compilation buffer"::
+	get(M, git_grep_command, GrepCommad),
+	send(M, compile,
+	     string(GrepCommad, GrepArgs),
+	     string('git-grep %s', GrepArgs)).
 
 
 shell(M) :->
@@ -755,10 +787,9 @@ shell(M) :->
 	    get(M, directory, Dir),
 	    send(P, directory, Dir),
 	    new(B, emacs_process_buffer(P, '*shell*')),
-	    send(B, pool, shell),
 	    send(B, directory, Dir),
 	    send(B, start_process),
-	    send(B, open)
+	    send(B, open, tab)
 	).
 
 better_shell('/bin/tcsh', '/bin/csh') :- !.
@@ -768,7 +799,6 @@ manual_entry(M, Spec:unix_manual_entry_for=name) :->
 	"Lookup Unix manual entry"::
 	send(M, has_processes),
 	new(B, emacs_buffer(@nil, Spec)),
-	send(B, pool, manual_entry),
 	send(B, (mode), man),
 	send(B, open),
 	send(B?editors?head, man, Spec).
@@ -810,7 +840,6 @@ annotate(M) :->
 	    send(B, contents, OB?contents)
 	),
 	send(B, (mode), annotate),
-	send(B, pool, wysiwyg),
 	send(B, open).
 
 
@@ -821,9 +850,10 @@ annotate(M) :->
 split_window(M) :->
 	"Create another window for this buffer"::
 	get(M, text_buffer, Buffer),
-	new(W2, emacs_frame(Buffer)),
-	send(W2?editor, caret, M?caret).
-
+	get(Buffer, open, window, Frame),
+	get(Frame, editor, NewEditor),
+	get(M, caret, Here),
+	send(NewEditor, caret, Here).
 
 only_window(M) :->
 	"Quit other windows on this buffer"::
@@ -875,9 +905,17 @@ what_line(M) :->
 	send(M, report, inform, 'line %d', CaretLine).
 
 
-goto_line(M, LineNo:int) :->
+goto_line(M, LineNo:int, Title:title=[char_array]*) :->
 	"Goto given line-number"::
-	send(M, point_to_top_of_file, LineNo).
+	(   (   Title == @nil
+	    ;	get(M, line_number, M?caret, CaretLine),
+		abs(LineNo-CaretLine) < 10
+	    )
+	->  send(M, point_to_top_of_file, LineNo)
+	;   send(@emacs, location_history),
+	    send(M, point_to_top_of_file, LineNo),
+	    send(@emacs, location_history, Title)
+	).
 
 
 count_lines_region(M) :->
