@@ -19,7 +19,7 @@
 
     You should have received a copy of the GNU Lesser General Public
     License along with this library; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 
     As a special exception, if you link this library with other files,
     compiled with a Free Software compiler, to produce an executable, this
@@ -112,7 +112,7 @@ update_label(B) :->
 	    send(DictItem, label, string('%s\t%s', Name, Indicator)),
 	    new(FrameLabel, string('%s [%s]', Name, Status)),
 	    send(B?editors, for_all,
-		 message(@arg1?frame, label, FrameLabel))
+		 message(@arg1, label, FrameLabel))
 	;   true
 	).
 
@@ -186,14 +186,14 @@ send_input(B, Caret:int) :->
 	(   Caret >= End
 	->  get(B, contents, End, Caret-End, Data),
 	    send(B, insert, Caret, @newline)
- 	;   get(B, scan, Caret, line, 0, start, SOL),
+	;   get(B, scan, Caret, line, 0, start, SOL),
 	    get(B, scan, Caret, line, 0, end, EOL),
 	    get(B, contents, SOL, EOL-SOL, Data),
 	    (   get(B, prompt_regex, Prompt), Prompt \== @nil,
 		get(Prompt, match, Data, Size)
 	    ->  send(Data, delete, 0, Size)
- 	    ;   true
- 	    ),
+	    ;   true
+	    ),
 	    send(B, append, Data),
 	    send(B, append, @newline),
 	    send(B?editors, for_all, message(@arg1, caret))
@@ -294,24 +294,12 @@ start_process(B) :->
 	).
 
 
-open(B, New:[bool]) :->
+open(B, Where:[{here,tab,window}]) :->
 	"Create window for buffer"::
-	(   New == @on
-	->  send(new(Window, emacs_frame(B)), open)
-	;   (	\+ send(B?editors, empty)
-	    ->	send(B?editors?head?frame, expose)
-	    ;	get(@emacs, free_window, B?pool, Window)
-	    ->	send(Window, buffer, B)
-	    ;	send(emacs_frame(B), open)
-	    )
-	),
-	(   nonvar(Window)
-	->  send(Window?editor, caret, @default)
-	;   true
-	),
+	get(B, open, Where, Frame),
+	send(Frame?editor, caret, @default),
 	send(B, update_label).
 
-
 :- pce_end_class.
 
 
@@ -494,7 +482,7 @@ end_of_file(E) :->
 	      new(chain(regex('(\\S+):\\s*(\\d+):'),        % gcc, grep
 			regex('"(\\S+)", line (\\d+):')))). % SUN cc
 :- pce_global(@emacs_cd_regexs,
-	      new(chain(regex('\\ycd\\s+((\\w|[_/+-.:])+)'),
+	      new(chain(regex('\\ycd\\s+((\\w|[_/+-.: ])+)'),
 			regex('Entering directory `([^\']+)\'')))).
 :- pce_global(@emacs_canonise_dir_regex,
 	      new(regex('[^/]+/\\.\\./'))).
@@ -613,10 +601,10 @@ goto_error(M) :->
 		and(message(@arg1, instance_of, emacs_link_fragment),
 		    message(@arg1, overlap, Caret)),
 		Fragment)
-	->  send(Fragment, follow)
+	->  send(Fragment, follow, tab)
 	;   get(M, error, Caret, tuple(File, Line))
 	->  new(B, emacs_buffer(File)),
-	    send(B, open),
+	    send(B, open, tab),
 	    send(B?editors?head, line_number, Line)
 	;   send(M, report, warning, 'No error on this line')
 	).
@@ -662,11 +650,11 @@ link(F, F2:emacs_link_fragment) :->
 	send(F2, slot, link, F).
 
 
-follow(F) :->
+follow(F, Where:[{here,tab,window}]) :->
 	"Goto other end of link"::
 	get(F, link, F2),
 	get(F2, text_buffer, B),
-	send(B, open),
+	send(B, open, Where),
 	send(B?editors?head, caret, F2?start).
 
 :- pce_end_class.
