@@ -3,9 +3,10 @@
     Part of XPCE --- The SWI-Prolog GUI toolkit
 
     Author:        Jan Wielemaker and Anjo Anjewierden
-    E-mail:        wielemak@science.uva.nl
+    E-mail:        J.Wielemaker@cs.vu.nl
     WWW:           http://www.swi-prolog/projects/xpce/
-    Copyright (C): 1985-2006, University of Amsterdam
+    Copyright (C): 1985-2011, University of Amsterdam
+			      VU University Amsterdam
 
     This program is free software; you can redistribute it and/or
     modify it under the terms of the GNU General Public License
@@ -19,7 +20,7 @@
 
     You should have received a copy of the GNU Lesser General Public
     License along with this library; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 
     As a special exception, if you link this library with other files,
     compiled with a Free Software compiler, to produce an executable, this
@@ -39,7 +40,8 @@
 	    get_tracer/2,		% :Goal, -Result
 	    in_debug_thread/2,		% +ObjOrThread, :Goal
 	    thread_debug_queue/2,	% +Thread, -Queue
-	    prolog_frame_attribute/4	% +GUI, +Frame, +Attr, -Value
+	    prolog_frame_attribute/4,	% +GUI, +Frame, +Attr, -Value
+	    prolog_choice_attribute/4	% +GUI, +Choice, +Attr, -Value
 	  ]).
 :- use_module(library(pce)).
 :- use_module(library(lists)).
@@ -62,6 +64,10 @@
 	user:prolog_event_hook/1,
 	user:message_hook/3.
 
+:- meta_predicate
+	in_debug_thread(+, 0),
+	send_pce(0).
+
 register_directories :-
 	(   member(SpyBase, ['icons/nospy', library('trace/icons/nospy')]),
 	    absolute_file_name(SpyBase,
@@ -114,15 +120,10 @@ prolog_tracer(Thread, Ref, Create) :-
 %%	break_level(-Level)
 %
 %	Current break-level.
-%
-%	@bug	Breaks only work in the main thread. We need to change I/O
-%		handling and where the break-level is stored to fix
-%		this.
 
 break_level(Level) :-
-	thread_self(main), !,
-	flag('$break_level', Level, Level).
-break_level(1).
+	current_prolog_flag(break_level, Level), !.
+break_level(-1).				% non-interactive thread.
 
 
 %%	send_tracer(+ThreadOrGUI, +Term) is semidet.
@@ -169,10 +170,6 @@ get_tracer(Thread, Term, Result) :-
 		 *    THREAD SYNCHRONISATION	*
 		 *******************************/
 
-:- meta_predicate
-	in_debug_thread(+, :),
-	send_pce(:).
-
 %%	thread_debug_queue(+Thread, -Queue) is det.
 %
 %	Queue is the debugging queue for Thread.  We do not use the main
@@ -193,19 +190,18 @@ thread_debug_queue(Thread, Queue) :-
 	user:prolog_event_hook/1.
 
 user:prolog_event_hook(thread_finished(TID)) :-
-	thread_debug_queue_store(TID, Queue),
-	message_queue_destroy(Queue),
+	forall(retract(thread_debug_queue_store(TID, Queue)),
+	       message_queue_destroy(Queue)),
 	fail.				% allow other hooks
 
 %%	send_pce(:Goal)
 %
-%	Run Goal in XPCE (main)  thread.   Wait  for  completion. In the
-%	meanwhile, allow the XPCE thread to call in_debug_thread/1.
-%
-%	@bug	XPCE thread does not _need_ to be =main=.
+%	Run Goal in XPCE thread. Wait  for completion. In the meanwhile,
+%	allow the XPCE thread to call in_debug_thread/1.
 
 send_pce(Goal) :-
-	thread_self(main), !,
+	thread_self(Me),
+	pce_thread(Me), !,
 	Goal.
 send_pce(Goal) :-
 	thread_self(Self),
@@ -280,6 +276,11 @@ prolog_frame_attribute(Thread, Frame, Attribute, Value) :-
 	in_debug_thread(Thread,
 			prolog_frame_attribute(Frame, Attribute, Value)).
 
+prolog_choice_attribute(GUI, Choice, Attribute, Value) :-
+	in_debug_thread(GUI,
+			prolog_choice_attribute(Choice, Attribute, Value)).
+
+
 
 		 /*******************************
 		 *     DEBUGGER APPLICATION	*
@@ -307,14 +308,17 @@ initialise(App) :->
 variable(source,	any,		both, "Source view").
 variable(break_level,	int,		get,  "Break-level I'm associated to").
 variable(thread,	'int|name*',	get,  "Associated thread").
-variable(current_frame, int*,   	both, "The most recent frame").
-variable(current_break,	tuple*,		both, "tuple(ClauseRef, PC)").
+variable(trap_frame,    int*,		get,  "Last trapped frame").
+variable(trap_port,     name*,		get,  "Last trapped port").
+variable(current_frame, int*,		both, "The most recent frame").
 variable(quitted,	bool := @off,   both, "Asked to quit").
-variable(mode,		name := created,both, "Current mode").
+variable(mode,		name := created,get,  "Current mode").
 
-initialise(F, Level:int, Thread0:'int|name') :->
-	assertion(thread_self(main)),
-	default(Thread0, main, Thread),
+running_in_pce_thread :-
+	pce_thread(Pce), thread_self(Pce).
+
+initialise(F, Level:int, Thread:'int|name') :->
+	assertion(running_in_pce_thread),
 	send(F, slot, break_level, Level),
 	send(F, slot, thread, Thread),
 	send_super(F, initialise, 'SWI-Prolog debugger', application := @prolog_gui),
@@ -411,7 +415,10 @@ fill_menu_bar(F) :->
 			      message(F, exceptions),
 			      end_group := @on),
 		    menu_item(toggle_edit_mode,
-			      message(F, edit))
+			      message(F, edit),
+			      end_group := @on),
+		    menu_item(copy_goal,
+			      message(F, copy_goal))
 		  ]),
 	send_list(View, append,
 		  [ menu_item(threads,
@@ -451,9 +458,18 @@ help(_) :->
         "Show window with help-text"::
         send(@helper, give_help, pltracer, main).
 
-show_frame(GUI, Frame:int, PC:'int|name') :->
+show_frame(GUI, Frame:int, PC:prolog) :->
 	"Show the variables of this frame"::
-	prolog_show_frame(Frame, [gui(GUI), pc(PC), source, bindings]).
+	(   get(GUI, trap_frame, Frame)		% the initial trapped port
+	->  get(GUI, trap_port, Style)
+	;   PC = choice(_)			% A choice-point
+	->  Style = choice
+	;   Style = frame			% Somewhere up the stack
+	),
+	prolog_show_frame(Frame, [ gui(GUI), pc(PC),
+				   source, bindings,
+				   style(Style)
+				 ]).
 
 show_stack(GUI, CallFrames:prolog, ChoiceFrames:prolog) :->
 	"Show the stack and choicepoints"::
@@ -465,6 +481,12 @@ show_threads(_GUI) :->
 	"Open Thread monitor"::
 	prolog_ide(thread_monitor).
 
+trapped_location(GUI, StartFrame:int, Frame:int, Port:name) :->
+	"The last trapped location"::
+	send(GUI, slot, trap_frame, Frame),
+	send(GUI, slot, trap_port, Port),
+	send(GUI, current_frame, StartFrame).
+
 
 		 /*******************************
 		 *	      EVENT		*
@@ -509,13 +531,14 @@ return(Frame, Result:any) :->
 	(   get(Frame, mode, wait_user)
 	->  get(Frame, thread, Thread),
 	    send(Frame, mode, replied),
-	    (	Thread == main
+	    (	pce_thread(Thread)
 	    ->	send_super(Frame, return, Result)
 	    ;   (   get(Frame, quitted, @on)
 		->  send(Frame, destroy)
 		;   true
 		),
-		debug(' ---> frame: result = ~p~n', [Result]),
+		debug(' ---> frame for thread = ~p: result = ~p~n',
+		      [Thread, Result]),
 		thread_debug_queue(Thread, Queue),
 		thread_send_message(Queue, '$trace'(action(Result)))
 	    )
@@ -608,8 +631,6 @@ nostop_or_spy(F) :->
 	"Clear spy-point"::
 	(   send(F?source, delete_selected_stop)
 	->  true
-	;   get(F, current_break, tuple(ClauseRef, PC))
-	->  '$break_at'(ClauseRef, PC, false)
 	;   (   get(F, current_frame, Frame)
 	    ;   get(F, selected_frame, Frame)
 	    ),
@@ -649,10 +670,50 @@ details(F) :->
 	get(F, member, bindings, Bindings),
 	send(Bindings, details).
 
+nodebug(F) :->
+	"User hit n(odebug)"::
+	(   setting(auto_close, true)
+	->  send(F, quitted, @on)
+	;   true
+	),
+	send(F, return, nodebug).
+
+abort(F) :->
+	"User hit a(bort)"::
+	(   setting(auto_close, true)
+	->  send(F, quitted, @on)
+	;   true
+	),
+	send(F, return, abort).
+
+interrupt(F) :->
+	"User hit t (interrupt, trace)"::
+	get(F, thread, Thread),
+	thread_signal(Thread, trace).
+
 query(_F) :->
 	"Enter and run a query"::
 	prolog_ide(open_query_window).
 
+copy_goal(F) :->
+	"Copy the current goal into the copy-buffer"::
+	get(F, selected_frame, Frame),
+	(   Frame \== @nil
+	->  true
+	;   send(F, report, warning, 'No current frame'),
+	    fail
+	),
+	prolog_frame_attribute(F, Frame, goal, Goal),
+	prolog_frame_attribute(F, Frame, predicate_indicator, PI),
+	(   numbervars(Goal, 0, _, [attvar(skip)]),
+	    format(string(Text), '~q', [Goal]),
+	    send(@display, copy, Text),
+	    fail
+	;   true
+	),
+	format(atom(PIA), '~q', [PI]),
+	send(F, report, inform, 'Copied goal (%s) to clipboard', PIA).
+
 :- pce_group(delegate).
 
 file(F, File:'name|emacs_buffer*') :->
@@ -711,6 +772,17 @@ query_finished(F, Message:char_array) :->
 	send(F, mode, query_finished),
 	send(F, report, status, Message).
 
+mode(F, Mode:name) :->
+	"Switch modes"::
+	send(F, slot, mode, Mode),
+	(   get(F, member, buttons, D)
+	->  (   Mode == wait_user
+	    ->  send(D, running, @off)
+	    ;   send(D, running, @on)
+	    )
+	;   true
+	).
+
 :- pce_end_class(prolog_debugger).
 
 		 /*******************************
@@ -732,9 +804,10 @@ button(finish,	       "f",   'finish.xpm',	     'Finish selected goal').
 button(gap,	       -,     -,		     -).
 button(retry,	       "r",   'retry.xpm',	     'Retry selected goal').
 button(gap,	       -,     -,		     -).
-button(nodebug,	       "n",   'nodebug.xpm',	     'Continue without debugging').
+button(+nodebug,       "n",   'nodebug.xpm',	     'Continue without debugging').
+button(+abort,	       "a",   'abort.xpm',	     'Abort to the Prolog toplevel').
+button(+interrupt,     "t",   'interrupt.xpm',	     'Interrupt (trace)').
 button(+query,	       "b",   'break.xpm',	     'Enter a query').
-button(abort,	       "a",   'abort.xpm',	     'Abort to the Prolog toplevel').
 button(fail,	       "F",   'fail.xpm',	     'Force query to fail').
 button(gap,	       -,     -,		     -).
 button(+up,	       "u",   'up.xpm',		     'Select parent frame').
@@ -771,11 +844,12 @@ initialise(D) :->
 	    (	Action == gap
 	    ->	send(TB, append, gap)
 	    ;   tag_balloon(Balloon0, Keys, Balloon),
-		make_message(Action, D, Message),
+		make_message(Action, Name, D, Message),
 	        send(TB, append,
 		     new(B, tool_button(Message,
 					image(Image),
-					Balloon))),
+					Balloon,
+				        name := Name))),
 	        chain_list(KL, Keys),
 	        send(B, attribute, keys, KL)
 	    ),
@@ -783,8 +857,8 @@ initialise(D) :->
 	;   true
 	).
 
-make_message(+Action, D, message(D?frame, Action)) :- !.
-make_message(Action,  D, message(D, return, Action)).
+make_message(+Action, Action, D, message(D?frame, Action)) :- !.
+make_message(Action,  Action, D, message(D, return, Action)).
 
 typed(D, Id:event_id, Delegate:[bool]) :->
 	"Handle typing"::
@@ -808,6 +882,27 @@ button(D, Name:name, Button:button) :<-
 	get(D, member, tool_bar, TB),
 	get(TB, member, Name, Button).
 
+running(D, Running:bool) :->
+	"Make some buttons (in)active"::
+	get(Running, negate, NotRunning),
+	forall(running_button(Name),
+	       (   get(D, button, Name, Button)
+	       ->  send(Button, active, NotRunning)
+	       ;   format('No button ~w~n', [Name])
+	       )),
+	(   get(D, button, interrupt, Interrupt)
+	->  send(Interrupt, active, Running)
+	;   true
+	).
+
+running_button(into).
+running_button(creep).
+running_button(skip).
+running_button(retry).
+running_button(finish).
+running_button(nodebug).
+running_button(abort).
+
 :- pce_end_class(prolog_button_dialog).
 
 
@@ -826,13 +921,15 @@ it.
 :- pce_begin_class(prolog_bindings_view, view,
 		   "Overview of bindings of the current frame").
 
-class_variable(font,	font, 	normal, "Font for bindings").
+class_variable(font,	font,	normal, "Font for bindings").
 class_variable(size,    size,   size(40,11), "Initial size").
 
 variable(prolog_frame, int*, both, "Frame who's variables we are showing").
 
 :- pce_global(@prolog_binding_recogniser,
 	      make_prolog_binding_recogniser).
+:- pce_global(@prolog_binding_popup,
+	      make_prolog_binding_popup).
 
 make_prolog_binding_recogniser(G) :-
 	new(View, @event?window),
@@ -841,8 +938,18 @@ make_prolog_binding_recogniser(G) :-
 			      message(View, on_click, Index))),
 	new(C2, click_gesture(left, '', double,
 			      message(View, details))),
-	new(G, handler_group(C1, C2)).
+	new(C3, popup_gesture(@prolog_binding_popup)),
+	send(@prolog_binding_popup, update_message,
+	     message(View, on_click, Index)),
+	new(G, handler_group(C1, C2, C3)).
 
+make_prolog_binding_popup(P) :-
+	new(P, popup),
+	send_list(P, append,
+		  [ menu_item(details, message(@arg1?window, details)),
+		    menu_item(copy,    message(@arg1?window, details,
+					       @default, copy))
+		  ]).
 
 initialise(B) :->
 	send_super(B, initialise),
@@ -854,6 +961,7 @@ initialise(B) :->
 	send(B?image, tab_stops, vector(Tab)),
 	send(B?image, recogniser, @prolog_binding_recogniser),
 	send(B, editable, @off),
+	send(B, style, constraint, style(colour := blue)),
 	send(B?text_cursor, displayed, @off),
 	send(B, ver_stretch, 0).
 
@@ -864,7 +972,7 @@ clear(B, Content:[bool]) :->
 	;   send_super(B, clear)
 	).
 
-details(B, Fragment:[prolog_frame_var_fragment]) :->
+details(B, Fragment:[prolog_frame_var_fragment], Action:[{view,copy}]) :->
 	"View details of the binding"::
 	get(B, prolog_frame, Frame),
 	(   Frame \== @nil
@@ -881,22 +989,33 @@ details(B, Fragment:[prolog_frame_var_fragment]) :->
 	    )
 	;   Frag = Fragment
 	),
-	get(Frag, var_name, VarName),
-	get(Frag, value, Value),
-	prolog_frame_attribute(B, Frame, level, Level),
-	prolog_frame_attribute(B, Frame, goal, Goal),
-	predicate_name(Goal, PredName),
-	(   integer(VarName)
-	->  VarType = 'Argument'
-	;   VarType = 'Variable'
-	),
-	format(string(Label), '~w ~w of frame at level ~d running ~w',
-		[ VarType, VarName, Level, PredName ]),
-	view_term(Value,
-		  [ comment(Label),
-		    source_object(Frag),
-		    expose(true)
-		  ]).
+	(   get(Frag, var_name, VarName)
+	->  get(Frag, value, Value),
+	    prolog_frame_attribute(B, Frame, level, Level),
+	    prolog_frame_attribute(B, Frame, goal, Goal),
+	    predicate_name(Goal, PredName),
+	    (   integer(VarName)
+	    ->  VarType = 'Argument'
+	    ;   VarType = 'Variable'
+	    ),
+	    format(string(Label), '~w ~w of frame at level ~d running ~w',
+		   [ VarType, VarName, Level, PredName ]),
+	    debug('Action ~w on ~w~n', [Action, Value]),
+	    (   Action == copy
+	    ->  (   numbervars(Value, 0, _, [attvar(skip)]),
+		    format(string(Text), '~q', [Value]),
+		    send(@display, copy, Text),
+		    fail
+		;   send(B, report, status, Label)
+		)
+	    ;   view_term(Value,
+			  [ comment(Label),
+			    source_object(Frag),
+			    expose(true)
+			  ])
+	    )
+	;   send(B, report, warning, 'Not a variable value')
+	).
 
 on_click(B, Index:int) :->
 	"Select fragment clicked"::
@@ -907,32 +1026,94 @@ on_click(B, Index:int) :->
 	;   send(B, selected_fragment, @nil)
 	).
 
+% Bindings is a list of Vars = Value,   where Vars is a list of variable
+% identifiers that take the form Name:ArgN,   were  Name is the variable
+% name (atom) and ArgN is the location in the frame.
+
 bindings(B, Bindings:prolog) :->
 	"Display complete list of bindings"::
+	(   term_attvars(Bindings, [])
+	->  Plain = Bindings,
+	    Constraints = []
+	;   copy_term(Bindings, Plain, Constraints)
+	),
+	bind_vars(Plain),
+	cycles(Plain, Template, Cycles, Plain),
 	send(B, background, white),
 	pce_open(B, write, Fd),
-	forall(member(Vars=Value, Bindings),
-	       send(B, append_binding, Vars, value(Value), Fd)),
+	(   forall(member(Vars=Value, Template),
+		   send(B, append_binding, Vars, value(Value), Fd)),
+	    forall(member(C, Constraints),
+		   send(B, append_extra, C, Fd, constraint)),
+	    forall(member(C, Cycles),
+		   send(B, append_extra, C, Fd, cycle)),
+	    fail
+	;   true
+	),
 	close(Fd),
 	send(B, caret, 0).
 
-append_binding(B, Names:prolog, ValueTerm:prolog, Fd:prolog) :->
+bind_vars([]).
+bind_vars([Vars=Value|T]) :-
+	(   var(Value)
+	->  Vars = [Name:_|_],
+	    Value = '$VAR'(Name)
+	;   true
+	),
+	bind_vars(T).
+
+cycles(Term, Template, Cycles, _) :-
+	acyclic_term(Term), !,
+	Template = Term,
+	Cycles = [].
+cycles(Term, Template, Cycles, Bindings) :-
+	'$factorize_term'(Term, Template, Factors),
+	bind_non_cycles(Factors, Cycles),
+	name_cycle_vars(Cycles, 1, Bindings).
+
+bind_non_cycles([], []).
+bind_non_cycles([V=Term|T], L) :-
+	unify_with_occurs_check(V, Term), !,
+	bind_non_cycles(T, L).
+bind_non_cycles([H|T0], [H|T]) :-
+	bind_non_cycles(T0, T).
+
+
+name_cycle_vars([], _, _).
+name_cycle_vars([H|T], I, Bindings) :-
+	H = (Var=_Value),
+	(   member(Vars=VarsValue, Bindings),
+	    VarsValue == Var,
+	    Vars = [Name:_|_]
+	->  I2 = I
+	;   atom_concat('_S', I, Name),
+	    I2 is I + 1
+	),
+	Var = '$VAR'(Name),
+	name_cycle_vars(T, I2, Bindings).
+
+
+append_binding(B, Names0:prolog, ValueTerm:prolog, Fd:prolog) :->
 	"Add a binding to the browser"::
-	ValueTerm = value(Value),	% protect :=, ?, etc.
-	(   var(Value),
+	ValueTerm = value(Value0),	% protect :=, ?, etc.
+	(   Value0 = '$VAR'(_), Names0 = [_],
 	    setting(show_unbound, false)
 	->  true
-	;   get(B, text_buffer, TB),
+	;   (   Value0 = '$VAR'(_), Names0 = [_,_|_]
+	    ->	append(Names, [VarN:_], Names0),
+		Value = '$VAR'(VarN)
+	    ;	Names = Names0,
+		Value = Value0
+	    ),
+	    get(B, text_buffer, TB),
 	    get(TB, size, S0),
 	    (   Names = VarName:ArgN
 	    ->	format(Fd, '~w', [VarName])
 	    ;	Names = [VarName:ArgN|_],
 	        write_varnames(Fd, Names)
 	    ),
-	    setting(attributes, Atts),
-	    current_prolog_flag(toplevel_print_options, Options0),
-	    delete(Options0, attributes(_), Options1),
-	    format(Fd, '\t= ~W~n', [Value, [attributes(Atts)|Options1]]),
+	    current_prolog_flag(toplevel_print_options, Options),
+	    format(Fd, '\t= ~W~n', [Value, Options]),
 	    flush_output(Fd),
 	    get(TB, size, S1),
 	    new(_, prolog_frame_var_fragment(TB, S0, S1, VarName, ArgN))
@@ -944,6 +1125,16 @@ write_varnames(Fd, [N:_|T]) :-
 	format(Fd, '~w = ', N),
 	write_varnames(Fd, T).
 
+append_extra(B, Constraint:prolog, Fd:prolog, Comment:name) :->
+	"Display current constraints"::
+	get(B, text_buffer, TB),
+	current_prolog_flag(toplevel_print_options, Options),
+	get(TB, size, S0),
+	format(Fd, '(~w)\t~W~n', [Comment, Constraint, Options]),
+	flush_output(Fd),
+	get(TB, size, S1),
+	new(_, prolog_frame_constraint_fragment(TB, S0, S1)).
+
 :- pce_end_class(prolog_bindings_view).
 
 
@@ -973,6 +1164,21 @@ value(F, Value:prolog) :<-
 :- pce_end_class(prolog_frame_var_fragment).
 
 
+:- pce_begin_class(prolog_frame_constraint_fragment, fragment,
+		   "Represent a contraint on a frame").
+
+initialise(F, TB:text_buffer, From:int, To:int) :->
+	Len is To-From,
+	send_super(F, initialise, TB, From, Len, constraint).
+
+var_name(_F, _Name:name) :<-
+	"Cannot show details"::
+	fail.
+
+:- pce_end_class(prolog_frame_constraint_fragment).
+
+
+
 		 /*******************************
 		 *	       EVENTS		*
 		 *******************************/
