@@ -19,7 +19,7 @@
 
     You should have received a copy of the GNU Lesser General Public
     License along with this library; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */
 
 #include <h/kernel.h>
@@ -65,7 +65,7 @@ unalloc(size_t n, Any p)
 }
 
 void
-initAlloc(void)
+pceInitAlloc(void)
 { wastedbytes = allocbytes = 0;
   allocTop  = 0L;
   allocBase = 0xffffffff;
@@ -87,7 +87,7 @@ allocRange(void *low, size_t size)
 }
 
 status
-listWastedCorePce(Pce pce, Bool ppcells)
+listWastedCorePce(Pce pce, BoolObj ppcells)
 { succeed;
 }
 
@@ -124,7 +124,7 @@ about  the same  relative numbers.    For  this reason  PCE addopts  a
 perfect fit strategy for memory allocation.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
-#define offset(structure, field) ((int) &(((structure *)NULL)->field))
+#define offset(structure, field) ((intptr_t) &(((structure *)NULL)->field))
 
 static inline Zone
 allocate(size_t size)
@@ -315,7 +315,7 @@ unalloc(size_t n, Any p)
 
 
 void
-initAlloc(void)
+pceInitAlloc(void)
 { int t;
 
   spaceptr  = NULL;
@@ -364,7 +364,7 @@ checkFreeChains()
 
 
 status
-listWastedCorePce(Pce pce, Bool ppcells)
+listWastedCorePce(Pce pce, BoolObj ppcells)
 { unsigned int n;
   Zone z;
   int total = 0;
