@@ -1,4 +1,4 @@
-/*  $Id: conversion.c,v 1.22 2007/01/14 20:16:01 jan Exp $
+/*  $Id$
 
     Part of XPCE --- The SWI-Prolog GUI toolkit
 
@@ -19,7 +19,7 @@
 
     You should have received a copy of the GNU Lesser General Public
     License along with this library; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */
 
 #include <h/kernel.h>
@@ -101,12 +101,12 @@ Real
 toReal(Any obj)
 { if ( instanceOfObject(obj, ClassReal) )
     return obj;
-  
+
   return getConvertReal(ClassReal, obj);
 }
 
 
-Bool
+BoolObj
 toBool(Any obj)
 { Int i;
   string s;
@@ -120,7 +120,7 @@ toBool(Any obj)
     else if ( i == ONE )
       return ON;
   }
-  
+
   if ( toString(obj, &s) && isstrA(&s) )
   { if       ( streq_ignore_case((char *)s.s_textA, "@on") ||
 	       streq_ignore_case((char *)s.s_textA, "true") ||
@@ -304,6 +304,9 @@ and without SIGBUS ...
 
 #ifdef __WINDOWS__
 
+#ifdef __MINGW32__
+#define __SEH_NOOP /* TBD: handle being passed wrong data on MinGW! */
+#endif
 #include <excpt.h>
 
 char *
@@ -434,7 +437,7 @@ expandFunction(Any obj)
 { while ( isFunction(obj) )
   { Function f = (Function) obj;
     Any rval = getExecuteFunction(f);
-    
+
     if ( rval == FAIL )
     { DEBUG(NAME_obtain, Cprintf("Function: %s\n", pp(f)));
       fail;
