@@ -19,7 +19,7 @@
 
     You should have received a copy of the GNU Lesser General Public
     License along with this library; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */
 
 #define UNICODE 1
@@ -43,6 +43,7 @@
 		 *******************************/
 
 HINSTANCE ThePceHInstance;		/* Global handle */
+DWORD	  ThePceThread;			/* Dispatching thread */
 
 BOOL WINAPI
 DllMain(HINSTANCE instance, DWORD reason, LPVOID reserved)
@@ -71,6 +72,18 @@ pceMTdetach()
 }
 
 
+unsigned					/* interface.h cannot depend */
+setPceThread(unsigned id)			/* on DWORD due to conflicts */
+{ DWORD old = ThePceThread;
+
+  assert(sizeof(unsigned) == sizeof(DWORD));
+
+  ThePceThread = id;
+
+  return old;
+}
+
+
 		 /*******************************
 		 *	      VERSIONS		*
 		 *******************************/
@@ -235,7 +248,7 @@ ws_user()
 }
 
 
-#include <Shlobj.h>
+#include <shlobj.h>
 
 Name
 ws_appdata(const char *sub)
@@ -613,7 +626,7 @@ getWinDirectoryDisplay(DisplayObj d,
     bi.lpszTitle = L"Choose folder";
   else
     bi.lpszTitle = nameToTCHAR(title);
-  bi.ulFlags = BIF_RETURNONLYFSDIRS;
+  bi.ulFlags = (BIF_RETURNONLYFSDIRS|BIF_USENEWUI);
   if ( notDefault(dir) )
   { wchar_t windir[MAXPATHLEN];
 
