@@ -19,7 +19,7 @@
 
     You should have received a copy of the GNU Lesser General Public
     License along with this library; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */
 
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
@@ -37,17 +37,6 @@ available test sets. The public goals are:
 
 % Load XPCE from the development environment
 
-:- initialization op(200, fy,  user:(@)).
-:- initialization op(250, yfx, user:(?)).
-:- initialization op(990, xfx, user:(:=)).
-
-user:file_search_path(foreign, '.').
-user:file_search_path(pce, '..').
-user:file_search_path(library, '../prolog/lib').
-:- use_module(library(pce)).
-:- use_module(library(pce_util)).
-
-
 :- get(@pce, version, V),
    format('XPCE/SWI-Prolog test suite. (XPCE version ~w)~n\
           To run all tests run ?- test.~n~n', [V]).
