@@ -1,4 +1,4 @@
-/*  $Id: regex.c,v 1.27 2007/01/23 19:22:17 jan Exp $
+/*  $Id$
 
     Part of XPCE --- The SWI-Prolog GUI toolkit
 
@@ -19,7 +19,7 @@
 
     You should have received a copy of the GNU Lesser General Public
     License along with this library; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */
 
 #include <h/kernel.h>
@@ -31,7 +31,7 @@ forwards Int getRegisterStartRegex(Regex, Int);
 NewClass(regex)
   CharArray		    pattern;	/* Pattern matched */
   Name			    syntax;	/* basic, extended, advanced */
-  Bool			    ignore_case;/* @on --> case insensitive */
+  BoolObj			    ignore_case;/* @on --> case insensitive */
   Int			    re_flags;	/* REG_* FLAGS */
   regex_t		   *compiled;	/* compiled regex */
   regmatch_t		   *registers;	/* \0-\9 matches */
@@ -40,7 +40,7 @@ End;
 
 static status
 initialiseRegex(Regex re, CharArray pattern,
-		Bool case_sensitive,
+		BoolObj case_sensitive,
 		Name syntax)
 { if ( isDefault(pattern) )
     pattern = (CharArray)NAME_;
@@ -136,7 +136,7 @@ patternRegex(Regex re, StringObj pattern)
 
 
 status
-ignoreCaseRegex(Regex re, Bool val)
+ignoreCaseRegex(Regex re, BoolObj val)
 { if ( re->ignore_case != val )
   { assign(re, ignore_case, val);
 
@@ -162,7 +162,7 @@ syntaxRegex(Regex re, Name syntax)
 
 
 status
-compileRegex(Regex re, Bool optimize)
+compileRegex(Regex re, BoolObj optimize)
 { succeed;				/* backward compatibility */
 }
 
@@ -270,7 +270,7 @@ search_string_regex(Regex re, String s)
 
   if ( !ensure_compiled_regex(re, RE_SEARCH) )
     fail;
-  
+
   rc = re_execW(re->compiled, IDX2PTR(0), s->size,
 		re_fetch_string, s,
 		NULL,
@@ -300,7 +300,7 @@ search_regex(Regex re, Any obj, Int start, Int end, int flags)
     closure = s;
   } else if ( instanceOfObject(obj, ClassTextBuffer) )
   { TextBuffer tb = obj;
-    
+
     len = tb->size;
     fetch = re_fetch_textbuffer;
     closure = tb;
@@ -313,7 +313,7 @@ search_regex(Regex re, Any obj, Int start, Int end, int flags)
     closure = frag;
   } else
     fail;
-   
+
   if ( isDefault(end) )
   { to = len;
   } else
@@ -436,7 +436,7 @@ getMatchRegex(Regex re, Any obj, Int start, Int end)
 { if ( search_regex(re, obj, start, end, RE_MATCH) )
     answer(toInt(re->registers[0].rm_eo -
 		 re->registers[0].rm_so));
-   
+
   fail;
 }
 
@@ -532,7 +532,7 @@ registerValueRegex(Regex re, Any obj, CharArray value, Int which)
       if ( re->registers[i].rm_eo >= start )
 	re->registers[i].rm_eo += shift;
     }
-    
+
     succeed;
   }
 
@@ -564,12 +564,12 @@ replaceRegex(Regex re, Any obj, CharArray value)
 	continue;
       } else
 	errorPce(re, NAME_noRegexRegister, reg, 0);
-    } 
+    }
     str_store(buf, o, c);
     o++;
   }
   buf->size = o;
-  
+
   repl = StringToScratchCharArray(buf);
   rval = registerValueRegex(re, obj, repl, ZERO);
   doneScratchCharArray(repl);
@@ -804,7 +804,7 @@ static getdecl get_regex[] =
 #define rc_regex NULL
 /*
 static classvardecl rc_regex[] =
-{ 
+{
 };
 */
 
