@@ -19,7 +19,7 @@
 
     You should have received a copy of the GNU Lesser General Public
     License along with this library; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */
 
 #ifdef __linux__
@@ -67,6 +67,9 @@ reduced to the facility to terminate the inferior process.
 #ifdef HAVE_UNISTD_H
 #include <unistd.h>
 #endif
+#ifdef HAVE_CRT_EXTERNS_H 
+#include <crt_externs.h>
+#endif
 
 #ifdef HAVE_TERMIOS_H
 #include <termios.h>
@@ -85,9 +88,12 @@ reduced to the facility to terminate the inferior process.
 #ifdef __WINDOWS__
 #define environ _environ	/* declared in STDLIB.H */
 #else
+#ifdef HAVE__NSGETENVIRON
+#define environ (*_NSGetEnviron())
+#else
 extern char **environ;		/* Unix version */
 #endif
-
+#endif
 					/* this fixes RS6000/AIX problems */
 #if !defined(TCGETS) && defined(HAVE_TERMIO_H)
 #include <termio.h>
@@ -800,7 +806,7 @@ endOfFileProcess(Process p)
 
 
 static status
-useTtyProcess(Process p, Bool val)
+useTtyProcess(Process p, BoolObj val)
 { if ( notNil(p->pid) )
     return errorPce(p, NAME_noChangeAfterOpen);
 
