@@ -1,4 +1,4 @@
-/*  $Id: window.c,v 1.70 2004/09/21 14:22:25 jan Exp $
+/*  $Id$
 
     Part of XPCE --- The SWI-Prolog GUI toolkit
 
@@ -19,7 +19,7 @@
 
     You should have received a copy of the GNU Lesser General Public
     License along with this library; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */
 
 #include <h/kernel.h>
@@ -28,6 +28,7 @@
 static status	uncreateWindow(PceWindow sw);
 static status   tileWindow(PceWindow sw, TileObj t);
 static status   updateScrollbarValuesWindow(PceWindow sw);
+static status   UpdateScrollbarValuesWindow(PceWindow sw);
 static status	visible_window(PceWindow sw, IArea a);
 extern void	unlink_changes_data_window(PceWindow sw);
 
@@ -106,7 +107,7 @@ uncreateWindow(PceWindow sw)
 
 
 status
-grabPointerWindow(PceWindow sw, Bool val)
+grabPointerWindow(PceWindow sw, BoolObj val)
 { DEBUG(NAME_focus,
 	Cprintf("FOCUS: grabPointerWindow(%s, %s)\n", pp(sw), pp(val)));
 
@@ -117,7 +118,7 @@ grabPointerWindow(PceWindow sw, Bool val)
 
 
 status
-grabKeyboardWindow(PceWindow sw, Bool val)
+grabKeyboardWindow(PceWindow sw, BoolObj val)
 { ws_grab_keyboard_window(sw, val);
 
   succeed;
@@ -171,7 +172,7 @@ unlinkWindow(PceWindow sw)
 		********************************/
 
 static status
-openWindow(PceWindow sw, Point pos, Bool normalise)
+openWindow(PceWindow sw, Point pos, BoolObj normalise)
 { if ( send(sw, NAME_create, EAV) &&
        send(getFrameWindow(sw, DEFAULT), NAME_open,
 	    pos, DEFAULT, normalise, EAV) )
@@ -182,9 +183,10 @@ openWindow(PceWindow sw, Point pos, Bool normalise)
 
 
 static status
-openCenteredWindow(PceWindow sw, Point pos)
-{ if ( send(sw, NAME_create, EAV) && 
-       send(getFrameWindow(sw, DEFAULT), NAME_openCentered, pos, EAV) )
+openCenteredWindow(PceWindow sw, Point pos, BoolObj grab, Monitor mon)
+{ if ( send(sw, NAME_create, EAV) &&
+       send(getFrameWindow(sw, DEFAULT), NAME_openCentered,
+	    pos, grab, mon, EAV) )
     succeed;
 
   fail;
@@ -192,7 +194,7 @@ openCenteredWindow(PceWindow sw, Point pos)
 
 
 static Any
-getConfirmWindow(PceWindow sw, Point pos, Bool grab, Bool normalise)
+getConfirmWindow(PceWindow sw, Point pos, BoolObj grab, BoolObj normalise)
 { TRY( send(sw, NAME_create, EAV) );
 
   answer(getConfirmFrame(getFrameWindow(sw, DEFAULT), pos, grab, normalise));
@@ -200,10 +202,11 @@ getConfirmWindow(PceWindow sw, Point pos, Bool grab, Bool normalise)
 
 
 static Any
-getConfirmCenteredWindow(PceWindow sw, Point pos, Bool grab)
+getConfirmCenteredWindow(PceWindow sw, Point pos, BoolObj grab, Monitor mon)
 { TRY( send(sw, NAME_create, EAV) );
 
-  answer(getConfirmCenteredFrame(getFrameWindow(sw, DEFAULT), pos, grab));
+  answer(getConfirmCenteredFrame(getFrameWindow(sw, DEFAULT),
+				 pos, grab, mon));
 }
 
 
@@ -236,7 +239,7 @@ createWindow(PceWindow sw, PceWindow parent)
       assign(sw, colour, parent->colour);
     if ( isDefault(sw->background) )
       assign(sw, background, parent->background);
-  } else 
+  } else
   { DisplayObj d;
 
     if ( notNil(sw->frame) )
@@ -409,7 +412,7 @@ deviceWindow(PceWindow sw, Device dev)
 
 
 static status
-displayedWindow(PceWindow sw, Bool val)
+displayedWindow(PceWindow sw, BoolObj val)
 { displayedGraphical(sw, val);
 
   if ( notNil(sw->decoration) )
@@ -441,7 +444,32 @@ resizeMessageWindow(PceWindow sw, Code msg)
   succeed;
 }
 
-  
+
+static Monitor
+getMonitorWindow(PceWindow sw)
+{ if ( isNil(sw->device) )
+  { DisplayObj d = getDisplayGraphical((Graphical)sw);
+
+    if ( d )
+    { FrameObj fr;
+      int dx, dy;
+      struct area a;
+
+      frame_offset_window(sw, &fr, &dx, &dy);
+      a = *fr->area;
+      a.x = toInt(valInt(a.x)+dx);
+      a.y = toInt(valInt(a.y)+dy);
+
+      answer(getMonitorDisplay(d, &a));
+    }
+
+    fail;
+  }
+
+  return getMonitorGraphical((Graphical)sw);
+}
+
+
 		/********************************
 		*           COMPUTING		*
 		********************************/
@@ -488,31 +516,31 @@ frame_offset_window(Any obj, FrameObj *fr, int *X, int *Y)
     { if ( notNil(w->device) )
       { PceWindow w2 = DEFAULT;
 	Int ox, oy;
-  
+
 	get_absolute_xy_graphical((Graphical)w, (Device *)&w2, &ox, &oy);
 	if ( instanceOfObject(w2, ClassWindow) )
 	{ int ox2, oy2;
-  
+
 	  offset_window(w2, &ox2, &oy2);
 	  x += valInt(ox) + ox2;
 	  y += valInt(oy) + oy2;
-  
+
 	  w = w2;
 	  continue;
 	}
       }
-  
+
       fail;
     }
-  
+
     x += valInt(w->area->x);
     y += valInt(w->area->y);
-  
+
     *fr = w->frame, *X = x, *Y = y;
     DEBUG(NAME_position,
 	  Cprintf("frame_offset_window(%s) --> fr = %s, offset = %d,%d\n",
 		  pp(obj), pp(*fr), x, y));
-  
+
     succeed;
   }
 }
@@ -579,7 +607,7 @@ inspectWindow(PceWindow sw, EventObj ev)
 
     for_cell(cell, d->inspect_handlers)
     { Handler h = cell->value;
-      
+
       if ( isAEvent(ev, h->event) )
 	return inspectDevice((Device) sw, ev);
     }
@@ -645,11 +673,11 @@ postEventWindow(PceWindow sw, EventObj ev)
 	    Cprintf("FOCUS: Directing focussed %s event to %s\n",
 		    pp(ev->id),
 		    isNil(sw->focus_recogniser) ? pp(sw->focus)
-		    				: pp(sw->focus_recogniser)));
+						: pp(sw->focus_recogniser)));
 
       rval = postEvent(ev, sw->focus,
 		       isNil(sw->focus_recogniser) ? DEFAULT
-		     				   : sw->focus_recogniser);
+						   : sw->focus_recogniser);
     }
 
     if ( isFreedObj(sw) )
@@ -676,10 +704,10 @@ postEventWindow(PceWindow sw, EventObj ev)
       obj = sb->object;
     else
       obj = sw;
-      
+
     rval = mapWheelMouseEvent(ev, obj);
   }
-    
+
   if ( !rval && isDownEvent(ev) )
     rval = postEvent(ev, (Graphical) sw, popupGesture());
 
@@ -702,13 +730,13 @@ destroyed:
 
 
 status
-typedWindow(PceWindow sw, EventId id, Bool delegate)
+typedWindow(PceWindow sw, EventId id, BoolObj delegate)
 { Name key = characterName(id);
   Graphical gr;
 
   for_chain(sw->graphicals, gr,
 	    if ( send(gr, NAME_key, key, EAV) )
-	      succeed); 
+	      succeed);
 
   if ( delegate == ON )
    { if ( notNil(sw->frame) )
@@ -726,7 +754,7 @@ typedWindow(PceWindow sw, EventId id, Bool delegate)
 		********************************/
 
 status
-inputFocusWindow(PceWindow sw, Bool val)
+inputFocusWindow(PceWindow sw, BoolObj val)
 { DEBUG(NAME_keyboard, Cprintf("inputFocusWindow(%s, %s)\n", pp(sw), pp(val)));
 
   if ( sw->input_focus != val )
@@ -735,12 +763,12 @@ inputFocusWindow(PceWindow sw, Bool val)
     if ( notNil(sw->keyboard_focus) )
       generateEventGraphical(sw->keyboard_focus,
 			     val == ON ? NAME_activateKeyboardFocus
-			   	       : NAME_deactivateKeyboardFocus);
+				       : NAME_deactivateKeyboardFocus);
   }
 
   if ( instanceOfObject(sw, ClassWindowDecorator) )
   { WindowDecorator dw = (WindowDecorator)sw;
-	
+
     sw = dw->window;
     inputFocusWindow(sw, val);
   }
@@ -758,10 +786,10 @@ status
 keyboardFocusWindow(PceWindow sw, Graphical gr)
 { if ( !isNil(gr) && sw->input_focus == OFF )
   { FrameObj fr = getFrameWindow(sw, OFF);
-  
+
     if ( fr )
       send(fr, NAME_keyboardFocus, sw, EAV);
-    
+
   }
 
   if ( sw->keyboard_focus != gr )
@@ -781,7 +809,7 @@ keyboardFocusWindow(PceWindow sw, Graphical gr)
     if ( notNil(gr) )
       generateEventGraphical(gr,
 			     sw->input_focus == ON ? NAME_activateKeyboardFocus
-			     			   : NAME_obtainKeyboardFocus);  }
+						   : NAME_obtainKeyboardFocus);  }
 
   succeed;
 }
@@ -814,7 +842,7 @@ focusWindow(PceWindow sw, Graphical gr, Recogniser recogniser,
     if ( notDefault(cursor) )
       assign(sw, focus_cursor, cursor);
     if ( isDefault(button) &&
-    	 notNil(sw->current_event) && isDownEvent(sw->current_event) )
+	 notNil(sw->current_event) && isDownEvent(sw->current_event) )
       assign(sw, focus_button, getButtonEvent(sw->current_event));
     else
       assign(sw, focus_button, button);
@@ -911,7 +939,7 @@ intersect_iarea(IArea a, IArea b)	/* shrink a with b; fail of empty */
 
   if ( w < 0 || h < 0 )
     fail;
-  
+
   a->x = x;
   a->y = y;
   a->w = w;
@@ -1066,7 +1094,7 @@ redrawWindow(PceWindow sw, Area a)
     succeed;
 
   compute_window(sw, &ox, &oy, &dw, &dh);
-  
+
   if ( isDefault(a) )
   { ia.x = 0;
     ia.y = 0;
@@ -1122,7 +1150,7 @@ RedrawWindow(PceWindow sw)
 
     a = sw->changes_data;
     sw->changes_data = NULL;		/* if we crash, data will be fine! */
- 					/* (only some bytes lost) */
+					/* (only some bytes lost) */
 
 
     DEBUG(NAME_changesData, Cprintf("%s:\n", pp(sw)));
@@ -1174,16 +1202,16 @@ RedrawAreaWindow(PceWindow sw, IArea a, int clear)
     oy += valInt(sw->scroll_offset->y);
 
     d_offset(ox, oy);
-    d_window(sw, a->x, a->y, a->w, a->h, clear, TRUE);
+    if ( d_window(sw, a->x, a->y, a->w, a->h, clear, TRUE) )
+    { assign(oa, x, toInt(a->x));
+      assign(oa, y, toInt(a->y));
+      assign(oa, w, toInt(a->w));
+      assign(oa, h, toInt(a->h));
 
-    assign(oa, x, toInt(a->x));
-    assign(oa, y, toInt(a->y));
-    assign(oa, w, toInt(a->w));
-    assign(oa, h, toInt(a->h));
+      qadSendv(sw, NAME_RedrawArea, 1, (Any *)&oa);
 
-    qadSendv(sw, NAME_RedrawArea, 1, (Any *)&oa);
-
-    d_done();
+      d_done();
+    }
     rewindAnswerStack(mark, NIL);
   }
 
@@ -1214,7 +1242,7 @@ redrawAreaWindow(PceWindow sw, Area a)
 		********************************/
 
 static status
-scrollWindow(PceWindow sw, Int x, Int y, Bool ax, Bool ay)
+scrollWindow(PceWindow sw, Int x, Int y, BoolObj ax, BoolObj ay)
 { int ox = valInt(sw->scroll_offset->x);
   int oy = valInt(sw->scroll_offset->y);
   int nx, ny;
@@ -1244,12 +1272,12 @@ scrollWindow(PceWindow sw, Int x, Int y, Bool ax, Bool ay)
 #ifdef WIN32_GRAPHICS
     ws_scroll_window(sw, nx-ox, ny-oy);
 #else
-  { int x, y, w, h; 
+  { int x, y, w, h;
     int p = valInt(sw->pen);
 
     compute_window(sw, &x, &y, &w, &h);
     x -= valInt(sw->scroll_offset->x) + p;
-    y -= valInt(sw->scroll_offset->y) + p; 
+    y -= valInt(sw->scroll_offset->y) + p;
 
 					/* should use block-move and only */
 					/* consider a small part changed */
@@ -1360,7 +1388,7 @@ normaliseWindow(PceWindow sw, Any obj, Name mode)
 	unionNormalisedArea(a, a2);
 	doneObject(a2);
       }
-    
+
     if ( a->w != ZERO && a->h != ZERO )
       normalise_window(sw, a, xy);
     considerPreserveObject(a);
@@ -1372,7 +1400,7 @@ normaliseWindow(PceWindow sw, Any obj, Name mode)
 
 static status
 scrollHorizontalWindow(PceWindow sw,
-		       Name dir, Name unit, Int amount, Bool force)
+		       Name dir, Name unit, Int amount, BoolObj force)
 { if ( force != ON &&
        (!instanceOfObject(sw->decoration, ClassWindowDecorator) ||
 	isNil(((WindowDecorator)sw->decoration)->horizontal_scrollbar)) )
@@ -1403,7 +1431,7 @@ scrollHorizontalWindow(PceWindow sw,
 
 static status
 scrollVerticalWindow(PceWindow sw,
-		     Name dir, Name unit, Int amount, Bool force)
+		     Name dir, Name unit, Int amount, BoolObj force)
 { if ( force != ON &&
        (!instanceOfObject(sw->decoration, ClassWindowDecorator) ||
 	isNil(((WindowDecorator)sw->decoration)->vertical_scrollbar)) )
@@ -1414,7 +1442,7 @@ scrollVerticalWindow(PceWindow sw,
 
     if ( dir == NAME_goto )
     { int h = ((valInt(bb->h)-valInt(sw->area->h)) * valInt(amount)) / 1000;
-      
+
       scrollWindow(sw, DEFAULT, toInt(h + valInt(bb->y)), ON, ON);
     }
   } else if ( unit == NAME_page )
@@ -1463,7 +1491,7 @@ view_region(int x, int w, int rx, int rw)
 
   return w < 0 ? 2 : w;
 }
-	
+
 
 static status				/* update bubble of scroll_bar */
 bubbleScrollBarWindow(PceWindow sw, ScrollBar sb)
@@ -1478,7 +1506,7 @@ bubbleScrollBarWindow(PceWindow sw, ScrollBar sb)
   x -= valInt(sw->scroll_offset->x);
   y -= valInt(sw->scroll_offset->y);
 					/* x, y, w, h: visible area */
-  
+
   view = view_region(start, length,
 		     hor ? -valInt(sw->scroll_offset->x)
 			 : -valInt(sw->scroll_offset->y),
@@ -1530,8 +1558,8 @@ getDisplayedCursorWindow(PceWindow sw)
       answer(sw->focus_cursor);
     if ( notNil(sw->focus->cursor) )
       answer(sw->focus->cursor);
-  }    
-  
+  }
+
   if ( (rval = getDisplayedCursorDevice((Device) sw)) &&
        notNil(rval) )
     answer(rval);
@@ -1618,7 +1646,7 @@ get_display_position_window(PceWindow sw, int *X, int *Y)
   TRY(frame_offset_window(sw, &fr, &x, &y));
   x += valInt(fr->area->x);
   y += valInt(fr->area->y);
-  
+
   *X = x; *Y = y;
 
   succeed;
@@ -1630,8 +1658,10 @@ visible_window(PceWindow sw, IArea a)
 { int p = valInt(sw->pen);
 
   compute_window(sw, &a->x, &a->y, &a->w, &a->h);
-  a->x -= valInt(sw->scroll_offset->x) + p;
-  a->y -= valInt(sw->scroll_offset->y) + p;
+  a->x -= valInt(sw->scroll_offset->x);
+  a->y -= valInt(sw->scroll_offset->y);
+  a->w -= 2*p;
+  a->h -= 2*p;
 
   succeed;
 }
@@ -1694,7 +1724,7 @@ for_all_tile(TileObj tile, SendFunc f, Any arg)
 
     for_chain(tile->members, st,
 	      TRY(for_all_tile(st, f, arg)));
-    
+
     succeed;
   }
 }
@@ -1757,15 +1787,15 @@ getTileWindow(PceWindow sw)
 
 
 FrameObj
-getFrameWindow(PceWindow sw, Bool create)
+getFrameWindow(PceWindow sw, BoolObj create)
 { PceWindow root = (PceWindow) getRootGraphical((Graphical) sw);
-  
+
   if ( instanceOfObject(root, ClassWindow) )
   { if ( create != OFF )
       frameWindow(root, DEFAULT);
     if ( notNil(root->frame) )
       answer(root->frame);
-  }    
+  }
 
   fail;
 }
@@ -1793,7 +1823,7 @@ mergeFramesWindow(PceWindow w1, PceWindow w2)
     frameWindow(w2, w1->frame);
   else
     frameWindow(w1, w2->frame);
-  
+
   succeed;
 }
 
@@ -1808,7 +1838,7 @@ relateWindow(PceWindow sw, Name how, Any to)
     return relateWindow(sw->decoration, how, to);
   if ( notNil(w2) && notNil(w2->decoration) )
     return relateWindow(sw, how, w2->decoration);
-  
+
   DeviceGraphical((Graphical)sw, NIL);
   if ( notNil(w2) )
   { DeviceGraphical((Graphical)w2, NIL);
@@ -1817,7 +1847,7 @@ relateWindow(PceWindow sw, Name how, Any to)
 
   if ( createdWindow(sw) && notNil(sw->frame) )
     send(sw->frame, NAME_delete, sw, EAV);
-  
+
   tileWindow(sw, DEFAULT);
 
   if ( isNil(wto) )
@@ -1844,7 +1874,7 @@ relateWindow(PceWindow sw, Name how, Any to)
 
     w2 = t2->object;
   }
-  
+
   mergeFramesWindow(sw, w2);
 
   if ( (fr=getFrameWindow(sw, OFF)) && createdFrame(fr) )
@@ -1891,7 +1921,7 @@ penWindow(PceWindow sw, Int pen)
     if ( ws_created_window(sw) )
     { int x, y, w, h;
       int pen = valInt(sw->pen);
-      
+
       x = valInt(sw->area->x);
       y = valInt(sw->area->y);
       w = valInt(sw->area->w);
@@ -1955,7 +1985,7 @@ getForegroundWindow(PceWindow sw)
 
 
 static status
-sensitiveWindow(PceWindow sw, Bool sensitive)
+sensitiveWindow(PceWindow sw, BoolObj sensitive)
 { if ( sw->sensitive != sensitive )
   { assign(sw, sensitive, sensitive);
 
@@ -1991,7 +2021,7 @@ status
 flashWindow(PceWindow sw, Area a, Int time)
 { if ( sw->displayed == ON && createdWindow(sw) )
   { int t;
-    
+
     if ( isDefault(time) )
       time = getClassVariableValueObject(sw, NAME_visualBellDuration);
     t = (isInteger(time) ? valInt(time) : 250);
@@ -2027,7 +2057,7 @@ exposeWindow(PceWindow sw)
   if ( notNil(sw->frame) )
     return exposeFrame(sw->frame);
 
-  ws_raise_window(sw); 
+  ws_raise_window(sw);
 
   succeed;
 }
@@ -2086,7 +2116,7 @@ catchAllWindowv(PceWindow sw, Name selector, int argc, Any *argv)
       return sendv(fr, selector, argc, argv);
     else
       fail;
-  }	
+  }
 
   if ( getSendMethodClass(ClassTile, selector) )
   { if ( notNil(sw->decoration) )
@@ -2126,7 +2156,7 @@ static char *T_scrollHV[] =
 static char *T_decorate[] =
         { "area=[{grow,shrink}]", "left_margin=[int]", "right_margin=[int]", "top_margin=[int]", "bottom_margin=[int]", "decorator=[window]" };
 static char *T_confirmCentered[] =
-        { "center=[point]", "grab=[bool]" };
+        { "center=[point]", "grab=[bool]", "monitor=[monitor]" };
 static char *T_typed[] =
         { "event|event_id", "delegate=[bool]" };
 static char *T_focus[] =
@@ -2270,9 +2300,11 @@ static senddecl send_window[] =
      NAME_layout, "Put me right of argument"),
   SM(NAME_create, 1, "[window]", createWindow,
      NAME_open, "Create associated X-window structure"),
+  SM(NAME_Create, 1, "[window]", createWindow,
+     NAME_open, "Create associated X-window structure (internal)"),
   SM(NAME_open, 2, T_open, openWindow,
      NAME_open, "Open associated frame on the display"),
-  SM(NAME_openCentered, 1, "[point]", openCenteredWindow,
+  SM(NAME_openCentered, 3, T_confirmCentered, openCenteredWindow,
      NAME_open, "Open frame centered around point"),
   SM(NAME_uncreate, 0, NULL, uncreateWindow,
      NAME_open, "Destroy associated X-window structure"),
@@ -2327,6 +2359,8 @@ static getdecl get_window[] =
      NAME_area, "New area representing visible part"),
   GM(NAME_size, 0, "size", NULL, getSizeGraphical,
      NAME_area, "New size representing size (avoid class-variable)"),
+  GM(NAME_monitor, 0, "monitor", NULL, getMonitorWindow,
+     NAME_organisation, "Monitor window is displayed on"),
   GM(NAME_displayedCursor, 0, "cursor*", NULL, getDisplayedCursorDevice,
      NAME_cursor, "Currently displayed cursor"),
   GM(NAME_confirm, 3, "any", T_confirm, getConfirmWindow,
@@ -2335,7 +2369,7 @@ static getdecl get_window[] =
   GM(NAME_winHandle, 0, "int", NULL, getWinHandleWindow,
      NAME_windows, "Fetch the MS-Windows HWND of the window (if any)"),
 #endif
-  GM(NAME_confirmCentered, 2, "any", T_confirmCentered, getConfirmCenteredWindow,
+  GM(NAME_confirmCentered, 3, "any", T_confirmCentered, getConfirmCenteredWindow,
      NAME_modal, "->confirm with frame centered around point"),
   GM(NAME_thread, 0, "int", NULL, getThreadWindow,
      NAME_thread, "Return system thread-id that owns the window")
