@@ -19,7 +19,7 @@
 
     You should have received a copy of the GNU Lesser General Public
     License along with this library; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */
 
 #ifndef _PCE_X11_INCLUDED
@@ -58,11 +58,6 @@
 #undef index				/* X11 defines this too */
 #undef String
 
-#ifndef XMalloc
-#define XMalloc(size) pceMalloc(size)
-#define XFree(ptr)    pceFree(ptr)
-#endif
-
 #if XT_VERSION == 10
 error XPCE cannot be build for X version 10.  Sorry.
 #endif
@@ -162,7 +157,7 @@ struct xpce_font_info
 { XFontSet	 font_set;		/* FontSet structure */
   char	       **missing;		/* Missing charsets */
   int		nmissing;		/* # missing charsets */
-  char 	        *def_string;		/* Default for missing glyphs */
+  char	        *def_string;		/* Default for missing glyphs */
 };
 
 #else /*USE_XFONTSET*/
