@@ -3,9 +3,10 @@
     Part of SWI-Prolog
 
     Author:        Jan Wielemaker
-    E-mail:        wielemak@science.uva.nl
+    E-mail:        J.Wielemaker@cs.vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (C): 1985-2007, University of Amsterdam
+    Copyright (C): 1985-2011, University of Amsterdam
+			      VU University Amsterdam
 
     This library is free software; you can redistribute it and/or
     modify it under the terms of the GNU Lesser General Public
@@ -19,7 +20,7 @@
 
     You should have received a copy of the GNU Lesser General Public
     License along with this library; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */
 
 :- module(test_arith, [test_arith/0]).
@@ -31,12 +32,46 @@
 */
 
 test_arith :-
-	run_tests([ rem,
+	run_tests([ div,
+		    gdiv,
+		    rem,
 		    mod,
+		    gcd,
+		    shift,
 		    errors,
-		    ar_builtin
+		    ar_builtin,
+		    eval
 		  ]).
 
+:- begin_tests(div).
+
+test(mod, true) :-
+	forall(between(-10, 10, X),
+	       forall((between(-10, 10, Y), Y =\= 0),
+		      (	  Q is div(X, Y),
+			  M is mod(X, Y),
+			  X =:= Y*Q+M
+		      ))).
+
+:- end_tests(div).
+
+:- begin_tests(gdiv).
+
+:- if(current_prolog_flag(bounded,false)).
+
+test(minint, X == 9223372036854775808) :-
+	X is -9223372036854775808 // -1.
+
+:- else.
+
+test(minint, error(evaluation_error(integer_overflow))) :-
+	X is -9223372036854775808 // -1,
+	writeln(X).
+
+:- endif.
+
+:- end_tests(gdiv).
+
 :- begin_tests(rem).
 
 test(small, R == 2) :-
@@ -83,6 +118,30 @@ test(big, [condition(current_prolog_flag(bounded, false)), R =:= 10^50-3]) :-
 
 :- end_tests(mod).
 
+:- begin_tests(shift).
+
+test(shift_right_large, X == 0) :-
+	X is 5>>64.
+test(shift_right_large, X == 0) :-
+	X is 5>>(1<<62).
+:- if(current_prolog_flag(bounded, false)).
+test(shift_right_large, X == 0) :-
+	X is 5>>(1<<100).
+test(shift_left_large, X == -18446744073709551616) :-
+	X is (-1<<40)<<24.
+:- endif.
+
+:- end_tests(shift).
+
+:- begin_tests(gcd).
+
+test(gcd, X == 4) :-
+	X is gcd(100, 24).
+test(gcd, X == 4) :-
+	X is gcd(24, 100).		% seems to be some argument ordering
+
+:- end_tests(gcd).
+
 :- begin_tests(errors).
 
 test(cyclic, [sto(rational_trees), error(type_error(expression, T))]) :-
@@ -111,3 +170,13 @@ test(a_fc_minus) :-
 	a2.
 
 :- end_tests(ar_builtin).
+
+
+:- begin_tests(eval).
+
+test(ref, R==6) :-			% Bug#12
+	term_to_atom(Expr, 'X = 2+3, Y is X+1'),
+	call(Expr),
+	Expr = (_,R is _).
+
+:- end_tests(eval).
