@@ -3,9 +3,10 @@
     Part of SWI-Prolog
 
     Author:        Jan Wielemaker
-    E-mail:        wielemak@science.uva.nl
+    E-mail:        J.Wielemaker@cs.vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (C): 1985-2008, University of Amsterdam
+    Copyright (C): 1985-2011, University of Amsterdam
+			      VU University Amsterdam
 
     This library is free software; you can redistribute it and/or
     modify it under the terms of the GNU Lesser General Public
@@ -19,7 +20,7 @@
 
     You should have received a copy of the GNU Lesser General Public
     License along with this library; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */
 
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
@@ -36,6 +37,7 @@ handling times must be cleaned, but that not only holds for this module.
 /*#define O_DEBUG 1*/
 /*#define O_DEBUG_MT 1*/
 
+#define NEEDS_SWINSOCK
 #include "pl-incl.h"
 #include "pl-ctype.h"
 #include "pl-utf8.h"
@@ -66,6 +68,11 @@ handling times must be cleaned, but that not only holds for this module.
 #undef LD				/* fetch LD once per function */
 #define LD LOCAL_LD
 
+/* there are two types of stream property functions. In the usual case,
+   they have an argument, but in a few cases they don't */
+typedef int (*property0_t)(IOSTREAM *s ARG_LD);
+typedef int (*property_t)(IOSTREAM *s, term_t prop ARG_LD);
+
 static int	bad_encoding(const char *msg, atom_t name);
 static int	noprotocol(void);
 static PL_blob_t stream_blob;
@@ -141,8 +148,7 @@ getStreamContext(IOSTREAM *s)
 { Symbol symb;
 
   if ( !(symb = lookupHTable(streamContext, s)) )
-  { GET_LD
-    stream_context *ctx = allocHeap(sizeof(*ctx));
+  { stream_context *ctx = allocHeapOrHalt(sizeof(*ctx));
 
     DEBUG(1, Sdprintf("Created ctx=%p for stream %p\n", ctx, s));
 
@@ -160,8 +166,7 @@ getStreamContext(IOSTREAM *s)
 
 void
 aliasStream(IOSTREAM *s, atom_t name)
-{ GET_LD
-  stream_context *ctx;
+{ stream_context *ctx;
   Symbol symb;
   alias *a;
 
@@ -173,7 +178,7 @@ aliasStream(IOSTREAM *s, atom_t name)
   addHTable(streamAliases, (void *)name, s);
   PL_register_atom(name);
 
-  a = allocHeap(sizeof(*a));
+  a = allocHeapOrHalt(sizeof(*a));
   a->next = NULL;
   a->name = name;
 
@@ -190,8 +195,7 @@ aliasStream(IOSTREAM *s, atom_t name)
 
 static void
 unaliasStream(IOSTREAM *s, atom_t name)
-{ GET_LD
-  Symbol symb;
+{ Symbol symb;
 
   if ( name )
   { if ( (symb = lookupHTable(streamAliases, (void *)name)) )
@@ -297,7 +301,7 @@ setFileNameStream(IOSTREAM *s, atom_t name)
   { PL_unregister_atom(ctx->filename);
     ctx->filename = NULL_ATOM;
   }
-  if ( name != NULL_ATOM )
+  if ( !(name == NULL_ATOM || name == ATOM_) )
     ctx->filename = name;
 }
 
@@ -452,6 +456,7 @@ typedef struct stream_ref
 static int
 write_stream_ref(IOSTREAM *s, atom_t aref, int flags)
 { stream_ref *ref = PL_blob_data(aref, NULL, NULL);
+  (void)flags;
 
   if ( ref->read && ref->write )
     Sfprintf(s, "<stream>(%p,%p)", ref->read, ref->write);
@@ -495,6 +500,7 @@ release_stream_ref(atom_t aref)
 static int
 save_stream_ref(atom_t aref, IOSTREAM *fd)
 { stream_ref *ref = PL_blob_data(aref, NULL, NULL);
+  (void)fd;
 
   return PL_warning("Cannot save reference to <stream>(%p,%p)",
 		    ref->read, ref->write);
@@ -503,7 +509,9 @@ save_stream_ref(atom_t aref, IOSTREAM *fd)
 
 static atom_t
 load_stream_ref(IOSTREAM *fd)
-{ return PL_new_atom("<saved-stream-ref>");
+{ (void)fd;
+
+  return PL_new_atom("<saved-stream-ref>");
 }
 
 
@@ -613,7 +621,7 @@ term_stream_handle(term_t t, IOSTREAM **s, int flags ARG_LD)
   if ( !PL_get_atom(t, &a) )
     return not_a_stream(t);
 
-  return get_stream_handle(a, s, SH_ERRORS|SH_ALIAS);
+  return get_stream_handle(a, s, flags);
 }
 
 
@@ -671,10 +679,8 @@ PL_unify_stream_or_alias(term_t t, IOSTREAM *s)
 
 int
 PL_unify_stream(term_t t, IOSTREAM *s)
-{ stream_context *ctx;
-
-  LOCK();
-  ctx = getStreamContext(s);
+{ LOCK();
+  (void)getStreamContext(s);		/* get stream known to Prolog */
   UNLOCK();
 
   return unify_stream_ref(t, s);
@@ -701,15 +707,46 @@ getOutputStream(term_t t, IOSTREAM **s)
     using releaseStream() or streamStatus().
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
-int
-getOutputStream(term_t t, IOSTREAM **stream)
-{ GET_LD
-  atom_t a;
+typedef enum
+{ S_DONTCARE = 0,
+  S_TEXT,
+  S_BINARY
+} s_type;
+
+
+static int
+checkStreamType(s_type text, IOSTREAM *s, atom_t *error ARG_LD)
+{ if ( text == S_DONTCARE || LD->IO.stream_type_check == ST_FALSE )
+    return TRUE;			/* no checking */
+
+					/* ok? */
+  if ( text == S_TEXT && (s->flags&SIO_TEXT) )
+    return TRUE;
+  if ( text == S_BINARY && !(s->flags&SIO_TEXT) )
+    return TRUE;
+					/* no */
+  if ( LD->IO.stream_type_check == ST_LOOSE )
+  { if ( text == S_TEXT )
+      return TRUE;
+    if ( s->encoding == ENC_ISO_LATIN_1 ||
+	 s->encoding == ENC_OCTET )
+      return TRUE;
+  }
+
+  *error = (text == S_TEXT ? ATOM_binary_stream : ATOM_text_stream);
+  return FALSE;
+}
+
+
+static int
+getOutputStream__LD(term_t t, s_type text, IOSTREAM **stream ARG_LD)
+{ atom_t a;
   IOSTREAM *s;
+  atom_t tp;
 
   if ( t == 0 )
-  { if ( (*stream = getStream(Scurout)) )
-      return TRUE;
+  { if ( (s = getStream(Scurout)) )
+      goto ok;
     return no_stream(t);
   }
 
@@ -717,33 +754,49 @@ getOutputStream(term_t t, IOSTREAM **stream)
     return not_a_stream(t);
 
   if ( a == ATOM_user )
-  { if ( (*stream = getStream(Suser_output)) )
-      return TRUE;
+  { if ( (s = getStream(Suser_output)) )
+      goto ok;
     return no_stream(t);
   }
 
   if ( !get_stream_handle(a, &s, SH_ERRORS|SH_ALIAS|SH_OUTPUT) )
     return FALSE;
 
+ok:
   if ( !(s->flags&SIO_OUTPUT) )
-  { releaseStream(s);
-    return PL_error(NULL, 0, NULL, ERR_PERMISSION,
-		    ATOM_output, ATOM_stream, t);
+  { tp = ATOM_stream;
+  } else if ( checkStreamType(text, s, &tp PASS_LD) )
+  { *stream = s;
+    return TRUE;
   }
 
-  *stream = s;
-  return TRUE;
+  releaseStream(s);
+  return PL_error(NULL, 0, NULL, ERR_PERMISSION,
+		  ATOM_output, tp, t);
 }
 
 
 int
-getInputStream__LD(term_t t, IOSTREAM **stream ARG_LD)
+getTextOutputStream__LD(term_t t, IOSTREAM **stream ARG_LD)
+{ return getOutputStream(t, S_TEXT, stream);
+}
+
+
+int
+getBinaryOutputStream__LD(term_t t, IOSTREAM **stream ARG_LD)
+{ return getOutputStream(t, S_BINARY, stream);
+}
+
+
+static int
+getInputStream__LD(term_t t, s_type text, IOSTREAM **stream ARG_LD)
 { atom_t a;
   IOSTREAM *s;
+  atom_t tp;
 
   if ( t == 0 )
-  { if ( (*stream = getStream(Scurin)) )
-      return TRUE;
+  { if ( (s = getStream(Scurin)) )
+      goto ok;
     return no_stream(t);
   }
 
@@ -751,22 +804,35 @@ getInputStream__LD(term_t t, IOSTREAM **stream ARG_LD)
     return not_a_stream(t);
 
   if ( a == ATOM_user )
-  { if ( (*stream = getStream(Suser_input)) )
-      return TRUE;
+  { if ( (s = getStream(Suser_input)) )
+      goto ok;
     return no_stream(t);
   }
 
   if ( !get_stream_handle(a, &s, SH_ERRORS|SH_ALIAS|SH_INPUT) )
     return FALSE;
 
-  if ( !(s->flags &SIO_INPUT) )
-  { releaseStream(s);
-    return PL_error(NULL, 0, NULL, ERR_PERMISSION,
-		    ATOM_input, ATOM_stream, t);
+ok:
+  if ( !(s->flags&SIO_INPUT) )
+  { tp = ATOM_stream;
+  } else if ( checkStreamType(text, s, &tp PASS_LD) )
+  { *stream = s;
+    return TRUE;
   }
 
-  *stream = s;
-  return TRUE;
+  releaseStream(s);
+  return PL_error(NULL, 0, NULL, ERR_PERMISSION,
+		  ATOM_input, tp, t);
+}
+
+int
+getTextInputStream__LD(term_t t, IOSTREAM **stream ARG_LD)
+{ return getInputStream__LD(t, S_TEXT, stream PASS_LD);
+}
+
+int
+getBinaryInputStream__LD(term_t t, IOSTREAM **stream ARG_LD)
+{ return getInputStream__LD(t, S_BINARY, stream PASS_LD);
 }
 
 
@@ -796,8 +862,8 @@ PRED_IMPL("stream_pair", 3, stream_pair, 0)
 	     PL_unify_stream_or_alias(A3, ref->write) );
   }
 
-  if ( getInputStream(A2, &in) &&
-       getOutputStream(A3, &out) )
+  if ( getInputStream(A2, S_DONTCARE, &in) &&
+       getOutputStream(A3, S_DONTCARE, &out) )
   { stream_ref ref;
 
     ref.read = in;
@@ -828,7 +894,7 @@ static int
 isConsoleStream(IOSTREAM *s)
 { int i = standardStreamIndexFromStream(s);
 
-  return i >= 0 && i < 3;
+  return i >= 1 && i < 3;			/* only output streams */
 }
 #else
 #define isConsoleStream(s) FALSE
@@ -878,7 +944,13 @@ reportStreamError(IOSTREAM *s)
 	} else
 	  op = ATOM_read;
       } else
-	op = ATOM_write;
+      { if ( (s->flags & SIO_TIMEOUT) )
+	{ PL_error(NULL, 0, NULL, ERR_TIMEOUT,
+		   ATOM_write, stream);
+	  return FALSE;
+	} else
+	  op = ATOM_write;
+      }
 
       if ( s->message )
       { msg = s->message;
@@ -1021,23 +1093,6 @@ closeFiles(int all)
 }
 
 
-/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-PL_cleanup_fork() must be called between  fork()   and  exec() to remove
-traces of Prolog that are not  supposed   to  leak into the new process.
-Note that we must be careful  here.   Notably,  the  code cannot lock or
-unlock any mutex as the behaviour of mutexes is undefined over fork().
-
-Earlier versions used the file-table to  close file descriptors that are
-in use by Prolog. This can't work as   the  table is guarded by a mutex.
-Now we use the FD_CLOEXEC flag in Snew();
-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
-
-void
-PL_cleanup_fork(void)
-{ stopItimer();
-}
-
-
 void
 protocol(const char *str, size_t n)
 { GET_LD
@@ -1058,9 +1113,9 @@ protocol(const char *str, size_t n)
 
 
 static int
-push_input_context()
+push_input_context(void)
 { GET_LD
-  InputContext c = allocHeap(sizeof(struct input_context));
+  InputContext c = allocHeapOrHalt(sizeof(struct input_context));
 
   c->stream           = Scurin;
   c->term_file        = source_file_name;
@@ -1073,7 +1128,7 @@ push_input_context()
 
 
 static int
-pop_input_context()
+pop_input_context(void)
 { GET_LD
   InputContext c = input_context_stack;
 
@@ -1105,9 +1160,9 @@ PRED_IMPL("$pop_input_context", 0, pop_input_context, 0)
 
 
 void
-pushOutputContext()
+pushOutputContext(void)
 { GET_LD
-  OutputContext c = allocHeap(sizeof(struct output_context));
+  OutputContext c = allocHeapOrHalt(sizeof(struct output_context));
 
   c->stream            = Scurout;
   c->previous          = output_context_stack;
@@ -1116,7 +1171,7 @@ pushOutputContext()
 
 
 void
-popOutputContext()
+popOutputContext(void)
 { GET_LD
   OutputContext c = output_context_stack;
 
@@ -1341,7 +1396,7 @@ getSingleChar(IOSTREAM *stream, int signals)
   int c;
   ttybuf buf;
 
-  debugstatus.suspendTrace++;
+  suspendTrace(TRUE);
   Slock(stream);
   Sflush(stream);
   PushTty(stream, &buf, TTY_RAW);	/* just donot prompt */
@@ -1368,7 +1423,7 @@ getSingleChar(IOSTREAM *stream, int signals)
     c = -1;
 
   PopTty(stream, &buf, TRUE);
-  debugstatus.suspendTrace--;
+  suspendTrace(FALSE);
   Sunlock(stream);
 
   return c;
@@ -1661,6 +1716,21 @@ PRED_IMPL("set_stream", 2, set_stream, 0)
 	else
 	  s->position = NULL;
 
+	goto ok;
+      } else if ( aname == ATOM_line_position )
+      { int lpos;
+
+	if ( !PL_get_integer_ex(a, &lpos) )
+	  goto error;
+
+	if ( s->position )
+	{ s->position->linepos = lpos;
+	} else
+	{ PL_error(NULL, 0, NULL, ERR_PERMISSION,
+		   ATOM_line_position, ATOM_stream, stream);
+	  goto error;
+	}
+
 	goto ok;
       } else if ( aname == ATOM_file_name ) /* file_name(Atom) */
       {	atom_t fn;
@@ -1785,7 +1855,7 @@ error:
   return FALSE;
 }
 
-#ifdef __WINDOWS__			/* defined in pl-nt.c */
+#ifdef _MSC_VER					/* defined in pl-nt.c */
 extern int ftruncate(int fileno, int64_t length);
 #define HAVE_FTRUNCATE
 #endif
@@ -1872,31 +1942,41 @@ PRED_IMPL("wait_for_input", 3, wait_for_input, 0)
 
 #else
 
+/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+Windows<->Unix note. This function uses the   Windows socket API for its
+implementation and defines the Unix API  in   terms  of the Windows API.
+This approach allows full support  of   the  restrictions of the Windows
+implementation. Because the Unix emulation is   more generic, this still
+supports  the  generic  facilities  of  Unix  select()  that  make  this
+predicate work on pipes, serial devices, etc.
+- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
+
+#ifndef __WINDOWS__
+typedef int SOCKET;
+#define INVALID_SOCKET -1
+#define Swinsock(s) Sfileno(s)
+#define NFDS(max) (max+1)			/* see also S__wait() */
+#else
+#define NFDS(n) 0
+#endif
+
 typedef struct fdentry
-{ int fd;
+{ SOCKET fd;
   term_t stream;
   struct fdentry *next;
 } fdentry;
 
 
-static inline term_t
-findmap(fdentry *map, int fd)
-{ for( ; map; map = map->next )
-  { if ( map->fd == fd )
-      return map->stream;
-  }
-  assert(0);
-  return 0;
-}
-
-
 static
 PRED_IMPL("wait_for_input", 3, wait_for_input, 0)
 { PRED_LD
   fd_set fds;
   struct timeval t, *to;
   double time;
-  int n, max = 0, ret, min = 1 << (INTBITSIZE-2);
+  int rc;
+#ifndef __WINDOWS__
+  SOCKET max = 0, min = INT_MAX;
+#endif
   fdentry *map     = NULL;
   term_t head      = PL_new_term_ref();
   term_t streams   = PL_copy_term_ref(A1);
@@ -1910,12 +1990,12 @@ PRED_IMPL("wait_for_input", 3, wait_for_input, 0)
   FD_ZERO(&fds);
   while( PL_get_list(streams, head, streams) )
   { IOSTREAM *s;
-    int fd;
+    SOCKET fd;
     fdentry *e;
 
     if ( !PL_get_stream_handle(head, &s) )
       return FALSE;
-    if ( (fd=Sfileno(s)) < 0 )
+    if ( (fd=Swinsock(s)) < 0 )
     { releaseStream(s);
       return PL_error("wait_for_input", 3, NULL, ERR_DOMAIN,
 		      PL_new_atom("file_stream"), head);
@@ -1935,16 +2015,13 @@ PRED_IMPL("wait_for_input", 3, wait_for_input, 0)
     e->next   = map;
     map       = e;
 
-#ifdef __WINDOWS__
-    FD_SET((SOCKET)fd, &fds);
-#else
     FD_SET(fd, &fds);
-#endif
-
+#ifndef __WINDOWS__
     if ( fd > max )
       max = fd;
     if( fd < min )
       min = fd;
+#endif
   }
   if ( !PL_get_nil(streams) )
     return PL_error("wait_for_input", 3, NULL, ERR_TYPE, ATOM_list, A1);
@@ -1984,7 +2061,7 @@ PRED_IMPL("wait_for_input", 3, wait_for_input, 0)
     to = &t;
   }
 
-  while( (ret=select(max+1, &fds, NULL, NULL, to)) == -1 &&
+  while( (rc=select(NFDS(max), &fds, NULL, NULL, to)) == -1 &&
 	 errno == EINTR )
   { fdentry *e;
 
@@ -1993,16 +2070,10 @@ PRED_IMPL("wait_for_input", 3, wait_for_input, 0)
 
     FD_ZERO(&fds);			/* EINTR may leave fds undefined */
     for(e=map; e; e=e->next)		/* so we rebuild it to be safe */
-    {
-#ifdef __WINDOWS__
-      FD_SET((SOCKET)e->fd, &fds);
-#else
       FD_SET(e->fd, &fds);
-#endif
-    }
   }
 
-  switch(ret)
+  switch(rc)
   { case -1:
       return PL_error("wait_for_input", 3, MSG_ERRNO, ERR_FILE_OPERATION,
 		      ATOM_select, ATOM_stream, A1);
@@ -2011,14 +2082,17 @@ PRED_IMPL("wait_for_input", 3, wait_for_input, 0)
       break;
 
     default: /* Something happend -> check fds */
-      for(n=min; n <= max; n++)
-      { if ( FD_ISSET(n, &fds) )
+    { fdentry *mp;
+
+      for(mp=map; mp; mp=mp->next)
+      { if ( FD_ISSET(mp->fd, &fds) )
 	{ if ( !PL_unify_list(available, ahead, available) ||
-	       !PL_unify(ahead, findmap(map, n)) )
+	       !PL_unify(ahead, mp->stream) )
 	    return FALSE;
 	}
       }
       break;
+    }
   }
 
   return PL_unify_nil(available);
@@ -2107,7 +2181,7 @@ PRED_IMPL("read_pending_input", 3, read_pending_input, 0)
 { PRED_LD
   IOSTREAM *s;
 
-  if ( getInputStream(A1, &s) )
+  if ( getInputStream(A1, S_DONTCARE, &s) )
   { char buf[MAX_PENDING];
     ssize_t n;
     int64_t off0 = Stell64(s);
@@ -2187,7 +2261,7 @@ PRED_IMPL("read_pending_input", 3, read_pending_input, 0)
 	  us += mbrtowc(&c, us, es-us, s->mbstate);
 	  if ( c == '\r' && skip_cr(s) )
 	    continue;
-    	  if ( s->position )
+	  if ( s->position )
 	    S__fupdatefilepos_getc(s, c);
 
 	  addSmallIntList(&ctx, c);
@@ -2225,7 +2299,7 @@ PRED_IMPL("read_pending_input", 3, read_pending_input, 0)
 	  us = utf8_get_char(us, &c);
 	  if ( c == '\r' && skip_cr(s) )
 	    continue;
-    	  if ( s->position )
+	  if ( s->position )
 	    S__fupdatefilepos_getc(s, c);
 
 	  addSmallIntList(&ctx, c);
@@ -2323,7 +2397,7 @@ put_byte(term_t stream, term_t byte ARG_LD)
 
   if ( !PL_get_integer(byte, &c) || c < 0 || c > 255 )
     return PL_error(NULL, 0, NULL, ERR_TYPE, ATOM_byte, byte);
-  if ( !getOutputStream(stream, &s) )
+  if ( !getBinaryOutputStream(stream, &s) )
     return FALSE;
 
   Sputc(c, s);
@@ -2355,7 +2429,7 @@ put_code(term_t stream, term_t chr ARG_LD)
 
   if ( !PL_get_char(chr, &c, FALSE) )
     return FALSE;
-  if ( !getOutputStream(stream, &s) )
+  if ( !getTextOutputStream(stream, &s) )
     return FALSE;
 
   Sputcode(c, s);
@@ -2400,7 +2474,7 @@ static foreign_t
 get_nonblank(term_t in, term_t chr ARG_LD)
 { IOSTREAM *s;
 
-  if ( getInputStream(in, &s) )
+  if ( getTextInputStream(in, &s) )
   { int c;
 
     for(;;)
@@ -2446,7 +2520,7 @@ skip(term_t in, term_t chr ARG_LD)
 
   if ( !PL_get_char(chr, &c, FALSE) )
     return FALSE;
-  if ( !getInputStream(in, &s) )
+  if ( !getTextInputStream(in, &s) )
     return FALSE;
 
   while((r=Sgetcode(s)) != c && r != EOF )
@@ -2502,7 +2576,7 @@ static foreign_t
 get_byte2(term_t in, term_t chr ARG_LD)
 { IOSTREAM *s;
 
-  if ( getInputStream(in, &s) )
+  if ( getBinaryInputStream(in, &s) )
   { int c = Sgetc(s);
 
     if ( PL_unify_integer(chr, c) )
@@ -2538,7 +2612,7 @@ static foreign_t
 get_code2(term_t in, term_t chr ARG_LD)
 { IOSTREAM *s;
 
-  if ( getInputStream(in, &s) )
+  if ( getTextInputStream(in, &s) )
   { int c = Sgetcode(s);
 
     if ( PL_unify_integer(chr, c) )
@@ -2573,7 +2647,7 @@ static foreign_t
 get_char2(term_t in, term_t chr ARG_LD)
 { IOSTREAM *s;
 
-  if ( getInputStream(in, &s) )
+  if ( getTextInputStream(in, &s) )
   { int c = Sgetcode(s);
 
     if ( PL_unify_atom(chr, c == -1 ? ATOM_end_of_file : codeToAtom(c)) )
@@ -2658,7 +2732,7 @@ PRED_IMPL("prompt", 2, prompt, 0)
   term_t new = A2;
 
   if ( PL_unify_atom(old, LD->prompt.current) &&
-       PL_get_atom(new, &a) )
+       PL_get_atom_ex(new, &a) )
   { if ( LD->prompt.current )
       PL_unregister_atom(LD->prompt.current);
     LD->prompt.current = a;
@@ -2720,11 +2794,11 @@ PrologPrompt()
 
 
 static int
-tab(term_t out, term_t spaces)
+tab(term_t out, term_t spaces ARG_LD)
 { int64_t count;
   IOSTREAM *s;
 
-  if ( !getOutputStream(out, &s) )
+  if ( !getTextOutputStream(out, &s) )
     return FALSE;
   if ( !PL_eval_expression_to_int64_ex(spaces, &count) )
     return FALSE;
@@ -2740,12 +2814,16 @@ tab(term_t out, term_t spaces)
 
 static
 PRED_IMPL("tab", 2, tab2, 0)
-{ return tab(A1, A2);
+{ PRED_LD
+
+  return tab(A1, A2 PASS_LD);
 }
 
 static
 PRED_IMPL("tab", 1, tab1, 0)
-{ return tab(0, A1);
+{ PRED_LD
+
+  return tab(0, A1 PASS_LD);
 }
 
 
@@ -2839,7 +2917,7 @@ static const opt_spec open4_options[] =
   { ATOM_lock,		 OPT_ATOM },
   { ATOM_wait,		 OPT_BOOL },
   { ATOM_encoding,	 OPT_ATOM },
-  { ATOM_bom,	 	 OPT_BOOL },
+  { ATOM_bom,		 OPT_BOOL },
   { NULL_ATOM,	         0 }
 };
 
@@ -2876,11 +2954,9 @@ openStream(term_t file, term_t mode, term_t options)
   { if ( mname == ATOM_write )
     { *h++ = 'w';
     } else if ( mname == ATOM_append )
-    { bom = FALSE;
-      *h++ = 'a';
+    { *h++ = 'a';
     } else if ( mname == ATOM_update )
-    { bom = FALSE;
-      *h++ = 'u';
+    { *h++ = 'u';
     } else if ( mname == ATOM_read )
     { *h++ = 'r';
     } else
@@ -2984,14 +3060,26 @@ openStream(term_t file, term_t mode, term_t options)
 	s->flags |= SIO_NOFEOF;
       else if ( eof_action == ATOM_error )
 	s->flags |= SIO_FEOF2ERR;
+      else
+      { term_t ex = PL_new_term_ref();
+	PL_put_atom(ex, eof_action);
+	PL_error(NULL, 0, NULL, ERR_DOMAIN, ATOM_eof_action, ex);
+	return NULL;
+      }
     }
   } else
   { if ( buffer != ATOM_full )
     { s->flags &= ~SIO_FBUF;
       if ( buffer == ATOM_line )
 	s->flags |= SIO_LBUF;
-      if ( buffer == ATOM_false )
+      else if ( buffer == ATOM_false )
 	s->flags |= SIO_NBUF;
+      else
+      { term_t ex = PL_new_term_ref();
+	PL_put_atom(ex, buffer);
+	PL_error(NULL, 0, NULL, ERR_DOMAIN, ATOM_buffer, ex);
+	return NULL;
+      }
     }
   }
 
@@ -3009,8 +3097,12 @@ openStream(term_t file, term_t mode, term_t options)
 	return NULL;
       }
     } else
-    { if ( SwriteBOM(s) < 0 )
-	goto bom_error;
+    { if ( mname == ATOM_write ||
+	   ( (mname == ATOM_append || mname == ATOM_update) &&
+	     Ssize(s) == 0 ) )
+      { if ( SwriteBOM(s) < 0 )
+	  goto bom_error;
+      }
     }
   }
 
@@ -3107,7 +3199,7 @@ ok:
 }
 
 int
-pl_seen()
+pl_seen(void)
 { GET_LD
   IOSTREAM *s = getStream(Scurin);
 
@@ -3230,19 +3322,28 @@ PRED_IMPL("told", 0, told, 0)
 
 static ssize_t
 Swrite_null(void *handle, char *buf, size_t size)
-{ return size;
+{ (void)handle;
+  (void)buf;
+
+  return size;
 }
 
 
 static ssize_t
 Sread_null(void *handle, char *buf, size_t size)
-{ return 0;
+{ (void)handle;
+  (void)buf;
+  (void)size;
+
+  return 0;
 }
 
 
 static long
 Sseek_null(void *handle, long offset, int whence)
-{ switch(whence)
+{ (void)handle;
+
+  switch(whence)
   { case SIO_SEEK_SET:
 	return offset;
     case SIO_SEEK_CUR:
@@ -3255,7 +3356,9 @@ Sseek_null(void *handle, long offset, int whence)
 
 static int
 Sclose_null(void *handle)
-{ return 0;
+{ (void)handle;
+
+  return 0;
 }
 
 
@@ -3269,7 +3372,7 @@ static const IOFUNCTIONS nullFunctions =
 
 static
 PRED_IMPL("open_null_stream", 1, open_null_stream, 0)
-{ int sflags = SIO_NBUF|SIO_RECORDPOS|SIO_OUTPUT;
+{ int sflags = SIO_NBUF|SIO_RECORDPOS|SIO_OUTPUT|SIO_TEXT;
   IOSTREAM *s = Snew((void *)NULL, sflags, (IOFUNCTIONS *)&nullFunctions);
 
   if ( s )
@@ -3368,8 +3471,10 @@ static int
 stream_file_name_propery(IOSTREAM *s, term_t prop ARG_LD)
 { atom_t name;
 
-  if ( (name = getStreamContext(s)->filename) )
-  { return PL_unify_atom(prop, name);
+  for(; s; s=s->downstream)
+  { if ( (name = getStreamContext(s)->filename) )
+    { return PL_unify_atom(prop, name);
+    }
   }
 
   return FALSE;
@@ -3399,13 +3504,17 @@ stream_mode_property(IOSTREAM *s, term_t prop ARG_LD)
 
 static int
 stream_input_prop(IOSTREAM *s ARG_LD)
-{ return (s->flags & SIO_INPUT) ? TRUE : FALSE;
+{ IGNORE_LD
+
+  return (s->flags & SIO_INPUT) ? TRUE : FALSE;
 }
 
 
 static int
 stream_output_prop(IOSTREAM *s ARG_LD)
-{ return (s->flags & SIO_OUTPUT) ? TRUE : FALSE;
+{ IGNORE_LD
+
+  return (s->flags & SIO_OUTPUT) ? TRUE : FALSE;
 }
 
 
@@ -3446,7 +3555,9 @@ stream_alias_prop(IOSTREAM *s, term_t prop ARG_LD)
 
 static int
 stream_position_prop(IOSTREAM *s, term_t prop ARG_LD)
-{ if ( s->position )
+{ IGNORE_LD
+
+  if ( s->position )
   { return PL_unify_term(prop,
 			 PL_FUNCTOR, FUNCTOR_stream_position4,
 			   PL_INT64, s->position->charno,
@@ -3462,8 +3573,7 @@ stream_position_prop(IOSTREAM *s, term_t prop ARG_LD)
 static int
 stream_end_of_stream_prop(IOSTREAM *s, term_t prop ARG_LD)
 { if ( s->flags & SIO_INPUT )
-  { GET_LD
-    atom_t val;
+  { atom_t val;
 
     if ( s->flags & SIO_FEOF2 )
       val = ATOM_past;
@@ -3512,7 +3622,7 @@ stream_reposition_prop(IOSTREAM *s, term_t prop ARG_LD)
     int fd = Sfileno(s);
     struct stat buf;
 
-    if ( fstat(fd, &buf) == 0 && S_ISREG(buf.st_mode) )
+    if ( fd != -1 && fstat(fd, &buf) == 0 && S_ISREG(buf.st_mode) )
       val = ATOM_true;
     else
       val = ATOM_false;
@@ -3528,7 +3638,9 @@ stream_reposition_prop(IOSTREAM *s, term_t prop ARG_LD)
 
 static int
 stream_close_on_abort_prop(IOSTREAM *s, term_t prop ARG_LD)
-{ return PL_unify_bool_ex(prop, !(s->flags & SIO_NOCLOSE));
+{ IGNORE_LD
+
+  return PL_unify_bool_ex(prop, !(s->flags & SIO_NOCLOSE));
 }
 
 
@@ -3551,7 +3663,9 @@ stream_file_no_prop(IOSTREAM *s, term_t prop ARG_LD)
 
 static int
 stream_tty_prop(IOSTREAM *s, term_t prop ARG_LD)
-{ if ( (s->flags & SIO_ISATTY) )
+{ IGNORE_LD
+
+  if ( (s->flags & SIO_ISATTY) )
     return PL_unify_bool_ex(prop, TRUE);
 
   return FALSE;
@@ -3560,7 +3674,9 @@ stream_tty_prop(IOSTREAM *s, term_t prop ARG_LD)
 
 static int
 stream_bom_prop(IOSTREAM *s, term_t prop ARG_LD)
-{ if ( (s->flags & SIO_BOM) )
+{ IGNORE_LD
+
+  if ( (s->flags & SIO_BOM) )
     return PL_unify_bool_ex(prop, TRUE);
 
   return FALSE;
@@ -3658,6 +3774,7 @@ stream_close_on_exec_prop(IOSTREAM *s, term_t prop ARG_LD)
 #else
    int fd_flags;
 #endif
+   IGNORE_LD
 
    if ( (fd = Sfileno(s)) < 0)
      return FALSE;
@@ -3683,21 +3800,21 @@ stream_close_on_exec_prop(IOSTREAM *s, term_t prop ARG_LD)
 
 typedef struct
 { functor_t functor;			/* functor of property */
-  int (*function)();			/* function to generate */
+  property_t function; /* function to generate */
 } sprop;
 
 
 static const sprop sprop_list [] =
 { { FUNCTOR_file_name1,	    stream_file_name_propery },
   { FUNCTOR_mode1,	    stream_mode_property },
-  { FUNCTOR_input0,	    stream_input_prop },
-  { FUNCTOR_output0,	    stream_output_prop },
+  { FUNCTOR_input0,	    (property_t)stream_input_prop },
+  { FUNCTOR_output0,	    (property_t)stream_output_prop },
   { FUNCTOR_alias1,	    stream_alias_prop },
   { FUNCTOR_position1,	    stream_position_prop },
   { FUNCTOR_end_of_stream1, stream_end_of_stream_prop },
   { FUNCTOR_eof_action1,    stream_eof_action_prop },
   { FUNCTOR_reposition1,    stream_reposition_prop },
-  { FUNCTOR_type1,    	    stream_type_prop },
+  { FUNCTOR_type1,	    stream_type_prop },
   { FUNCTOR_file_no1,	    stream_file_no_prop },
   { FUNCTOR_buffer1,	    stream_buffer_prop },
   { FUNCTOR_buffer_size1,   stream_buffer_size_prop },
@@ -3756,7 +3873,7 @@ PRED_IMPL("stream_property", 2, stream_property,
 			    ATOM_stream_property, property);
 	}
 
-	pe = allocHeap(sizeof(*pe));
+	pe = allocForeignState(sizeof(*pe));
 
 	pe->e = newTableEnum(streamContext);
 	pe->s = NULL;
@@ -3774,7 +3891,7 @@ PRED_IMPL("stream_property", 2, stream_property,
       { functor_t f;
 
 	if ( PL_is_variable(property) )	/* generate properties */
-	{ pe = allocHeap(sizeof(*pe));
+	{ pe = allocForeignState(sizeof(*pe));
 
 	  pe->e = NULL;
 	  pe->s = s;
@@ -3794,7 +3911,7 @@ PRED_IMPL("stream_property", 2, stream_property,
 
 	      switch(arityFunctor(f))
 	      { case 0:
-		  rval = (*p->function)(s PASS_LD);
+		  rval = (*(property0_t)p->function)(s PASS_LD);
 		  break;
 		case 1:
 		{ term_t a1 = PL_new_term_ref();
@@ -3832,7 +3949,7 @@ PRED_IMPL("stream_property", 2, stream_property,
       { if ( pe->e )
 	  freeTableEnum(pe->e);
 
-	freeHeap(pe, sizeof(*pe));
+	freeForeignState(pe, sizeof(*pe));
       }
       return TRUE;
     }
@@ -3848,7 +3965,7 @@ PRED_IMPL("stream_property", 2, stream_property,
     if ( pe->e )
       freeTableEnum(pe->e);
 
-    freeHeap(pe, sizeof(*pe));
+    freeForeignState(pe, sizeof(*pe));
     return FALSE;
   }
 
@@ -3869,7 +3986,7 @@ PRED_IMPL("stream_property", 2, stream_property,
 
 	  switch(arityFunctor(pe->p->functor))
 	  { case 0:
-	      rval = (*pe->p->function)(pe->s PASS_LD);
+	      rval = (*(property0_t)pe->p->function)(pe->s PASS_LD);
 	      break;
 	    case 1:
 	    { _PL_get_arg(1, property, a1);
@@ -3922,7 +4039,7 @@ PRED_IMPL("stream_property", 2, stream_property,
     { if ( pe->e )
 	freeTableEnum(pe->e);
 
-      freeHeap(pe, sizeof(*pe));
+      freeForeignState(pe, sizeof(*pe));
       return FALSE;
     }
   }
@@ -3952,10 +4069,10 @@ PRED_IMPL("is_stream", 1, is_stream, 0)
 
 
 static int
-flush_output(term_t out)
+flush_output(term_t out ARG_LD)
 { IOSTREAM *s;
 
-  if ( getOutputStream(out, &s) )
+  if ( getOutputStream(out, S_DONTCARE, &s) )
   { Sflush(s);
     return streamStatus(s);
   }
@@ -3965,12 +4082,16 @@ flush_output(term_t out)
 
 static
 PRED_IMPL("flush_output", 0, flush_output, PL_FA_ISO)
-{ return flush_output(0);
+{ PRED_LD
+
+  return flush_output(0 PASS_LD);
 }
 
 static
 PRED_IMPL("flush_output", 1, flush_output1, PL_FA_ISO)
-{ return flush_output(A1);
+{ PRED_LD
+
+  return flush_output(A1 PASS_LD);
 }
 
 
@@ -4122,7 +4243,7 @@ PRED_IMPL("set_input", 1, set_input, PL_FA_ISO)
 { PRED_LD
   IOSTREAM *s;
 
-  if ( getInputStream(A1, &s) )
+  if ( getInputStream(A1, S_DONTCARE, &s) )
   { Scurin = s;
     releaseStream(s);
     return TRUE;
@@ -4137,7 +4258,7 @@ PRED_IMPL("set_output", 1, set_output, PL_FA_ISO)
 { PRED_LD
   IOSTREAM *s;
 
-  if ( getOutputStream(A1, &s) )
+  if ( getOutputStream(A1, S_DONTCARE, &s) )
   { Scurout = s;
     releaseStream(s);
     return TRUE;
@@ -4163,7 +4284,8 @@ PRED_IMPL("current_output", 1, current_output, PL_FA_ISO)
 
 static
 PRED_IMPL("byte_count", 2, byte_count, 0)
-{ IOSTREAM *s;
+{ PRED_LD
+  IOSTREAM *s;
 
   if ( getStreamWithPosition(A1, &s) )
   { int64_t n = s->position->byteno;
@@ -4178,7 +4300,8 @@ PRED_IMPL("byte_count", 2, byte_count, 0)
 
 static
 PRED_IMPL("character_count", 2, character_count, 0)
-{ IOSTREAM *s;
+{ PRED_LD
+  IOSTREAM *s;
 
   if ( getStreamWithPosition(A1, &s) )
   { int64_t n = s->position->charno;
@@ -4239,7 +4362,7 @@ static int
 at_end_of_stream(term_t stream ARG_LD)
 { IOSTREAM *s;
 
-  if ( getInputStream(stream, &s) )
+  if ( getInputStream(stream, S_DONTCARE, &s) )
   { int rval = Sfeof(s);
 
     if ( rval < 0 )
@@ -4277,9 +4400,9 @@ peek(term_t stream, term_t chr, int how ARG_LD)
 { IOSTREAM *s;
   int c;
 
-  if ( !getInputStream(stream, &s) )
+  if ( !getInputStream(stream, how == PL_BYTE ? S_BINARY : S_TEXT, &s) )
     return FALSE;
-  if ( true(s, SIO_NBUF) || (s->bufsize && s->bufsize < MB_LEN_MAX) )
+  if ( true(s, SIO_NBUF) || (s->bufsize && s->bufsize < PL_MB_LEN_MAX) )
   { releaseStream(s);
     return PL_error(NULL, 0, "stream is unbuffered", ERR_PERMISSION,
 		    ATOM_peek, ATOM_stream, stream);
@@ -4366,20 +4489,23 @@ ssize_t
 Sread_user(void *handle, char *buf, size_t size)
 { GET_LD
   wrappedIO *wio = handle;
+  ssize_t rc;
 
   if ( LD->prompt.next && ttymode != TTY_RAW )
     PL_write_prompt(TRUE);
   else
     Sflush(Suser_output);
 
-  size = (*wio->wrapped_functions->read)(wio->wrapped_handle, buf, size);
-  if ( size == 0 )			/* end-of-file */
+  rc = (*wio->wrapped_functions->read)(wio->wrapped_handle, buf, size);
+  if ( rc == 0 )			/* end-of-file */
   { Sclearerr(Suser_input);
     LD->prompt.next = TRUE;
-  } else if ( size > 0 && buf[size-1] == '\n' )
+  } else if ( rc == 1 && buf[0] == 04 )
+  { rc = 0;				/* Map ^D to end-of-file */
+  } else if ( rc > 0 && buf[rc-1] == '\n' )
     LD->prompt.next = TRUE;
 
-  return size;
+  return rc;
 }
 
 
@@ -4427,6 +4553,7 @@ PRED_IMPL("set_prolog_IO", 3, set_prolog_IO, 0)
   IOSTREAM *in = NULL, *out = NULL, *error = NULL;
   int rval = FALSE;
   int wrapin = FALSE;
+  int i;
 
   if ( !term_stream_handle(A1, &in, SH_ERRORS|SH_ALIAS|SH_UNLOCKED PASS_LD) ||
        !term_stream_handle(A2, &out, SH_ERRORS|SH_ALIAS PASS_LD) )
@@ -4464,6 +4591,11 @@ PRED_IMPL("set_prolog_IO", 3, set_prolog_IO, 0)
     LD->prompt.next = TRUE;
   }
 
+  for(i=0; i<3; i++)
+  { LD->IO.streams[i]->position = &LD->IO.streams[0]->posbuf;
+    LD->IO.streams[i]->flags |= SIO_RECORDPOS;
+  }
+
   UNLOCK();
   rval = TRUE;
 
@@ -4488,7 +4620,7 @@ PRED_IMPL("$size_stream", 2, size_stream, 0)
   if ( !PL_get_stream_handle(A1, &s) )
     return FALSE;
 
-  rval = PL_unify_integer(A2, Ssize(s));
+  rval = PL_unify_int64(A2, Ssize(s));
   PL_release_stream(s);
 
   return rval;
@@ -4507,9 +4639,9 @@ copy_stream_data(term_t in, term_t out, term_t len ARG_LD)
   int c;
   int count = 0;
 
-  if ( !getInputStream(in, &i) )
+  if ( !getInputStream(in, S_DONTCARE, &i) )
     return FALSE;
-  if ( !getOutputStream(out, &o) )
+  if ( !getOutputStream(out, S_DONTCARE, &o) )
   { releaseStream(i);
     return FALSE;
   }
