@@ -19,7 +19,7 @@
 
     You should have received a copy of the GNU Lesser General Public
     License along with this library; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */
 
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
@@ -33,7 +33,7 @@ source should also use format() to produce error messages, etc.
 #include <ctype.h>
 
 static char *	formatNumber(bool split, int div, int radix,
-			     bool small, Number n, Buffer out);
+			     bool smll, Number n, Buffer out);
 static char *	formatFloat(int how, int arg, Number f, Buffer out);
 
 #define MAXRUBBER 100
@@ -53,9 +53,9 @@ typedef struct
   struct rubber rub[MAXRUBBER];
 } format_state;
 
-#define BUFSIZE 	1024
-#define DEFAULT 	(-1)
-#define SHIFT   	{ argc--; argv++; }
+#define BUFSIZE		1024
+#define DEFAULT		(-1)
+#define SHIFT		{ argc--; argv++; }
 #define NEED_ARG	{ if ( argc <= 0 ) \
 			  { FMT_ERROR("not enough arguments"); \
 			  } \
@@ -189,7 +189,8 @@ outtext(format_state *state, PL_chars_t *txt)
 #define format_predicates (GD->format.predicates)
 
 static int	update_column(int, Char);
-static bool	do_format(IOSTREAM *fd, PL_chars_t *fmt, int ac, term_t av);
+static bool	do_format(IOSTREAM *fd, PL_chars_t *fmt,
+			  int ac, term_t av, Module m);
 static void	distribute_rubber(struct rubber *, int, int);
 static int	emit_rubber(format_state *state);
 
@@ -201,15 +202,17 @@ static int	emit_rubber(format_state *state);
 word
 pl_format_predicate(term_t chr, term_t descr)
 { int c;
-  Procedure proc;
+  predicate_t proc = NULL;
   Symbol s;
+  int arity;
 
   if ( !PL_get_char_ex(chr, &c, FALSE) )
     fail;
 
   if ( !get_procedure(descr, &proc, 0, GP_CREATE) )
     fail;
-  if ( proc->definition->functor->arity == 0 )
+  PL_predicate_info(proc, NULL, &arity, NULL);
+  if ( arity == 0 )
     return PL_error(NULL, 0, "arity must be > 0", ERR_DOMAIN,
 		    PL_new_atom("format_predicate"),
 		    descr);
@@ -255,8 +258,7 @@ pl_current_format_predicate(term_t chr, term_t descr, control_t h)
   }
   while( (s=advanceTableEnum(e)) )
   { if ( PL_unify_integer(chr, (intptr_t)s->name) &&
-	 unify_definition(contextModule(LD->environment),
-			  descr, ((Procedure)s->value)->definition, 0, 0) )
+	 PL_unify_predicate(descr, (predicate_t)s->value, 0) )
     { PL_close_foreign_frame(fid);
       ForeignRedoPtr(e);
     }
@@ -271,7 +273,7 @@ pl_current_format_predicate(term_t chr, term_t descr, control_t h)
 
 
 static word
-format_impl(IOSTREAM *out, term_t format, term_t Args)
+format_impl(IOSTREAM *out, term_t format, term_t Args, Module m)
 { GET_LD
   term_t argv;
   int argc = 0;
@@ -306,7 +308,7 @@ format_impl(IOSTREAM *out, term_t format, term_t Args)
       break;
   }
 
-  rval = do_format(out, &fmt, argc, argv);
+  rval = do_format(out, &fmt, argc, argv, m);
   PL_free_text(&fmt);
   if ( !endCritical )
     return FALSE;
@@ -317,11 +319,17 @@ format_impl(IOSTREAM *out, term_t format, term_t Args)
 
 word
 pl_format3(term_t out, term_t format, term_t args)
-{ redir_context ctx;
+{ GET_LD
+  redir_context ctx;
   word rc;
+  Module m = NULL;
+  term_t list = PL_new_term_ref();
+
+  if ( !PL_strip_module(args, &m, list) )
+    return FALSE;
 
   if ( (rc=setupOutputRedirect(out, &ctx, FALSE)) )
-  { if ( (rc = format_impl(ctx.stream, format, args)) )
+  { if ( (rc = format_impl(ctx.stream, format, list, m)) )
       rc = closeOutputRedirect(&ctx);
     else
       discardOutputRedirect(&ctx);
@@ -356,7 +364,7 @@ get_chr_from_text(const PL_chars_t *t, int index)
 		********************************/
 
 static bool
-do_format(IOSTREAM *fd, PL_chars_t *fmt, int argc, term_t argv)
+do_format(IOSTREAM *fd, PL_chars_t *fmt, int argc, term_t argv, Module m)
 { GET_LD
   format_state state;			/* complete state */
   int tab_stop = 0;			/* padded tab stop */
@@ -419,20 +427,23 @@ do_format(IOSTREAM *fd, PL_chars_t *fmt, int argc, term_t argv)
 					/* Check for user defined format */
 	  if ( format_predicates &&
 	       (s = lookupHTable(format_predicates, (void*)((intptr_t)c))) )
-	  { Procedure proc = (Procedure) s->value;
-	    FunctorDef fdef = proc->definition->functor;
-	    term_t av = PL_new_term_refs(fdef->arity);
+	  { predicate_t proc = (predicate_t) s->value;
+	    int arity;
+	    term_t av;
 	    char buf[BUFSIZE];
 	    char *str = buf;
 	    size_t bufsize = BUFSIZE;
-	    unsigned int i;
+	    int i;
+
+	    PL_predicate_info(proc, NULL, &arity, NULL);
+	    av = PL_new_term_refs(arity);
 
 	    if ( arg == DEFAULT )
 	      PL_put_atom(av+0, ATOM_default);
 	    else
 	      PL_put_integer(av+0, arg);
 
-	    for(i=1; i<fdef->arity; i++)
+	    for(i=1; i < arity; i++)
 	    { NEED_ARG;
 	      PL_put_term(av+i, argv);
 	      SHIFT;
@@ -486,7 +497,10 @@ do_format(IOSTREAM *fd, PL_chars_t *fmt, int argc, term_t argv)
 	      case 'g':			/* shortest of 'f' and 'e' */
 	      case 'G':			/* shortest of 'f' and 'E' */
 		{ number n;
+		  union {
 		  tmp_buffer b;
+		    buffer b1;
+		  } u;
 
 		  NEED_ARG;
 		  if ( !valueExpression(argv, &n PASS_LD) )
@@ -498,11 +512,11 @@ do_format(IOSTREAM *fd, PL_chars_t *fmt, int argc, term_t argv)
 		  }
 		  SHIFT;
 
-		  initBuffer(&b);
-		  formatFloat(c, arg, &n, (Buffer)&b);
+		  initBuffer(&u.b);
+		  formatFloat(c, arg, &n, &u.b1);
 		  clearNumber(&n);
-		  outstring0(&state, baseBuffer(&b, char));
-		  discardBuffer(&b);
+		  outstring0(&state, baseBuffer(&u.b, char));
+		  discardBuffer(&u.b);
 		  here++;
 		  break;
 		}
@@ -680,7 +694,7 @@ do_format(IOSTREAM *fd, PL_chars_t *fmt, int argc, term_t argv)
 		  { FMT_ERROR("not enough arguments");
 		  }
 		  tellString(&str, &bufsize, ENC_UTF8);
-		  rval = callProlog(NULL, argv, PL_Q_CATCH_EXCEPTION, &ex);
+		  rval = callProlog(m, argv, PL_Q_CATCH_EXCEPTION, &ex);
 		  toldString();
 		  oututf8(&state, str, bufsize);
 		  if ( str != buf )
@@ -853,7 +867,7 @@ emit_rubber(format_state *state)
  ** Fri Aug 19 22:26:41 1988  jan@swivax.UUCP (Jan Wielemaker)  */
 
 static char *
-formatNumber(bool split, int div, int radix, bool small, Number i,
+formatNumber(bool split, int div, int radix, bool smll, Number i,
 	     Buffer out)
 { switch(i->type)
   { case V_INTEGER:
@@ -888,7 +902,7 @@ formatNumber(bool split, int div, int radix, bool small, Number i,
 	  }
 	  if ( split && before && (digits++ % 3) == 0 && digits != 1 )
 	    *--s = ',';
-	  *--s = digitName((int)(n % radix), small);
+	  *--s = digitName((int)(n % radix), smll);
 	  n /= radix;
 	}
 	if ( negative )
@@ -913,7 +927,7 @@ formatNumber(bool split, int div, int radix, bool small, Number i,
 	buf = tmp;
 
       mpz_get_str(buf, radix, i->value.mpz);
-      if ( !small && radix > 10 )
+      if ( !smll && radix > 10 )
       { char *s;
 
 	for(s=buf; *s; s++)
@@ -1008,7 +1022,8 @@ formatFloat(int how, int arg, Number f, Buffer out)
       while(written >= size)
       { size = written+1;
 
-	growBuffer(out, size);		/* reserve for -.e<null> */
+	if ( !growBuffer(out, size) )	/* reserve for -.e<null> */
+	  outOfCore();
 	written = gmp_snprintf(baseBuffer(out, char), size, tmp, mpf);
       }
       mpf_clear(mpf);
@@ -1029,7 +1044,8 @@ formatFloat(int how, int arg, Number f, Buffer out)
       while(written >= size)
       { size = written+1;
 
-	growBuffer(out, size);
+	if ( !growBuffer(out, size) )
+	  outOfCore();
 	written = snprintf(baseBuffer(out, char), size, tmp, f->value.f);
       }
       out->top = out->base + written;
