@@ -19,7 +19,7 @@
 
     You should have received a copy of the GNU Lesser General Public
     License along with this library; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */
 
 #define UNICODE 1
@@ -245,7 +245,8 @@ _xos_os_filenameW(const char *cname, wchar_t *osname, size_t len)
     q += 3;
   }
 
-  if ( q[0] == '/' || q[0] == '\\' )	/* deal with //host/share */
+  if ( (q[0] == '/' || q[0] == '\\') &&
+       (q[1] == '/' || q[1] == '\\') )	/* deal with //host/share */
   { if ( s+1 >= e )
     { errno = ENAMETOOLONG;
       return NULL;
