@@ -3,9 +3,10 @@
     Part of SWI-Prolog
 
     Author:        Jan Wielemaker
-    E-mail:        wielemak@science.uva.nl
+    E-mail:        J.Wielemaker@cs.vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (C): 1985-2007, University of Amsterdam
+    Copyright (C): 1985-2011, University of Amsterdam
+			      VU University Amsterdam
 
     This library is free software; you can redistribute it and/or
     modify it under the terms of the GNU Lesser General Public
@@ -19,12 +20,12 @@
 
     You should have received a copy of the GNU Lesser General Public
     License along with this library; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */
 
 /*#define O_DEBUG 1*/
 #include "pl-incl.h"
-#include "pl-ctype.h"
+#include "os/pl-ctype.h"
 
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 Implementation issues
@@ -227,10 +228,10 @@ PL_find_blob_type(const char *name)
 
 int
 PL_unregister_blob_type(PL_blob_t *type)
-{ unsigned int i;
+{ size_t index;
+  int i, last=FALSE;
   PL_blob_t **t;
   int discarded = 0;
-  Atom *ap;
 
   PL_LOCK(L_MISC);
   for(t = &GD->atoms.types; *t; t = &(*t)->next)
@@ -244,13 +245,19 @@ PL_unregister_blob_type(PL_blob_t *type)
   PL_register_blob_type(&unregistered_blob_atom);
 
   LOCK();
+  for(index=1, i=0; !last; i++)
+  { size_t upto = (size_t)2<<i;
+    Atom *b = GD->atoms.array.blocks[i];
 
-  ap = GD->atoms.array;
-  for(i=0; i < GD->atoms.count; i++, ap++ )
-  { Atom atom;
+    if ( upto >= GD->atoms.highest )
+    { upto = GD->atoms.highest;
+      last = TRUE;
+    }
 
-    if ( (atom = *ap) )
-    { if ( atom->type == type )
+    for(; index<upto; index++)
+    { Atom atom = b[index];
+
+      if ( atom && atom->type == type )
       { atom->type = &unregistered_blob_atom;
 
 	atom->name = "<discarded blob>";
@@ -279,6 +286,33 @@ static const ccharp atoms[] = {
 };
 #undef ATOM
 
+/* Note that we use PL_malloc_uncollectable() here because the pointer in
+   our block is not the real memory pointer.  Probably it is better to
+   have two pointers; one to the allocated memory and one with the
+   necessary offset.
+*/
+
+static void
+putAtomArray(size_t where, Atom a)
+{ int idx = MSB(where);
+
+  assert(where >= 0);
+
+  if ( !GD->atoms.array.blocks[idx] )
+  { PL_LOCK(L_MISC);
+    if ( !GD->atoms.array.blocks[idx] )
+    { size_t bs = (size_t)1<<idx;
+      Atom *newblock = PL_malloc_uncollectable(bs*sizeof(Atom));
+
+      GD->atoms.array.blocks[idx] = newblock-bs;
+    }
+    PL_UNLOCK(L_MISC);
+  }
+
+  GD->atoms.array.blocks[idx][where] = a;
+}
+
+
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 It might be wise to  provide  for   an  option  that does not reallocate
 atoms. In that case accessing a GC'ed   atom  causes a crash rather then
@@ -287,43 +321,50 @@ another atom.
 
 static void
 registerAtom(Atom a)
-{ size_t n = GD->atoms.count;
+{ size_t index;
 #ifdef O_ATOMGC				/* try to find a hole! */
-  Atom *ap = GD->atoms.array;
-  Atom *ep = ap+n;
-  Atom *p;
+  int i;
+  int last = FALSE;
 
-  for(p = &ap[GD->atoms.no_hole_before]; p < ep; p++)
-  { if ( *p == NULL )
-    { n = p - ap;
-      *p = a;
-      a->atom = (n<<LMASK_BITS)|TAG_ATOM;
-      if ( indexAtom(a->atom) != (uintptr_t)n )
-      {	/* TBD: user-level exception */
-	fatalError("Too many (%d) atoms", n);
+  for(index=GD->atoms.no_hole_before, i=MSB(index); !last; i++)
+  { size_t upto = (size_t)2<<i;
+    Atom *b = GD->atoms.array.blocks[i];
+
+    if ( upto >= GD->atoms.highest )
+    { upto = GD->atoms.highest;
+      last = TRUE;
+    }
+
+    for(; index<upto; index++)
+    { if ( b[index] == NULL )
+      { a->atom = (index<<LMASK_BITS)|TAG_ATOM;
+	b[index] = a;
+	GD->atoms.no_hole_before = index+1;
+
+	return;
       }
-      GD->atoms.no_hole_before = n+1;
-
-      return;
     }
   }
-  GD->atoms.no_hole_before = n+1;
+  GD->atoms.no_hole_before = index+1;
+#else
+  index = GD->atoms.highest;
 #endif /*O_ATOMGC*/
 
-  a->atom = (n<<LMASK_BITS)|TAG_ATOM;
-  if ( n >= GD->atoms.array_allocated )
-  { size_t newcount = GD->atoms.array_allocated * 2;
-    size_t newsize  = newcount*sizeof(Atom);
-    Atom *np = PL_malloc(newsize);
+  a->atom = (index<<LMASK_BITS)|TAG_ATOM;
+  if ( indexAtom(a->atom) != index )	/* TBD: user-level exception */
+    fatalError("Too many (%d) atoms", index);
+  putAtomArray(index, a);
+  GD->atoms.highest = index+1;
+}
 
-    memcpy(np, ap, newsize/2);
-    GD->atoms.array = np;
-    GD->atoms.array_allocated = newcount;
-    PL_free(ap);
-    ap = np;
+
+static size_t
+paddingBlob(PL_blob_t *type)
+{ if ( true(type, PL_BLOB_TEXT) )
+  { return true(type, PL_BLOB_WCHAR) ? sizeof(pl_wchar_t) : sizeof(char);
+  } else
+  { return 0;
   }
-  ap[n++] = a;
-  GD->atoms.count = n;
 }
 
 
@@ -346,51 +387,49 @@ treat the signal as bogus if agc has already been performed.
 
 word
 lookupBlob(const char *s, size_t length, PL_blob_t *type, int *new)
-{ GET_LD
-  unsigned int v0, v;
-  uintptr_t oldheap;
+{ unsigned int v0, v;
   Atom a;
 
   if ( !type->registered )		/* avoid deadlock */
     PL_register_blob_type(type);
-
-  startCritical;
-  LOCK();
   v0 = MurmurHashAligned2(s, length, MURMUR_SEED);
+
+  LOCK();
   v  = v0 & (atom_buckets-1);
-  DEBUG(0, lookups++);
+  DEBUG(MSG_HASH_STAT, lookups++);
 
   if ( true(type, PL_BLOB_UNIQUE) )
   { if ( false(type, PL_BLOB_NOCOPY) )
     { for(a = atomTable[v]; a; a = a->next)
-      { DEBUG(0, cmps++);
+      { DEBUG(MSG_HASH_STAT, cmps++);
 	if ( length == a->length &&
 	     type == a->type &&
 	     memcmp(s, a->name, length) == 0 )
 	{
 #ifdef O_ATOMGC
 	  if ( indexAtom(a->atom) >= GD->atoms.builtin )
-	    a->references++;
+	  { if ( a->references++ == 0 )
+	      GD->atoms.unregistered--;
+	  }
 #endif
           UNLOCK();
-	  endCritical;
 	  *new = FALSE;
 	  return a->atom;
 	}
       }
     } else
     { for(a = atomTable[v]; a; a = a->next)
-      { DEBUG(0, cmps++);
+      { DEBUG(MSG_HASH_STAT, cmps++);
 
 	if ( length == a->length &&
 	     type == a->type &&
 	     s == a->name )
 	{
 #ifdef O_ATOMGC
-	  a->references++;
+	  if ( a->references++ == 0 )
+	    GD->atoms.unregistered--;
 #endif
           UNLOCK();
-	  endCritical;
 	  *new = FALSE;
 	  return a->atom;
 	}
@@ -398,14 +437,22 @@ lookupBlob(const char *s, size_t length, PL_blob_t *type, int *new)
     }
   }
 
-  oldheap = GD->statistics.heap;
-  a = allocHeap(sizeof(struct atom));
+  a = allocHeapOrHalt(sizeof(struct atom));
   a->length = length;
   a->type = type;
   if ( false(type, PL_BLOB_NOCOPY) )
-  { a->name = allocHeap(length+1);
-    memcpy(a->name, s, length);
-    a->name[length] = EOS;
+  { if ( true(type, PL_BLOB_TEXT) )
+    { size_t pad = paddingBlob(type);
+
+      a->name = PL_malloc_atomic(length+pad);
+      memcpy(a->name, s, length);
+      memset(a->name+length, 0, pad);
+      GD->statistics.atom_string_space += length+pad;
+    } else
+    { a->name = PL_malloc(length);
+      memcpy(a->name, s, length);
+      GD->statistics.atom_string_space += length;
+    }
   } else
   { a->name = (char *)s;
   }
@@ -424,8 +471,8 @@ lookupBlob(const char *s, size_t length, PL_blob_t *type, int *new)
 
 #ifdef O_ATOMGC
   if ( GD->atoms.margin != 0 &&
-       GD->statistics.atoms >= GD->atoms.non_garbage + GD->atoms.margin )
-  { intptr_t x = GD->statistics.atoms - (GD->atoms.non_garbage + GD->atoms.margin);
+       GD->atoms.unregistered >= GD->atoms.non_garbage + GD->atoms.margin )
+  { intptr_t x = GD->atoms.unregistered - (GD->atoms.non_garbage + GD->atoms.margin);
 
     if ( x % 128 == 0 )			/* see (*) above */
       PL_raise(SIG_ATOM_GC);
@@ -435,14 +482,11 @@ lookupBlob(const char *s, size_t length, PL_blob_t *type, int *new)
   if ( atom_buckets * 2 < GD->statistics.atoms )
     rehashAtoms();
 
-  GD->statistics.atomspace += (GD->statistics.heap - oldheap);
-
   UNLOCK();
 
   *new = TRUE;
   if ( type->acquire )
     (*type->acquire)(a->atom);
-  endCritical;
 
   return a->atom;
 }
@@ -513,7 +557,7 @@ _PL_debug_atom_value(atom_t a)
     Sdprintf("*** No atom at index (#%d) ***", i);
     trap_gdb();
 
-    atom = allocHeap(sizeof(*atom));
+    atom = allocHeapOrHalt(sizeof(*atom));
     Ssprintf(buf, "***(#%d)***", i);
     atom->name = store_string(buf);
     atom->length = strlen(atom->name);
@@ -553,10 +597,10 @@ To be used after loading the program,   so we won't traverse the program
 atoms each pass.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
-void
-lockAtoms()
-{ GD->atoms.builtin     = GD->atoms.count;
-  GD->atoms.non_garbage = GD->atoms.builtin;
+static void
+lockAtoms(void)
+{ GD->atoms.builtin      = GD->atoms.highest;
+  GD->atoms.unregistered = 0;
 }
 
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
@@ -568,12 +612,12 @@ markAtom(atom_t a)
 { size_t i = indexAtom(a);
   Atom ap;
 
-  if ( i >= GD->atoms.count )
+  if ( i >= GD->atoms.highest )
     return;				/* not an atom */
   if ( i < GD->atoms.builtin )
     return;				/* locked range */
 
-  ap = GD->atoms.array[i];
+  ap = fetchAtomArray(i);
 
   if ( ap )
   {
@@ -593,21 +637,21 @@ atoms that should *not* be subject to AGC.   If we find one collected we
 know we trapped a bug.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
-static void
-destroyAtom(Atom *ap, uintptr_t mask ARG_LD)
+static int
+destroyAtom(Atom *ap, uintptr_t mask)
 { Atom a = *ap;
   Atom *ap2 = &atomTable[a->hash_value & mask];
 
   if ( a->type->release )
   { if ( !(*a->type->release)(a->atom) )
-      return;
+      return FALSE;
   } else if ( GD->atoms.gc_hook )
   { if ( !(*GD->atoms.gc_hook)(a->atom) )
-      return;				/* foreign hooks says `no' */
+      return FALSE;				/* foreign hooks says `no' */
   }
 
 #if 0
-  if ( strncmp(a->name, "xxxx", 4) == 0 ) 	/* (*) see above */
+  if ( strncmp(a->name, "xxxx", 4) == 0 )	/* (*) see above */
   { Sdprintf("Deleting %s\n", a->name);
     assert(0);
   }
@@ -618,56 +662,72 @@ destroyAtom(Atom *ap, uintptr_t mask ARG_LD)
     Sfprintf(atomLogFd, "Deleted `%s'\n", a->name);
 #endif
 
-  for( ; ; ap2 = &(*ap2)->next )
-  { assert(*ap2);			/* MT: TBD: failed a few times!? */
+  if ( true(a->type, PL_BLOB_UNIQUE) )
+  { for( ; ; ap2 = &(*ap2)->next )
+    { assert(*ap2);		/* MT: TBD: failed a few times!? */
 
-    if ( *ap2 == a )
-    { *ap2 = a->next;
-      break;
+      if ( *ap2 == a )
+      { *ap2 = a->next;
+        break;
+      }
     }
   }
 
   *ap = NULL;			/* delete from index array */
-  GD->atoms.collected++;
-  GD->statistics.atoms--;
-
   if ( false(a->type, PL_BLOB_NOCOPY) )
-    freeHeap(a->name, a->length+1);
+  { size_t slen = a->length+paddingBlob(a->type);
+    GD->statistics.atom_string_space -= slen;
+    GD->statistics.atom_string_space_freed += slen;
+    PL_free(a->name);
+  }
   freeHeap(a, sizeof(*a));
+
+  return TRUE;
 }
 
 
-static void
+static size_t
 collectAtoms(void)
-{ GET_LD
-  Atom *ap0 = GD->atoms.array;
-  Atom *ap  = ap0 + GD->atoms.builtin;
-  Atom *ep  = ap0 + GD->atoms.count;
-  int hole_seen = FALSE;
-  uintptr_t mask = atom_buckets-1;
+{ int hole_seen = FALSE;
+  size_t reclaimed = 0;
+  size_t index;
+  int i, last=FALSE;
 
-  ap--;
-  while(++ap < ep)
-  { Atom a = *ap;
+  for(index=GD->atoms.builtin, i=MSB(index); !last; i++)
+  { size_t upto = (size_t)2<<i;
+    Atom *b = GD->atoms.array.blocks[i];
 
-    if ( !a )
-    { if ( !hole_seen )
-      { hole_seen = TRUE;
-	GD->atoms.no_hole_before = ap-ap0;
-      }
-      continue;
+    if ( upto >= GD->atoms.highest )
+    { upto = GD->atoms.highest;
+      last = TRUE;
     }
 
-    if ( a->references == 0 )
-    { destroyAtom(ap, mask PASS_LD);
-      if ( !hole_seen && *ap == NULL )
-      { hole_seen = TRUE;
-	GD->atoms.no_hole_before = ap-ap0;
+    for(; index<upto; index++)
+    { Atom a = b[index];
+
+      if ( !a )
+      { if ( !hole_seen )
+	{ hole_seen = TRUE;
+	  GD->atoms.no_hole_before = index;
+	}
+	continue;
+      }
+
+      if ( a->references == 0 )
+      { if ( destroyAtom(&b[index], atom_buckets-1) )
+	{ reclaimed++;
+	  if ( !hole_seen )
+	  { hole_seen = TRUE;
+	    GD->atoms.no_hole_before = index;
+	  }
+	}
+      } else
+      { a->references &= ~ATOM_MARKED_REFERENCE;
       }
-    } else
-    { a->references &= ~ATOM_MARKED_REFERENCE;
     }
   }
+
+  return reclaimed;
 }
 
 
@@ -680,11 +740,11 @@ GC. These issues are described with enterGC() in pl-gc.c.
 word
 pl_garbage_collect_atoms()
 { GET_LD
-  intptr_t oldheap, freed;
   int64_t oldcollected;
   int verbose;
   double t;
   sigset_t set;
+  size_t reclaimed;
 
   PL_LOCK(L_GC);
   if ( gc_status.blocked )		/* Tricky things; avoid problems. */
@@ -693,8 +753,8 @@ pl_garbage_collect_atoms()
   }
 
 #ifdef O_PLMT
-  if ( GD->gc.active ) 			/* GC in progress: delay */
-  { DEBUG(2, Sdprintf("GC active; delaying AGC\n"));
+  if ( GD->gc.active )			/* GC in progress: delay */
+  { DEBUG(MSG_AGC, Sdprintf("GC active; delaying AGC\n"));
     GD->gc.agc_waiting = TRUE;
     PL_UNLOCK(L_GC);
     succeed;
@@ -707,10 +767,11 @@ pl_garbage_collect_atoms()
   {
 #ifdef O_DEBUG_ATOMGC
 /*
+    access_level_t old;
     Sdprintf("Starting ATOM-GC.  Stack:\n");
-    systemMode(TRUE);
-    backTrace(NULL, 5);
-    systemMode(FALSE);
+    old = setAccessLevel(ACCESS_LEVEL_SYSTEM);
+    backTrace(5);
+    setAccessLevel(old);
 */
 #endif
     printMessage(ATOM_informational,
@@ -729,15 +790,14 @@ pl_garbage_collect_atoms()
   forThreadLocalData(markAtomsOnStacks, 0);
 #endif
   oldcollected = GD->atoms.collected;
-  oldheap = GD->statistics.heap;
-  collectAtoms();
-  GD->atoms.non_garbage = GD->statistics.atoms;
+  reclaimed = collectAtoms();
+  GD->atoms.collected += reclaimed;
+  GD->statistics.atoms -= reclaimed;
+  GD->atoms.unregistered -= reclaimed;
+  GD->atoms.non_garbage = GD->atoms.unregistered;
   t = CpuTime(CPU_USER) - t;
   GD->atoms.gc_time += t;
   GD->atoms.gc++;
-  freed = oldheap - GD->statistics.heap;
-  GD->statistics.atomspacefreed += freed;
-  GD->statistics.atomspace -= freed;
   unblockSignals(&set);
   UNLOCK();
   PL_UNLOCK(L_AGC);
@@ -787,8 +847,9 @@ PL_register_atom(atom_t a)
   { Atom p;
 
     LOCK();
-    p = GD->atoms.array[index];
-    p->references++;
+    p = fetchAtomArray(index);
+    if ( p->references++ == 0 )
+      GD->atoms.unregistered--;
     UNLOCK();
   }
 #endif
@@ -804,8 +865,9 @@ PL_unregister_atom(atom_t a)
   { Atom p;
 
     LOCK();
-    p = GD->atoms.array[index];
-    p->references--;
+    p = fetchAtomArray(index);
+    if ( --p->references == 0 )
+      GD->atoms.unregistered++;
     if ( p->references == (unsigned)-1 )
     { Sdprintf("OOPS: -1 references to '%s'\n", p->name);
       trap_gdb();
@@ -823,34 +885,43 @@ PL_unregister_atom(atom_t a)
 		 *******************************/
 
 static void
-rehashAtoms()
-{ GET_LD
-  Atom *oldtab   = atomTable;
+rehashAtoms(void)
+{ Atom *oldtab   = atomTable;
   int   oldbucks = atom_buckets;
-  size_t mx = GD->atoms.count;
   uintptr_t mask;
-  Atom *ap, *ep;
+  size_t index;
+  int i, last=FALSE;
 
-  startCritical;
   atom_buckets *= 2;
   mask = atom_buckets-1;
-  atomTable = allocHeap(atom_buckets * sizeof(Atom));
+  atomTable = allocHeapOrHalt(atom_buckets * sizeof(Atom));
   memset(atomTable, 0, atom_buckets * sizeof(Atom));
 
-  DEBUG(0, Sdprintf("rehashing atoms (%d --> %d)\n", oldbucks, atom_buckets));
+  DEBUG(MSG_HASH_STAT,
+	Sdprintf("rehashing atoms (%d --> %d)\n", oldbucks, atom_buckets));
 
-  for(ap = GD->atoms.array, ep = ap+mx;
-      ap < ep;
-      ap++)
-  { Atom a = *ap;
-    size_t v = a->hash_value & mask;
+  for(index=1, i=0; !last; i++)
+  { size_t upto = (size_t)2<<i;
+    Atom *b = GD->atoms.array.blocks[i];
 
-    a->next = atomTable[v];
-    atomTable[v] = a;
+    if ( upto >= GD->atoms.highest )
+    { upto = GD->atoms.highest;
+      last = TRUE;
+    }
+
+    for(; index<upto; index++)
+    { Atom a = b[index];
+
+      if ( a )
+      { size_t v = a->hash_value & mask;
+
+	a->next = atomTable[v];
+	atomTable[v] = a;
+      }
+    }
   }
 
   freeHeap(oldtab, oldbucks * sizeof(Atom));
-  endCritical;
 }
 
 
@@ -870,15 +941,15 @@ pl_atom_hashstat(term_t idx, term_t n)
 
 
 static void
-registerBuiltinAtoms()
-{ GET_LD
-  int size = sizeof(atoms)/sizeof(char *) - 1;
-  Atom a = allocHeap(size * sizeof(struct atom));
+registerBuiltinAtoms(void)
+{ int size = sizeof(atoms)/sizeof(char *) - 1;
+  Atom a;
   const ccharp *s;
 
+  GD->atoms.builtin_array = PL_malloc(size * sizeof(struct atom));
   GD->statistics.atoms = size;
 
-  for(s = atoms; *s; s++, a++)
+  for(s = atoms, a = GD->atoms.builtin_array; *s; s++, a++)
   { size_t len = strlen(*s);
     unsigned int v0 = MurmurHashAligned2(*s, len, MURMUR_SEED);
     unsigned int v = v0 & (atom_buckets-1);
@@ -901,8 +972,11 @@ registerBuiltinAtoms()
 
 #if O_DEBUG
 static void
-exitAtoms(int status, void *arg)
-{ Sdprintf("hashstat: %d lookupAtom() calls used %d strcmp() calls\n",
+exitAtoms(int status, void *context)
+{ (void)status;
+  (void)context;
+
+  Sdprintf("hashstat: %d lookupAtom() calls used %d strcmp() calls\n",
 	   lookups, cmps);
 }
 #endif
@@ -911,14 +985,13 @@ exitAtoms(int status, void *arg)
 void
 initAtoms(void)
 { LOCK();
-  if ( !atomTable )
-  { GET_LD
-    atom_buckets = ATOMHASHSIZE;
-    atomTable = allocHeap(atom_buckets * sizeof(Atom));
-
+  if ( !atomTable )			/* Atom hash table */
+  { atom_buckets = ATOMHASHSIZE;
+    atomTable = allocHeapOrHalt(atom_buckets * sizeof(Atom));
     memset(atomTable, 0, atom_buckets * sizeof(Atom));
-    GD->atoms.array_allocated = 4096;
-    GD->atoms.array = PL_malloc(GD->atoms.array_allocated*sizeof(Atom));
+
+    GD->atoms.highest = 1;
+    GD->atoms.no_hole_before = 1;
     registerBuiltinAtoms();
 #ifdef O_ATOMGC
     GD->atoms.margin = 10000;
@@ -926,21 +999,81 @@ initAtoms(void)
 #endif
     PL_register_blob_type(&text_atom);
 
-    DEBUG(0, PL_on_halt(exitAtoms, NULL));
+    DEBUG(MSG_HASH_STAT, PL_on_halt(exitAtoms, NULL));
   }
   UNLOCK();
 }
 
 
+/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+cleanupAtoms() is called at shutdown. There are three possible scenarios
+these days: (1) do not cleanup at  all, (2) cleanup the main structures,
+leaving the rest to GC or (3) cleanup the whole thing.
+- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
+
 void
 cleanupAtoms(void)
-{ PL_free(GD->atoms.array);
+{ int i;
+  int builtin_count = sizeof(atoms)/sizeof(char *) - 1;
+  Atom builtin_start = GD->atoms.builtin_array;
+  Atom builtin_end   = builtin_start+builtin_count;
+  Atom *ap0;
+
+  for(i=0; (ap0=GD->atoms.array.blocks[i]); i++)
+  { size_t bs = (size_t)1<<i;
+    size_t upto = (size_t)2<<i;
+    Atom *ap, *ep;
+
+    ap0 += bs;
+    ap = ap0;
+    ep = ap+bs;
+    if ( upto > GD->atoms.highest )
+      ep -= upto-GD->atoms.highest;
+
+    for(; ap<ep; ap++)
+    { if ( *ap )
+      { Atom a = *ap;
+
+	if ( !(a>=builtin_start && a<builtin_end) )
+	{ if ( a->type->release )
+	    (*a->type->release)(a->atom);
+	  else if ( GD->atoms.gc_hook )
+	    (*GD->atoms.gc_hook)(a->atom);
+
+	  if ( false(a->type, PL_BLOB_NOCOPY) )
+	    PL_free(a->name);
+	  freeHeap(a, sizeof(*a));
+	}
+      }
+    }
+
+    GD->atoms.array.blocks[i] = NULL;
+    PL_free(ap0);
+  }
+
+  PL_free(builtin_start);
+
+  for(i=0; i<256; i++)			/* char-code -> char-atom map */
+  { atom_t *p;
+
+    if ( (p=GD->atoms.for_code[i]) )
+    { GD->atoms.for_code[i] = NULL;
+      PL_free(p);
+    }
+  }
+
+  if ( atomTable )
+  { freeHeap(atomTable, atom_buckets * sizeof(Atom));
+    atomTable = NULL;
+  }
 }
 
 
 static word
-current_blob(term_t a, term_t type, frg_code call, intptr_t i ARG_LD)
+current_blob(term_t a, term_t type, frg_code call, intptr_t state ARG_LD)
 { atom_t type_name = 0;
+  size_t index;
+  int i, last=0;
 
   switch( call )
   { case FRG_FIRST_CALL:
@@ -955,12 +1088,13 @@ current_blob(term_t a, term_t type, frg_code call, intptr_t i ARG_LD)
 	succeed;
       }
       if ( !PL_is_variable(a) )
-	return PL_error(NULL, 0, NULL, ERR_DOMAIN, ATOM_atom, a);
+	return FALSE;
 
-      i = 0;
+      index = 1;
       break;
     }
     case FRG_REDO:
+      index = state;
       break;
     case FRG_CUTTED:
     default:
@@ -973,30 +1107,38 @@ current_blob(term_t a, term_t type, frg_code call, intptr_t i ARG_LD)
       fail;
   }
 
-  for( ; i < (int)GD->atoms.count; i++ )
-  { Atom atom;
+  PL_LOCK(L_AGC);
+  for(i=MSB(index); !last; i++)
+  { size_t upto = (size_t)2<<i;
+    Atom *b = GD->atoms.array.blocks[i];
 
-    if ( (atom = GD->atoms.array[i]) )
-    {
-#ifdef LIFE_GC				/* see linkVal__LD() check */
-      if ( atom->atom == ATOM_garbage_collected )
-	continue;
-#endif
+    if ( upto >= GD->atoms.highest )
+    { upto = GD->atoms.highest;
+      last = TRUE;
+    }
 
-      if ( type )
-      { if ( type_name && type_name != atom->type->atom_name )
+    for(; index<upto; index++)
+    { Atom atom = b[index];
+
+      if ( atom &&
+	   atom->atom != ATOM_garbage_collected )
+      { if ( type )
+	{ if ( type_name && type_name != atom->type->atom_name )
+	    continue;
+
+	  PL_unify_atom(type, atom->type->atom_name);
+	} else if ( false(atom->type, PL_BLOB_TEXT) )
 	  continue;
 
-	PL_unify_atom(type, atom->type->atom_name);
-      } else if ( false(atom->type, PL_BLOB_TEXT) )
-	continue;
-
-      PL_unify_atom(a, atom->atom);
-      ForeignRedoInt(i+1);
+	PL_unify_atom(a, atom->atom);
+	PL_UNLOCK(L_AGC);
+	ForeignRedoInt(index+1);
+      }
     }
   }
+  PL_UNLOCK(L_AGC);
 
-  fail;
+  return FALSE;
 }
 
 
@@ -1015,6 +1157,24 @@ PRED_IMPL("current_atom", 1, current_atom, PL_FA_NONDETERMINISTIC)
   return current_blob(A1, 0, CTX_CNTRL, CTX_INT PASS_LD);
 }
 
+
+/** blob(@Term, ?Type) is semidet
+
+Type-test for a blob.
+*/
+
+static
+PRED_IMPL("blob", 2, blob, 0)
+{ PRED_LD
+  PL_blob_t *bt;
+
+  if ( PL_is_blob(A1, &bt) )
+    return PL_unify_atom(A2, bt->atom_name);
+
+  return FALSE;
+}
+
+
 static
 PRED_IMPL("$atom_references", 2, atom_references, 0)
 { PRED_LD
@@ -1038,15 +1198,15 @@ PRED_IMPL("$atom_references", 2, atom_references, 0)
 #define stringMatch(m)	((m)->name->name)
 
 typedef struct match
-{ Atom	name;
-  int	length;
+{ Atom		name;
+  size_t	length;
 } *Match;
 
 
 static bool
-allAlpha(const char *s)
-{ for( ; *s; s++)
-  { if ( !isAlpha(*s) )
+allAlpha(const char *s, size_t len)
+{ for( ; --len>=0; s++)
+  { if ( !*s || !isAlpha(*s) )
       fail;
   }
   succeed;
@@ -1055,28 +1215,39 @@ allAlpha(const char *s)
 
 static int
 extendAtom(char *prefix, bool *unique, char *common)
-{ intptr_t i, mx = GD->atoms.count;
-  Atom a;
+{ size_t index;
+  int i, last=FALSE;
   bool first = TRUE;
-  int lp = (int) strlen(prefix);
-  Atom *ap = GD->atoms.array;
+  size_t lp = strlen(prefix);
 
   *unique = TRUE;
 
-  for(i=0; i<mx; i++, ap++)
-  { if ( (a=*ap) && strprefix(a->name, prefix) )
-    { if ( strlen(a->name) >= LINESIZ )
-	continue;
-      if ( first == TRUE )
-      { strcpy(common, a->name+lp);
-	first = FALSE;
-      } else
-      { char *s = common;
-	char *q = a->name+lp;
-	while( *s && *s == *q )
-	  s++, q++;
-	*s = EOS;
-	*unique = FALSE;
+  for(index=1, i=0; !last; i++)
+  { size_t upto = (size_t)2<<i;
+    Atom *b = GD->atoms.array.blocks[i];
+
+    if ( upto >= GD->atoms.highest )
+    { upto = GD->atoms.highest;
+      last = TRUE;
+    }
+
+    for(; index<upto; index++)
+    { Atom a = b[index];
+
+      if ( a && a->type == &text_atom &&
+	   strprefix(a->name, prefix) &&
+	   strlen(a->name) < LINESIZ )
+      { if ( first == TRUE )
+	{ strcpy(common, a->name+lp);
+	  first = FALSE;
+	} else
+	{ char *s = common;
+	  char *q = a->name+lp;
+	  while( *s && *s == *q )
+	    s++, q++;
+	  *s = EOS;
+	  *unique = FALSE;
+	}
       }
     }
   }
@@ -1092,7 +1263,7 @@ pl_complete_atom(term_t prefix, term_t common, term_t unique)
   char buf[LINESIZ];
   char cmm[LINESIZ];
 
-  if ( !PL_get_chars_ex(prefix, &p, CVT_ALL) )
+  if ( !PL_get_chars(prefix, &p, CVT_ALL|CVT_EXCEPTION) )
     fail;
   strcpy(buf, p);
 
@@ -1118,29 +1289,37 @@ compareMatch(const void *m1, const void *m2)
 
 static bool
 extend_alternatives(char *prefix, struct match *altv, int *altn)
-{ intptr_t i, mx = GD->atoms.count;
-  Atom a, *ap;
-  char *as;
-  int l;
+{ size_t index;
+  int i, last=FALSE;
 
   *altn = 0;
-  ap = GD->atoms.array;
-  for(i=0; i<mx; i++, ap++)
-  { if ( !(a=*ap) )
-      continue;
+  for(index=1, i=0; !last; i++)
+  { size_t upto = (size_t)2<<i;
+    Atom *b = GD->atoms.array.blocks[i];
 
-    as = a->name;
-    if ( strprefix(as, prefix) &&
-	 allAlpha(as) &&
-	 (l = (int)strlen(as)) < ALT_SIZ )
-    { Match m = &altv[(*altn)++];
-      m->name = a;
-      m->length = l;
-      if ( *altn > ALT_MAX )
-	break;
+    if ( upto >= GD->atoms.highest )
+    { upto = GD->atoms.highest;
+      last = TRUE;
+    }
+
+    for(; index<upto; index++)
+    { Atom a = b[index];
+
+      if ( a && a->type == &text_atom &&
+	   strprefix(a->name, prefix) &&
+	   a->length < ALT_SIZ &&
+	   allAlpha(a->name, a->length) )
+      { Match m = &altv[(*altn)++];
+
+	m->name = a;
+	m->length = a->length;
+	if ( *altn > ALT_MAX )
+	  goto out;
+      }
     }
   }
 
+out:
   qsort(altv, *altn, sizeof(struct match), compareMatch);
 
   succeed;
@@ -1158,7 +1337,7 @@ pl_atom_completions(term_t prefix, term_t alternatives)
   term_t alts = PL_copy_term_ref(alternatives);
   term_t head = PL_new_term_ref();
 
-  if ( !PL_get_chars_ex(prefix, &p, CVT_ALL) )
+  if ( !PL_get_chars(prefix, &p, CVT_ALL|CVT_EXCEPTION) )
     fail;
   strcpy(buf, p);
 
@@ -1189,7 +1368,7 @@ thread.
 static pthread_key_t key;
 #endif
 
-#define is_signalled() (LD && LD->pending_signals != 0)
+#define is_signalled() unlikely(LD && LD->signal.pending != 0)
 
 static int
 alnum_text(PL_chars_t *txt)
@@ -1224,8 +1403,8 @@ alnum_text(PL_chars_t *txt)
 static int
 atom_generator(PL_chars_t *prefix, PL_chars_t *hit, int state)
 { GET_LD
-  intptr_t i, mx = GD->atoms.count;
-  Atom *ap;
+  size_t index;
+  int i, last=FALSE;
 
 #ifdef O_PLMT
   if ( !key )
@@ -1233,37 +1412,43 @@ atom_generator(PL_chars_t *prefix, PL_chars_t *hit, int state)
 #endif
 
   if ( !state )
-    i = 0;
-  else
+  { index = 1;
+  } else
   {
 #ifdef O_PLMT
-    i = (intptr_t)pthread_getspecific(key);
+    index = (size_t)pthread_getspecific(key);
 #else
-    i = LD->atoms.generator;
+    index = LD->atoms.generator;
 #endif
   }
 
-  for(ap=&GD->atoms.array[i]; i<mx; i++, ap++)
-  { Atom a;
+  for(i=MSB(index); !last; i++)
+  { size_t upto = (size_t)2<<i;
+    Atom *b = GD->atoms.array.blocks[i];
 
-    if ( !(a = *ap) )
-      continue;
+    if ( upto >= GD->atoms.highest )
+    { upto = GD->atoms.highest;
+      last = TRUE;
+    }
 
-    if ( is_signalled() )		/* Notably allow windows version */
-      PL_handle_signals();		/* to break out on ^C */
+    for(; index<upto; index++)
+    { Atom a = b[index];
 
-    if ( get_atom_ptr_text(a, hit) &&
+      if ( is_signalled() )		/* Notably allow windows version */
+	PL_handle_signals();		/* to break out on ^C */
+
+      if ( a && get_atom_ptr_text(a, hit) &&
 	 hit->length < ALT_SIZ &&
 	 PL_cmp_text(prefix, 0, hit, 0, prefix->length) == 0 &&
 	 alnum_text(hit) )
-    {
+      {
 #ifdef O_PLMT
-      pthread_setspecific(key, (void *)(i+1));
+        pthread_setspecific(key, (void *)(index+1));
 #else
-      LD->atoms.generator = i+1;
+        LD->atoms.generator = index+1;
 #endif
-
-      return TRUE;
+        return TRUE;
+      }
     }
   }
 
@@ -1331,5 +1516,6 @@ PL_atom_generator_w(const pl_wchar_t *prefix,
 BeginPredDefs(atom)
   PRED_DEF("current_blob",  2, current_blob, PL_FA_NONDETERMINISTIC)
   PRED_DEF("current_atom", 1, current_atom, PL_FA_NONDETERMINISTIC)
+  PRED_DEF("blob", 2, blob, 0)
   PRED_DEF("$atom_references", 2, atom_references, 0)
 EndPredDefs
