@@ -19,7 +19,7 @@
 
     You should have received a copy of the GNU Lesser General Public
     License along with this library; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */
 
 
@@ -81,6 +81,7 @@ rewrite_callable(atom_t *expected, term_t actual)
 int
 PL_error(const char *pred, int arity, const char *msg, PL_error_code id, ...)
 { GET_LD
+  char msgbuf[50];
   Definition caller;
   term_t except, formal, swi;
   va_list args;
@@ -88,6 +89,9 @@ PL_error(const char *pred, int arity, const char *msg, PL_error_code id, ...)
   fid_t fid;
   int rc;
 
+  if ( exception_term )			/* do not overrule older exception */
+    return FALSE;
+
   if ( environment_frame )
     caller = environment_frame->predicate;
   else
@@ -122,12 +126,11 @@ PL_error(const char *pred, int arity, const char *msg, PL_error_code id, ...)
     case ERR_UNINSTANTIATION:
     { int argn = va_arg(args, int);
       term_t bound = va_arg(args, term_t);
-      char buf[50];
 
       if ( !msg && argn > 0 )
-      { Ssprintf(buf, "%d-%s argument",
+      { Ssprintf(msgbuf, "%d-%s argument",
 		 argn, argn == 1 ? "st" : argn == 2 ? "nd" : "th");
-	msg = buf;
+	msg = msgbuf;
       }
 
       rc = PL_unify_term(formal,
@@ -139,6 +142,7 @@ PL_error(const char *pred, int arity, const char *msg, PL_error_code id, ...)
     { atom_t expected = va_arg(args, atom_t);
       term_t actual   = va_arg(args, term_t);
 
+    case_type_error:
       if ( expected == ATOM_callable )
 	rewrite_callable(&expected, actual);
       if ( PL_is_variable(actual) && expected != ATOM_variable )
@@ -149,6 +153,16 @@ PL_error(const char *pred, int arity, const char *msg, PL_error_code id, ...)
 			   PL_ATOM, expected,
 			   PL_TERM, actual);
       break;
+    case ERR_PTR_TYPE:			/* atom_t, Word */
+      { Word ptr;
+
+	expected = va_arg(args, atom_t);
+	ptr      = va_arg(args, Word);
+	actual   = PL_new_term_ref();
+
+	*valTermRef(actual) = *ptr;
+	goto case_type_error;
+      }
     }
     case ERR_CHARS_TYPE:		/* ERR_INSTANTIATION if var(actual) */
     { const char *expected = va_arg(args, const char*);
@@ -175,6 +189,18 @@ PL_error(const char *pred, int arity, const char *msg, PL_error_code id, ...)
 			    PL_TERM, actual));
       break;
     }
+    case ERR_AR_DOMAIN:
+    { atom_t domain = va_arg(args, atom_t);
+      Number num    = va_arg(args, Number);
+      term_t actual = PL_new_term_ref();
+
+      rc = (_PL_put_number(actual, num) &&
+	    PL_unify_term(formal,
+			  PL_FUNCTOR, FUNCTOR_domain_error2,
+			    PL_ATOM, domain,
+			    PL_TERM, actual));
+      break;
+    }
     case ERR_AR_UNDEF:
     { rc = PL_unify_term(formal,
 			 PL_FUNCTOR, FUNCTOR_evaluation_error1,
@@ -334,8 +360,10 @@ PL_error(const char *pred, int arity, const char *msg, PL_error_code id, ...)
 
       rc = PL_unify_term(formal,
 			 PL_FUNCTOR, FUNCTOR_occurs_check2,
-			   PL_TERM, wordToTermRef(p1),
-			   PL_TERM, wordToTermRef(p2));
+			   PL_TERM, pushWordAsTermRef(p1),
+			   PL_TERM, pushWordAsTermRef(p2));
+      popTermRef();
+      popTermRef();
 
       break;
     }
@@ -597,25 +625,15 @@ PL_error(const char *pred, int arity, const char *msg, PL_error_code id, ...)
 }
 
 
-char *
-tostr(char *buf, const char *fmt, ...)
-{ va_list args;
-
-  va_start(args, fmt);
-  Svsprintf(buf, fmt, args);
-  va_end(args);
-
-  return buf;
-}
-
-
 		 /*******************************
 		 *	  TYPICAL ERRORS	*
 		 *******************************/
 
 int
 PL_instantiation_error(term_t actual)
-{ return PL_error(NULL, 0, NULL, ERR_INSTANTIATION);
+{ (void)actual;
+
+  return PL_error(NULL, 0, NULL, ERR_INSTANTIATION);
 }
 
 int
@@ -667,6 +685,23 @@ PL_permission_error(const char *op, const char *type, term_t obj)
 }
 
 
+int
+PL_resource_error(const char *resource)
+{ atom_t r = PL_new_atom(resource);
+  int rc = PL_error(NULL, 0, NULL, ERR_RESOURCE, r);
+
+  PL_unregister_atom(r);
+
+  return rc;
+}
+
+
+int
+PL_no_memory(void)
+{ return PL_error(NULL, 0, NULL, ERR_RESOURCE, ATOM_memory);
+}
+
+
 		 /*******************************
 		 *	PRINTING MESSAGES	*
 		 *******************************/
@@ -736,6 +771,16 @@ PL_get_atom_ex__LD(term_t t, atom_t *a ARG_LD)
 }
 
 
+#undef PL_get_atom_ex
+int
+PL_get_atom_ex(term_t t, atom_t *a)
+{ GET_LD
+
+  return PL_get_atom_ex__LD(t, a PASS_LD);
+}
+#define PL_get_atom_ex(t, a)	PL_get_atom_ex__LD(t, a PASS_LD)
+
+
 int
 PL_get_integer_ex(term_t t, int *i)
 { GET_LD
