@@ -19,7 +19,7 @@
 
     You should have received a copy of the GNU Lesser General Public
     License along with this library; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */
 
 #ifndef PL_GLOBAL_H_INCLUDED
@@ -85,6 +85,7 @@ struct PL_global_data
   cleanup_status cleaning;		/* Inside PL_cleanup() */
   int		bootsession;		/* -b boot compilation */
   int		debug_level;		/* Maintenance debugging: 0..9 */
+  struct bit_vector *debug_topics;	/* debug topics enabled */
   void *	resourceDB;		/* program resource database */
 
 #ifdef HAVE_SIGNAL
@@ -104,24 +105,24 @@ struct PL_global_data
   } cmdline;
 
   struct
-  { char *	executable;		/* Running executable */
+  { char *	CWDdir;
+    size_t	CWDlen;
+    char *	executable;		/* Running executable */
 #ifdef __WINDOWS__
     char *	module;			/* argv[0] module passed */
 #endif
   } paths;
 
   struct
-  { size_t	heap;			/* heap in use */
-    size_t	atoms;			/* No. of atoms defined */
-    size_t	atomspace;		/* # bytes used to store atoms */
+  { size_t	atoms;			/* No. of atoms defined */
+    size_t	atom_string_space;	/* # bytes used to store atoms */
+    size_t	atom_string_space_freed;/* # bytes in freed atoms */
     size_t	stack_space;		/* # bytes on stacks */
-#ifdef O_ATOMGC
-    size_t	atomspacefreed;		/* Freed atom-space */
-#endif
     int		functors;		/* No. of functors defined */
     int		predicates;		/* No. of predicates defined */
     int		modules;		/* No. of modules in the system */
-    intptr_t	codes;			/* No. of byte codes generated */
+    size_t	clauses;		/* No. clauses */
+    size_t	codes;			/* No. of VM codes generated */
     double	start_time;		/* When Prolog was started */
     double	user_cputime;		/* User CPU time (whole process) */
     double	system_cputime;		/* Kernel CPU time (whole process) */
@@ -153,11 +154,11 @@ struct PL_global_data
   } arith;
 
   struct
-  { Atom *	array;			/* index --> atom */
-    size_t	count;			/* elements in array */
-    size_t	array_allocated;	/* allocated size of array */
+  { size_t	highest;		/* Highest atom index */
+    atom_array	array;
     unsigned int buckets;		/* # buckets in char * --> atom */
     Atom *	table;			/* hash-table */
+    Atom	builtin_array;		/* Builtin atoms */
     int		lookups;		/* # atom lookups */
     int		cmps;			/* # string compares for lookup */
 #ifdef O_ATOMGC
@@ -218,7 +219,8 @@ struct PL_global_data
   } prolog_flag;
 
   struct
-  { buffer	array;			/* index --> functor */
+  { size_t	highest;		/* Next index to handout */
+    functor_array array;		/* index --> functor */
     int		buckets;		/* # buckets in atom --> functor */
     FunctorDef* table;			/* hash-table */
   } functors;
@@ -303,7 +305,6 @@ struct PL_global_data
     int			enabled;	/* threads are enabled */
     Table		mutexTable;	/* Name --> mutex table */
     int			mutex_next_id;	/* next id for anonymous mutexes */
-    struct pl_mutex*	MUTEX_load;	/* The $load mutex */
 #ifdef __WINDOWS__
     HINSTANCE		instance;	/* Win32 process instance */
 #endif
@@ -333,12 +334,9 @@ struct PL_local_data
 #endif
   pl_stacks_t   stacks;			/* Prolog runtime stacks */
   uintptr_t	bases[STG_MASK+1];	/* area base addresses */
-#if defined(O_SECURE) || defined(SECURE_GC)
-  unsigned int  incr_seed;		/* Seed for random stack increments */
-#endif
   int		alerted;		/* Special mode. See updateAlerted() */
   int		critical;		/* heap is being modified */
-  abort_type	aborted;		/* !ABORT_NONE: abort in Critical */
+  int		break_level;		/* current break level */
   Stack		outofstack;		/* thread is out of stack */
   int		trim_stack_requested;	/* perform a trim-stack */
 #ifdef O_PLMT
@@ -366,6 +364,12 @@ struct PL_local_data
     jmp_buf	context;		/* context of longjmp() */
   } pipe;
 
+  struct
+  { char       *getstr_buffer;		/* getString() buffer */
+    size_t	getstr_buffer_size;	/* size of getstr_buffer */
+    struct wic_state *current_state;	/* qlf-creation state */
+  } qlf;
+
   struct
   { atom_t	current;		/* current global prompt */
     atom_t	first;			/* how to prompt first line */
@@ -481,12 +485,14 @@ struct PL_local_data
 #endif
 
   struct
-  { char *	_CWDdir;
-    size_t	_CWDlen;
+  {
 #ifdef __BEOS__
     status_t	dl_error;		/* dlopen() emulation in pl-beos.c */
 #endif
     int		rand_initialised;	/* have we initialised random? */
+#ifdef O_DDE
+    unsigned	dde_instance;		/* Actually DWORD */
+#endif
   } os;
 
   struct
@@ -494,6 +500,7 @@ struct PL_local_data
     pl_features_t mask;			/* Masked access to booleans */
     int		  write_attributes;	/* how to write attvars? */
     occurs_check_t occurs_check;	/* Unify and occurs check */
+    access_level_t access_level;	/* Current access level */
   } prolog_flag;
 
   struct
@@ -535,38 +542,6 @@ struct PL_local_data
   } depth_info;
 #endif
 
-  struct
-  { intptr_t _total_marked;		/* # marked global cells */
-    intptr_t _trailcells_deleted;	/* # garbage trailcells */
-    intptr_t _relocation_chains;	/* # relocation chains (debugging) */
-    intptr_t _relocation_cells;		/* # relocation cells */
-    intptr_t _relocated_cells;		/* # relocated cells */
-    intptr_t _needs_relocation;		/* # cells that need relocation */
-    intptr_t _local_marked;		/* # marked local -> global ptrs */
-    intptr_t _marks_swept;		/* # marks swept */
-    intptr_t _marks_unswept;		/* # marks swept */
-    intptr_t _alien_relocations;	/* # alien_into_relocation_chain() */
-    intptr_t _local_frames;		/* frame count for debugging */
-    intptr_t _choice_count;		/* choice-point count for debugging */
-    int  *_start_map;			/* bitmap with legal global starts */
-    sigset_t saved_sigmask;		/* Saved signal mask */
-#if defined(O_SECURE) || defined(SECURE_GC)
-    intptr_t _trailtops_marked;		/* # marked trailtops */
-    Word *_mark_base;			/* Array of marked cells addresses */
-    Word *_mark_top;			/* Top of this array */
-    Table _check_table;			/* relocation address table */
-    Table _local_table;			/* marked local variables */
-    int  _relocated_check;		/* Verify relocated addresses? */
-#endif
-    int64_t inferences;			/* #inferences at last GC */
-
-    pl_gc_status_t	status;		/* Garbage collection status */
-#ifdef O_CALL_RESIDUE
-    int			marked_attvars;	/* do not GC attvars */
-#endif
-    int active;				/* GC is running in this thread */
-  } gc;
-
   pl_shift_status_t shift_status;	/* Stack shifter status */
   pl_debugstatus_t _debugstatus;	/* status of the debugger */
   struct btrace *btrace_store;		/* C-backtraces */
@@ -586,6 +561,41 @@ struct PL_local_data
 
   struct alloc_pool alloc_pool;		/* Thread allocation pool */
 #endif
+
+  struct
+  { intptr_t _total_marked;		/* # marked global cells */
+    intptr_t _trailcells_deleted;	/* # garbage trailcells */
+    intptr_t _relocation_chains;	/* # relocation chains (debugging) */
+    intptr_t _relocation_cells;		/* # relocation cells */
+    intptr_t _relocated_cells;		/* # relocated cells */
+    intptr_t _needs_relocation;		/* # cells that need relocation */
+    intptr_t _local_marked;		/* # marked local -> global ptrs */
+    intptr_t _marks_swept;		/* # marks swept */
+    intptr_t _marks_unswept;		/* # marks swept */
+    intptr_t _alien_relocations;	/* # alien_into_relocation_chain() */
+    intptr_t _local_frames;		/* frame count for debugging */
+    intptr_t _choice_count;		/* choice-point count for debugging */
+    int  *_start_map;			/* bitmap with legal global starts */
+    sigset_t saved_sigmask;		/* Saved signal mask */
+    int64_t inferences;			/* #inferences at last GC */
+    pl_gc_status_t	status;		/* Garbage collection status */
+#ifdef O_CALL_RESIDUE
+    int			marked_attvars;	/* do not GC attvars */
+#endif
+    int active;				/* GC is running in this thread */
+					/* These must be at the end to be */
+					/* able to define O_DEBUG in only */
+					/* some modules */
+#if defined(O_DEBUG) || defined(SECURE_GC)
+    intptr_t _trailtops_marked;		/* # marked trailtops */
+    Word *_mark_base;			/* Array of marked cells addresses */
+    Word *_mark_top;			/* Top of this array */
+    Table _check_table;			/* relocation address table */
+    Table _local_table;			/* marked local variables */
+    int  _relocated_check;		/* Verify relocated addresses? */
+    unsigned int incr_seed;		/* Seed for random stack increments */
+#endif
+  } gc;
 };
 
 GLOBAL PL_global_data_t PL_global_data;
