@@ -19,7 +19,7 @@
 
     You should have received a copy of the GNU Lesser General Public
     License along with this library; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */
 
 /*#define O_DEBUG 1*/
@@ -77,7 +77,8 @@ mp_alloc(size_t bytes)
   if ( TOO_BIG_GMP(bytes) ||
        !(mem = malloc(sizeof(mp_mem_header)+bytes)) )
   { gmp_too_big();
-    abortProlog(ABORT_THROW);
+    abortProlog();
+    PL_rethrow();
     return NULL;			/* make compiler happy */
   }
 
@@ -111,7 +112,8 @@ mp_realloc(void *ptr, size_t oldsize, size_t newsize)
   if ( TOO_BIG_GMP(newsize) ||
        !(newmem = realloc(oldmem, sizeof(mp_mem_header)+newsize)) )
   { gmp_too_big();
-    abortProlog(ABORT_THROW);
+    abortProlog();
+    PL_rethrow();
     return NULL;			/* make compiler happy */
   }
 
@@ -300,7 +302,8 @@ addMPZToBuffer(Buffer b, mpz_t mpz)
   long hdrsize;
   size_t count;
 
-  growBuffer(b, size+4);
+  if ( !growBuffer(b, size+4) )
+    outOfCore();
   if ( mpz_sgn(mpz) < 0 )
     hdrsize = -(long)size;
   else
@@ -750,9 +753,8 @@ put_number(Word at, Number n, int flags ARG_LD)
 
 
 int
-PL_unify_number(term_t t, Number n)
-{ GET_LD
-  Word p = valTermRef(t);
+PL_unify_number__LD(term_t t, Number n ARG_LD)
+{ Word p = valTermRef(t);
 
   deRef(p);
 
@@ -1053,7 +1055,8 @@ PL_get_mpq(term_t t, mpq_t mpq)
 
 int
 PL_unify_mpz(term_t t, mpz_t mpz)
-{ number n;
+{ GET_LD
+  number n;
   int rc;
 
   n.type = V_MPZ;
@@ -1069,7 +1072,8 @@ PL_unify_mpz(term_t t, mpz_t mpz)
 
 int
 PL_unify_mpq(term_t t, mpq_t mpq)
-{ number n;
+{ GET_LD
+  number n;
   int rc;
 
   n.type = V_MPQ;
