@@ -1,4 +1,4 @@
-/*  $Id: pl-list.c,v 1.27 2007/05/07 19:29:31 jan Exp $
+/*  $Id$
 
     Part of SWI-Prolog
 
@@ -19,7 +19,7 @@
 
     You should have received a copy of the GNU Lesser General Public
     License along with this library; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */
 
 #include "pl-incl.h"
@@ -57,32 +57,60 @@ pl_length(term_t list, term_t l)
 
   if ( PL_is_variable(l) )
   { intptr_t n;
-  
+
     if ( (n=lengthList(list, FALSE)) >= 0 )
       return PL_unify_integer(l, n);
 
     fail;			/* both variables: generate in Prolog */
   }
-  
+
 				/* large integer; we cannot represent anyway */
   if ( PL_is_integer(l) )
+  { number i;
+    Word p =  valTermRef(l);
+
+    deRef(p);
+    get_integer(*p, &i);
+    if ( ar_sign_i(&i) < 0 )
+      fail;
+
     return outOfStack((Stack)&LD->stacks.global, STACK_OVERFLOW_RAISE);
+  }
 
   return PL_error("length", 2, NULL, ERR_TYPE, ATOM_integer, l);
-}  
+}
 
 
-word
-pl_memberchk(term_t e, term_t list)
+static
+PRED_IMPL("memberchk", 2, memberchk, 0)
 { GET_LD
   term_t h = PL_new_term_ref();
-  term_t l = PL_copy_term_ref(list);
+  term_t l = PL_copy_term_ref(A2);
+  fid_t fid;
+
+  if ( !(fid=PL_open_foreign_frame()) )
+    return FALSE;
 
   for(;;)
-  { TRY(PL_unify_list(l, h, l));
-      
-    if ( PL_unify(e, h) )
-      succeed;
+  { if ( !PL_unify_list(l, h, l) )
+    { PL_close_foreign_frame(fid);
+      fail;
+    }
+
+    if ( PL_unify(A1, h) )
+    { term_t ex = 0;
+
+      if ( foreignWakeup(&ex PASS_LD) )
+      { PL_close_foreign_frame(fid);
+	succeed;
+      } else
+      { if ( ex )
+	  return PL_raise_exception(ex);
+	PL_rewind_foreign_frame(fid);
+      }
+    } else
+    { PL_rewind_foreign_frame(fid);
+    }
   }
 }
 
@@ -137,7 +165,7 @@ just plain not using qsort().
 
 /*  ITEM	The type of an individual item.
     COMPARE	Compares two items given their addresses (allows ITEM to be
-    		large and avoids pass by copy).  Return <0, =0, or >0.
+		large and avoids pass by copy).  Return <0, =0, or >0.
     COMPARE_KEY	Compares the keys of two items given the addresses of the
 		entire items.
     FREE	Frees a List_Record including its ITEM.
@@ -148,6 +176,7 @@ typedef struct
   Word key;
 } ITEM;
 
+					/* TBD: handle CMP_ERROR */
 #ifndef COMPARE
 #define COMPARE(x,y) compareStandard((x)->term, (y)->term, FALSE PASS_LD)
 #endif
@@ -187,7 +216,7 @@ nat_sort(list data, int remove_dups, int compare_keys)
   remove_dups = !remove_dups;		/* 0 -> do, 1 -> don't */
   while ((p = data) != NIL)
   { /* pick up a run from the front of data, setting */
-    /* p = (pointer to beginning of run), data = (rest of data) */ 
+    /* p = (pointer to beginning of run), data = (rest of data) */
     if ((q = p->next) != NIL)
     { compare(c, p, q);
 
@@ -208,7 +237,7 @@ nat_sort(list data, int remove_dups, int compare_keys)
 	if (c > 0)
 	  break;
 	if (c == remove_dups)
-	{ s = r->next;		
+	{ s = r->next;
 	  FREE(r);
 	  r = s;
 	} else
@@ -287,32 +316,36 @@ will be.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
 static int
-prolog_list_to_sort_list(term_t t, int key, list *lp, Word *end)
+prolog_list_to_sort_list(term_t t, int remove_dups, int key,
+			 list *lp, Word *end)
 { GET_LD
-  int n = lengthList(t, TRUE);
-  Word l;
+  Word l, tail;
   list p;
-  intptr_t minfree;
+  intptr_t len;
+  int rc;
 
-  if ( n < 0 )
-    fail;				/* not a proper list */
-  minfree = sizeof(word)*n*3;
+  l = valTermRef(t);
+  len = skip_list(l, &tail PASS_LD);
+  if ( !(isNil(*tail) ||			/* proper list */
+	 (isList(*tail) && remove_dups)) )	/* sort/2 on cyclic list */
+  {
+    if ( isVar(*tail) )
+      return PL_error(NULL, 0, NULL, ERR_INSTANTIATION);
+    else
+      return PL_error(NULL, 0, NULL, ERR_TYPE, ATOM_list, t);
+  }
 
-#ifdef O_SHIFT_STACKS
-  if ( roomStack(global) < minfree )
-  { if ( !growStacks(NULL, NULL, NULL, 0, minfree, 0) )
-      fail;
+  if ( !hasGlobalSpace(len*3) )
+  { if ( (rc=ensureGlobalSpace(len*3, ALLOW_GC)) != TRUE )
+      return raiseStackOverflow(rc);
+    l = valTermRef(t);			/* may be shifted */
+    deRef(l);
   }
-#else
-  requireStack(global, minfree);
-#endif
 
   p = (list)gTop;
   *lp = p;
-  l = valTermRef(t);
-  deRef(l);
 
-  for(;;)
+  while(len-- > 0)
   { p->item.term = HeadList(l);
     deRef(p->item.term);
     if ( key )
@@ -322,23 +355,25 @@ prolog_list_to_sort_list(term_t t, int key, list *lp, Word *end)
       { p->item.key = argTermP(w, 0);
 	deRef(p->item.key);
       } else
-      { term_t t = wordToTermRef(p->item.term);
-
-	return PL_error("keysort", 2, NULL,
-			ERR_TYPE, ATOM_key_value_pair, t);
+      { PL_error("keysort", 2, NULL, ERR_TYPE,
+		 ATOM_pair, pushWordAsTermRef(p->item.term));
+	popTermRef();
+	return FALSE;
       }
     }
     l = TailList(l);
     deRef(l);
-    if ( isList(*l) )
-    { p->next = p+1;
+    if ( len > 0 )
+    { assert(isList(*l));
+      p->next = p+1;
       p++;
-    } else
-    { p->next = NULL;
-      *end = (Word)(p+1);
-      succeed;
     }
   }
+
+  p->next = NULL;
+  *end = (Word)(p+1);
+
+  succeed;
 }
 
 
@@ -373,11 +408,11 @@ pl_nat_sort(term_t in, term_t out, int remove_dups, int compare_keys ARG_LD)
 { if ( PL_get_nil(in) )
     return PL_unify_atom(out, ATOM_nil);
   else
-  { list l;
+  { list l = 0;
     term_t tmp = PL_new_term_ref();
     Word top = NULL;
 
-    if ( prolog_list_to_sort_list(in, compare_keys, &l, &top) )
+    if ( prolog_list_to_sort_list(in, remove_dups, compare_keys, &l, &top) )
     { l = nat_sort(l, remove_dups, compare_keys);
       put_sort_list(tmp, l);
       gTop = top;
@@ -420,7 +455,8 @@ PRED_IMPL("keysort", 2, keysort, 0)
 
 BeginPredDefs(list)
   PRED_DEF("is_list", 1, is_list, 0)
-  PRED_DEF("sort", 2, sort, 0)
+  PRED_DEF("memberchk", 2, memberchk, 0)
+  PRED_DEF("sort", 2, sort, PL_FA_ISO)
   PRED_DEF("msort", 2, msort, 0)
-  PRED_DEF("keysort", 2, keysort, 0)
+  PRED_DEF("keysort", 2, keysort, PL_FA_ISO)
 EndPredDefs
