@@ -19,7 +19,7 @@
 
     You should have received a copy of the GNU Lesser General Public
     License along with this library; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */
 
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
@@ -27,7 +27,7 @@ Get the ball rolling.  The main task of  this  module  is  command  line
 option  parsing,  initialisation  and  handling  of errors and warnings.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
-#define O_DEBUG 1
+/* #define O_DEBUG 1 */
 
 #include "rc/rc.h"
 #include "pl-incl.h"
@@ -117,8 +117,7 @@ longopt(const char *opt, int argc, const char **argv)
 
 static int
 opt_append(opt_list **l, char *s)
-{ GET_LD
-  opt_list *n = allocHeap(sizeof(*n));
+{ opt_list *n = allocHeapOrHalt(sizeof(*n));
 
   n->opt_val = s;
   n->next = NULL;
@@ -345,7 +344,54 @@ initPaths(int argc, const char **argv)
 
 
 static void
-initDefaults()
+cleanupStringP(char **loc)
+{ char *s;
+
+  if ( (s=*loc) )
+  { *loc = NULL;
+    remove_string(s);
+  }
+}
+
+static void
+cleanupOptListP(opt_list **listp)
+{ opt_list *l, *n;
+
+  if ( (l=*listp) )
+  { *listp = NULL;
+
+    for(; l; l=n)
+    { n = l->next;
+
+      remove_string(l->opt_val);
+      freeHeap(l, sizeof(*l));
+    }
+  }
+}
+
+
+static void
+cleanupPaths(void)
+{ cleanupStringP(&GD->paths.executable);
+  cleanupStringP(&systemDefaults.home);
+  cleanupStringP(&systemDefaults.startup);
+  cleanupStringP(&GD->options.systemInitFile);
+  cleanupStringP(&GD->options.compileOut);
+  cleanupStringP(&GD->options.goal);
+  cleanupStringP(&GD->options.topLevel);
+  cleanupStringP(&GD->options.initFile);
+  cleanupStringP(&GD->options.saveclass);
+  cleanupStringP(&GD->os.myhome);
+#ifdef __WINDOWS__
+  cleanupStringP(&GD->paths.module);
+#endif
+
+  cleanupOptListP(&GD->options.scriptFiles);
+}
+
+
+static void
+initDefaults(void)
 { GET_LD
 
   systemDefaults.arch	     = PLARCH;
@@ -501,7 +547,7 @@ parseCommandLineOptions(int argc0, char **argv, int *compile)
     while(*s)
     { switch(*s)
       { case 'd':	if (argc > 1)
-			{ GD->debug_level = atoi(argv[1]);
+			{ prolog_debug_from_string(argv[1], TRUE);
 			  argc--, argv++;
 			} else
 			  return -1;
@@ -709,8 +755,7 @@ stack-parameters.
 
 static void
 script_argv(int argc, char **argv)
-{ GET_LD
-  FILE *fd;
+{ FILE *fd;
   int i;
 
   DEBUG(1,
@@ -743,7 +788,7 @@ script_argv(int argc, char **argv)
 
 	    DEBUG(1, Sdprintf("Got script %s\n", argv[i+1]));
 
-	    av = allocHeap(sizeof(char*)*(argc+2));
+	    av = allocHeapOrHalt(sizeof(char*)*(argc+2));
 	    for(j=0; j<=i+1; j++)
 	      av[j] = argv[j];
 	    av[j] = "--";
@@ -806,7 +851,7 @@ script_argv(int argc, char **argv)
 	    continue;
 	}
 #endif
-	av[an] = allocHeap(o-start+1);
+	av[an] = allocHeapOrHalt(o-start+1);
 	strncpy(av[an], start, o-start);
 	av[an][o-start] = EOS;
 	if ( ++an >= MAXARGV )
@@ -827,7 +872,7 @@ script_argv(int argc, char **argv)
       av[an++] = argv[i];
     GD->cmdline.argc = an;
     av[an++] = NULL;
-    GD->cmdline.argv = allocHeap(sizeof(char *) * an);
+    GD->cmdline.argv = allocHeapOrHalt(sizeof(char *) * an);
     memcpy(GD->cmdline.argv, av, sizeof(char *) * an);
 
     fclose(fd);
@@ -853,8 +898,6 @@ PL_initialise(int argc, char **argv)
   initPrologThreads();			/* initialise thread system */
   SinitStreams();			/* before anything else */
 
-  GD->debug_level = 0;			/* 1-9: debug, also -d <level> */
-
   script_argv(argc, argv);		/* hande #! arguments */
   argc = GD->cmdline.argc;
   argv = GD->cmdline.argv;
@@ -920,7 +963,8 @@ PL_initialise(int argc, char **argv)
     argv += done;
   }
 
-  setupProlog();
+  if ( !setupProlog() )
+    return FALSE;
 #ifdef O_PLMT
   aliasThread(PL_thread_self(), ATOM_main);
   enableThreads(TRUE);
@@ -928,7 +972,7 @@ PL_initialise(int argc, char **argv)
   PL_set_prolog_flag("resource_database", PL_ATOM|FF_READONLY, rcpath);
   initialiseForeign(GD->cmdline.argc, /* PL_initialise_hook() functions */
 		    GD->cmdline.argv);
-  systemMode(TRUE);
+  setAccessLevel(ACCESS_LEVEL_SYSTEM);
 
   if ( GD->bootsession )
   { IOSTREAM *s = SopenRC(GD->resourceDB, "$state", "$prolog", RC_WRONLY);
@@ -975,7 +1019,7 @@ PL_initialise(int argc, char **argv)
   debugstatus.styleCheck = (LONGATOM_CHECK|
 			    SINGLETON_CHECK|
 			    DISCONTIGUOUS_STYLE);
-  systemMode(FALSE);
+  setAccessLevel(ACCESS_LEVEL_USER);
   GD->initialised = TRUE;
   registerForeignLicenses();
 
@@ -1023,6 +1067,9 @@ usage()
     "    --nodebug        Omit generation of debug info\n",
     "    --quiet          Quiet operation (also -q)\n",
     "    --home=DIR       Use DIR as SWI-Prolog home\n",
+#ifdef O_DEBUG
+    "    -d level|topic   Enable maintenance debugging\n",
+#endif
     NULL
   };
   const cline *lp = lines;
@@ -1031,7 +1078,7 @@ usage()
   if ( GD->cmdline.argc > 0 )
     prog = BaseName(GD->cmdline.argv[0]);
   else
-    prog = "pl";
+    prog = "swipl";
 
   for(lp = lines; *lp; lp++)
     Sfprintf(Serror, *lp, prog);
@@ -1169,23 +1216,41 @@ struct on_halt
 void
 PL_on_halt(halt_function f, void *arg)
 { if ( !GD->os.halting )
-  { GET_LD
-    OnHalt h = allocHeap(sizeof(struct on_halt));
+  { OnHalt h = allocHeapOrHalt(sizeof(struct on_halt));
 
     h->function = f;
     h->argument = arg;
-    startCritical;
     h->next = GD->os.on_halt_list;
     GD->os.on_halt_list = h;
-    endCritical;
+  }
+}
+
+static void
+run_on_halt(int rval)
+{ OnHalt h, next;
+
+  for(h = GD->os.on_halt_list; h; h=next)
+  { next = h->next;
+    (*h->function)(rval, h->argument);
+    freeHeap(h, sizeof(*h));
   }
 }
 
 
+/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+Cleanup Prolog. The reclaim_memory  argument   says  whether  the system
+tries to reclaim memory. This  is   true  when called from PL_cleanup().
+Ideally, this would allow for  restarting   the  system  without loosing
+memory. In practice, this is hard,   especially if foreign libraries are
+loaded.
+
+When called from PL_halt(),  reclaim_memory   is  FALSE, unless compiled
+with -DGC_DEBUG.
+- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
+
 int
-PL_cleanup(int rval)
+cleanupProlog(int rval, int reclaim_memory)
 { GET_LD
-  OnHalt h;
   int rc = TRUE;
 
   if ( GD->cleaning != CLN_NORMAL )
@@ -1212,7 +1277,7 @@ PL_cleanup(int rval)
   GD->cleaning = CLN_PROLOG;
 
   qlfCleanup();				/* remove errornous .qlf files */
-  if ( GD->initialised && !LD->aborted )
+  if ( GD->initialised )
   { fid_t cid = PL_open_foreign_frame();
     predicate_t proc = PL_predicate("$run_at_halt", 0, "system");
 
@@ -1222,12 +1287,7 @@ PL_cleanup(int rval)
   }
 
   GD->cleaning = CLN_FOREIGN;
-
-					/* run PL_on_halt() hooks */
-  if ( !LD->aborted )
-  { for(h = GD->os.on_halt_list; h; h = h->next)
-      (*h->function)(rval, h->argument);
-  }
+  run_on_halt(rval);
 
 #ifdef __WINDOWS__
   if ( rval != 0 && !hasConsole() )
@@ -1239,7 +1299,7 @@ PL_cleanup(int rval)
 
   GD->cleaning = CLN_SHARED;
 
-  if ( GD->initialised && !LD->aborted )
+  if ( GD->initialised )
   { fid_t cid = PL_open_foreign_frame();
     predicate_t proc = PL_predicate("unload_all_foreign_libraries", 0,
 				    "shlib");
@@ -1258,38 +1318,56 @@ PL_cleanup(int rval)
   }
 
   cleanupSignals();
-  freeStacks(PASS_LD1);
 #ifdef HAVE_DMALLOC_H
   dmalloc_verify(0);
 #endif
-  freePrologLocalData(LD);
-  cleanupSourceFiles();
-  cleanupAtoms();
-  cleanupFunctors();
-  cleanupArith();
-  cleanupInitialiseHooks();
-  cleanupExtensions();
-  cleanupOs();
-  Scleanup();
+
+  if ( reclaim_memory )
+  { freeStacks(PASS_LD1);
+    cleanupLocalDefinitions(LD);
+    freePrologLocalData(LD);
+    cleanupSourceFiles();
+    cleanupModules();
+    cleanupPrologFlags();
+    cleanupFlags();
+    cleanupRecords();
+    cleanupTerm();
+    cleanupAtoms();
+    cleanupFunctors();
+    cleanupArith();
+    cleanupInitialiseHooks();
+    cleanupExtensions();
+    cleanupOs();
+    Scleanup();
 #ifdef O_PLMT
-  cleanupThreads();
+    cleanupThreads();
 #endif
-  cleanupForeign();
-  cleanupCodeToAtom();
-  if ( rc )
-    cleanupMemAlloc();
+    cleanupForeign();
+    cleanupPaths();
+    cleanupCodeToAtom();
 #ifdef O_GMP
-  cleanupGMP();
+    cleanupGMP();
 #endif
+    cleanupDebug();
+  }
 
   UNLOCK();				/* requires GD->thread.enabled */
 
-  memset(&PL_global_data, 0, sizeof(PL_global_data));
-  memset(&PL_local_data,  0, sizeof(PL_local_data));
+  if ( reclaim_memory )
+  { memset(&PL_global_data, 0, sizeof(PL_global_data));
+    memset(&PL_local_data,  0, sizeof(PL_local_data));
+  }
 
   return TRUE;
 }
 
+
+int
+PL_cleanup(int rc)
+{ return cleanupProlog(rc, TRUE);
+}
+
+
 		 /*******************************
 		 *	ERRORS AND WARNINGS	*
 		 *******************************/
@@ -1358,9 +1436,9 @@ vsysError(const char *fm, va_list args)
   }
 
 #if defined(O_DEBUGGER)
-  systemMode(TRUE);
+  setAccessLevel(ACCESS_LEVEL_SYSTEM);
   Sfprintf(Serror, "\n\nPROLOG STACK:\n");
-  backTrace(NULL, 10);
+  backTrace(10);
   Sfprintf(Serror, "]\n");
 #endif /*O_DEBUGGER*/
 
@@ -1377,10 +1455,7 @@ action:
   ResetTty();
 
   switch(getSingleChar(Sinput, FALSE))
-  { case 'a':
-      abortProlog(ABORT_FATAL);
-      break;
-    case EOF:
+  { case EOF:
       Sfprintf(Serror, "EOF: exit\n");
     case 'e':
       PL_halt(3);
@@ -1388,14 +1463,11 @@ action:
     default:
       Sfprintf(Serror,
 	       "Unknown action.  Valid actions are:\n"
-	      "\ta\tabort to toplevel\n"
 	      "\te\texit Prolog\n");
       goto action;
   }
 
-  abortProlog(ABORT_FATAL);
-  PL_halt(3);
-  PL_fail;
+  return FALSE;					/* not reached */
 }
 
 
