@@ -19,7 +19,7 @@
 
     You should have received a copy of the GNU Lesser General Public
     License along with this library; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */
 
 #include "pl-incl.h"
@@ -80,9 +80,8 @@ typedef struct _opdef			/* predefined and enumerated */
 
 static void
 copyOperatorSymbol(Symbol s)
-{ GET_LD
-  operator *op = s->value;
-  operator *o2 = allocHeap(sizeof(*o2));
+{ operator *op = s->value;
+  operator *o2 = allocHeapOrHalt(sizeof(*o2));
 
   *o2 = *op;
 }
@@ -90,17 +89,16 @@ copyOperatorSymbol(Symbol s)
 
 static void
 freeOperatorSymbol(Symbol s)
-{ GET_LD
-  operator *op = s->value;
+{ operator *op = s->value;
 
   PL_unregister_atom((atom_t) s->name);
   freeHeap(op, sizeof(*op));
 }
 
 
-Table
-newOperatorTable()
-{ Table t = newHTable(8);
+static Table
+newOperatorTable(int size)
+{ Table t = newHTable(size);
 
   t->copy_symbol = copyOperatorSymbol;
   t->free_symbol = freeOperatorSymbol;
@@ -149,7 +147,7 @@ defOperator(Module m, atom_t name, int type, int priority, int force)
 
   LOCK();
   if ( !m->operators )
-    m->operators = newOperatorTable();
+    m->operators = newOperatorTable(8);
 
   if ( (s = lookupHTable(m->operators, (void *)name)) )
   { op = s->value;
@@ -157,7 +155,7 @@ defOperator(Module m, atom_t name, int type, int priority, int force)
   { UNLOCK();				/* already inherited: do not change */
     return TRUE;
   } else
-  { op = allocHeap(sizeof(*op));
+  { op = allocHeapOrHalt(sizeof(*op));
 
     op->priority[OP_PREFIX]  = -1;
     op->priority[OP_INFIX]   = -1;
@@ -504,7 +502,7 @@ current_op(Module m, int inherit,
       } else
 	return FALSE;
 
-      e = allocHeap(sizeof(*e));
+      e = allocHeapOrHalt(sizeof(*e));
       b = &e->buffer;
       initBuffer(b);
       e->index = 0;
@@ -659,6 +657,8 @@ void
 initOperators(void)
 { const opdef *op;
 
+  MODULE_system->operators = newOperatorTable(32);
+
   for( op = operators; op->name; op++ )
     defOperator(MODULE_system, op->name, op->type, op->priority, TRUE);
 }
