@@ -19,7 +19,7 @@
 
     You should have received a copy of the GNU Lesser General Public
     License along with this library; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */
 
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
@@ -216,10 +216,8 @@ static int	rlc_between(RlcData b, int f, int t, int v);
 static void	free_user_data(RlcData b);
 
 static RlcQueue	rlc_make_queue(int size);
-static void	rlc_free_queue(RlcQueue q);
 static int	rlc_from_queue(RlcQueue q);
 static int	rlc_is_empty_queue(RlcQueue q);
-static void	rlc_empty_queue(RlcQueue q);
 
 extern int	main();
 
@@ -389,7 +387,7 @@ rlc_main(HANDLE hInstance, HANDLE hPrevInstance,
 { TCHAR *	    argv[100];
   int		    argc;
   TCHAR		    program[MAXPATHLEN];
-  TCHAR	 	    progbase[100];
+  TCHAR		    progbase[100];
   RlcData           b;
   rlc_console_attr  attr;
 
@@ -734,7 +732,7 @@ static HKEY
 rlc_option_key(rlc_console_attr *attr, int create)
 { TCHAR Prog[256];
   TCHAR *address[] = { _T("Software"),
-  		      RLC_VENDOR,
+		      RLC_VENDOR,
 		      Prog,
 		      _T("Console"),
 		      (TCHAR *)attr->key,	/* possible secondary key */
@@ -985,7 +983,7 @@ rlc_wnd_proc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
 
       if ( placement.showCmd == SW_SHOWNORMAL )
       { b->win_x = placement.rcNormalPosition.left;
-  	b->win_y = placement.rcNormalPosition.top;
+	b->win_y = placement.rcNormalPosition.top;
 
 	b->modified_options |= OPT_POSITION;
       }
@@ -1411,7 +1409,7 @@ rlc_is_word_char(int chr)
 		 *	    SELECTION		*
 		 *******************************/
 
-#define SelLT(l1, c1, l2, c2) ((l1) < (l2) || (l1) == (l2) && (c1) < (c2))
+#define SelLT(l1, c1, l2, c2) ((l1) < (l2) || ((l1) == (l2) && (c1) < (c2)))
 #define SelEQ(l1, c1, l2, c2) ((l1) == (l2) && (c1) == (c2))
 
 static int
@@ -1565,19 +1563,6 @@ rlc_start_selection(RlcData b, int x, int y)
 }
 
 
-static void
-rlc_end_selection(RlcData b, int x, int y)
-{ int l, c;
-
-  rlc_translate_mouse(b, x, y, &l, &c);
-  if ( SelLT(l, c, b->sel_org_line, b->sel_org_char) )
-    rlc_set_selection(b, l, c, b->sel_org_line, b->sel_org_char);
-  else if ( SelLT(b->sel_org_line, b->sel_org_char, l, c) )
-    rlc_set_selection(b, b->sel_org_line, b->sel_org_char, l, c);
-  rlc_set_selection(b, l, c, l, c);
-}
-
-
 static int				/* v >= f && v <= t */
 rlc_between(RlcData b, int f, int t, int v)
 { int h = rlc_count_lines(b, b->first, v);
@@ -1616,6 +1601,8 @@ rlc_word_selection(RlcData b, int x, int y)
 static void
 rlc_extend_selection(RlcData b, int x, int y)
 { int l, c;
+  int el = b->sel_org_line;
+  int ec = b->sel_org_char;
 
   rlc_translate_mouse(b, x, y, &l, &c);
   if ( SelLT(l, c, b->sel_org_line, b->sel_org_char) )
@@ -1627,9 +1614,16 @@ rlc_extend_selection(RlcData b, int x, int y)
 	  for(; c > 0 && rlc_is_word_char(tl->text[c-1]); c--)
 	    ;
       }
+      if ( rlc_between(b, b->first, b->last, el) )
+      { TextLine tl = &b->lines[el];
+
+	if ( ec < tl->size && rlc_is_word_char(tl->text[ec]) )
+	  for(; ec < tl->size && rlc_is_word_char(tl->text[ec]); ec++)
+	    ;
+      }
     } else if ( b->sel_unit == SEL_LINE )
       c = 0;
-    rlc_set_selection(b, l, c, b->sel_end_line, b->sel_end_char);
+    rlc_set_selection(b, l, c, el, ec);
   } else if ( SelLT(b->sel_org_line, b->sel_org_char, l, c) )
   { if ( b->sel_unit == SEL_WORD )
     { if ( rlc_between(b, b->first, b->last, l) )
@@ -1639,9 +1633,16 @@ rlc_extend_selection(RlcData b, int x, int y)
 	  for(; c < tl->size && rlc_is_word_char(tl->text[c]); c++)
 	    ;
       }
+      if ( rlc_between(b, b->first, b->last, el) )
+      { TextLine tl = &b->lines[el];
+
+	if ( ec < tl->size && rlc_is_word_char(tl->text[ec]) )
+	  for(; ec > 0 && rlc_is_word_char(tl->text[ec-1]); ec--)
+	    ;
+      }
     } else if ( b->sel_unit == SEL_LINE )
       c = b->width;
-    rlc_set_selection(b, b->sel_start_line, b->sel_start_char, l, c);
+    rlc_set_selection(b, el, ec, l, c);
   }
 }
 
@@ -1652,7 +1653,7 @@ rlc_read_from_window(RlcData b, int sl, int sc, int el, int ec)
   TCHAR *buf;
   int i = 0;
 
-  if ( el < sl || el == sl && ec < sc )
+  if ( el < sl || (el == sl && ec < sc) )
     return NULL;			/* invalid region */
   if ( !(buf = rlc_malloc(bufsize * sizeof(TCHAR))) )
     return NULL;			/* not enough memory */
@@ -2121,7 +2122,7 @@ rlc_queryfont(RlcData b)
   cf.hwndOwner   = b->window;
   cf.lpLogFont   = &lf;
   cf.Flags       = CF_SCREENFONTS|
-    		   CF_NOVERTFONTS|
+		   CF_NOVERTFONTS|
 		   CF_NOSIMULATIONS|
 		   CF_FORCEFONTEXIST|
 		   CF_INITTOLOGFONTSTRUCT;
@@ -2159,8 +2160,8 @@ rlc_make_buffer(int w, int h)
   b->changed	    = CHG_CARET|CHG_CHANGED|CHG_CLEAR;
   b->imode	    = IMODE_COOKED;	/* switch on first rlc_read() call */
   b->imodeswitch    = FALSE;
-  b->lhead 	    = NULL;
-  b->ltail 	    = NULL;
+  b->lhead	    = NULL;
+  b->ltail	    = NULL;
 
   memset(b->lines, 0, sizeof(text_line) * h);
   for(i=0; i<h; i++)
@@ -2330,7 +2331,7 @@ rlc_adjust_line(RlcData b, int line)
 
   if ( tl->text && !tl->adjusted )
   { tl->text = rlc_realloc(tl->text, tl->size == 0
-			   	? sizeof(TCHAR)
+				? sizeof(TCHAR)
 				: tl->size * sizeof(TCHAR));
     tl->adjusted = TRUE;
   }
@@ -2877,7 +2878,7 @@ window_loop(LPVOID arg)
   if ( b->closing <= 2 )
   { MSG msg;
     TCHAR *waiting = _T("\r\nWaiting for Prolog. ")
-      		     _T("Close again to force termination ..");
+		     _T("Close again to force termination ..");
 
     rlc_write(b, waiting, _tcslen(waiting));
 
@@ -3032,16 +3033,6 @@ rlc_make_queue(int size)
 }
 
 
-void
-rlc_free_queue(RlcQueue q)
-{ if ( q )
-  { if ( q->buffer )
-      rlc_free(q->buffer);
-    rlc_free(q);
-  }
-}
-
-
 static int
 rlc_resize_queue(RlcQueue q, int size)
 { TCHAR *newbuf;
@@ -3098,12 +3089,6 @@ rlc_is_empty_queue(RlcQueue q)
 }
 
 
-void
-rlc_empty_queue(RlcQueue q)
-{ q->first = q->last = 0;
-}
-
-
 static int
 rlc_from_queue(RlcQueue q)
 { if ( q->first != q->last )
@@ -3354,8 +3339,6 @@ rlc_clearprompt(rlc_console c)
 		 *	    MISC STUFF		*
 		 *******************************/
 
-static TCHAR current_title[RLC_TITLE_MAX];
-
 void
 rlc_title(rlc_console c, TCHAR *title, TCHAR *old, int size)
 { RlcData b = rlc_get_data(c);
